/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: app2d.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:45:58 stolz>
 **************************************************************************
 *  MODULE              : app2d
 *
 *  IMPLEMENTATION
 *
 *	Hold's predefined appereances and the new/free function
 *	for the appearance-structure. (quite simple)
 *
 *  NOTE
 *      The module specification and definition is given in app2d.h!
 *
 *  HISTORY
 *  28-Jun-94   Split app2d from xdraw.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "app2d.h"
#include "error.h"
#include "basic.h"


/**************************************************************************
 * exported appearances
 */
static vector3	def_2d_mob_points[17] = {
    {0.000,  0.000, 0.0},			/* 0 */

    {-0.284, -0.240, 0.0},  		/* 1- mob */
    {-0.184, -0.340, 0.0},
    {+0.641, -0.340, 0.0},
    {+0.741, -0.240, 0.0},
    {+0.741, +0.240, 0.0},
    {+0.641, +0.340, 0.0},
    {-0.184, +0.340, 0.0},
    {-0.284, +0.240, 0.0},

/* old
   {-0.284, -0.277, 0.0},  		* 1- mob *
   {-0.221, -0.340, 0.0},
   {+0.678, -0.340, 0.0},
   {+0.741, -0.277, 0.0},
   {+0.741, +0.277, 0.0},
   {+0.678, +0.340, 0.0},
   {-0.221, +0.340, 0.0},
   {-0.284, +0.277, 0.0},
*/

    {-0.120,	+0.327, 0.0},			/* 9- left rear wheel */
    {+0.120,	+0.327, 0.0},
    {+0.120,	+0.269, 0.0},
    {-0.120,	+0.269, 0.0},

    {-0.120,	-0.269, 0.0},			/* 13- right rear wheel */
    {+0.120,	-0.269, 0.0},
    {+0.120,	-0.327, 0.0},
    {-0.120,	-0.327, 0.0}
};

static vector3	def_2d_wheel_points[6] = {
    {0.000,	 0.000, 0.0},
    {-0.045,  0.000, 0.0},			/* front wheel */
    {-0.106, +0.019, 0.0},
    {+0.016, +0.019, 0.0},
    {+0.016, -0.019, 0.0},
    {-0.106, -0.019, 0.0}
};

static vector3	def_2d_us_sensor_points[4] = {
    {0.0, +0.015, -0.020},			/* sensor */
    {0.0, +0.015, +0.000},
    {0.0, -0.015, +0.000},
    {0.0, -0.015, -0.020}
};

static vector3	def_2d_camera_points[8] = {
    {-0.100, +0.050, 0.0},			/* camera */
    {+0.080, +0.050, 0.0},
    {+0.080, -0.050, 0.0},
    {-0.100, -0.050, 0.0},
    {+0.080, +0.020, 0.0},
    {+0.120, +0.020, 0.0},
    {+0.120, -0.020, 0.0},
    {+0.080, -0.020, 0.0}
};

static int	def_2d_mob_code[] = {
/*    POINTS_2D,	0, */
    POLYLINE_2D, 1, 2, 3, 4, 5, 6, 7, 8, 1,	/* mob-border */
    POLYLINE_2D, 9, 10, 11, 12, 9,		/* left rear wheel */
    POLYLINE_2D, 13, 14, 15, 16, 13,            /* right rear wheel */
    END_2D
};
static int	def_2d_wheel_code[] = {
    POLYLINE_2D, 0, 1,
    POLYLINE_2D, 2, 3, 4, 5, 2,
    END_2D
};
static int	def_2d_us_sensor_code[] = {
    POLYLINE_2D, 0, 1, 2, 3,
    END_2D
};
static int	def_2d_camera_code[] = {
    POLYLINE_2D, 0, 1, 2, 3, 0,
    POLYLINE_2D, 4, 5, 6, 7,
    END_2D
};


app_2d def_2d_mob = { 17, def_2d_mob_points, def_2d_mob_code };
app_2d def_2d_wheel = { 6, def_2d_wheel_points, def_2d_wheel_code };
app_2d def_2d_us_sensor = { 4, def_2d_us_sensor_points, def_2d_us_sensor_code };
app_2d def_2d_camera = { 8, def_2d_camera_points, def_2d_camera_code };
app_2d def_2d_empty = { 0, 0, 0 };






void app2d_free(app_2d *app)
/**************************************************************************
 */
{
    if (app && (app!=&def_2d_empty) && (app!=&def_2d_mob) &&
	(app!=&def_2d_wheel) && (app!=&def_2d_us_sensor) &&
	(app!=&def_2d_camera)) {

	if (app->point) free(app->point);
	if (app->code) free(app->code);
	free(app);
    }
}



app_2d *app2d_new(int nr_points, int code_length)
/**************************************************************************
 */
{
    app_2d *app;

    if (nr_points<=0 || code_length<=1) return &def_2d_empty;

    app = malloc(sizeof(app_2d));
    if (app==0) error("app2d: app2d_new() - can't alloc memory for app2d-struct");

    app->nr_points = nr_points;

    app->point = malloc(nr_points*sizeof(vector3));
    if (app->point==0)
	error("app2d: app2d_new() - can't alloc memory for app2d-struct");

    app->code = malloc((code_length+1)*sizeof(int));
    if (app->code==0)
	error("app2d: app2d_new() - can't alloc memory for app2d-struct");


    app->code[0] = END_2D;
    app->code[code_length] = END_2D;

    return app;
}


int app2d_code_length(app_2d *app)
/**************************************************************************
 */
{
    if (app && app->code) {
	int len = 0;

	while(app->code[len] != END_2D)
	    len++;

	return len+1;
    }
    return 0;
}
