/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: app2d.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 16:50:17 stolz>
 **************************************************************************
 *  MODULE              : xdraw
 *
 *  FUNCTION
 *	This module provide a simple structure which can hold a
 *	representation of on object-appearance in form of points and
 *      lines. The main goal is to hold the 2D-Form of some objects
 *	for further drawing with the xdraw-module, and to allow a
 *	simple form for a 2D object representation.
 *
 *      Such representaton consist of a array of points and a
 *      description of what to draw with this points. This discription
 *      is a simple array of point-numbers mixed with special
 *	command codes, which say what is to do.
 *
 *	This codes have the following meaning:
 *         POINTS_2D	Simply draw a point for each following points
 *                      the next code is reached.
 *         POLYLINE_2D	Connect the following points-numbers until the
 *                      next code with lines.
 *	   END_2D	End of the description.
 *
 *      The point-numbers are simple the index to the point-array.
 *	Therfore must a point number >= 0.
 *
 *  USAGE
 *
 *
 *  NOTE
 *
 *  HISTORY
 *  26-Jun-94   Split 2D-Appearance from xdraw.h
 *
 **************************************************************************/
#ifndef APP2D_H
#define APP2D_H


#ifndef VECTOR_H
#include "vector.h"
#endif


/**************************************************************************
 * Draw-Codes
 */
enum draw_codes {
	END_2D      = -1,
	POINTS_2D   = -2,
	POLYLINE_2D = -3
};


/**************************************************************************
 * Apperance-structure definition.
 */
typedef struct {
    int		nr_points;
    vector3	*point;
    int		*code;
} app_2d;



/**************************************************************************
 * predefined structures.
 *
 */
extern app_2d def_2d_empty;	/* an empty structure! */
extern app_2d def_2d_mob;	/* an bounding box of a mobile Robot */
extern app_2d def_2d_wheel;	/* one movable front-wheel */
extern app_2d def_2d_us_sensor; /* one US-Sensor */
extern app_2d def_2d_camera;	/* one Camera */



app_2d *app2d_new(int nr_points, int code_length);
/****************************************************************************
 * NAME
 *	app2d_new(...)	allocate a appearance structure
 *
 * FUNCTION
 *	This routine allocate the memory the the
 *	appearance-structure inclusive point-array and code-array.
 *      It also initialize the nr_points field of the structure.
 *	The first and last element of the code-array is initialized with
 *	END_2D, so that app2d_code_length provide an empty code-
 *	description.
 *
 *      It's up to the user of this function to take care of the
 *	point-array-size (parameter nr_points) and the code-array-
 *	size (parameter code_length). The allocation is done
 *      only for this sizes. But it's is not necessary to use
 *	the whole point-array or code-array.
 *
 *	The size of the code-array will be code_length+1 to hold
 *	the last END_2D code!
 *
 * PARAMETERS
 *	nr_points               Maximum size of the point-array.
 *      code_length		Maximum size of the code-array.
 *
 * RESULTS
 *	return the pointer to the appereance structure.
 *
 ****************************************************************************
 */


void    app2d_free(app_2d *app2d);
/****************************************************************************
 * NAME
 *	app2d_free(...)		free the appearance structure.
 *
 * FUNCTION
 *	This routine frees the appearance-structure which was inclusive the
 *	point-array and the code code-array. It take als care if
 *      the app2d-structures is one of the predefined. If the
 *	parameter app2d is zero, it do nothing. app2d_free() frees also
 *	only the pointers which are not 0, so you don't have to use app2d_new()
 *	if you set the app2d-structure correctly.
 *
 * PARAMETERS
 *	app2d		pointer to a appearance-structure.
 *
 ****************************************************************************
 */


int	app2d_code_length(app_2d *app);
/****************************************************************************
 * NAME
 *	app2d_code_length(...)	determine the code-length(without END-code)
 *
 * FUNCTION
 *      The number of codes and indices are counted until the
 *	first END_2D code is examined.
 *	This routine take care of an zero parameter app, or an
 *      zero app->code pointer.
 *
 * PARAMETERS
 *	app2d		pointer to a appearance-structure.
 *
 * RESULTS
 *	returns the number of point-numbers or codes upto the first
 *	END_2D code. (END_2D don't count!)
 *
 ****************************************************************************
 */

#endif
