/********************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: basic.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 16:51:14 stolz>
 ********************************************************************************
 *  MODULE              : basic
 *
 *  FUNCTION
 *	Some usefull addition to the normal ansi-types and functions.
 *	Necessary
 *
 *	This include a boolean-type, usefull defines and macros for
 *	the computation with floating points and a random-function.
 *
 *  NOTE
 *	Include this header-file after the normal <> includes.
 *
 *  HISTORY
 *  07-Mar-94   Creation date.
 *  14-Jun-94	Implement dcut() and dround().
 *  16-Jun-94   Add drand() function.
 *
 ********************************************************************************/
#ifndef BASIC_H
#define BASIC_H



/**************************************************************************
 * definition for a zero-pointer
 */
#ifndef NULL
#define NULL 0
#endif



/**************************************************************************
 * a boolean-type, with defines for TRUE and FALSE.
 */

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define and	||
#define or	&&
#define not(x)	(!(x))

typedef int	boolean;



/**************************************************************************
 * usefull macros for all types of numbers.
 */

#ifndef max
#define max(a,b) (((a)>(b))?(a):(b))
#endif

#ifndef min
#define min(a,b) (((a)<(b))?(a):(b))
#endif

#define is_positive(x) ((x)>=0)
#define is_negative(x) ((x)<0)

#define has_same_sign(a,b) ((is_positive(a)&&is_positive(b))||(is_negative(a)&&is_negative(b)))



/**************************************************************************
 * definitions and functions for the usage of floating point numbers.
 */

#ifndef PI
#define PI 3.14159265358979323846
#endif

#define PIx2 6.2831853717958647692


#define ZERO_MAX	1e-6		/* max limit for a zero-detection */

#define is_zero(x) (fabs(x)<ZERO_MAX)
#define is_not_zero(x) (fabs(x)>=ZERO_MAX)

#define in_range(low,x,high) (((x)>=(low))&&((x)<=(high)))
#define not_in_range(low,x,high) (((x)<(low))||((x)>(high)))


double	sign_heritage(double sgn, double val);
/****************************************************************************
 * NAME
 *	sign_heritage(...)	returns a number with the sign of another.
 *
 * FUNCTION
 *	This routine returns val with the sign of sgn, or zero if sgn was
 *	zero.
 *
 * PARAMETERS
 *      sgn		Number from which val should herite the sign.
 *	val             Value.
 *
 * RESULTS
 *	Returns val with the sign of sgn.
 *
 * EXAMPLE
 *	sign_heritage(-123456789.0, +10.0) should return -10.
 *	sign_heritage(0.0, -10.0) should return 0.
 *
 ****************************************************************************
 */


long	dround(double x);
/****************************************************************************
 * NAME
 *	dround(...)		round the number x
 *
 * FUNCTION
 *      Round the number x to the next higher number and convert it
 *	to a long. This means that |dround(x)| = |x| or |x+1|.
 *      The convertions is done by adding 0.5 in respect of the
 *	sign of x and then converting it to long.
 *
 * PARAMETERS
 *      x               any floating point number.
 *
 * RESULTS
 *	Returns normalized angle in the range [-PI,+PI].
 *
 * EXAMPLES
 *	dround(0.50001)	should return 1.
 *	dround(-5.7)	should return 6.
 *      dround(13.1)	should return 13.
 *
 ****************************************************************************
 */


double	dcut(double x, double fac);
/****************************************************************************
 * NAME
 *	dcut(...)		x DIV fac for floating point numbers.
 *
 * FUNCTION
 *	This routine returns how often fac fit into x.
 *
 * PARAMETERS
 *      x               any number.
 *	fac		any number.
 *
 * RESULTS
 *	Returns x DIV fac.
 *
 ****************************************************************************
 */


double	dmod(double x, double fac);
/****************************************************************************
 * NAME
 *	dmod(...)		compute the modulo of x % fac.
 *
 * FUNCTION
 *	Compute the modulo of two floating point numbers.
 *
 * PARAMETERS
 *      x               any number.
 *	fac		any number.
 *
 * RESULTS
 *	Returns x % f, so that  dmod(x, f) + dcut(x, f) = x.
 *
 * EXAMPLE
 *      dmod(17.3, 5.0)	should return 2.3.
 *
 ****************************************************************************
 */


double	norm_angle(double angle);
/****************************************************************************
 * NAME
 *	norm_angle(...)		normalize the angle
 *
 * FUNCTION
 *	This routine normalize the angle to the range [-PI,+PI].
 *      So that the trigonometric functions sin/cos/tan/...
 *      returns the same results.
 *
 * PARAMETERS
 *      angle		Angle in rad.
 *
 * RESULTS
 *	Returns normalized angle in the range [-PI,+PI].
 *
 ****************************************************************************
 */


double	drand();
/****************************************************************************
 * NAME
 *	drand(...)		random number generator.
 *
 * FUNCTION
 *	This routine returns a normal-distributed random number
 *	in the range [0,1].
 *
 * RESULTS
 *	Returns a random number in the range [0.0, 1.0].
 *
 ****************************************************************************
 */


#endif
