/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: clipping.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:51:34 stolz>
 **************************************************************************
 *  MODULE              : clipping
 *  IMPORT		: vector.o
 *
 *  NOTE
 *      The module specification and definition is given in clipping.h!
 *
 *  HISTORY
 *  16-May-94   Creation date
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "clipping.h"
#include "basic.h"

int clip_points(vector3 *points, vector3 start, vector3 line,
		      int xyz_select, double min, double max)
/****************************************************************************
 * FUNCTION
 *	Clips the line (start->+line) in the x or y or z-coordinates at
 *	the values min and max. The clipped points are added to the
 *      vector-array points. Only clipped point which are not identicall
 *	to the start or end of the line are returned!
 *
 * PREREQUISITES
 *	min < max
 *      0 <= xyz_select <= 2  (x=0, y=1, z=2)
 *
 *	start is the starting point of the line.
 *	the vector line is the index of the line, so the end-point
 *	is start+line!
 *
 * RESULTS
 *	The clipped points are stored in the points-array. The order
 *	of saving is from lines-start to line-end.
 *      The number of clipped points is returned.
 *	->   0 <= return value <= 2
 *
 */
{
    double	alpha_min, alpha_max, f;
    int		index = 0;

#ifdef DEBUG_DETAIL
    fprintf(stderr, "clip points (%f %f %f) -> (%f %f %f)"
		    "\non index %d in region [%f, %f]\n",
                    start[0], start[1], start[2],
                    line[0], line[1], line[2],
                    xyz_select, min, max);
#endif

    /* compute factors to reach the min/max-values.
     * sort this to alpha_min/max. (alpha_min < alpha_max)
     */
    if (is_zero(line[xyz_select])) {	/* line lies in xyz-direction */

	return 0;
    }
    else {
	alpha_min = (min - start[xyz_select]) / line[xyz_select];
	f = (max - start[xyz_select]) / line[xyz_select];
	if (f<alpha_min) {
	    alpha_max = alpha_min;
	    alpha_min = f;
	}
	else {
	    alpha_max = f;
	}
    }

#ifdef DEBUG_DETAIL
    fprintf(stderr, "alpha_min = %f, alpha_max = %f\n", alpha_min, alpha_max);
#endif

    /* 1. clipping point without the start/end-point of the line
     */
    if (in_range(ZERO_MAX, alpha_min, 1.0-ZERO_MAX)) {

	v3_addscal(points[0], start, alpha_min, line);
	index++;
    }

    /* 2. clipping point without the start/end-point of the line
     */
    if (in_range(ZERO_MAX, alpha_max, 1.0-ZERO_MAX) &&
	is_not_zero(alpha_max-alpha_min)) {

	v3_addscal(points[index], start, alpha_max, line);
	index++;
    }

    return index;
}


