/********************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: com_motn.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/03/02 19:18:06 stolz>
 ********************************************************************************
 *  MODULE              : com_motn
 *
 *  NOTE
 *      The module specification and definition is given in com_motn.h.
 *
 *  HISTORY
 *  24-Mar-94   Creation date
 *  02-Mar-95   remove parameter-ordering of motn le/ri/fo/ba.
 *
 ********************************************************************************/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "mob_private.h"
#include "com_scanner.h"
#include "basic.h"
#include "mob_limits.h"
#include "error.h"


static char *motn_foba(mob m, long option, char *args)
/*************************************************************************
 * forward/backward-movement
 */
{
    long par,
         lv, ld=0;  /* velocity and distance in mm/s and mm */
    double v,   /* velocity in m/s */
           d,   /* distance in m */
	   a,	/* linear acceleration in m/s^2 */
	   u,   /* distance which can be drived with constant speed */
	   t,   /* acceleration and deceleration time */
	   t2;  /* time to drive with constant speed */
    mob_motion_state ms;

    /* check if robot is in motion */
    if (motion_type(m) != MT_STOP) return ERR_ROBOT_IN_MOTION;


    /* parse args for D= and V= in this order to set the velocity and distance
     * ld=0 -> infinite distance.
     */
    lv = LDEFAULT_LINEAR_VEL;

    while ((par = get_word(&args, 2)) != WORD_EMPTY) switch (par) {
    case VAL_DISTANCE:
	if (get_long(&args, &ld)) return ERR_INVALID_PARAMETER;
	break;
    case VAL_VELOCITY:
	if (get_long(&args, &lv)) return ERR_INVALID_PARAMETER;
	if (lv<LMIN_LINEAR_VEL || lv>LMAX_LINEAR_VEL) return ERR_ILLEGAL_VELOCITY;
	break;
    default:
	return ERR_INVALID_PARAMETER;
	break;
    }

    /* set velocity to forward or backward motion
     */
    if (option==OPT_BACKWARD) lv = -lv;

    /* compute linear velocity in m/s */
    v = (double)lv/1000.0;
    /* compute acceleration time for default acc to reach velocity  */
    a = (lv>=0)? DEFAULT_LINEAR_ACC : -DEFAULT_LINEAR_ACC;
    t = fabs(v)/fabs(a);

    /* is this motion limited by a distance? */
    if (ld) { /* yes */

	/* check if distance can be obtained via default acceleration */

	d = (double)ld/1000.0;
	if ((u = d - fabs(a) * t * t) < 0.0) {
	    /* distance to short for default acc, make t smaller.
	     * so the specified velocity may not be reached.
	     */
	    t = sqrt(d/fabs(a));
	    t2 = 0.0;
	}
	else t2 = u/fabs(v);

	motion_add(m, t, a, 0.0);
	if (!is_zero(t2))
	    motion_add(m, t2, 0.0, 0.0);
	motion_add(m, t, -a, 0.0);
    }
    else {
	/* unlimited movement */
        motion_add(m, t, a, 0.0);
    }

    mob_get_intern_state(m, &ms);

    m->motion.motn.x_start = ms.x;
    m->motion.motn.y_start = ms.y;
    m->motion.motn.linear_motion = 1;

    return 0;
}




static char *motn_leri(mob m, long option, char *args)
/*************************************************************************
 * rotation left/right
 */
{
    long par,
         lw, la=0;  /* angular velocity and angle in deg/s and deg */
    double w,   /* angular velocity in rad/s */
           a,   /* angle in rad */
	   u,	/* angular acceleration in rad/s^2 */
	   rr,  /* distance which can be drived with constant speed */
	   t,   /* acceleration and deceleration time */
	   t2;  /* time to rotate with constant speed */
    mob_motion_state ms;


    /* check if robot is in motion */
    if (motion_type(m) != MT_STOP) return ERR_ROBOT_IN_MOTION;


    /* parse args for A= and V= in this order to set the angle and velocity
     * la=0 -> infinite angle.
     */
    lw = DTOI_RAD_TO_GRAD(DEFAULT_ANGULAR_VEL);

    while ((par = get_word(&args, 2))!=WORD_EMPTY) switch(par) {
    case VAL_ANGLE:
	if (get_long(&args, &la)) return ERR_INVALID_PARAMETER;
	break;
    case VAL_VELOCITY:
	if (get_long(&args, &lw)) return ERR_INVALID_PARAMETER;
	if (not_in_range(LMIN_ANGULAR_VEL, lw, LMAX_ANGULAR_VEL))
	    return ERR_ILLEGAL_VELOCITY;
	break;
    default:
	return ERR_INVALID_PARAMETER;
    }

    /* set velocity to forward or backward motion
     */
    if (option==OPT_RIGHT) lw = -lw;

    /* compute linear velocity in mdeg/s */
    w = (double)lw / 1000.0;

    /* compute acceleration time for default acc to reach velocity  */
    u = (lw >= 0) ? DEFAULT_ANGULAR_ACC : -DEFAULT_ANGULAR_ACC;
    t = fabs(w/u);

    /* is this motion limited by a distance? */
    if (la) { /* yes */

	/* check if distance can be obtained via default acceleration */
	a = PI/180.0 * (double)la;
	if ((rr = a - fabs(u) * t * t) < 0.0) {
	    /* distance to short for default acc, make t smaller.
	     * so the specified velocity may not be reached.
	     */
	    t = sqrt(fabs(a/u));
	    t2 = 0.0;
	}
	else t2 = rr/fabs(w);

	motion_add(m, t, 0.0, u);
	if (!is_zero(t2))
	    motion_add(m, t2, 0.0, 0.0);
	motion_add(m, t, 0.0, -u);
    }
    else {
	/* unlimited movement */
        motion_add(m, t, 0.0, u);
    }

    mob_get_intern_state(m, &ms);

    m->motion.motn.angle_start = ms.angle;
    m->motion.motn.linear_motion = 0;

    return 0;
}

char *motn(mob m, char *args)
/*************************************************************************
 * MOTN-emulation
 */
{
    long option = get_word(&args, 2);
    double d;   /* temp variable for distance of status information */
    long ld;
    static char status_info[] = "0 123456789123456789";
    mob_motion_state ms;

    /* check if the correct or none motionsserver is active.
     * if the option is ON, set motion server to MS_MOTN.
     */
    if (option != OPT_ON) {
	if (motion_server(m) == MS_NONE) {              /* motion server inactive */
	    return ERR_SERVER_INACTIVE;
	} else if (motion_server(m) != MS_MOTN) {   /* wrong motion server */
	    return ERR_SERVER_ALLOCATED;
	}
    }
    else {  /* MOTN ON */
        if (get_word(&args, 2)!=WORD_EMPTY) return ERR_INVALID_PARAMETER;
	if (motion_server(m) != MS_NONE && motion_server(m) != MS_MOTN) {
	    return ERR_SERVER_ALLOCATED;
	}
	change_motion_server(m, MS_MOTN);
	
	return 0;
    }
    /* Motion server is now MS_MOTN! */
    
    /* Argument parsing  
     */
    switch(option) {

    case OPT_OFF:
	if (get_word(&args, 2)!=WORD_EMPTY) return ERR_INVALID_PARAMETER;
	/* check if robot is in motion */
        if (motion_type(m) != MT_STOP) return ERR_ROBOT_IN_MOTION;
	change_motion_server(m, MS_NONE);
        break;

    case OPT_FORWARD:
    case OPT_BACKWARD:
	return motn_foba(m, option, args);
	break;

    case OPT_LEFT:
    case OPT_RIGHT:
	return motn_leri(m, option, args);
	break;

    case OPT_KILL:
	if (get_word(&args, 2)!=WORD_EMPTY) return ERR_INVALID_PARAMETER;
	motion_kill(m);
	break;

    case OPT_WAIT:
	if (get_word(&args, 2)!=WORD_EMPTY) return ERR_INVALID_PARAMETER;
	d = motion_stop_time(m);
	if (d!=0.0) mob_command_delay(m, d+1e-4);
	break;

    case OPT_STATUS:
	if (get_word(&args, 2)!=WORD_EMPTY) return ERR_INVALID_PARAMETER;

        status_info[0] = (motion_type(m) == MT_STOP) ? '0' : '1';

	mob_get_intern_state(m, &ms);

	switch (m->motion.motn.linear_motion) {
	case 1:
	    d = sqrt((ms.x) * (ms.x) + (ms.y)*(ms.y));
	    d -= sqrt(pow(m->motion.motn.x_start,2.0) +
		      pow(m->motion.motn.y_start,2.0));
	    ld = dround(d*1000.0);

	    /* print out linear distance in mm */
	    sprintf(&status_info[2], "%ld", ld);
	    mob_output(m, status_info);
	    break;

	case 0:
	    d = ms.angle - m->motion.motn.angle_start;
	    ld = dround(d*180.0/PI);

	    /* print out angular distance in grad */
	    sprintf(&status_info[2], "%ld", ld);
	    mob_output(m, status_info);
	    break;

	default:
	    break;
	}
	break;

    default:
	return ERR_INVALID_PARAMETER;
    }
    return 0;
}
