/********************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: com_motv.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/03/07 16:05:54 stolz>
 ********************************************************************************
 *  MODULE              : com_motv
 *
 *  IMPLEMENTATION
 *	The Queue-mechansim of the MOTV-command is emulated
 *	by an motion-queue is accessed from the timepoint-function
 *	motv_changed().
 *
 *  NOTE
 *      The module specification and definition is given in com_motv.h!
 *
 *  HISTORY
 *  24-Mar-94   Creation date
 *  29-Apr-94   motv ad/la working fine!
 *  02-Mar-95   remove parameter-ordering of motv ad/la
 *
 ********************************************************************************/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "com_motv.h"
#include "mob_private.h"
#include "mob_limits.h"
#include "com_scanner.h"
#include "basic.h"
#include "queue.h"
#include "error.h"




static void calc_limit_acc(double *ta, double *aa, double *uu,
			   double v0, double w0, double v1, double w1,
			   double a_max, double u_max)
/***************************************************************************
 * computes the time ta and the linear and angular acceleration
 * aa and uu to change the velocities from v0,w0 to v1,w1.
 * the acceleration values are smaller or equal than a_max, u_max!
 */
{
    double t;


    if (is_zero(a_max)) *ta = 0.0; else *ta = fabs((v1-v0)/a_max);
    if (is_zero(u_max)) t = 0.0; else t = fabs((w1-w0)/u_max);

    if (t > (*ta)) {
	*ta = t;
    }

    if (is_zero(*ta)) {
	*aa = 0.0;
	*uu = 0.0;
    }
    else {
	*aa = (v1-v0) / (*ta);
	*uu = (w1-w0) / (*ta);
    }
}


static void print_motion(motv_motion *m)
/***************************************************************************
 * test-function prints the a motv-motion onto stderr.
 */
{
    fprintf(stderr, "MOTV-Motion\nType: ");

    switch(m->type) {
    case MVT_VELOCITY:
	fprintf(stderr, "VELOCITY\n");
	break;
    case MVT_DISTANCE:
	fprintf(stderr, "DISTANCE\nd=%fm\n", m->distance);
	break;
    case MVT_ANGLE:
	fprintf(stderr, "ANGLE\nangle=%frad\n", m->angle);
	break;
    case MVT_TIME:
	fprintf(stderr, "TIME\ntime=%fs\n", m->time);
	break;
    case MVT_INFINITE:
	fprintf(stderr, "INFINITE\n");
	break;
    default:
	fprintf(stderr, "unkown\n");
	break;
    }

    fprintf(stderr, "v=%fm/s\n"
	    "w=%frad/s\n"
	    "a=%fm/s^2\n"
	    "u=%frad/s^2\n"
	    "ta=%fs/s\n"
	    "tc=%fs/s\n",
	    m->v, m->w, m->a, m->u, m->ta, m->tc);
}


/*
 * compute the acc time ta with the linear and angular accelerations aa and uu
 * and the constant vel time tc to fullfill the deceleration phase of the
 * last motion with type "type". The acc. phase transforms from velocity
 * v0,w0 to the new one v and w. The acc.-limits for this are in the parameters
 * a and u.
 */
static void
calc_times_acc(double *ta, double *tc, double *aa, double *uu,
	      double v0, double w0, double ta0, int type,
	      double v, double w, double a, double u)
/*
 * ta0 >=0, v0/w0=v/w=a/u
 */
{
    double t;

    *ta = 0.0;
    *tc = 0.0;

    if ((t=v-v0)>0.0) *aa = a; else *aa = -a;
    if (is_not_zero(*aa)) *ta = fabs(t / (*aa));

    if ((t=w-w0)>0.0) *uu = u; else *uu = -u;
    if (is_not_zero(*uu)) *ta = fabs(t / (*uu));

    if (ta0==0.0) return;

    switch(type) {
    case MVT_DISTANCE:
	if (is_zero(v)) *tc = 0.0;
	else *tc = (0.5*fabs(a)*ta0*ta0-fabs(v0)*(*ta)-0.5*fabs(*aa)*(*ta)*(*ta))/fabs(v);
	break;
    case MVT_ANGLE:
	if (is_zero(w)) *tc = 0.0;
	else *tc = (0.5*fabs(u)*ta0*ta0-fabs(w0)*(*ta)-0.5*fabs(*uu)*(*ta)*(*ta))/fabs(w);
	break;
    case MVT_TIME:
	*tc = ta0 - *ta;
	break;
    case MVT_INFINITE:
	*tc = 0.0;
	break;
    default:
	error("com_motv: calc_times_acc() - unknown motion type\n");
	break;
    }
}


static void acc_prop_to_vel(motv_motion *motion)
/***************************************************************************
 * sets the acceleration of a motv-motion proportional
 * with correct +/- to the veclocities.
 */
{
    if (is_zero(motion->v)) {
	motion->u = fabs(motion->u);
    }
    else {
	motion->u = fabs(motion->w * motion->a / motion->v);
    }

    if (is_zero(motion->v)) {
	motion->a = 0.0;
    }
    else {
	motion->a = fabs(motion->a);
    }
}

static void calc_times(motv_motion *motion)
/***************************************************************************
 * Calculate the acceleration time ta and the time for constant velocity
 * tc of an motv-motion out of the given v,w,a,u and also compute the
 * really reached velocities wx, vx. (with wx<w, vx<v)
 */
{
    if (is_not_zero(motion->a)) motion->ta = fabs(motion->v / motion->a);
    else if (is_not_zero(motion->u)) motion->ta = fabs(motion->w / motion->u);
    else motion->ta = 0.0;

    switch(motion->type) {
    case MVT_DISTANCE:
        motion->tc = fabs(motion->distance / motion->v) - motion->ta;
	if (motion->tc < 0.0) {
	    motion->ta = sqrt(fabs(motion->distance / motion->a));
	    motion->tc = 0.0;
	}
	break;
    case MVT_ANGLE:
        motion->tc = fabs(motion->angle / motion->w) - motion->ta;
	if (motion->tc < 0.0) {
	    motion->ta = sqrt(fabs(motion->angle / motion->u));
	    motion->tc = 0.0;
	}
	break;
    case MVT_TIME:
	motion->tc = motion->time - motion->ta;
	break;
    case MVT_INFINITE:
	motion->tc = 1e9;
	break;
    default:
	error("com_motv: calc_times() - unknown motion type\n");
	break;
    }
    if (motion->tc < 0.0) {

	motion->tc = 0.0;
    }

    motion->v_max = sign_heritage(motion->v, motion->a * motion->ta);
    motion->w_max = sign_heritage(motion->w, motion->u * motion->ta);

}

static void calc_via_velocities(double *vx, double *wx,
				double v0, double w0, double v, double w)
/***************************************************************************
 * sets the via velocities vx,vw to (v,w) if the v/w=v0/w0 or 
 * else to (vx,wx)=(0,0).
 */
{
    if (v0*v < 0.0 || w0*w < 0.0 || is_not_zero(v0*w-v*w0)) {
	*vx = 0.0;
	*wx = 0.0;
    }
    else {
	*vx = v;
	*wx = w;
    }
}



/*
 * computes the linear and angular acceleration aa and uu so that
 * v0,w0 -> vx,wx -> v1,w1 and vx,wx in [v0,v1],[w0,w1] and vx=wx=0
 * if v0/w0 != v1/w1.
 */
static void calc_via_accs(double *aa, double *uu,
			  double v0, double w0, double v1, double w1,
			  double a, double u)
{
    double tv, tw;

    tv = v1 * w0;
    tw = v0 * w1;

    if (is_zero(tv - tw)) {

	*aa = sign_heritage(v1-v0, a);
	*uu = sign_heritage(w1-w0, u);
    }
    else {
	*aa = -sign_heritage(v0, a);
	*uu = -sign_heritage(w0, u);
    }
}


void calc_ac_times(double *ta, double *tc, double *a, double *u,
	      double v0, double w0, double ta0, int type,
	      double v1, double w1)
/***************************************************************************
 * calculate the acceleration time ta and the time of constant drive tc
 * for the motion change from the old motion with the velocities v0,w0
 * the acc-time ta0 and the motion-type type, to the new velocities
 * v1,w1 with the accelerations a,u.
 *
 */
{
    double a0, u0, ta_max1, ta_max2;

    if (ta0<0.0) error("com_motv: calc_ac_times() - ta0 is zero or negative!");

    /* accelerations to reach from 0 the velocities v0,w0 in the time ta0.
     */
    a0 = v0 / ta0;
    u0 = w0 / ta0;

    calc_via_accs(a, u, v0, w0, v1, w1, a0, u0);

    if (has_same_sign(a0, *a) && has_same_sign(u0, *u)) {
	ta_max1 = ta0 * (sqrt(2.0)-1.0);
    }
    else {
	ta_max1 = ta0;
    }

    if (is_not_zero(*a) && is_zero(*u)) ta_max2 = (v1 - v0) / (*a);
    else if (is_not_zero(*u) && is_zero(*a)) ta_max2 = (w1 - w0) / (*u);
    else ta_max2 = ta_max1;

    if (ta_max2 >= 0.0 && ta_max2 < ta_max1 )
	*ta = ta_max2;
    else *ta = ta_max1;

    switch(type) {
    case MVT_DISTANCE:
	if (is_zero(v0)) *tc = 0.0;
	else *tc = (0.5*a0*ta0*ta0-v0*(*ta)-0.5*(*a)*(*ta)*(*ta))/v0;
	break;
    case MVT_ANGLE:
	if (is_zero(w0)) *tc = 0.0;
	else *tc = (0.5*u0*ta0*ta0-w0*(*ta)-0.5*(*u)*(*ta)*(*ta))/w0;
	break;
    case MVT_TIME:
	*tc = ta0 - *ta;
	break;
    case MVT_INFINITE:
	*tc = 1e9;
	break;
    default:
	error("com_motv: calc_ac_times() - unknown motion type\n");
	break;
    }

    if (is_zero(*ta)) *ta = 0.0;
    if (is_zero(*tc)) *tc = 0.0;

    if (is_negative(*tc))
	error("com_motv: calc_ac_times() - tc is negative!");
}


static void motv_change(scheduler *s, timepoint *tp, double t)
/***************************************************************************
 * Timepoint-function for an motion-changes.
 * The function get a motv-motion out of the motion-queue and invoke
 * a change between the old and the new motion, which is done in
 * 4 Phases:
 *		old-motion(acc)
 *		old-motion(const-drive)
 *   Phase 1:	old-motion fuhrter constant drive
 *   Phase 2:	old-motion acc/dec to valid changing velocities.
 *      - now the aim of the old motion is reached
 *   Phase 3:	old-motion acc/dec to valid changing velocities.
 *   Phase 4:	new-motion const-drive
 *	- now the aim of the acc-phase of the new-motion is reached.
 *		new-motion(const-drive)
 *		new-motion(decel)
 *
 * Please note, a isolated motion is done by:
 *   acceleration->contant-drive->deceleration
 */
{
    mob m = tp->workspace;
    motv_info *mi = &(m->motion.motv);
    motv_motion motion;
    mob_motion_state ms;
    double ta, tc, a, u, vx, wx;



    if (queue_first(&motion, mi->motions)) {

	/* a new motion was found in the queue,
	 * so finish the last motion and begin the
	 * new one.
	 */

	/* request current motion state
	 */
        mob_get_intern_state(m, &ms);

#ifdef DEBUG_DETAIL
	fprintf(stderr, "motion-change at v=%fm/s, w=%frad/s\n, t=%fs",
		ms.v, ms.w, t);
#endif
	/* PHASE 1&2
	 */
	calc_ac_times(&ta, &tc, &a, &u,
		      ms.v, ms.w, mi->last_motion.ta, mi->last_motion.type,
		      motion.v_max, motion.w_max);

#ifdef DEBUG_DETAIL
	fprintf(stderr, "phase1&2: a=%fm/s^2, u=%frad/s^2\n"
		"          ta=%fs, tc=%fs\n",
		a, u, ta, tc);
#endif

	if (is_not_zero(tc))
	    motion_add(m, tc, 0.0, 0.0);

	if (is_not_zero(ta))
	    motion_add(m, ta, a, u);

	t += ta + tc;

	vx = ms.v + ta * a;
	wx = ms.w + ta * u;

#ifdef DEBUG_DETAIL
	fprintf(stderr, "via velocities are: vx=%fm/s, wx=%frad/s\n",
		vx, wx);
#endif

	/* PHASE 3&4
	 */
	calc_ac_times(&ta, &tc, &a, &u,
		      motion.v_max, motion.w_max, motion.ta, motion.type,
		      vx, wx);


#ifdef DEBUG_DETAIL
	fprintf(stderr, "phase3&4: a=%fm/s^2, u=%frad/s^2\n"
		"          ta=%fs, tc=%fs\n",
		-a, -u, ta, tc);
#endif

	if (is_not_zero(ta))
	    motion_add(m, ta, -a, -u);

	if (is_not_zero(tc))
	    motion_add(m, tc, 0.0, 0.0);

	t += ta + tc;

	t += motion.tc;

#ifdef DEBUG_DETAIL
	fprintf(stderr, "next scheduler point at %fs\n", t);
#endif
        scheduler_point_add(s, tp, t);

	mi->last_motion = motion;
    }
    else {

	/* stop motion from current velocity in the time last_motion.ta
	 * to 0.
	 */

	mob_get_intern_state(m, &ms);

	if ((is_not_zero(ms.v) || is_not_zero(ms.w)) &&
	    is_not_zero(mi->last_motion.ta)) {

	    a = - ms.v / mi->last_motion.ta;
	    u = - ms.w / mi->last_motion.ta;

#ifdef DEBUG_DETAIL
	    fprintf(stderr, "decellaration with ta=%fs, a=%fm/s^2, u=%frad/s^2\n",
		    mi->last_motion.ta, a, u);
#endif
	    motion_add(m, mi->last_motion.ta, a, u);

#ifdef DEBUG_DETAIL
	    fprintf(stderr, "next scheduler point at %fs\n", t+mi->last_motion.ta);
#endif
	    scheduler_point_add(s, tp, t+mi->last_motion.ta);
	}
#ifdef DEBUG_DETAIL
	else fprintf(stderr, "motv_change? mob already stopped.\n");
#endif
    }
}


static void launch_motions(mob m, motv_motion *motion)
/***************************************************************************
 * launch a motion or invoke the execution of the motion-queue
 */
{
    mob_motion_state ms;
    motv_info *mi = &(m->motion.motv);
    double ta, tc, a, u, vx, wx, t;


    empty_motion_queue(m);

    mob_get_intern_state(m, &ms);

    motion->t_start = scheduler_time(m->scheduler);
    motion->x_start = ms.x;
    motion->y_start = ms.y;
    motion->angle_start = ms.angle;

    mi->last_motion = *motion;

    if (MT_STOP==motion_type(m)) {

	a = (motion->v_max >=0.0 ? motion->a : - motion->a);
	u = (motion->w_max >=0.0 ? motion->u : - motion->u);

#ifdef DEBUG_DETAIL
	fprintf(stderr, "Launch from rest with:\n");
	fprintf(stderr, "ta=%fs, tc=%fs\na=%fm/s^2, u=%frad/s^2\n",
		motion->ta, motion->tc, a, u);
	fprintf(stderr, "next tp in %fs at %fs\n",
		motion->ta+motion->tc,
		scheduler_time(m->scheduler)+motion->ta+motion->tc);
#endif

	if (is_not_zero(motion->ta))
	    motion_add(m, motion->ta, a, u);

	scheduler_point_unlink(m->scheduler, &(mi->change_tp));
	scheduler_point_add(m->scheduler, &(mi->change_tp),
			    scheduler_time(m->scheduler) +
			    motion->ta + motion->tc );

    }
    else {

	calc_via_velocities(&vx, &wx, ms.v, ms.w, motion->v_max, motion->w_max);

	if (is_not_zero(vx-ms.v) || is_not_zero(wx-ms.w)) {

	    calc_limit_acc(&t, &a, &u, ms.v, ms.w, vx, wx,
			   DEFAULT_LINEAR_ACC, DEFAULT_ANGULAR_ACC);

#ifdef DEBUG_DETAIL
	    fprintf(stderr, "decel to via velocities vx=%fm/s, wx=%frad/s\n",
		    vx, wx);
	    fprintf(stderr, "in %fs with a=%fm/s^2, u=%fm/s^2\n",
		    t, a, u);
#endif
		    
	    if (is_not_zero(t))
		motion_add(m, t, a, u);
	}
	else t = 0.0;


	calc_times_acc(&ta, &tc, &a, &u,
		       motion->v_max, motion->w_max, motion->ta, motion->type,
		       vx, wx, motion->a, motion->u);
	
	t += ta + tc + motion->tc;

#ifdef DEBUG_DETAIL
	fprintf(stderr, "accelerate to v=%fm/s, w=%frad/s in %fs\n",
		motion->v_max, motion->w_max, ta);
	fprintf(stderr, "with a=%fm/s^s, u=%frad/s^s\n", -a, -u);

	fprintf(stderr, "next tp in %fsec at %f\n", t,
				scheduler_time(m->scheduler) + t);
#endif
	t += scheduler_time(m->scheduler);


	if (is_not_zero(ta))
	    motion_add(m, ta, -a, -u);

	scheduler_point_unlink(m->scheduler, &(mi->change_tp));
	scheduler_point_add(m->scheduler, &(mi->change_tp), t);

    }

}



static char *motv_adla(long option, mob m, char *args)
/***************************************************************************
 */
{
    long par =0, l;
    motv_motion mot;
    char *err=0;

    /* default motion is straight line with 100mm/s.
     */
    mot.type = MVT_INFINITE;
    mot.v = DEFAULT_LINEAR_VEL;
    mot.a = DEFAULT_LINEAR_ACC;
    mot.w = 0.0;
    mot.u = LDEFAULT_ANGULAR_ACC;
    mot.distance = 0.0;
    mot.time = 0.0;
    mot.angle = 0.0;

    /* check for motion server */
    if (motion_server(m) != MS_MOTV) return ERR_SERVER_INACTIVE;


    par = get_word(&args, 4);
    if (option==OPT_LAUNCH && WORD_EMPTY==par) {
	if (queue_first(&mot, m->motion.motv.motions))
	    launch_motions(m, &mot);
	else return ERR_QUEUE_EMPTY;
        
	return 0;
    }


    while(par!=WORD_EMPTY) switch(par) {
    case VAL_VELOCITY:
    case VAL_POS_LIN_VEL:
    case VAL_NEG_LIN_VEL:
	/* linear velocity parameter
	 */
	err = get_value(&l, &args, &par,
			VAL_VELOCITY, VAL_POS_LIN_VEL, VAL_NEG_LIN_VEL,
			LMAX_LINEAR_VEL, 0, LDEFAULT_LINEAR_VEL,
			ERR_LIN_VEL_RANGE, ERR_INVALID_PARAMETER);
	if (err) return err;
	else {
	    mot.v = ITOD_DIV1000(l);
	}
	break;
	
    case VAL_ANGULAR_VEL:
    case VAL_POS_ANG_VEL:
    case VAL_NEG_ANG_VEL:
	/* angular velocity parameter
	 */
	err = get_value(&l, &args, &par,
			VAL_ANGULAR_VEL, VAL_POS_ANG_VEL, VAL_NEG_ANG_VEL,
			LMAX_ANGULAR_VEL, LMIN_ANGULAR_VEL, 0,
			ERR_ANG_VEL_RANGE, ERR_INVALID_PARAMETER);
	if (err) return err;
	else {
	    mot.w = ITOD_DIV1000(l);
	}
	break;

    case VAL_DISTANCE:
	if (get_long(&args, &l)) return ERR_INVALID_PARAMETER;
	if (not_in_range(0, l, LMAX_DISTANCE))
 	    return ERR_ILLEGAL_DISTANCE;
	mot.distance = ITOD_DIV1000(l);
	mot.type = MVT_DISTANCE;
	par = get_word(&args, 2);
	break;

    case VAL_ANGLE:
	if (get_long(&args, &l)) return ERR_INVALID_PARAMETER;
	if (not_in_range(0, l, LMAX_ANGLE))
	    return ERR_ILLEGAL_ANGLE;
	mot.angle = ITOD_GRAD_TO_RAD(l);
	mot.type = MVT_ANGLE;
	par = get_word(&args, 2);
	break;

    case VAL_TIME:
	if (get_long(&args, &l)) return ERR_INVALID_PARAMETER;
	if (not_in_range(0, l, LMAX_TIME))
	    return ERR_ILLEGAL_TIME;
	mot.time = (double)l * 0.040;        /* 1 unit == 40ms */
	mot.type = MVT_TIME;
	par = get_word(&args, 2);
	break;

    case OPT_INFINITE:
	mot.type = MVT_INFINITE;
	par = get_word(&args, 2);
	break;

    case VAL_LIN_ACC:
    case VAL_POS_LIN_ACC:
	/* linear acceleration option
	 */
	err = get_value(&l, &args, &par, 
			VAL_LIN_ACC, VAL_POS_LIN_ACC, 0,
			LMAX_LINEAR_ACC, LMIN_LINEAR_ACC, 
			(mot.v!=0.0) ? LDEFAULT_LINEAR_ACC : 0,
			ERR_ILLEGAL_LIN_ACC, ERR_INVALID_PARAMETER);
	if (err) return err;
	else {
	    mot.a = ITOD_DIV1000(l);
	}
	break;

    case VAL_ANG_ACC:
    case VAL_POS_ANG_ACC:
	/* angular acceleration option
	 */
	err = get_value(&l, &args, &par,
			VAL_ANG_ACC, VAL_POS_ANG_ACC, 0,
			LMAX_ANGULAR_ACC, LMIN_ANGULAR_ACC,
			(mot.w!=0.0) ? LDEFAULT_ANGULAR_ACC : 0,
			ERR_ILLEGAL_ANG_ACC, ERR_INVALID_PARAMETER);
	if (err) return err;
	else {
	    mot.u = ITOD_DIV1000(l);
	}
	break;

    default:
	return ERR_INVALID_PARAMETER;
    }
    
    if (mot.type == MVT_ANGLE && is_not_zero(mot.v))
	return ERR_ILLEGAL_VELOCITY;
    if (is_not_zero(mot.v) && fabs(mot.v/mot.w) <SMALLEST_RADIUS)
	return ERR_RADIUS_RANGE;

    if (is_zero(mot.v)) mot.a = 0.0;
    if (is_zero(mot.w)) mot.u = 0.0;
    acc_prop_to_vel(&mot);
    calc_times(&mot);

#ifdef DEBUG_DETAIL
    print_motion(&mot);
#endif

    if (option==OPT_LAUNCH) {
	launch_motions(m, &mot);
    } else {
	if (queue_size(m->motion.motv.motions) == MOTV_QUEUE_SIZE)
	    return ERR_QUEUE_FULL;

	queue_add(m->motion.motv.motions, &mot);
    }

    return err;
}




static char *motv_ms(mob m, char *args)
/***************************************************************************
 */
{
    motv_info *mi = &(m->motion.motv);
    int i;
    motv_motion mot;
    char motion_info_text[80];


    /* check for motion server */
    if (motion_server(m) != MS_MOTV) return ERR_SERVER_INACTIVE;

    if (WORD_EMPTY != get_word(&args, 4)) return ERR_INVALID_PARAMETER;

    for (i=queue_size(mi->motions); i>0; i--) {
        queue_first(&mot, mi->motions);
        queue_add(mi->motions, &mot);

	switch(mot.type) {
	case MVT_VELOCITY:
	    sprintf(motion_info_text, "v=%5ld  w=%5ld  ??????????  l=%5ld  k=%5ld",
		    DTOI_X1000(mot.v),
		    DTOI_X1000(mot.w),
		    DTOI_X1000(mot.a),
		    DTOI_X1000(mot.u));
	    mob_output(m, motion_info_text);
            break;
	case MVT_DISTANCE:
	    sprintf(motion_info_text, "v=%5ld  w=%5ld  d=%8ld  l=%5ld  k=%5ld",
		    DTOI_X1000(mot.v),
		    DTOI_X1000(mot.w),
		    DTOI_X1000(mot.distance),
		    DTOI_X1000(mot.a),
		    DTOI_X1000(mot.u));
	    mob_output(m, motion_info_text);
            break;
	case MVT_ANGLE:
	    sprintf(motion_info_text, "v=%5ld  w=%5ld  a=%8ld  l=%5ld  k=%5ld",
		    DTOI_X1000(mot.v),
		    DTOI_X1000(mot.w),
		    DTOI_RAD_TO_GRAD(mot.angle),
		    DTOI_X1000(mot.a),
		    DTOI_X1000(mot.u));
	    mob_output(m, motion_info_text);
            break;
	case MVT_TIME:
	    sprintf(motion_info_text, "v=%5ld  w=%5ld  t=%8ld  l=%5ld  k=%5ld",
		    DTOI_X1000(mot.v),
		    DTOI_X1000(mot.w),
		    DTOI_X1000(mot.time),
		    DTOI_X1000(mot.a),
		    DTOI_X1000(mot.u));
	    mob_output(m, motion_info_text);
            break;
	case MVT_INFINITE:
	    sprintf(motion_info_text, "v=%5ld  w=%5ld              l=%5ld  k=%5ld",
		    DTOI_X1000(mot.v),
		    DTOI_X1000(mot.w),
		    DTOI_X1000(mot.a),
		    DTOI_X1000(mot.u));
	    mob_output(m, motion_info_text);
            break;
	default:
            error("com_motv : motv_ms() - unknown motion type");
	    break;
	}
    }

    return 0;
}


static char *motv_ns(mob m, char *args)
/***************************************************************************
 */
{
    motv_info *mi = &(m->motion.motv);
    char motion_info_text[10];

    /* check for motion server */
    if (motion_server(m) != MS_MOTV) return ERR_SERVER_INACTIVE;

    if (WORD_EMPTY != get_word(&args, 4)) return ERR_INVALID_PARAMETER;

    sprintf(motion_info_text, "%d", queue_size(mi->motions));

    mob_output(m, motion_info_text);

    return 0;
}


static char *motv_kill(mob m, char *args)
/***************************************************************************
 */
{
    long par =0, l;
    char *err;
    double a, u, t;
    mob_motion_state ms;

    /* check for motion server */
    if (motion_server(m) != MS_MOTV) return ERR_SERVER_INACTIVE;

    /* linear acceleration option
     */
    err = get_value(&l, &args, &par, 
		    VAL_LIN_ACC, VAL_POS_LIN_ACC, 0,
		    LMAX_LINEAR_ACC, LMIN_LINEAR_ACC, 
		    LDEFAULT_LINEAR_ACC,
		    ERR_ILLEGAL_LIN_ACC, ERR_INVALID_PARAMETER);
    if (err) return err;
    else {
	a = ITOD_DIV1000(l);
    }

    /* angular acceleration option
     */
    err = get_value(&l, &args, &par,
		    VAL_ANG_ACC, VAL_POS_ANG_ACC, 0,
		    LMAX_ANGULAR_ACC, LMIN_ANGULAR_ACC,
		    LDEFAULT_ANGULAR_ACC,
		    ERR_ILLEGAL_ANG_ACC, ERR_INVALID_PARAMETER);
    if (err) return err;
    else {
	u = ITOD_DIV1000(l);
    }

    if (par != WORD_EMPTY) return ERR_INVALID_PARAMETER;

    mob_get_intern_state(m, &ms);

    calc_limit_acc(&t, &a, &u, ms.v, ms.w, 0.0, 0.0, a, u);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "decel to 0 in %fs with a=%fm/s^2, u=%fm/s^2\n",
	    t, a, u);
#endif
    
    empty_motion_queue(m);

    if (is_not_zero(t))
	motion_add(m, t, a, u);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "unlink tp\n");
#endif
    scheduler_point_unlink(m->scheduler, &(m->motion.motv.change_tp));

    return 0;
}

static char *motv_reset(mob m, char *args)
/***************************************************************************
 */
{
    long par;
    motv_motion mot;

    /* check for motion server */
    if (motion_server(m) != MS_MOTV) return ERR_SERVER_INACTIVE;

    par = get_word(&args, 4);
    if (par != WORD_EMPTY) return ERR_INVALID_PARAMETER;


    if (motion_type(m)==MT_STOP) {
	while (queue_first(&mot, m->motion.motv.motions))
	    ;
    }
    else return ERR_ROBOT_IN_MOTION;

    return 0;
}


char *motv(mob m, char *args)
/***************************************************************************
 */
{
    long option = get_word(&args, 2);

    /* Argument parsing
     */
    switch(option) {
    case OPT_ON:
	if ((motion_server(m) != MS_MOTV) && (motion_server(m) != MS_NONE))
	    return ERR_SERVER_ALLOCATED;

	if (get_word(&args, 2)!=WORD_EMPTY) return ERR_INVALID_PARAMETER;

	change_motion_server(m, MS_MOTV);
	m->motion.motv.change_tp.task = 0;
	m->motion.motv.change_tp.workspace = m;
	m->motion.motv.change_tp.free = 0;
	m->motion.motv.change_tp.function = motv_change;

	return 0;
	break;

    case OPT_OFF:
	if (motion_server(m) != MS_MOTV)        /* motion server active? */
	    return ERR_SERVER_INACTIVE;

	if (get_word(&args, 2)!=WORD_EMPTY) return ERR_INVALID_PARAMETER;

	/* check if robot is in motion */
        if (motion_type(m) != MT_STOP) return ERR_ROBOT_IN_MOTION;
	change_motion_server(m, MS_NONE);
        break;

    case OPT_ADD: case OPT_LAUNCH:
	return motv_adla(option, m, args);

	break;

    case OPT_CHANGE:
	break;

    case OPT_KILL:
	return motv_kill(m, args);
	break;

    case OPT_RESET:
	return motv_reset(m, args);
	break;

    case OPT_MOTIONPOS:
	break;

    case OPT_STATUS:
	break;

    case OPT_MOTIONQUEUE:
	return motv_ms(m, args);
	break;

    case OPT_QUEUESIZE:
	return motv_ns(m, args);
	break;

    default:
	return ERR_INVALID_PARAMETER;
    }

    return 0;
}


