/********************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: command.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:52:56 stolz>
 ********************************************************************************
 *  MODULE              : command
 *
 *  IMPLEMENTATION
 *      This module is the top-parser of the Robuter-Commands.
 *	It calls the specific Command-parser/implementation of
 *	the modules com_motv, com_motn, coms_us.
 *	
 *	The implementation of the odometry-Commands ODOM, ODOS
 *	and ODOF will be found in this module!
 *
 *  NOTE
 *      The module specification and definition is given in command.h!
 *	Also the function specification insist in command.h.
 *
 *  HISTORY
 *  13-Mar-94   Creation date
 *  20-Mar-94   Keyword test via a LONG-word
 *  20-Apr-94	Add bumper and odom commands
 *  10-Jun-94   Add essential us-sensor-commands
 *  12-Jul-94   Call output_open() and ..close() to identify the output made
 *		by an command.
 *
 ********************************************************************************/
#include <stdio.h>
#include <ctype.h>
#include <math.h>

#include "mob_private.h"
#include "command.h"
#include "com_scanner.h"
#include "com_motn.h"
#include "com_motv.h"
#include "coms_us.h"
#include "error.h"


static char *odom(mob m, char *opt)
/**********************************************************************
 */
{
    long option = get_word(&opt, 2);
    mob_motion_state ms;
    char odom_output[100];

    switch(option) {
    case WORD_EMPTY:
	if (m->odom_on) {
	    double sinx, cosx, x, y, angle;

            mob_get_intern_state(m, &ms);

	    sinx = sin(m->odom_offset_angle);
	    cosx = cos(m->odom_offset_angle);
	    x = (ms.x - m->odom_offset_x) * cosx +
                (ms.y - m->odom_offset_y) * sinx;
	    y = (ms.y - m->odom_offset_y) * cosx -
                (ms.x - m->odom_offset_x) * sinx;
	    angle = ms.angle - m->odom_offset_angle;

	    sprintf(odom_output, "%ld %ld %ld", DTOI_X1000(x), DTOI_X1000(y),
		    DTOI_RAD_TO_GRAD(angle));

	    mob_output(m, odom_output);
        }
	else return ERR_ODOM_INACTIVE;
	break;
    case OPT_ON:
	mob_get_intern_state(m, &ms);
	m->odom_offset_x = ms.x;
	m->odom_offset_y = ms.y;
	m->odom_offset_angle = ms.angle;
	m->odom_on = TRUE;
	option=get_word(&opt, 2);
	break;

    case OPT_OFF: m->odom_on = FALSE; option=get_word(&opt, 2); break;
    default:
	return ERR_INVALID_PARAMETER;
    }

    if (option==WORD_EMPTY) return 0;
    else return ERR_INVALID_PARAMETER;
}


static char *read_coord(char **opt, double *x, double *y, double *angle)
/**********************************************************************
 */
{
    long str = get_word(opt, 2);
    long lx, ly, langle;

    if (str!=VAL_COORDINATES ||
	get_long(opt, &lx) || get_comma(opt) ||
	get_long(opt, &ly) || get_comma(opt) ||
	get_long(opt, &langle) || WORD_EMPTY!=get_word(opt,2)) {

	return ERR_INVALID_PARAMETER;
    }
    else {
	*x = ITOD_DIV1000(lx);
	*y = ITOD_DIV1000(ly);
	*angle = ITOD_GRAD_TO_RAD(langle);

        return 0;
    }
}


static char *odos(mob m, char *opt)
/**********************************************************************
 */
{
    mob_motion_state ms;
    double x, y, angle, sinx, cosx;
    char *err;

    err=read_coord(&opt, &x, &y, &angle);
    if (err) return err;

    mob_get_intern_state(m, &ms);

    m->odom_offset_angle = ms.angle - angle;

    sinx = sin(m->odom_offset_angle);
    cosx = cos(m->odom_offset_angle);

    m->odom_offset_x = ms.x - (x*cosx-y*sinx);
    m->odom_offset_y = ms.y - (x*sinx+y*cosx);

    return 0;
}


static char *odof(mob m, char *opt)
/**********************************************************************
 */
{
    double sinx, cosx, x, y, angle;
    char *err;

    err=read_coord(&opt, &x, &y, &angle);
    if (err) return err;

    sinx = sin(m->odom_offset_angle);
    cosx = cos(m->odom_offset_angle);
    m->odom_offset_x += x * cosx - y * sinx;
    m->odom_offset_y += y * sinx + y * cosx;
    m->odom_offset_angle += angle;

    return 0;
}


void mob_do_command(mob m, double t, char *com)
/**********************************************************************
 */
{
    long name;
    char *err = 0;

    /* read command-keyword */
    name = get_word(&com, 4);

    mob_output_open(m);

    if (motion_type(m)==MT_HALT) {
	switch(name) {
	case COM_RESET: mob_reset(m); break;
	case COM_RESUME: mob_resume(m); break;
	case COM_HALT: break;
	default: err = ERR_HALT;
	}
    }
    else switch(name) {
    case COM_MOTN: err = motn(m, com); break;
    case COM_MOTV: err = motv(m, com); break;
    case COM_ODOM: err = odom(m, com); break;
    case COM_ODOS: err = odos(m, com); break;
    case COM_ODOF: err = odof(m, com); break;
    case COM_RESET: mob_reset(m); break;
    case COM_RNOD: err = rnod(m, com); break;
    case COM_SSET: err = sset(m, com); break;
    case COM_READ: err = us_read(m, com); break;

    case COM_MOTP:
    case COM_SNOD:
    case COM_SNEW:
    case COM_LCOM:
    case COM_PROMPT:
	mob_output(m, "HaHaHaHaHa - you're joking!");
	err = ERR_ILLEGAL_COMMAND;
	break;

    case 0x4655434B:
	mob_output(m, "Fuck yourself");
	err = ERR_ILLEGAL_COMMAND;
	break;

    case 0x53484954:
	mob_output(m, "Bullshit");
	err = ERR_ILLEGAL_COMMAND;
	break;

    case 0x50495353:
	mob_output(m, "Do you need a doctor?");
	err = ERR_ILLEGAL_COMMAND;
	break;

    case COM_BUMP:
	switch(name = get_word(&com, 2)) {
	case OPT_ON: m->bumper_on = TRUE; name = get_word(&com, 2); break;
	case OPT_OFF: m->bumper_on = FALSE; name = get_word(&com, 2); break;
	default: break;
        }
	if (name!=WORD_EMPTY) err = ERR_INVALID_PARAMETER;
	break;

    case COM_RESUME: break;
    case COM_HALT:   mob_halt(m); break;

    case WORD_EMPTY:
	break;

    default:
	err = ERR_ILLEGAL_COMMAND;
	break;
    }
    if (err) mob_output(m, err);
    mob_output_close(m); 
}
