/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: coms_us.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:53:17 stolz>
 **************************************************************************
 *  MODULE              : coms_us
 *
 *  IMPLEMENTATION
 *	This implementation of the READ, RNOD and SSET commands which
 *	are necessary for the use of the ultrasonic sensors.
 *
 *
 *  NOTE
 *      The module specification and definition is given in coms_us.h!
 *      The real US-Sensor reading is done at scheduler-timepoint in the
 *	routine mob_update_us_sensors() of the module mob.
 *
 *  HISTORY
 *  16-Jun-94   Creation date.
 *  16-Jun-94   Implement sset, rnod, read.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "mob_private.h"
#include "com_scanner.h"
#include "coms_us.h"
#include "error.h"

#define	US_TIMER_FREQ	(420000.0)      	/* 420kHz */



static void output_us_results(mob m);


char *rnod(mob m, char *args)
/**************************************************************************
 */
{
    long option = get_word(&args, 2);
    int i;

#ifdef DEBUG_DETAIL
    fprintf(stderr, "RNODE %s\n", args);
#endif

    switch (option) {
    case VAL_CONFIG: {
	unsigned long field;
	int len;
	int node;
	int sensor, sensor_nr, i;
	char new_scan_mode;


	if (m->us_mode == SM_REQUEST) return ERR_REQ_MODE_ACTIVE;
	len = get_bitfield2(&field, &args);
	if (len<=0 || len>=9) return  ERR_IN_CONFIGURATION;
	if (get_word(&args,2)!=WORD_EMPTY) return ERR_INVALID_PARAMETER;
	    
#ifdef DEBUG_DETAIL
	fprintf(stderr, "C=%lu\n", field);
#endif
	
	for (i=0; i<MAX_NR_US_SENSORS; i++)
	    m->us_sensor[i].scan_mode = USSM_INACTIVE;

	node = 0;
	while (len>0) {
	    switch(field&3) {
            case 0: new_scan_mode = USSM_LOW; break;
            case 3: new_scan_mode = USSM_INACTIVE; break;
            default: new_scan_mode = USSM_HIGH; break;
	    }

	    /* for all adresses in this node
	     * set the scan mode for the installed sensor to new_scan_mode.
             */
	    sensor = node_adr(node, 0);
	    for (i=0; i<MAX_NR_SENSORS_PER_NODE; i++, sensor++) {
		sensor_nr = m->sensor_nr[sensor];
		if (sensor_nr >= 0)
		    m->us_sensor[sensor_nr].scan_mode = new_scan_mode;
            }
	    
	    len--;
            node++;
	    field>>=2;
	}
	
	/* determine number of active sensors
	 */
	sensor = 0;
	for (i=0; i<MAX_NR_US_SENSORS; i++)
	    if (m->us_sensor[i].scan_mode != USSM_INACTIVE) sensor++;

#ifdef DEBUG_DETAIL
	fprintf(stderr, "nr of active sensors: %d\n", sensor);
#endif

	if (sensor) {
	    m->us_mode = SM_AUTOMATIC;
	    m->invoke_us_read = TRUE;
	}
	else {
	    m->us_mode = SM_VOID;
	    m->invoke_us_read = FALSE;
	    m->read_node = 0;
	}

	} break;

    case VAL_R:
	if (m->us_mode == SM_AUTOMATIC) return ERR_AUTO_MODE_ACTIVE;
	switch (get_word(&args, 2)) {
        case OPT_ON:
	    if (get_word(&args,2)!=WORD_EMPTY) return ERR_INVALID_PARAMETER;
	    m->us_mode = SM_REQUEST;
	    m->invoke_us_read = FALSE;
	    for (i=0; i<MAX_NR_US_SENSORS; i++)
		m->us_sensor[i].scan_mode = USSM_INACTIVE;
            break;

	case OPT_OFF:
	    if (get_word(&args,2)!=WORD_EMPTY) return ERR_INVALID_PARAMETER;
	    m->us_mode = SM_VOID;
	    m->invoke_us_read = FALSE;
	    for (i=0; i<MAX_NR_US_SENSORS; i++)
		m->us_sensor[i].scan_mode = USSM_INACTIVE;
	    break;

	default:
            return ERR_INVALID_PARAMETER;
	    break;
	}
	break;

    default:
	return ERR_INVALID_PARAMETER;
	break;
    }
    return 0;
}



char *sset(mob m, char *args)
/**************************************************************************
 */
{
    long option = get_word(&args, 2);
    long sensor_nr = 0,
         angle = -1,
         direction = -1,
         value;
    int nadr = -1, node, adr;


#ifdef DEBUG_DETAIL
    fprintf(stderr, "SSET %s\n", args);
#endif

    /* parse S=
     */
    if (option != VAL_S) return ERR_INVALID_PARAMETER;
    if (get_long(&args, &sensor_nr)) return ERR_INVALID_DESCRIPTOR;
    sensor_nr--;
    if (sensor_nr<0 || sensor_nr >=MAX_NR_US_SENSORS) return ERR_INVALID_DESCRIPTOR;


    option = get_word(&args, 2);

    /* parse N=
     */
    if (option == VAL_N) {

	if (get_long(&args, &value) || value<0) return ERR_INVALID_SENSOR_NR;

        if (value) {
	    node = value / 10 - 1;
	    adr = value % 10 - 1;

            if (node<0 || node >= MAX_NR_NODES ||
		adr<0  || adr  >= MAX_NR_SENSORS_PER_NODE)
		return ERR_INVALID_SENSOR_NR;

	    nadr = node_adr(node, adr);
        }
	else nadr = -1;

	option = get_word(&args, 2);
    }

    /* parse A=
     */
    if (option == VAL_ANGLE) {

	if (get_long(&args, &angle) || angle <=0 ) return ERR_ILLEGAL_ANGLE;

	option = get_word(&args, 2);
    }

    /* parse D=
     */
    if (option == VAL_D) {

	if (get_long(&args, &direction) || (direction&(~15)) ||
	    ((direction&(USD_FRONT|USD_BACK))==(USD_FRONT|USD_BACK)) ||
	    ((direction&(USD_LEFT|USD_RIGHT))==(USD_LEFT|USD_RIGHT))) {

	    return ERR_IN_SCAN_DIR;
        }

	option = get_word(&args, 2);
    }

    if (option != WORD_EMPTY) return ERR_INVALID_PARAMETER;


#ifdef DEBUG_DETAIL
    fprintf(stderr, "sensor_nr=%ld, node+adr=%d\n", sensor_nr, nadr);
    fprintf(stderr, "set direction=%ld, angle=%ld\n", direction, angle);
#endif

    if (direction>=0) m->us_sensor[sensor_nr].direction = direction;
    if (angle>=0) m->us_sensor[sensor_nr].emission_angle = angle;
    if (nadr<0) m->us_sensor[sensor_nr].scan_mode = USSM_INACTIVE;
    else m->sensor_nr[nadr] = sensor_nr;

    m->sensor_node_adr[sensor_nr] = nadr;

    return 0;
}


char *us_read(mob m, char *args)
/**************************************************************************
 */
{
    long node;
    int old_node;
    int len;
    unsigned long mask, dummy;

    switch(get_word(&args, 2)) {
    case OPT_S:
	/* request results
	 */
	if (get_word(&args, 2) != WORD_EMPTY) return ERR_INVALID_PARAMETER;

	if (m->us_mode != SM_REQUEST) return ERR_REQ_MODE_INACTIVE;
	if (!(m->invoke_us_read)) return ERR_NO_READ_REQUEST;
	if (!(m->us_read_complete)) return ERR_READ_NOT_FINISHED;

	output_us_results(m);
/*   m->invoke_us_read = FALSE; */
	break;

    case VAL_CONFIG:
	/* parse node-number
	 */
	if (get_long(&args, &node) || (--node)<0 || node>=MAX_NR_NODES)
	    return ERR_ILLEGAL_NODE;

	if (get_comma(&args)) return ERR_IN_CONFIGURATION;

	/* parse sensor-selection
	 */
        len = get_bitfield(&mask, &args);
        if (len <=0 || len > MAX_NR_SENSORS_PER_NODE) return ERR_IN_CONFIGURATION;

	if (get_comma(&args)) return ERR_IN_CONFIGURATION;

        len = get_bitfield(&dummy, &args);
        if (len <=0 || len > MAX_NR_SENSORS_PER_NODE) return ERR_IN_CONFIGURATION;

        if (get_word(&args, 2)!=WORD_EMPTY) return ERR_INVALID_PARAMETER;

	old_node = m->read_node;
	m->invoke_us_read = TRUE;
	m->read_node = node;
	m->sensor_selection = mask;

	if (m->us_mode == SM_AUTOMATIC ) {
	    output_us_results(m);
	    m->read_node = old_node;
	}
	else if (m->us_mode != SM_REQUEST) return ERR_REQ_MODE_INACTIVE;
	else {
	    int i, j;


	    output_us_results(m);

	    for (i=0; i<MAX_NR_US_SENSORS; i++)
		m->us_sensor[i].scan_mode = USSM_INACTIVE;

	    for (i=0; i<MAX_NR_SENSORS_PER_NODE; i++) {
		j = m->sensor_nr[node_adr(m->read_node, i)];
		
		if (j>=0 && (mask&1)) {
		    m->us_sensor[j].scan_mode = USSM_LOW;
		} 
		mask >>= 1;
	    }
	}
	break;

    default:
	return ERR_INVALID_PARAMETER;
	break;
    }
    return 0;
}



static void output_us_results(mob m)
/**************************************************************************
 * output the results of the sensors from the m->read_node to the mob.
 */
{
    int i, sensor_nr;
    unsigned long sensor_mask = m->sensor_selection;
    static char result_string[MAX_NR_SENSORS_PER_NODE*40];
    char *result = result_string;
    long tof_result;

    if (sensor_mask == 0) return;

    result_string[0] = '\0';

    for (i=0; i<MAX_NR_SENSORS_PER_NODE; i++, sensor_mask>>=1) {
	if (sensor_mask&1) {
	    sensor_nr = m->sensor_nr[node_adr(m->read_node, i)];

	    if (m->us_read_complete&& sensor_nr >=0 &&
                sensor_nr < MAX_NR_US_SENSORS &&
		m->us_sensor[sensor_nr].scan_mode != USSM_INACTIVE) {

		tof_result = dround(m->us_sensor[sensor_nr].tof * US_TIMER_FREQ);
                tof_result = max(tof_result, 0);
		result[0] = '\0';
		(void)sprintf(result, " %ld", tof_result);
#ifdef DEBUG_DETAIL
		fprintf(stderr, "sensor %d, node %d, nr %d -> %ld\n",
			sensor_nr+1, m->read_node+1, i+1, tof_result);
#endif
		result += strlen(result);
	    }
            else {
		result[0] = ' ';
		result[1] = '0';
		result[2] = '\0';
		result += 2;
	    }
	} else {
	    result[0] = ' ';
	    result[1] = '0';
	    result[2] = '\0';
	    result += 2;
	}
    }

    if (result>&result_string[MAX_NR_SENSORS_PER_NODE*40])
	error("coms_us: output_us_results() - result string exceeded");

    mob_output(m, result_string);
}
