/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: darray.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:53:38 stolz>
 **************************************************************************
 *  MODULE              : darray
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in darray.h!
 *
 *  HISTORY
 *  29-Aug-94   Creation date
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "darray.h"
#include "error.h"


typedef struct adt_darray adt_darray;
typedef struct adt_darray {
    int size;
    int max_size;
    int elem_size;
    char *elements;
};


darray darray_new(int elem_size)
/*************************************************************************
 */
{
    darray da = malloc(sizeof(adt_darray));

    if (da) {
	da->size = 0;
	da->max_size = 0;
	da->elem_size = elem_size;
	da->elements = 0;
    }
    else error("darray: darray_new() - can't alloc memory for adt darray");

    return da;
}


void darray_free(darray da)
/*************************************************************************
 */
{
    if (da) {
	if (da->elements) free(da->elements);
	free(da);
    }
}


void darray_clear(darray da)
/*************************************************************************
 */
{
    da->size = 0;
}


int darray_size(darray da)
/*************************************************************************
 */
{
    return da->size;
}


void darray_resize(darray da, int index)
/*************************************************************************
 */
{
    if (index >= da->max_size) {

	da->max_size = index + 32 + da->max_size >> 2;

	if (da->elements)
	    da->elements = (char *)realloc(da->elements, da->elem_size * da->max_size);
	else
	    da->elements = (char *)malloc(da->elem_size * da->max_size);

	if (da->elements==0)
	    error("darray: darray_resize() - can't alloc memory for elements");
    }
}

void darray_set(darray da, int index, darray_elem elem)
/*************************************************************************
 */
{
    if (index>=0) {
        if (index >= da->max_size) darray_resize(da, index);

	memcpy(&(da->elements[da->elem_size * index]),
               elem, da->elem_size);
    }
    else error("darray: darray_set() - index < 0");
}


void darray_get(darray_elem elem, darray da, int index)
/*************************************************************************
 */
{
    if (index>=0) {

        if (index >= da->size)
	    error("darray: darray_get() - access on a unitialized element");

	memcpy(elem, (void *)(da->elements + da->elem_size * index), da->elem_size);

    }
    else error("darray: darray_get() - index < 0");
}


void darray_copy(darray da, int target, int source)
/*************************************************************************
 */
{
    if (source < 0 || source >=da->size)
	error("darray: darray_copy() - source-index out of range");

    if (target<0)
	error("darray: darray_copy() - target-index < 0");

    if (target >= da->max_size) darray_resize(da, target);

    memcpy(&(da->elements[da->elem_size * target]),
	   &(da->elements[da->elem_size * source]),
           da->elem_size);
}


void darray_swap(darray da, int a, int b)
/*************************************************************************
 */
{
    char *source, *target, *target_end, c;

    if (a < 0 || a >=da->size)
	error("darray: darray_swap() - 1.index out of range");

    if (b < 0 || b >=da->size)
	error("darray: darray_copy() - 2.index out of range");

    source = &(da->elements[da->elem_size * b]);
    target = &(da->elements[da->elem_size * a]);
    target_end = target + da->elem_size;

    for (; target<target_end; target++, source++) {
	c = *target;
	*target = *source;
	*source = c;
    }
}


void darray_exchange(darray_elem elem, darray da, int index)
/*************************************************************************
 */
{
    char *source, *source_end, c;
    char *target = (char *)elem;

    if (index < 0 || index >= da->size)
	error("darray: darray_copy() - index out of range");

    source = &(da->elements[da->elem_size * index]);
    source_end = source + da->elem_size;

    for (; source<source_end; target++, source++) {
	c = *target;
	*target = *source;
	*source = c;
    }
}


darray_elem darray_ref(darray da, int index)
/*************************************************************************
 */
{
    return (darray_elem)(&(da->elements[da->elem_size * index]));
}


void darray_for_all(darray da, void (*f)(darray_elem))
/*************************************************************************
 */
{
    int i;

    if (f)
	for (i=0; i<da->size; i++)
	    f(&(da->elements[da->elem_size * i]));
}


void darray_print(darray da)
/*************************************************************************
 */
{
    if (da) {
	printf("DARRAY {\nelem_size = %d\nsize = %ld\nmax_size = %ld\n}\n",
	       da->elem_size, da->size, da->max_size);
    }
}
