/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: darray.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 16:53:44 stolz>
 **************************************************************************
 *  MODULE              : darray
 *
 *  FUNCTION
 *	A abstract dynamic array for constant-size elements
 *	is provided of this module.
 *
 *	The dynamic array is created by a call of darray_new() with
 *	the element-size as option. Now elements can accessed in
 *      index-range [0, MAX_INT].
 *
 *	The elements can be set by darray_set(). Then the element is
 *	copied into the darray. The contents of the array can be
 *	requested by value with darray_get() or by reference with
 *	darray_ref(). darray_swap() swaps two elements in the array.
 *	darray_exchange() exchange an array-element with the element
 *      of the parameter.
 *
 *	If the accessed index is greater than the actual array, an
 *	automatic resize is performed. The actual size can requested
 *	and set by darray_size() and darray_resize(). Note that an
 *	resize or access of on element which was not first set
 *	by darray_set() do to an random unitialized element.
 *
 *	To clear the array and set the size to zero use darray_clear().
 *
 *	For further convience with darray_for_all(), a function
 *	will be called for every element of the array.
 *
 *	Finaly free the abstract data type with darray_free().
 *
 *
 *  EXAMPLE
 *
 *      struct adr {
 *              char name[80];
 *		int age;
 *	};
 *
 *      print_adr(void *v)
 *	{
 *	struct *adr = v;
 *	printf("Name=%s, age=%d\n", adr->name, adr->age);
 *	}
 *
 *	main
 *	{
 *	struct adr p1 = {"Thomas", 16"};
 *	struct adr p2 = {"Judith", 26"};
 *
 *	darray persons = darray_new(sizeof(struct adr));
 *      darray_set(persons, 0, &p1);    -- [0] = Thomas
 *      darray_set(persons, 1, &p2);    -- [1] = Judith
 *      darray_get(&p2, persons, 1);    -- p2 <= Judith
 *	darray_copy(persons, 1, 3);	-- element 2 is uninitialized
 *                                      -- now also [3] = Judith
 *	darray_swap(persons, 0, 3);     -- result: [0]=Judith [3]=Thomas
 *	darray_exchange(&p1, persons, 1);  -- result:p1=Judith,[1]=Thomas
 *
 *	int size=darray_size(persons);   --  should be 4
 *	darray_resize(persons, 2);	-- set size to 2
 *
 *
 *	darray_for_all(persons, print_adr);
 *	-- Result-output:
 *	-- Judith, 26
 *	-- Thomas, 16
 *
 *	darray_free(persons);
 *	}
 *
 *
 *  NOTE
 *
 *  HISTORY
 *  29-Aug-94   Creation date.
 *
 **************************************************************************/
#ifndef DARRAY_H
#define DARRAY_H

/* abstract type definition
 */
typedef struct adt_darray*	darray;
typedef void*			darray_elem;

/* create, clear or free a d-array.
 */
darray		darray_new(int elem_size);
void		darray_clear(darray);
void		darray_free(darray);

/* request size and resize array
 */
int		darray_size(darray);
void		darray_resize(darray, int max_size);

/* access an element or perform changes
 */
void		darray_set(darray da, int index, darray_elem);
void		darray_get(darray_elem, darray, int index);
void		darray_copy(darray, int target, int source);
void		darray_swap(darray, int a, int b);
void		darray_exchange(darray_elem, darray , int index);

/* request the pointer to the element
 */
darray_elem	darray_ref(darray, int index);

/* perform a function for every element
 */
void		darray_for_all(darray, void (*f)(darray_elem));

/* print the status of a darray on stdout.
 * Mainly for debugging purposes.
 */
void		darray_print(darray);

#endif
