/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: integration.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:54:13 stolz>
 **************************************************************************
 *  MODULE              : integration
 *  IMPORT              : error
 *
 *  IMPLEMENTATION
 *
 *	The integrator implements a Runge-Kutta with 4.order.
 *	The temporare variables which are necessary are allocated
 *	dynamically if the nr_eq is greater than the max_nr_eq.
 *
 *  NOTE
 *      The module specification and definition is given in integration.h!
 *
 *  HISTORY
 *  29-May-94   Creation date. Module test passed.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#include "integration.h"
#include "error.h"


void integrator(double *y_end,
		double *y_start,
		int	nr_eq,
		double t_start, double t_end, double dt,
		void (*f)(double *dy, double *y_t, double t))
/**************************************************************************
 * FUNCTION
 *	integrates a equation-system wit "nr_eq" equations from time t_start
 *	to the time t_end with the stepsize dt.
 *      The parameter y_start should hold a array with the initial values
 *      from the time t_start!
 *
 * RESULT
 *      The procedure copy the result y(t_end) to the array y_end.
 *	It's possible to hold the start and end-values in the same array,
 *	so y_end=y_start is no problem.
 *
 */
{
    static int max_nr_eq = 0;
    static double *k0, *k1, *k2, *k3, *y_temp;
    double temp;
    int i;

    if (nr_eq <= 0) error("integration: integrator() - nr of equations <= 0!");

    if (f==0) error("integration: integrator() - nr of equations <= 0!");

    /* allocate memory for arrays k0, k1, k2 and k3 and y_temp.
     */
    if (nr_eq > max_nr_eq) {
	if (max_nr_eq != 0) {
	    free(k0);
	    free(k1);
	    free(k2);
	    free(k3);
	    free(y_temp);
	}
	max_nr_eq = nr_eq;

	k0 = malloc(sizeof(double)*max_nr_eq);
	k1 = malloc(sizeof(double)*max_nr_eq);
	k2 = malloc(sizeof(double)*max_nr_eq);
	k3 = malloc(sizeof(double)*max_nr_eq);
	y_temp = malloc(sizeof(double)*max_nr_eq);

	if (k0==0 || k1==0 || k2==0 || k3==0 || y_temp==0)
            error("integration: integrator() - can't alloc memory for temp-vars");
    }


    /* loop until end-time reached
     */
    while (t_start < t_end) {

	/* variate dt to reach t_end exact.
	 */
	if (fabs(t_start + dt - t_end) < 1.1 * dt) {
	    dt = t_end - t_start;
	    temp = t_end;
	}
	else {
	    temp = t_start + dt;
	}

	/* 4.order Runge-Kutta,
         * integrate y(t) to y(t+h)! Note that dy/dt = f(t,y)
         *
         *  k0 = h * f(t, y)
         *  k1 = h * f(t + h/2, y + k0/2)
         *  k1 = h * f(t + h/2, y + k1/2)
         *  k2 = h * f(t + h, y + k2)
         *  y(t+h) =  y + 1/6*(k0+2*k1+2*k2+k3)
         */

	/* compute k0
	 */
	f(k0, y_start, t_start);

	/* compute k1
	 */
        for (i=0; i<nr_eq; i++) y_temp[i] = y_start[i] + 0.5*dt*k0[i];
	f(k1, y_temp, t_start + 0.5 * dt);

	/* compute k2
	 */
        for (i=0; i<nr_eq; i++) y_temp[i] = y_start[i] + 0.5*dt*k1[i];
	f(k2, y_temp, t_start + 0.5 * dt);

	/* compute k3
	 */
        for (i=0; i<nr_eq; i++) y_temp[i] = y_start[i] + dt*k2[i];
	f(k3, y_temp, temp);

	/* compute y_end for t+dt out of y_start,
	 */
        for (i=0; i<nr_eq; i++)
	    y_end[i] = y_start[i] + dt * (k0[i]/6.0 + k1[i]/3.0 + k2[i]/3.0 + k3[i]/6.0);

	y_start = y_end;

#ifdef DEBUG_DETAIL
	fprintf(stderr, "y(%f) =", temp);
        for (i=0; i<nr_eq; i++)
	    fprintf(stderr, "%f ", y_end[i]);
	fprintf(stderr, "\n");
#endif
	t_start = temp;
    }
}



/****************************************************************************
 * MODULE-TEST-routines
 */

#ifdef MODULE_TEST


static void f(double *dy, double *y, double t)
{
    dy[0] = -2.0 * t * y[0] * y[0];
    dy[1] = sin(t);
    dy[2] = 1;
}


int integration_test()
{
    int failed, errors = 0;
    char index;
    double y[3] = {1.0, 0.0, 0.0};
    double y1[3];

    y1[0] = 0.5;
    y1[1] = cos(0.0) - cos(1.0);
    y1[2] = 1.0;


    fprintf(stderr, "INTEGRATION-TEST from %s %s\n",
	    __DATE__, __TIME__);

    index = 'a';
    fprintf(stderr, "I. Integrate dx(t)=-2*t*x(t)*x(t), dy(t)=sin(t) and dz(t)=1\n"
		    "   from t=0 to t=1 with the stepsize 0.1.\n");

    fprintf(stderr,"Inititial Values x(0)=%f, y(0)=%f, z(0)=%f\n", y[0], y[1], y[2]);

    integrator(y, y, 3, 0.0, 1.0, 0.1, f);

    fprintf(stderr,"%c) Integration result: x(1)=%f, y(1)=%f, z(1)=%f\n",
		index++, y[0], y[1], y[2]);

    fprintf(stderr,"   correct result should be: x(1)=%f, y(1)=%f, z(1)=%f\n",
	    y1[0], y1[1], y1[2]);

    fprintf(stderr,"   differences: Dx(1)=%f, Dy(1)=%f, Dz(1)=%f\n",
	    y1[0]-y[0], y1[1]-y[1], y1[2]-y[2]);


    failed = 0;
    if (fabs(y[0]-y1[0])>0.0001) {failed = 1; errors++;}
    if (fabs(y[1]-y1[1])>0.0001) {failed = 1; errors++;}
    if (fabs(y[2]-y1[2])>0.0001) {failed = 1; errors++;}
    fprintf(stderr, "   integration test : %s\n",
			(failed ? "FAILED" : "PASSED"));

    fprintf(stderr, "\nTotal amount of errors: %d. Test %s!\n\n",
			errors, errors ? "FAILED" : "PASSED");


    return errors;
}


int main(int argc, char **argv)
{
    return integration_test();
}

#endif
