/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: integration.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 16:54:18 stolz>
 **************************************************************************
 *  MODULE              : integration
 *
 *  FUNCTION
 *
 *	Provides a Integration-Routine to solve ordinary differential
 *	equations or other problems which can be solved by integration.
 *
 *  HISTORY
 *  29-May-94   Remove tricky code in vehicle. Module Test ok!
 *
 **************************************************************************/
#ifndef INTEGRATION_H
#define INTEGRATION_H


void integrator(double *y_end,
		double *y_start,
		int	nr_eq,
		double t_start, double t_end, double dt,
		void (*f)(double *dy, double *y_t, double t) );
/**************************************************************************
 * FUNCTION
 *	Integrates a equation-system y with nr_eq variables y[0]..y[nr_eq-1]
 *	from time t_start to the time t_end with the stepsize dt. The
 *	meter y_start should hold a array with the initial values for the
 *	time t_start! The user must provide a function f, which computes
 *	the derivation dy/dt out of y(t) and t. It's also necessary that
 *	the function f fills all elements of dy[0]..dy[nr_eq-1]!
 *
 * RESULT
 *      The procedure copy the result y(t_end) to the array y_end.
 *	It's possible to hold the start and end-values in the same array,
 *	so y_end=y_start is no problem.
 *
 * EXAMPLE
 *
 *	To solve the equation system:
 *		dy0/dt = -2 * t * y0(t)		y0(0) = 1
 *		dy1/dt = dy0/dy + y1(t)		y1(0) = 0
 *	you could formulate it as the programm:
 *
 *	#include "integration.h"
 *
 *	void f(double *dy, double *y, double t)
 *	{
 *	    dy[0] = - 2 * t * y[0] * y[0];
 *	    dy[1] = dy[0] + y[1];
 *	}
 *
 *      void main()
 *	{
 *	    double y[2] = {1.0, 0.0};
 *
 *	    // integrate from t=0 to t=5.0 with stepsize 0.1
 *	    integrator(y, y, 2, 0.0, 5.0, 0.1, f);
 *
 *          printf("Result y0(5)=%f, y1(5)=%f\n", y[0], y[1]);
 *	}
 *
 */


#endif
