/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: ivd.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/03/02 16:54:51 stolz>
 **************************************************************************
 *  MODULE              : ivd
 *
 *  IMPLEMENTATION
 *	The communication between the iv-display-server
 *	and this module is done via PVM.
 *	Look into ivdp_obj.h for the protocoll.
 *
 *
 *  NOTE
 *      The module specification and definition is given in ivd!
 *
 *  HISTORY
 *  23-Mar-94   Creation date
 *  31-Jan-95   mi_open try to use the parent_tid as simulator-tid
 *  13-Jan-95   ivd_new starts only the ivd-server with an given servername
 *  02-Mar-95   a closed ivd-server don't make a deadlock-situation
 **************************************************************************/

#include <stdio.h>

#include "ivd.h"
#include "ivdp_obj.h"
#include "error.h"

#include "pvm3.h"

#include "mob_world_private.h"

/*
#define DEBUG_DETAIL 1
*/

static int ivd_server_tid = -1;
static int ivd_tid = -1;
static unsigned int obj_changes = OC_ALL;



void ivd_object_new(mw_object obj)
/*************************************************************************
 */
{
#ifndef NO_PVM
    int id;
    int info;

#if 0
    printf("ivd_object_new()\n");
#endif

    if (ivd_server_tid<0) return;

    /* request unique id from server process
     */
    pvm_initsend(PvmDataDefault);
    info = pvm_send(ivd_server_tid, IVDP_REQ_OBJ_ID);
    if (info<0) {
	ivd_server_tid = -1;
	return;
    }

    /* receive id from server.
     */

    pvm_recv(ivd_server_tid, IVDP_OBJ_ID);
    pvm_upkint(&id, 1, 1);

#if 0
    printf("obj-id = %d\n", id);
#endif
    mw_object_set_id(obj, id);

    ivd_object_change(obj);
#endif
}


void ivd_object_free(mw_object obj)
/*************************************************************************
 */
{
#ifndef NO_PVM
    int id;
    int info;

#if 0
    printf("ivd_object_free()\n");
#endif

    if (ivd_server_tid<0) return;

    id = mw_object_id(obj);

    /* send free-packet to server
     */
    pvm_initsend(PvmDataDefault);
    pvm_pkint(&id, 1, 1);
    info = pvm_send(ivd_server_tid, IVDP_OBJ_FREE);
    if (info<0) {
	ivd_server_tid = -1;
	return;
    }

#endif
}

void ivd_object_change(mw_object obj)
/*************************************************************************
 */
{
#ifndef NO_PVM
    int id, info, parent, type;
    mw_object head;
    vector3    v;
    vector3    v3[3];
    quaternion q;
    float f[9];
    char *app3d;

#if 0
    printf("ivd_object_change()\n");
#endif

    if (ivd_server_tid<0) return;

    /* send object definition to server
     */
    pvm_initsend(PvmDataDefault);

    id = mw_object_id(obj);
    pvm_pkint(&id, 1, 1);

    pvm_pkuint(&obj_changes, 1, 1);

    head = mw_object_head(obj);
    if (head) parent = mw_object_id(head);
    else parent = -1;
    pvm_pkint(&parent, 1, 1);

    type = mw_object_type(obj);
    pvm_pkint(&type, 1, 1);

    mw_object_position(&v, &q, obj);
    f[0] = (float)v[0];
    f[1] = (float)v[1];
    f[2] = (float)v[2];
    f[3] = (float)q[0];
    f[4] = (float)q[1];
    f[5] = (float)q[2];
    f[6] = (float)q[3];
    pvm_pkfloat(&f[0], 7, 1);

    app3d = (char *)mw_object_get_3d(obj);
    if (app3d)
	pvm_pkstr(app3d);
    else
	pvm_pkstr("");


    if (type != MWO_GROUP) {

	mw_object_get_color(&f[0], &f[1], &f[2], obj);
	pvm_pkfloat(&f[0], 3, 1);

	switch(type) {
        case MWO_CUBOID:
	    mwo_cuboid_get(&v, obj);
            f[0] = (float)v[0];
            f[1] = (float)v[1];
            f[2] = (float)v[2];
	    pvm_pkfloat(&f[0], 3, 1);
	    break;

	case MWO_RECTANGLE:
	    mwo_rectangle_get(&v[0], &v[1], obj);
            f[0] = (float)v[0];
            f[1] = (float)v[1];
	    pvm_pkfloat(&f[0], 2, 1);
	    break;

	case MWO_TRIANGLE:
	    mwo_triangle_get(v3, obj);
            f[0] = (float)v3[0][0];
            f[1] = (float)v3[0][1];
            f[2] = (float)v3[0][2];
            f[3] = (float)v3[1][0];
            f[4] = (float)v3[1][1];
            f[5] = (float)v3[1][2];
            f[6] = (float)v3[2][0];
            f[7] = (float)v3[2][1];
            f[8] = (float)v3[2][2];
	    pvm_pkfloat(&f[0], 9, 1);
	    break;

	default:
            error("ivd: ivd_object_change() - unknown object type");
	}
    }

    info = pvm_send(ivd_server_tid, IVDP_OBJ_DEF);
    if (info<0) {
	ivd_server_tid = -1;
	return;
    }
#endif
    obj_changes = OC_ALL;
}

void ivd_object_changes(mw_object obj, unsigned int changes)
/*************************************************************************
 */
{
    obj_changes = changes;
    ivd_object_change(obj);
}


void ivd_new(char *ivd_server, char *ivd_server_host, char *ivd_display)
{
    int info;
    if (ivd_server_host==0) return;

#ifndef NO_PVM
    ivd_tid = pvm_mytid();
    if (ivd_tid<0) {
	warning("ivd() - ivd_new() PVM not started! ");
	ivd_tid = -1;
    }

#ifdef DEBUG_DETAIL
    printf("ivd_new() - try to spawn process '%s' on '%s'.\n",
	   ivd_server ? ivd_server : IVD_SERVER,
	   ivd_server_host);
#endif

    /* start remote ivd-server
     */
    if (pvm_spawn(ivd_server ? ivd_server : IVD_SERVER,
	      0, (ivd_server_host ? PvmTaskHost : PvmTaskDefault),
		  ivd_server_host, 1,  &ivd_server_tid) != 1) {

	ivd_server_tid = -1;
	warning("ivd: ivd_new() - can't start remote ivd-server");
	return;
    }

    pvm_initsend(PvmDataDefault);
    pvm_pkstr(ivd_display);
    info = pvm_send(ivd_server_tid, IVDP_OPEN_SERVER);
    if (info<0) {
	ivd_server_tid = -1;
	return;
    }

#endif
}


void ivd_new2(int server_tid, char *ivd_display)
/*************************************************************************
 */
{
#ifndef NO_PVM
    int info;

    ivd_tid = pvm_mytid();
    if (ivd_tid<0) {
	warning("ivd() - ivd_new() PVM not started! ");
	ivd_tid = -1;
    }

    ivd_server_tid = server_tid;

    pvm_initsend(PvmDataDefault);
    pvm_pkstr(ivd_display);
    info = pvm_send(ivd_server_tid, IVDP_OPEN_SERVER);
    if (info<0) {
	ivd_server_tid = -1;
	return;
    }
#endif
}


void ivd_free()
/*************************************************************************
 */
{
    if (ivd_server_tid >=0) {
#ifndef NO_PVM
	pvm_initsend(PvmDataDefault);
	pvm_send(ivd_server_tid, IVDP_CLOSE_SERVER);
#endif

	pvm_kill(ivd_server_tid);


    }
    ivd_tid = -1;
    ivd_server_tid = -1;
}


void ivd_send_zero_frame(int tid)
/*************************************************************************
 */
{
#ifndef NO_PVM
    long frame_size = 0;
    double dt = 0.0;


    if (tid<=0) return;

#ifdef DEBUG_DETAIL
    printf("ivd_send_zero_frame() - to tid %d\n", tid);
#endif

    pvm_initsend(PvmDataDefault);
    pvm_pkdouble(&dt, 1, 1);
    pvm_pklong(&frame_size, 1, 1);
    pvm_send(tid, IVDP_FRAME);
#endif
}

void ivd_req_camera_frame(mw_object camera, int tid, 
			  int width, int height, int color)
/*************************************************************************
 */
{
    vector3 v;
    quaternion q;
    vector3 cam_pos = {0.0, 0.0, 0.0};
    quaternion cam_dir = {1.0, 0.0, 0.0, 0.0};
    vector3 cam_axis = {0.0, -1.0, 0.0};
    vector3 cam_axis2 = {0.0, 0.0, 1.0};
    float   f[10];
    int info;

#ifndef NO_PVM
    if (ivd_server_tid < 0) {
	if (tid>0) ivd_send_zero_frame(tid);
	return;
    }


#ifdef DEBUG_DETAIL
    printf("ivd_req_camera_frame() - camera=%p, tid=%d, width=%d, height=%d, color=%d\n",
	   camera, tid, width, height, color);
#endif


    pvm_initsend(PvmDataDefault);

    pvm_pkint(&tid, 1, 1);
    pvm_pkint(&width, 1, 1);
    pvm_pkint(&height, 1, 1);
    pvm_pkint(&color, 1, 1);

    mw_object_get_abs_v3(&v, camera, cam_pos);
    f[0] = (float)v[0];
    f[1] = (float)v[1];
    f[2] = (float)v[2];
    
    q_set_av(&cam_dir, 0.5*PI, cam_axis);

    q_set_av(&q, 0.5*PI, cam_axis2);
    q_mul(&cam_dir, q, cam_dir);
/*
 * development-code
    {
	int x;

	printf("*** axis ?"); scanf("%d", &x);
	v3_set(cam_axis, 0.0, 0.0, 0.0);
	if (x>0 && x<4) cam_axis[x-1] = 1.0;
	if (x<0 && x>-4) cam_axis[-x-1] = -1.0;
	if (x!=0) {
	    printf("*** angle ?"); scanf("%d", &x);
	    q_set_av(&q, ((double)x)*PI/180.0, cam_axis);
	    q_mul(&cam_dir, q, cam_dir);
	}
    }
*/
    mw_object_get_abs_q(&q, camera, cam_dir);

    f[3] = (float)q[0];
    f[4] = (float)q[1];
    f[5] = (float)q[2];
    f[6] = (float)q[3];

    f[7]  = camera->func.camera.angle;
    f[8]  = camera->func.camera.focus_distance;
    f[9]  = camera->func.camera.aspect_ratio;

    pvm_pkfloat(&f[0], 10, 1);

    info = pvm_send(ivd_server_tid, IVDP_FRAME_REQ);
    if (info<0) {
	ivd_server_tid = -1;
	ivd_send_zero_frame(tid);
	return;
    }

#endif
}


void ivd_insert_file(char *name)
{
#ifndef NO_PVM
    int info;
    if (ivd_server_tid <0) return;

    pvm_initsend(PvmDataDefault);
    if (name) pvm_pkstr(name); else pvm_pkstr("");
    info = pvm_send(ivd_server_tid, IVDP_INSERT_IV);
    if (info<0) ivd_server_tid = -1;
#endif
}

void ivd_save_file(char *name)
{
#ifndef NO_PVM
    int info;

    if (ivd_server_tid <0) return;

    pvm_initsend(PvmDataDefault);
    if (name) pvm_pkstr(name); else pvm_pkstr("");
    info = pvm_send(ivd_server_tid, IVDP_SAVE_IV);
    if (info<0) ivd_server_tid = -1;
#endif
}

void ivd_save_image(char *name)
{
#ifndef NO_PVM
    int info;

    if (ivd_server_tid <0) return;

    pvm_initsend(PvmDataDefault);
    if (name) pvm_pkstr(name); else pvm_pkstr("");
    info = pvm_send(ivd_server_tid, IVDP_SAVE_IMAGE);
    if (info<0) ivd_server_tid = -1;
#endif
}


void ivd_lock()
{
#ifndef NO_PVM
    int info;

    if (ivd_server_tid <0) return;

    pvm_initsend(PvmDataDefault);
    info = pvm_send(ivd_server_tid, IVDP_LOCK);
    if (info<0) ivd_server_tid = -1;
#endif
}

void ivd_unlock()
{
#ifndef NO_PVM
    int info;

    if (ivd_server_tid <0) return;

    pvm_initsend(PvmDataDefault);
    info = pvm_send(ivd_server_tid, IVDP_UNLOCK);
    if (info<0) ivd_server_tid = -1;
#endif
}

void ivd_camera_view(mw_object camera)
{
    ivd_req_camera_frame(camera, -1, 0, 0, 0);
}

