/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: ivd_server.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/03/17 17:38:08 stolz>
 **************************************************************************
 *  MODULE              : ivd_server
 *
 *  FUNCTION
 *
 *	Display server for MOBS (Mobile Robot Simulator) communicates
 *	via PVM and use the Inventor-Library from Silicon Graphics
 *	to generate/display 3D-Views.
 *
 *      Camera-views are created on request and send back to an application-
 *	prozess. Also a 3D-control window is provided via the
 *	Inventor-SoXtExaminerViewer.
 *
 *	The objects (ivdp_obj) which comes from the simulator
 *      are transformed to an object-tree of inventor-object.
 *
 *  HISTORY
 *  04-Nov-94   Add comments.
 *  11-Nov-94   Add support for an iv-file-visualisation of an object
 *  31-Jan-95   Add support for changing the DISPLAY.
 *  20-Feb-95   Support lock/unlock and empty-req_frames.
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>

#include <X11/Intrinsic.h>
#include <Inventor_c/SbColor.h>
#include <Inventor_c/SoOutput.h>
#include <Inventor_c/nodes/SoPerspectiveCamera.h>
#include <Inventor_c/nodes/SoMaterial.h>
#include <Inventor_c/nodes/SoDirectionalLight.h>
#include <Inventor_c/nodes/SoCone.h>
#include <Inventor_c/nodes/SoSeparator.h>
#include <Inventor_c/nodes/SoMaterial.h>
#include <Inventor_c/nodes/SoCube.h>
#include <Inventor_c/nodes/SoTransform.h>
#include <Inventor_c/nodes/SoCoordinate3.h>
#include <Inventor_c/nodes/SoFaceSet.h>
#include <Inventor_c/nodes/SoShapeHints.h>
#include <Inventor_c/Xt/SoXt.h>
#include <Inventor_c/Xt/SoXtRenderArea.h>
#include <Inventor_c/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor_c/actions/SoWriteAction.h>
#include <Inventor_c/SoDB.h>
#include <Inventor_c/SoInput.h>
#include <Inventor_c/SoOutput.h>
#include <Inventor_c/SoOffscreenRenderer.h>
#include <Inventor_c/SbViewportRegion.h>
#include <Inventor_c/nodes/SoTexture2.h>
#include <Inventor_c/fields/SoSFImage.h>

#include "ivd.h"
#include "ivdp_obj.h"
#include "mw_object.h"
#include "error.h"
#include "pvm3.h"
#include "mobs_interface_def.h"



#define DEBUG_DETAIL 0
/*
*/

/***************************************************************************
 * local variables
 ***************************************************************************
 */

/* for PVM-communication with the MOBS-Prozess or the robot-applications
 */
static int my_tid, parent_tid;     /* pvm task-identifiers */



/***************************************************************************
 * for PVM-communication with the MOBS-Prozess or the robot-applications
 */
static Widget       mainWindow;
static SoXtRenderArea *renderArea;
static SoSep  *root;
static SoNode *node;
static SoXtExamVwr  *viewer;


/***************************************************************************
 * MANAGMENT of ivdp-objects with an dynamic array.
 * The ivdp-objects are stored in objs[], the corresponding
 * inventor-node can be found in id_node[].
 * The uniq id-number is stored in the ids-array. free_ids is a
 * list of free id-numbers.
 * The dynamic arrays objs, id_node and ids has all the size ids_size.
 */
static ivdp_obj *objs = 0;       /* dynamic arrays for ivd-objects */
static SoNode **id_node = 0;     /* and the corresponding iv-node */

/* PRIVATE use of ids_sizeby id_new(), id_free() */
static int    *ids = 0;          /* dynamic array of id-numbers */
static int ids_size = 0;         /* size of the array */
static int free_ids = -1;        /* list of free ids */


/***************************************************************************
 * functions
 ***************************************************************************
 */

/***************************************************************************
 * NAME
 *	id_new(...)       generate a uniq id-number for an ivdp-object
 *
 * FUNCTION
 *	This routine returns the index to a free element in the dynamic 
 *      array for ivdp-objects and the corresponding iv-node. Now the user
 *      can put the contents of his ivdp-object into objs[nr], and the
 *      node-pointer to id_node[nr].
 *      This index will be allocated. Free it by id_free().
 *      If necessary, the size of the dynamic array will grow.
 *
 * RESULTS
 *	Return the index to a newly allocted element.
 *      This number is therefor >=0 and < ids_size.
 *
 * EXAMPLE
 *	int nr;
 *	nr = id_new();
 *
 * NOTE
 *	The element objs[] is not initialized by
 *      id_new(), but id_node[] ist set to zero!
 *
 * SEE ALSO
 *	id_free(), parent_of(), id_node[], objs[]
 *
 ***************************************************************************
 */
int id_new()
{
    int id;

    if (free_ids<0 || free_ids>=ids_size ) {

	int new_size = (ids_size<<1) + 100;

	if (ids_size) {
	    ids     = (int *)     realloc(ids,     new_size*sizeof(int));
	    id_node = (SoNode **) realloc(id_node, new_size*sizeof(SoNode *));
	    objs    = (ivdp_obj *)realloc(objs,    new_size*sizeof(ivdp_obj));
	}
	else {
	    ids     = (int *)     malloc(new_size*sizeof(int));
	    id_node = (SoNode **) malloc(new_size*sizeof(SoNode *));
	    objs    = (ivdp_obj *)malloc(new_size*sizeof(ivdp_obj));
	}

	if (ids == 0 || id_node == 0 || objs == 0)
	    error("ivd_server: id_new() can't alloc memory for requested id");

	for (id = ids_size; id<new_size; id++) {
	    ids[id] = id + 1;
	    id_node[id] = (SoNode *)0;
	}

	free_ids = ids_size;
	ids_size = new_size;
    }

    id = free_ids;
    free_ids = ids[id];

    return id;
}


/***************************************************************************
 * NAME
 *	parent_of(id)   determine parent of an object
 *
 * FUNCTION
 *	This routine determine the parent iv-node of the object with
 *      the id-number <id>. Therefor it look into id_node[id].parent
 *      for the parent. The function takes also care of valid id-numbers.
 *
 * RESULTS
  *	return 0 or the iv-node of the parent object.
  *
  ***************************************************************************
  */
SoNode *parent_of(int id)
{
    if (id>=0 && id<ids_size) {
	id = objs[id].parent;
	if (id>=0 && id<ids_size)
	    return id_node[id];
	else
	    return (SoNode *)root;
    }
    else return (SoNode *)0;
}



 /***************************************************************************
  * NAME
  *	id_free(id)   frees an element of the ivdp-obj-arrays.
  *
  * FUNCTION
  *	This routine marks the object with the nr <id> as free, and
  *      frees the removes the iv-node from the parent node.
  *
  ***************************************************************************
  */
 void id_free(int id)
 {
     if (id>=0 && id < ids_size) {
	 ids[id] = free_ids;
	 free_ids = id;

	 if (id_node[id]) {
	     SoSepRemoveChildNode(parent_of(id), id_node[id]);
	     SoSepUnref(id_node[id]);
	     id_node[id] = (SoNode *)0;
	 }
     }
     else 
	 error("ivd_server: id_free() - illegal id-number");
 }





 /***************************************************************************
  * NAME
  *	set_material_rgb(...)     sets the elements of a material node
  *
  * FUNCTION
  *	This material node with his elements for specular, diffuse,
  *      ambient and emissive color, shininess and transparency are
  *      set corresponding to the rgb-array.
  *      The RGB-array should contain color-values in the range [0,1].
  *
  ***************************************************************************
  */
 void set_material_rgb(SoMtl *material, float rgb[3])
 {
     static float zero3[3] = {0.0, 0.0, 0.0};
     static float f3[3];

     SoMColSetRGB( &(material->diffuseColor), rgb);
     f3[0] = rgb[0]*0.6;
     f3[1] = rgb[1]*0.6;
     f3[2] = rgb[2]*0.6;
     SoMColSetRGB( &(material->specularColor), rgb);
     f3[0] = rgb[0]*0.3;
     f3[1] = rgb[1]*0.3;
     f3[2] = rgb[2]*0.3;
     SoMColSetRGB( &(material->ambientColor), f3);
     SoMColSetRGB( &(material->emissiveColor), zero3);
     SoMFloatSet( &(material->shininess), 0.3);
     SoMFloatSet( &(material->transparency), 0.0);
 }



 /***************************************************************************
  * NAME
  *	generateTextureMap(...)	 render a scene for an texture offscreen.
  *
  * FUNCTION
  *	This routine renders the scene in <root> off-screen an writes
  *      the result with the specific <width> and <hight> into the
  *      texture-image.
  *
  * RESULTS
  *	returns TRUE if the rendering was successfull, 
  *      The rendered image will be found in the texture-image-field.
  *
  ***************************************************************************
  */
 SbBool generateTextureMap (SoNode *root, SoTex2 *texture, 
    short textureWidth, short textureHeight)
 {
    SbVPReg  *myViewport;
    SoOffRnd *myRenderer;
    SbColor  myCol;
    SbVec2s  texSize;

    myViewport = SbVPRegCreateWH(textureWidth, textureHeight);

    /* Render the scene 
     */
    myRenderer = SoOffRndCreate(myViewport);
    SbColSetR_G_B(myCol, 0.0, 0.0, 0.0);
    SoOffRndSetBkgCol(myRenderer, &myCol);

    if (!SoOffRndRnd(myRenderer, root)) {
       SoOffRndDelete(myRenderer);
       return FALSE;
    }

    /* Generate the texture
     */
    SbV2sSetX_Y(texSize, textureWidth, textureHeight);
    SoSImgSet(&(texture->image), &texSize, 
	      /* SO_OFF_RND_LUMINANCE, for color: */
	      SO_OFF_RND_RGB, 
	      SoOffRndGetBuff(myRenderer));

    SoOffRndDelete(myRenderer);

    return TRUE; 
 }



 /***************************************************************************
  * NAME
  *	save_iv(...)    writes the root-scene into a file
  *
  * FUNCTION
  *	Writes the scene under the root-node into a file.
  *
  * OUTPUT
  *	A warning message on stderr will be printed in case of an file-error.
  *
  * EXAMPLE
  *	save_iv("root_scene.iv");
  *
  ***************************************************************************
  *
  */
 void save_iv(SoNode *node, char *name)
 {
     SoWriteAction *wa;
     SoOutput      *out;

     out = SoOutCreate();
     if (SoOutOpenFile(out, name)) {
	 wa = SoWriteActCreateOut(out);
	 SoWriteActApply(wa, node);
	 SoOutCloseFile(out);
     }
     else warning("ivd_server: iv_save_iv() - can't open file for writing");

     SoOutDelete(out);
 }



 /***************************************************************************
  * NAME
  *	save_image(...)    saves a image of the root-scene into a file
  *
  * FUNCTION
  *	Writes the image, which is viewed from <myCamera>, as
  *      PPM-File. The Images size is <width> and <height>.
  *
  * OUTPUT
  *	A warning message on stderr will be printed in case of an file-error.
  *
  * EXAMPLE
  *	save_image("root_scene.ppm", node, camera, 128, 128);
  *
  ***************************************************************************
  */
 void save_image(char *filename, SoNode *root, SoPerspCam *myCamera,
		    short width, short height)
 {
     SoTex2      *texture;
     SbVec2s      size;
     int nc;
     const unsigned char *image;
     int i, count;
     short x, y;
     FILE *fh;
     SoSep       *texRoot;

     texRoot = SoSepCreate();
     SoSepRef(texRoot);

     SoSepAddChild(texRoot, (SoNode *)myCamera);
     SoSepAddChild(texRoot, (SoNode *)root);

     texture = SoTex2Create();
     SoTex2Ref(texture);

     if (generateTextureMap((SoNode *)texRoot, texture, width, height))
	 {
 #ifdef DEBUG_DETAIL
	 printf ("Successfully generated texture map\n");
 #endif
     }
     else {
 #ifdef DEBUG_DETAIL
	 printf ("Could not generate texture map\n");
 #endif
     }

     image = SoSImgGet(&(texture->image), &size, &nc);
     SbV2sGetX_Y(width, height, size);

 #ifdef DEBUG_DETAIL
     printf("image(%p), width=%d, height=%d, nc = %d\n", image, width, height, nc);
     printf("image = ");
     for (i=0; i<25; i++) printf("%d", image[i]);
     printf("\n");
 #endif

     fh = fopen(filename, "wb");
     if (fh==0) return;

     count = width * height * nc;

     if (nc==1) {
	 fprintf(fh, "P5\n%d %d\n255\n", width, height);

	 for (y=height-1; y>=0; y--)
	     fwrite(&image[y*width], width, 1, fh);
     }
     else if (nc==3) {

	 fprintf(fh, "P6\n%d %d\n255\n", width, height);

	 for (y=height-1; y>=0; y--)
	     fwrite(&image[y*width*3], width*3, 1, fh);
     }
     else warning("can't write image!");

     fclose(fh);

     SoSepRemoveChildNode(texRoot, (SoNode *)myCamera);
     SoSepRemoveChildNode(texRoot, (SoNode *)root);
     SoTex2Unref(texture);
     SoSepUnref(texRoot);
 }


 /***************************************************************************
  * NAME
  *	add_iv(...)	inserts a iv-file to a seperator-node
  *
  * FUNCTION
  *	This loads a inventor-file into a node an hang the node onto
  *	a seperator-node.
  *
  * OUTPUT
  *	A error-text will be printed on stderr in case of an file-error.
  *
  * EXAMPLE
  *	add_iv(root, "lights.iv");
  *
  ***************************************************************************
  */
 void add_iv(SoSep *sep, char *filename)
 {
     SoIn        *in;
     SoNode      *result;

 #ifdef DEBUG_DETAIL
     printf("add_iv() - add a iv-file to a node\n");
 #endif    
     in = SoInCreate();
     if (SoInOpenFile(in, filename, FALSE)) {
	 SoDBRead(in, &result);
	 SoInCloseFile(in);
	 SoSepAddChild(sep, result);
     }
     else warning("ivd_server: add_iv() - can't open file for reading");
     SoInDelete(in);
 }


 /***************************************************************************
  * NAME
  *	send_zero_frame(...)	sends a empty frame to the robo-application
  *
  * FUNCTION
  *	Sends a zero-frame via PVM to the robot-application prozess
  *	with the specified tid.
  *
  ***************************************************************************
  */
 void send_zero_frame(int tid)
 {
 #ifndef NO_PVM
     long frame_size = 0;
     double dt = 0.0;

 #ifdef DEBUG_DETAIL
     printf("idd_send_zero_frame() - to tid %d\n", tid);
 #endif

     pvm_initsend(PvmDataDefault);
     pvm_pkdouble(&dt, 1, 1);
     pvm_pklong(&frame_size, 1, 1);
     pvm_send(tid, IVDP_FRAME);
 #endif
 }


 /***************************************************************************
  * NAME
  *	send_frame(...)	sends a rendered frame to the spec. pvm-prozess.
  *
  * FUNCTION
  *	Renders the Object under the root-node from the camera point of
  *	view and sends the generated image with the size widh*height to
  *	the specified tid-number with the IVDP_FRAME tag.
  *      The image is grayscale (parameter color=0) or color. The ordering
  *	of the sended image is top->bottom, left->right. In a color image
  *	a pixel consist of the pair (red,green,blue).
  *
  *	The sended images has therefor the size:
  *	   width*height*(color ? 3 : 1)
  *
  * NOTE
  *      A zero-frame is send, if the renderer fails.
  *
  ***************************************************************************
  */
 void send_frame(int tid,
		 SoNode *root, SoPerspCam *myCamera,
		 short width, short height, int color)

 {
     SoTex2      *texture;
     SbVec2s      size;
     int nc;
     const unsigned char *image;
     int i, count;
     short x, y;
     long frame_size = 0;
     double dt = 0.0;
     SoSep       *texRoot;


     texRoot = SoSepCreate();
     SoSepRef(texRoot);

     SoSepAddChild(texRoot, (SoNode *)myCamera);
     SoSepAddChild(texRoot, (SoNode *)root);

     texture = SoTex2Create();
     SoTex2Ref(texture);

     if (generateTextureMap((SoNode *)texRoot, texture, width, height))
	 {
 #ifdef DEBUG_DETAIL
	 printf ("send_frame() - Successfully generated texture map\n");
 #endif
     }
     else {
 #ifdef DEBUG_DETAIL
	 printf ("send_frame() - Could not generate texture map\n");
 #endif
	 send_zero_frame(tid);
	 SoSepRemoveChildNode(texRoot, (SoNode *)myCamera);
	 SoSepRemoveChildNode(texRoot, (SoNode *)root);
	 SoTex2Unref(texture);
	 SoSepUnref(texRoot);
	 return;
     }


     image = SoSImgGet(&(texture->image), &size, &nc);
     SbV2sGetX_Y(width, height, size);


 #ifdef DEBUG_DETAIL
     printf("send_frame() - image(%p), width=%d, height=%d, nc = %d\n",
	    image, width, height, nc);
 #endif

     if (nc!=3) 
	 error("ivd_server: send_frame() - cannot render a color-image?!");

     count = width * height * nc;

 #ifndef NO_PVM

 #ifdef DEBUG_DETAIL
     printf("send_frame() - send frame to tid %d\n", tid);
 #endif

     pvm_initsend(PvmDataDefault);
     pvm_pkdouble(&dt, 1, 1);

     if (color) {
	 frame_size = width * height * 3;
	 pvm_pklong(&frame_size, 1, 1);
	 for (y=height-1; y>=0; y--)
	     pvm_pkbyte((char *)(&(image[y*width*3])), width*3, 1);

     } else {
	 register unsigned char *frame;
	 register unsigned char *row;
	 register int x, gray;

	 frame_size = width * height;
	 pvm_pklong(&frame_size, 1, 1);

	 frame = (unsigned char *)malloc(width*sizeof(unsigned char));
	 if (frame==0) error("ivd_server: send_frame() - can't alloc memory for frame-row");

	 for (y=height-1; y>=0; y--) {
	     row = &(image[y*width*3]);
	     for (x=0; x<width; x++) {
		 gray = *(row++) * 299;
		 gray += *(row++) * 587;
		 gray += *(row++) * 114;
		 gray /= 1000;
		 frame[x] = gray;
	     }

	     pvm_pkbyte(&frame[0], width, 1);
	 }
	 free(frame);
     }
     pvm_send(tid, IVDP_FRAME);
 #endif


     SoSepRemoveChildNode(texRoot, (SoNode *)myCamera);
     SoSepRemoveChildNode(texRoot, (SoNode *)root);
     SoTex2Unref(texture);
     SoSepUnref(texRoot);
 }




 /***************************************************************************
  * NAME
  *	set_camera(...)	  sets the values of a PerspCamera-Node
  *
  * FUNCTION
  *	Sets the position/orientation, angle, focus and aspect-ration
  *      of a Persp-Camera-node. 
  *
  ***************************************************************************
  */
 void set_camera(SoPerspCam  *myCamera,
		 float pos[3], float orient[4], 
		 float angle, float focus_distance, float aspect_ratio)
 {
     float len;
     float rotangle;
     SbVec3f axis;

     SoSV3fSetXYZ(&(myCamera->position), pos);

     len = sqrt(orient[1] * orient[1] + 
		orient[2] * orient[2] +
		orient[3] * orient[3]);
     if (len<1e-9) len=1.0;
     orient[1] /= len;
     orient[2] /= len;
     orient[3] /= len;
     SbV3fSetX_Y_Z(axis, orient[1], orient[2], orient[3]);

     rotangle = -2.0 * acos(orient[0]);

     SoSRotSetAxisAngle(&(myCamera->orientation), 
			&axis, rotangle);

     SoSFloatSet(&(myCamera->heightAngle), angle);
     SoSFloatSet(&(myCamera->focalDistance), focus_distance);
     SoSFloatSet(&(myCamera->aspectRatio), aspect_ratio);

     SoSFloatSet(&(myCamera->nearDistance), 0.06);
     SoSFloatSet(&(myCamera->farDistance), 100.00);
 }



 /**************************************************************************
  * NAME
  *	remove_shape(...)	remove the shape from an iv-object
  *
  * FUNCTION
  *	Removes the shape-nodes, except for an group-object, from a
  *	iv-object, which is given by his Seperator-Node.
  *
  * PARAMETERS
  *	SoSep *group	Separator-Node of the iv-object
  *	ivdp_obj *obj	Object specification
  *
  * NOTE
  *      This routine is based on the iv-object-creation of
  *	draw_obj() and add_shape()!
  *
  * SEE ALSO
  *      add_shape(), change_shape(), draw_obj()
  *
  **************************************************************************
  */
 void remove_shape(SoSep *group, ivdp_obj *obj)
 {
 #ifdef DEBUG_DETAIL
     printf("remove_shape()\n");
 #endif

     switch (obj->type) {
     case MWO_GROUP:
	 break;

     case MWO_CUBOID:
	 SoSepRemoveChild(group, 3);
	 break;

     case MWO_RECTANGLE:
     case MWO_TRIANGLE:
	 SoSepRemoveChild(group, 4);
	 SoSepRemoveChild(group, 5);
	 SoSepRemoveChild(group, 6);
	 break;

     default:
	 error("ivd_server: remove_shape() - unknown object type");
	 break;
     }
 }


 /**************************************************************************
  * NAME
  *	change_shape(...)	change the shape of an object
  *
  * FUNCTION
  *	Change the shape-parmeters(size/edges) of an object.
  *
  * PARAMETERS
  *	SoSep *group	Separator-Node of the iv-object
  *	ivdp_obj *obj	Object specification with changed obj-parameters.
  *
  * NOTE
  *      This routine is based on the iv-object-creation of
  *	draw_obj() and add_shape()!
  *
  * SEE ALSO
  *      remove_shape(), add_shape(), draw_obj()
  *
  **************************************************************************
  */
 void change_shape(SoSep *group, ivdp_obj *obj)
 {
 #ifdef DEBUG_DETAIL
     printf("change_shape()\n");
 #endif

     switch (obj->type) {
     case MWO_GROUP:
	 break;

     case MWO_CUBOID:
	 {
	     SoCube *cube = (SoCube *)SoSepGetChild(group, 3);

	     SoSFloatSet(&(cube->width), (obj->cuboid_edge_length[0]));
	     SoSFloatSet(&(cube->height), (obj->cuboid_edge_length[1]));
	     SoSFloatSet(&(cube->depth), (obj->cuboid_edge_length[2]));
	 }
	 break;

     case MWO_RECTANGLE:
	 {
	     SoCoord3 *coord = (SoCoord3 *)SoSepGetChild(group, 4);
	     SbVec3f poly[4];
	     float x_size = 0.5 * (obj->rectangle_xy_size[0]);
	     float y_size = 0.5 * (obj->rectangle_xy_size[1]);

	     SbV3fSetX_Y_Z(poly[0], -x_size, -y_size, 0.0);
	     SbV3fSetX_Y_Z(poly[1], -x_size,  y_size, 0.0);
	     SbV3fSetX_Y_Z(poly[2],  x_size,  y_size, 0.0);
	     SbV3fSetX_Y_Z(poly[3],  x_size, -y_size, 0.0);
	     SoMV3fSetN(&(coord->point), 0, 4, poly);
	 }
	 break;

     case MWO_TRIANGLE:
	 {
	     SoCoord3 *coord = (SoCoord3 *)SoSepGetChild(group, 4);
	     SbVec3f poly[3];

	     SbV3fSetXYZ(poly[0], obj->triangle_edges[0]);
	     SbV3fSetXYZ(poly[1], obj->triangle_edges[1]);
	     SbV3fSetXYZ(poly[2], obj->triangle_edges[2]);
	     SoMV3fSetN(&(coord->point), 0, 3, poly);
	 }
	 break;

     default:
	 error("ivd_server: change_shape() - unknown object type");
	 break;
     }
 }


 /**************************************************************************
  * NAME
  *	add_shape(...)	add the shape-nodes to an iv-object
  *
  * FUNCTION
  *	Adds the shape-nodes to the seperator-node of the object.
  *
  * PARAMETERS
  *	SoSep *group	Separator-Node of the iv-object
  *	ivdp_obj *obj	Object specification
  *
  * NOTE
  *      This routine is based on the iv-object-creation of
  *	draw_obj()!
  *
  * SEE ALSO
  *      remove_shape(), change_shape(), draw_obj()
  *
  **************************************************************************
  */
 void add_shape(SoSep *group, ivdp_obj *obj)
 {
 #ifdef DEBUG_DETAIL
     printf("add_shape()\n");
 #endif

     switch (obj->type) {
     case MWO_GROUP:
	 SoSepAddChild(group, (SoNode *) SoSepCreate());
	 break;

     case MWO_CUBOID:
	 {
	     SoCube *cube = SoCubeCreate();

	     SoSFloatSet(&(cube->width), (obj->cuboid_edge_length[0]));
	     SoSFloatSet(&(cube->height), (obj->cuboid_edge_length[1]));
	     SoSFloatSet(&(cube->depth), (obj->cuboid_edge_length[2]));

	     SoSepAddChild(group, (SoNode *)cube);
	 }
	 break;

     case MWO_RECTANGLE:
	 {
	     SoShapeHints *shape_hints = SoShapeHintsCreate();
	     SoCoord3 *coord = SoCoord3Create();
	     SoFaceSet *faceset = SoFSetCreate();
	     SbVec3f poly[4];
	     long numvertices = 4;
	     float x_size = 0.5 * (obj->rectangle_xy_size[0]);
	     float y_size = 0.5 * (obj->rectangle_xy_size[1]);

	     SoSEnumSet(&(shape_hints->faceType),
			SO_SHAPE_HINTS_CONVEX);
	     SoSEnumSet(&(shape_hints->vertexOrdering),
			SO_SHAPE_HINTS_COUNTERCLOCKWISE);
	     SoSEnumSet(&(shape_hints->shapeType),
			SO_SHAPE_HINTS_UNKNOWN_FACE_TYPE);
	     SoSepAddChild(group, (SoNode *)shape_hints);

	     SbV3fSetX_Y_Z(poly[0], -x_size, -y_size, 0.0);
	     SbV3fSetX_Y_Z(poly[1], -x_size,  y_size, 0.0);
	     SbV3fSetX_Y_Z(poly[2],  x_size,  y_size, 0.0);
	     SbV3fSetX_Y_Z(poly[3],  x_size, -y_size, 0.0);

	     SoMV3fSetN(&(coord->point), 0, 4, poly);
	     SoSepAddChild(group, (SoNode *)coord);

	     SoMLongSetN(&(faceset->numVertices), 0, 1, &numvertices);
	     SoSepAddChild(group, (SoNode *)faceset);
	 }
	 break;

     case MWO_TRIANGLE:
	 {
	     SoShapeHints *shape_hints = SoShapeHintsCreate();
	     SoCoord3 *coord = SoCoord3Create();
	     SoFaceSet *faceset = SoFSetCreate();
	     SbVec3f poly[3];
	     long numvertices = 3;


	     SoSEnumSet(&(shape_hints->faceType), 
			SO_SHAPE_HINTS_CONVEX);
	     SoSEnumSet(&(shape_hints->vertexOrdering), 
			    SO_SHAPE_HINTS_COUNTERCLOCKWISE);
	     SoSEnumSet(&(shape_hints->shapeType), 
			SO_SHAPE_HINTS_UNKNOWN_FACE_TYPE);
	     SoSepAddChild(group, (SoNode *)shape_hints);

	     SbV3fSetXYZ(poly[0], obj->triangle_edges[0]);
	     SbV3fSetXYZ(poly[1], obj->triangle_edges[1]);
	     SbV3fSetXYZ(poly[2], obj->triangle_edges[2]);

	     SoMV3fSetN(&(coord->point), 0, 3, poly);
	     SoSepAddChild(group, (SoNode *)coord);

	     SoMLongSetN(&(faceset->numVertices), 0, 1, &numvertices);
	     SoSepAddChild(group, (SoNode *)faceset);
	 }
	 break;

     default:
	 error("ivd_server: add_shape() - unknown object type");
	 break;
     }
 }



 /**************************************************************************
  * NAME
  *	set_transform(...)	set position/orientation
  *
  * FUNCTION
  *	Change the Position/orienation field of the transform-node
  *	to the new values.
  *
  * PARAMETERS
  *	SoTransform *transform	Tranformation node
  *	float position[3]	Position-Vector
  *	float orientation[4]    Orientation as Quaternion
  *
  **************************************************************************
  */
 void set_transform(SoTransform *transform, float position[3], float orientation[4])
 {
     SbVec3f axis;
     float len;

 #ifdef DEBUG_DETAIL
     printf("set_transform() - set position & orientation\n");
 #endif

     SoSV3fSetXYZ(&(transform->translation), position);

     len = sqrt(orientation[1] * orientation[1] +
		orientation[2] * orientation[2] +
		orientation[3] * orientation[3]);
     if (len<1e-9) len=1.0;
     SbV3fSetX_Y_Z(axis, orientation[1]/len,
		   orientation[2]/len,
		   orientation[3]/len);
     SoSRotSetAxisAngle(&(transform->rotation),
			&axis, -2.0 * acos(orientation[0]));
 }



 /**************************************************************************
  * NAME
  *	object_changes(...)	determine the changes between two objects
  *
  * FUNCTION
  *	Determine the changes between two objects with the same type!
  *      This can be a change in position, orientation, parent-object,
  *	color, shape or visual.
  *
  * PARAMETERS
  *	ivdp_obj *obj		New object
  *	ivdp_obj *old_obj	Old object
  *
  * RESULT
  *      int	bitfield of the enum obj_changes (look into file ivd.h!)
  *
  **************************************************************************
  */
 int object_changes(ivdp_obj *obj, ivdp_obj *old_obj)
 {
     int changes = OC_UNCHANGED;

     if (obj==0 || old_obj==0) return;

     if (obj->type != old_obj->type)
	 error("ivd_server: object_changes() - can't change a object type!");

     if (obj->parent != old_obj->parent)
	 changes |= OC_PARENT;

     if (obj->position[0] != old_obj->position[0] ||
	 obj->position[1] != old_obj->position[1] ||
	 obj->position[2] != old_obj->position[2]) {

	 changes |= OC_POSITION;
     }

     if (obj->orientation[0] != old_obj->orientation[0] ||
	 obj->orientation[1] != old_obj->orientation[1] ||
	 obj->orientation[2] != old_obj->orientation[2] ||
	 obj->orientation[3] != old_obj->orientation[3]) {

	 changes |= OC_ORIENTATION;
     }

     if (obj->type != MWO_GROUP &&
	 (obj->rgb[0] != old_obj->rgb[0] ||
	  obj->rgb[1] != old_obj->rgb[1] ||
	  obj->rgb[2] != old_obj->rgb[2])) {

	 changes |= OC_COLOR;
     }

     if (strcmp(obj->iv_file, old_obj->iv_file)!=0)
	 changes |= OC_VISUAL;

     switch (obj->type) {
     case MWO_GROUP:
	 break;

     case MWO_CUBOID:
	 if (obj->cuboid_edge_length[0] != old_obj->cuboid_edge_length[0] ||
	     obj->cuboid_edge_length[1] != old_obj->cuboid_edge_length[1] ||
	     obj->cuboid_edge_length[2] != old_obj->cuboid_edge_length[2]) {

	     changes |= OC_SHAPE;
	 }
	 break;

     case MWO_RECTANGLE:
	 if (obj->rectangle_xy_size[0] != old_obj->rectangle_xy_size[0] ||
	     obj->rectangle_xy_size[1] != old_obj->rectangle_xy_size[1]) {

	     changes |= OC_SHAPE;
	 }
	 break;

     case MWO_TRIANGLE:
	 changes |= OC_SHAPE;
	 break;

     default:
	 error("ivd_server: object_changes() - unknown object type");
	 break;
     }

 #ifdef DEBUG_DETAIL
     printf("object_changes() - old->new: ");
     if (changes & OC_PARENT) printf("parent ");
     if (changes & OC_POSITION) printf("position ");
     if (changes & OC_ORIENTATION) printf("orient ");
     if (changes & OC_COLOR) printf("color ");
     if (changes & OC_VISUAL) printf("visual ");
     if (changes & OC_SHAPE) printf("shape ");
     if (changes == OC_UNCHANGED) printf("unchanged");
     printf("\n");
 #endif

     return changes;
 }



 /**************************************************************************
  * NAME
  *	draw_obj(...)	create/change the iv-object from a sim-object
  *
  * FUNCTION
  *	This routine transform a ivdp-object into a iv-object.
  *	If the object exist, the changes will be performed.
  *
  *      The Object from the simulator is transform into a Inventor-object,
  *	consisting of the following nodes:
  *
  *	  iv-object:
  *                        Seperator
  *                           |
  *	      --------------------------------
  *	      |          |         |         |
  *        Transform  Seperator  Material   Shape
  *
  *      A Pointer of this iv-object is stored in the array id_node[]
  *	and added to the parent-Node (or root-Node).
  *
  *	Also the obj-Parameters are stored in the array objs[]. This is
  *      corresponding to the the id-number of the object!
  *
  *      The Seperator after the Transform node is for adding a
  *	iv-object-structure via file to an object. If this is done,
  *      the Shape-Nodes will disappear! Also note that no material-node
  *	will be created for an group-object.
  *
  *	The different object-types are reflected in a different shape:
  *
  *	  Type  ||  Group     | Cuboid | Triangle or Rectangle
  *	  -------------------------------------------------------------
  *	  Shape ||  Seperator | Cube   | Shape-Hints + Coord3 + FaceSet
  *
  *      The Seperator-Node of a group-object is taken as Parent-Node 
  *	of the child-objects.
  *
  * PARAMETERS
  *      ivdp_obj *obj           Object from the simulator
  *
  * SEE ALSO
  *      id_new(), id_free(), parent_of(), objs[], iv_node[], root,
  *	add_shape(), remove_shape(), change_shape(),
  *	set_transform(), object_changes()
  *
  **************************************************************************
  */
 void draw_obj(ivdp_obj *obj)
 {
     if (obj==0) return;

     if (obj->id < 0 && obj->id >= ids_size) {
	 error("ivd_server: draw_obj() - id not in current limits");
     }

     if (id_node[obj->id]) {
	 ivdp_obj *old_obj;
	 SoSep *group = (SoSep *)id_node[obj->id];

	 /* perform changes of the object
	  */

 #ifdef DEBUG_DETAIL
	 printf("draw_obj() - change object\n");
 #endif

	 /* don't rely on extern changes-flag.
	  */
	 old_obj = &objs[obj->id];
	 obj->changes = object_changes(obj, old_obj);

	 /* change parent
	  */
	 if (obj->changes & OC_PARENT) {

	     if (obj->parent>=0 && obj->parent < ids_size)
		 SoSepAddChild(id_node[obj->parent], (SoNode *)group);
	     else
		 SoSepAddChild(root, (SoNode *)group);

	     SoSepRemoveChildNode(parent_of(obj->id), (SoNode *)group);
	 }

	 /* change position / orientation
	  */
	 if (obj->changes & (OC_POSITION|OC_ORIENTATION)) {
	     SoTransform *transform = (SoTransform *)SoSepGetChild(group, 0);

	     set_transform(transform, obj->position, obj->orientation);
	 }

	 /* change color
	  */
	 if (obj->type != MWO_GROUP && obj->changes & OC_COLOR) {
	     SoMtl        *material = (SoMtl *)SoSepGetChild(group, 2);

	     set_material_rgb(material, obj->rgb);
	 }

	 /* change iv-file
	  */
	 if (obj->changes & OC_VISUAL) {
	     if (obj->iv_file[0]!='\0' && old_obj->iv_file[0]=='\0') {
		 remove_shape(group, obj);
		 add_iv((SoSep *)SoSepGetChild(group, 1), obj->iv_file);
	     }
	     if (obj->iv_file[0]=='\0' && old_obj->iv_file[0]!='\0') {
		 SoSepReplaceChild(group, 1, (SoNode *)SoSepCreate());
		 add_shape(group, obj);
	     }
	     else {
		 SoSepReplaceChild(group, 1, (SoNode *)SoSepCreate());
		 add_iv((SoSep *)SoSepGetChild(group, 1), obj->iv_file);
	     }
	 }

	 /* change shape
	  */
	 if (obj->iv_file[0]==0 && obj->changes & OC_SHAPE)
	     change_shape(group, obj);

	 /* update the object information
	  */
	 objs[obj->id] = *obj;
     }
     else {
	 SoSep *group = SoSepCreate();
	 SoSep *extra = SoSepCreate();
	 SoTransform  *transform = SoXfCreate();


	 SoSepRef(group);
	 
	 /* create object
	  * out of the iv-nodes: translation, extra-sep, [material], [shape]...
	  */

 #ifdef DEBUG_DETAIL
	 printf("draw_obj() - create object\n");
 #endif

	 set_transform(transform, obj->position, obj->orientation);
	 SoSepAddChild(group, (SoNode *)transform);

	 SoSepAddChild(group, (SoNode *)extra);

	 if (obj->type != MWO_GROUP) {
	     SoMtl        *material = SoMtlCreate();

	     set_material_rgb(material, obj->rgb);
	     SoSepAddChild(group, (SoNode *)material);
	 }

	 if (obj->iv_file[0]!='\0') {
	     add_iv((SoSep *)SoSepGetChild(group, 1), obj->iv_file);
	     if (obj->type == MWO_GROUP) add_shape(group, obj);
	 }
	 else
	     add_shape(group, obj);

	 /* update information in objs[] and id_node[] and
	  * hang object under parent.
	  */
	 objs[obj->id] = *obj;
	 id_node[obj->id] = (SoNode *)group;
	 SoSepAddChild(parent_of(obj->id), (SoNode *)group);
     }
 }



 /****************************************************************************
  * Callback-Funktion for PVM-Messages from the Simulator MOBS.
  * The messages described in the file ivd.h are recognized.
  */
 static void ServerLoop_CB(XtPointer cli, int *src, XtInputId id)
 {
     int bufid, bytes, msgtag, client_tid;

     while(pvm_probe(-1, -1)) {

	 bufid = pvm_recv(-1,-1);
	 pvm_bufinfo(bufid, &bytes, &msgtag, &client_tid);

	 switch(msgtag) {
	 case IVDP_OBJ_ID:
	     {
		 int id;

 #ifdef DEBUG_DETAIL
		 printf("IVDP_OBJECT_ID\n");
 #endif
		 pvm_upkint(&id, 1, 1);
	     }
	     break;

	 case IVDP_CLOSE_SERVER:
 #ifdef DEBUG_DETAIL
	     printf("IVDP_CLOSE_SERVER\n");
 #endif
	     exit(0);
	     break;

	 case IVDP_REQ_OBJ_ID:
	     {
		 int id;

		 id = id_new();

 #ifdef DEBUG_DETAIL
		 printf("IVDP_REQ_OB_ID\n");
 #endif

		 pvm_initsend(PvmDataDefault);
		 pvm_pkint(&id, 1, 1);
		 pvm_send(client_tid, IVDP_OBJ_ID);
	     }
	     break;

	 case IVDP_OBJ_DEF:
	     {
		 ivdp_obj obj;
		 char iv_file[300];

 #ifdef DEBUG_DETAIL
		 printf("IVDP_OBJECT_DEF\n");
 #endif
		 pvm_upkint(&(obj.id), 1, 1);
		 pvm_upkuint(&(obj.changes), 1, 1);
		 pvm_upkint(&(obj.parent), 1, 1);
		 pvm_upkint(&(obj.type), 1, 1);

		 pvm_upkfloat(&(obj.position[0]), 3, 1);
		 pvm_upkfloat(&(obj.orientation[0]), 4, 1);

		 pvm_upkstr(&(obj.iv_file[0]));
		 printf("iv-file: %s\n", obj.iv_file);

		 switch(obj.type) {
		 case MWO_GROUP:
		     break;
		 case MWO_CUBOID:
		     pvm_upkfloat(&(obj.rgb[0]), 3, 1);
		     pvm_upkfloat(&(obj.cuboid_edge_length[0]), 3, 1);
		     break;
		 case MWO_RECTANGLE:
		     pvm_upkfloat(&(obj.rgb[0]), 3, 1);
		     pvm_upkfloat(&(obj.rectangle_xy_size[0]), 2, 1);
		     break;
		 case MWO_TRIANGLE:
		     pvm_upkfloat(&(obj.rgb[0]), 3, 1);
		     pvm_upkfloat(&(obj.triangle_edges[0][0]), 9, 1);
		     break;
		 default:
		     break;
		 }
		 draw_obj(&obj);
	     }
	     break;

	 case IVDP_OBJ_FREE:
	     {
		 int id;

 #ifdef DEBUG_DETAIL
		 printf("IVDP_OBJ_FREE\n");
 #endif
		 pvm_upkint(&id, 1, 1);

		 id_free(id);
	     }
	     break;

	 case IVDP_FRAME_REQ:
	     {
		 SoPerspCam  *myCamera;
		 int tid, width, height, color;
		 float pos[3];
		 float orient[4];
		 float angle, focus_distance, aspect_ratio;
		 float len;
		 float rotangle;
		 SbVec3f axis;

#ifdef DEBUG_DETAIL
		 printf("IVDP_FRAME_REQ\n");
#endif
		 pvm_upkint(&tid, 1, 1);
		 pvm_upkint(&width, 1, 1);
		 pvm_upkint(&height, 1, 1);
		 pvm_upkint(&color, 1, 1);

		 pvm_upkfloat(&(pos[0]), 3, 1);
		 pvm_upkfloat(&(orient[0]), 4, 1);
		 pvm_upkfloat(&angle, 1, 1);
		 pvm_upkfloat(&focus_distance, 1, 1);
		 pvm_upkfloat(&aspect_ratio, 1, 1);

#ifdef DEBUG_DETAIL
		 printf(": tid=%d, width=%d, height=%d, color=%d\n", 
			tid, width, height, color);
		 printf(": angle=%f, focus_dist=%f, aspect_ratio=%f\n",
			angle, focus_distance, aspect_ratio);

		 printf(": position = %f %f %f\n",
			pos[0], pos[1], pos[2]);

		 printf(": orientation = angle %f', axis %f %f %f\n",
			(-2.0*acos(orient[0]))*180.0/M_PI,
			orient[1], orient[2], orient[3]);
#endif


		 myCamera = (SoPerspCam *)SoXtExamVwrGetCam(viewer);
		 set_camera(myCamera, pos, orient, 
			    angle, focus_distance, aspect_ratio);
		 SoXtExamVwrSetCam(viewer, (SoCamera *)myCamera);

		 myCamera = SoPerspCamCreate();
		 SoPerspCamRef(myCamera);
		 set_camera(myCamera, pos, orient, angle,
			    focus_distance, aspect_ratio);
		 if (tid>0) 
		     send_frame(tid, (SoNode *)root, myCamera, 
				width, height, color);
		 SoPerspCamUnref(myCamera);



 /* obsolent test-code:
		 save_image("snapshot.ppm", (SoNode)root, myCamera,
			    width, height);
		 send_zero_frame(tid);
  */
	     }
	     break;

	 case IVDP_SAVE_IV:
	     {
		 char filename[300];

		 pvm_upkstr(filename);
 #ifdef DEBUG_DETAIL
		 printf("IVDP_SAVE_IV_FREE '%s'\n", filename);
 #endif
		 save_iv((SoNode *)root, filename);
	     }
	     break;

	 case IVDP_INSERT_IV:
	     {
		 char filename[300];

 #ifdef DEBUG_DETAIL
		 printf("IVDP_INSERT_IV_FREE\n");
 #endif
		 pvm_upkstr(filename);
		 add_iv(root, filename);
	     }
	     break;

	 case IVDP_SAVE_IMAGE:
	     {
		 char filename[300];
		 SoPerspCam  *myCamera;

 #ifdef DEBUG_DETAIL
		 printf("IVDP_SAVE_IMAGE\n");
 #endif
		 pvm_upkstr(filename);
		 myCamera = (SoPerspCam *)SoXtExamVwrGetCam(viewer);
		 save_image(filename, (SoNode *)root, myCamera, 512, 512);
	     }
	     break;

	 case IVDP_LOCK:
#ifdef DEBUG_DETAIL
	     printf("IVDP_LOCK\n");
#endif
	     SoXtExamVwrSetAutoRedraw(viewer, FALSE);
	     break;

	 case IVDP_UNLOCK:
#ifdef DEBUG_DETAIL
	     printf("IVDP_UNLOCK\n");
#endif
	     SoXtExamVwrSetAutoRedraw(viewer, TRUE);
	     SoSepTouch(root);
	     break;

	 default:
	     printf("ivd_server: Ignore unknown message with tag %d.\n", msgtag);
	     break;
	 }
     }
 }



 /****************************************************************************
  */

int main(int argc, char **argv)
{
    XtAppContext app_context;
    int n, *fds;
    char *iv_file_dir = getenv("MOBS_HOME");
    int bufid, bytes, msgtag, client_tid;

/* debug only
    freopen("ivd_server.out", "w", stdout);
*/

     /* Determine own pvm-tid.
      */
     my_tid = pvm_mytid();
     parent_tid = pvm_parent();

     if (my_tid < 0)
	 error("pvm not running.");

     if (parent_tid < 0) {
	 printf("%s: no parent process, this make no sense!\n", argv[0]);
	 printf("%s: my pvm-tid is %d\n", argv[0], my_tid);
	 printf("%s: WAITING for connection-request of MOBS!!!\n", argv[0]);
     }


    bufid = pvm_recv(-1,-1);
    pvm_bufinfo(bufid, &bytes, &msgtag, &client_tid);
    if (msgtag == IVDP_OPEN_SERVER) {
	static char display[300] = "DISPLAY=";
	
#ifdef DEBUG_DETAIL
		 printf("IVDP_OPEN_SERVER\n");
#endif
	pvm_upkstr(&(display[8]));
	if (display[8])
	    (void)putenv(display);
	
    }
    else {
	error("extern client try to connect with wrong message-tag.");
    }


#ifdef DEBUG_DETAIL
     printf("main() - my_tid = %d, parent_tid = %d\n", my_tid, parent_tid);
#endif

     /* change dir to MOBS-Root-directory
      */
    if (iv_file_dir) {
#ifdef DEBUG_DETAIL
     printf("main() - change dir to %s.\n", iv_file_dir);
#endif
	chdir(iv_file_dir);
    }
    

    /* Initialize Xt and Inventor,
     * add callbacks to process PVM-messages
     */
    mainWindow = SoXtInit( argv[0], "Inventor" );
    if ( mainWindow == NULL ) exit( 1 );

    app_context = SoXtGetAppContext();

    n = pvm_getfds(&fds);

    while(n-- > 0) {
	XtAppAddInput(app_context, fds[n], (XtPointer) XtInputReadMask,
		  ServerLoop_CB, (XtPointer) 0);
    }


    /* Create root-node for all objects.
     */
    root = SoSepCreate();
    node = (SoNode *)root;
    SoSepRef( root );


    /* Setup viewer with the root.
     */
    viewer = SoXtExamVwrCreateStd(mainWindow, NULL);
    SoXtExamVwrSetScene( viewer, (SoNode *) root );
    SoXtExamVwrSetTitle(viewer, "MOBS Camera Viewer");
    SoXtExamVwrScheduleRedraw(viewer);
    SoXtExamVwrShow( viewer );

    /* Initialize the Inventor-Database (for load/save-operations).
     */
    SoDBInit();

    SoXtShow(mainWindow);

    SoXtMainLoop();
}
