/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: list.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:55:38 stolz>
 **************************************************************************
 *  MODULE              : list
 *  IMPORT		: basic, error
 *
 *  IMPLEMENTATION
 *
 *      The adt list is implemented by a list-node which holds the element
 *	(pointer to the user-datastructure). This nodes are single-linked.
 *
 *      The allocations and freeing of the node-structures is done by
 *	malloc() and free() calls. Perhaps, this can lead to performance
 *	problems.
 *
 *
 *  NOTE
 *      The module specification and definition is given in list.h!
 *
 *	A main()-function with a call of the module-test can be generated
 *	by defining MODULE_TEST.
 *
 *  HISTORY
 *  02-May-94   Creation date
 *  03-May-94   Module test passed. list_remove_current() is obsolent!
 *  05-May-94   l->current changed to type list_node*. Seg-fault? HiHo
 *  19-Jul-94   Add list_for_all() routines.
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>

#include "list.h"

#include "error.h"


/****************************************************************************
 */

typedef struct list_elem list_node;
struct list_elem {
    void*	element;
    list_node*	next;
};

typedef struct adt_list adt_list;
struct adt_list {
    int size;
    list_node	*first, *current;
};



/****************************************************************************
 */
list	list_new()
{
    list l;

    l = (list)malloc(sizeof(adt_list));
    if (l) {
	l->size    = 0;
	l->first   = 0;
	l->current = 0;
    }
    else error("list: list_new() - can't alloc memory for adt list");

    return l;
}



/****************************************************************************
 */
void	list_free(list l)
{
    list_node *n, *t;

    for (n=l->first; n; n=t) {
        t = n->next;
	free(n);
    }

    free(l);
}


/****************************************************************************
 */
void    list_add(list l, list_elem e)
{
    list_node *n;

    n = (list_node *)malloc(sizeof(list_node));
    if (n) {
	if (l->size == INT_MAX)
	    error("list: list_add() - max number of elements reached");

	n->element = e;
        n->next = l->first;
	l->size = l->size + 1;
        l->first = n;
    }
    else error("list: list_add() - can't alloc memory for a list-node");
}


/****************************************************************************
 */
void list_remove_all(list l)
{
    list_node *n, *t;

    for (n=l->first; n; n=t) {
        t = n->next;
	free(n);
    }

    l->size    = 0;
    l->first   = 0;
    l->current = 0;
}

/****************************************************************************
 */
void	list_remove(list l, list_elem e)
{
    list_node **np, *t;

    for (np=&(l->first); *np; np=&((*np)->next)) {
	if (e==(*np)->element) {
	    t = *np;
	    if (t == l->current) l->current = t->next;
            *np = t->next;
	    free(t);
	    l->size = l->size - 1;
	    return;
	}
    }
}


/****************************************************************************
 */
#ifdef MODULE_TEST
void list_remove_current(list l)
{
    list_node *n;

    n = *(l->current);
    if (n) {
        *(l->current) = n->next;
	free(n);
	l->size = l->size - 1;
    }
}
#endif

/****************************************************************************
 */
list_elem	list_first(list l)
{
    l->current = l->first;

    if (l->current) {

        return l->current->element;
    }
    else return 0;
}


/****************************************************************************
 */
list_elem	list_next(list l)
{
    if (l->current) {
	l->current = l->current->next;

	if (l->current) {
	    return l->current->element;
	}
	else return 0;
    }
    else return 0;
}


/****************************************************************************
 */
void		list_for_all(list l, void (*f)(list_elem))
{
    list_node *current;

    if (f) {
	for (current = l->first; current; current = current->next)
	    f(current->element);
    }
}


/****************************************************************************
 */
void		list_for_all2(list l, void (*f)(list_elem, void *userdata),
			      void *userdata)
{
    list_node *current;

    if (f) {
	for (current = l->first; current; current = current->next)
	    f(current->element, userdata);
    }
}


/****************************************************************************
 */
boolean list_is_empty(list l)
{
    return (0 == l->size);
}


/****************************************************************************
 */
boolean list_is_not_empty(list l)
{
    return (0 != l->size);
}


/****************************************************************************
 */
int	list_size(list l)
{
    return l->size;
}



boolean lists_equal(list l1, list l2)
/****************************************************************************
 */
{
    list_node *n1, *n2;
    int count1, count2;

    if (l1->size != l2->size) return FALSE;

    for (n1=l1->first; n1; n1=n1->next) {
	count1 = 0;
	for (n2=l1->first; n2; n2=n2->next)
	    if (n1->element == n2->element) count1++;

	count2 = 0;
	for (n2=l2->first; n2; n2=n2->next)
	    if (n1->element == n2->element) count2++;

	if (count1 != count2) return FALSE;
    }

    return TRUE;
}



/****************************************************************************
 * MODULE-TEST-routines
 */

#ifdef MODULE_TEST

#define N       10

static int test_data[N];


int list_test()
{
    int failed, errors = 0;
    char index;
    list l;
    int i;
    int *ip;

    index = 'a';
    fprintf(stderr, "LIST-TEST from %s %s\n"
            "  %c) generate an empty list via list_new(..)\n",
	    __DATE__, __TIME__, index++);

    l = list_new();

    if (failed = (!list_is_empty(l))) errors++;
    fprintf(stderr, "  %c) test function list_is_empty(..) on empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (list_is_not_empty(l))) errors++;
    fprintf(stderr, "  %c) test function list_is_not_empty(..) on empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (0 != list_size(l))) errors++;
    fprintf(stderr, "  %c) test function list_size(..) on empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (0!=list_first(l))) errors++;
    fprintf(stderr, "  %c) use function list_first(..) on empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));


    fprintf(stderr, "  Add %d items with numbers %d..0 to the empty list\n", N, N-1);
    for (i=N-1; i>=0; i--) {
	test_data[i] = i;
	list_add(l, &test_data[i]);
    }

    fprintf(stderr, "  Print out contents of list using list_first() and ..next()\n");
    fprintf(stderr, "  list = <");
    for (ip = list_first(l); ip; ip=list_next(l)) {
	fprintf(stderr, "%d ", *ip);
    }
    fprintf(stderr, ">\n");


    if (failed = (list_is_empty(l))) errors++;
    fprintf(stderr, "  %c) test function list_is_empty(..) on not empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (!list_is_not_empty(l))) errors++;
    fprintf(stderr, "  %c) test function list_is_not_empty(..) on not empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (N != list_size(l))) errors++;
    fprintf(stderr, "  %c) test function list_size(..) on not empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (list_first(l) != &test_data[0]) {
	failed = TRUE;
	errors++;
    }

    fprintf(stderr, "  %c) use function list_first(..) on not empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (list_next(l) != &test_data[1]) {
	failed = TRUE;
	errors++;
    }
    fprintf(stderr, "  %c) use function list_next(..) on not empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (N != list_size(l))) errors++;
    fprintf(stderr, "  %c) test function list_size(..) after examinination of 2 elems: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));


    list_remove_current(l);
    if (failed = (N-1 != list_size(l))) errors++;
    fprintf(stderr, "  %c) list_size(..) after removing the current element(1): %s\n",
	    index++, (failed ? "FAILED" : "PASSED"));


    for (i=3; i<N; i++) {
	if (list_next(l) != &test_data[i]) {
	    failed = TRUE;
	    errors++;
        }
	fprintf(stderr, "  %c.%d) call of list_next(..) : %s\n", index, i-1,
			(failed ? "FAILED" : "PASSED"));
    }
    index++;

    list_remove(l, &test_data[0]);
    list_remove(l, &test_data[3]);
    if (failed = (N-3 != list_size(l))) errors++;
    fprintf(stderr, "  %c) removing 1. and 4. element from list via list_remove()\n"
		    "     test to length %d via list_size(): %s\n",
	    index++, N-3, (failed ? "FAILED" : "PASSED"));

    list_remove_current(l);
    if (failed = (N-4 != list_size(l))) errors++;
    fprintf(stderr, "  %c) remove current(should be 9, the last) element\n"
		    "     and test to length %d via list_size(): %s\n",
	    index++, N-4, (failed ? "FAILED" : "PASSED"));

    if (list_next(l) != 0) {
	failed = TRUE;
	errors++;
    }
    fprintf(stderr, "  %c) list_next(..) should now be zero: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (list_first(l) != &test_data[2]) {
	failed = TRUE;
	errors++;
    }
    fprintf(stderr, "  %c) list_first(..) should point to elem 2 in the new list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (list_next(l) != &test_data[4]) {
	failed = TRUE;
	errors++;
    }
    fprintf(stderr, "  %c) list_next(..) should return elem 4: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));


    fprintf(stderr, "  list = <");
    for (ip = list_first(l); ip; ip=list_next(l)) {
	fprintf(stderr, "%d ", *ip);
    }
    fprintf(stderr, ">\n");

    list_free(l);


    fprintf(stderr, "Total amount of errors: %d. Test %s!\n\n",
			errors, errors ? "FAILED" : "PASSED");


    return errors;
}

int main(int argc, char **argv)
{
    return list_test();
}

#endif
