/********************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: list.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 16:55:45 stolz>
 ********************************************************************************
 *  MODULE              : list
 *  IMPORT		: basic
 *
 *  FUNCTION
 *
 *	This module make a abstract datatype "list" available. The
 *      porpose of this type is to collect data-structures not with
 *	their contents but as pointers. So the structures can be
 *	consistent members of more than one list. The size of the
 *	data-structures is also meaningless.
 *
 *      A list must be initialized via list_new(). After that, elements
 *      (pointers) can be added and removed via list_add() and
 *	list_remove(). To empty the whole list, use list_remove_all().
 *
 *	The size of the list and if it is empty or not can be
 *      aquired by list_size(), list_is_empty(), list_is_not_empty().
 *
 *      To examine what is in the list use the function list_first()
 *	and for the next element list_next(). A zero as return value
 *      indicates that no element was found. Don't mix this functions
 *	with the ..add/remove.. calls.
 *	Another way to examine a list are the list_for_all() functions.
 *	They call for each element in the list the function f. The 
 *	list_for_all2() function allows the calling of a function f
 *	with a pointer to user-specific data. It's also forbidden to
 *	call list_add/remove/free in the user-function f!
 *	But it's possible to nest list_for_all() calls for the same
 *	list.
 *
 *	The list is freed by a call of list_free().
 *
 *      For test-purposes, the function lists_equal() can be used to
 *      check if two list has the same elements. This is indicated
 *	by the return value TRUE.
 *
 *
 *  EXAMPLE
 *
 *	#include "list.h"
 *
 *	struct elem {
 *	    int nr;
 *	    char sig;
 *      };
 *
 *      void print_elem(struct elem *ep)
 *      {
 *	    printf("Element with nr=%d and sig=%c\n", ep->nr, ep->sig);
 *      }
 *
 *      void main()
 *      {
 *	    int i;
 *	    list l;
 *          struct elem e[4];
 *          struct elem *ep;
 *
 *	    l = list_new();
 *
 *	    // insert 4 elements
 *	    for (i=3; i>=0; i++) {
 *		e[i].nr = i+1;
 *		e[i].sig = 'a' + i;
 *              queue_add(l, &e[i]);
 *	    }
 *
 *          if (list_is_empty(l)) printf("error, empty list?\n");
 *
 *          printf("%d elements in my list.\n", list_size(l));
 *
 *	    // examine the list
 *	    for (ep=list_first(l); ep; ep=list_next(l)) {
 *              printf("elem with nr=%d, sig=%c\n", ep->nr, ep->sig);
 *	    }
 *
 *          // another way to print the contents of the list.
 *	    list_for_all(l, print_elem);
 *
 *          // remove 3. element
 *	    list_remove(l, &e[2]);
 *	    if (3 != list_size(l))
 *              printf("hoppla, error something is wrong!\n");
 *
 *          // to empty the list for further usage
 *	    list_remove_all(l);
 *          // ... do something with list
 *
 *          // freeing adt list
 *	    list_free(l);
 *      }
 *
 *
 *  NOTE
 *
 *  HISTORY
 *  02-May-94   Creation date.
 *  03-May-94   Module test passed. Adding list_remove_all().
 *  27-May-94   add lists_equal().
 *  19-Jul-94   Add list_for_all() routines.
 *
 ********************************************************************************/
#ifndef LIST_H
#define LIST_H

#ifndef BASIC_H
#include "basic.h"
#endif

/****************************************************************************
 * abstract datatype queue is hidden behind a opac pointer.
 */

typedef struct adt_list*	list;
typedef void*			list_elem;

/****************************************************************************
 * new and free a list.
 */
list	list_new();
void	list_free(list);

/****************************************************************************
 * add/remove element(s)
 */
void    list_add(list, list_elem);
void	list_remove(list, list_elem);
void    list_remove_all(list l);


/****************************************************************************
 * examination, don't mix with add/remove!
 */
list_elem	list_first(list);
list_elem	list_next(list);
void		list_for_all(list, void (*f)()); /* void (*f)(list_elem) */
void		list_for_all2(list, void (*f)(list_elem, void *userdata), 
			      void *userdata);



/****************************************************************************
 * list characteristics
 */
boolean list_is_empty(list);
boolean list_is_not_empty(list);
int	list_size(list);
boolean lists_equal(list l1, list l2);

#endif
