/********************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: mob.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:55:52 stolz>
 ********************************************************************************
 *  MODULE              : mob
 *
 *  IMPLEMENTATION
 *	A mobile Robot based on the vehicle-module is implemented.
 *	The command-interpretion is done in peridoc time-point-functions
 *	(from the scheduler). Also the us-sensor-measuring is done in
 *	another time-point-function.
 *
 *  NOTE
 *      The specification and definition is given in mob.h!
 *     
 *
 *  HISTORY
 *  20-Apr-94   create this out of old mobs and robot module.
 *  29-May-94	removing obsolent vehicle parameters r1 and r2.
 *  16-Jun-94   Add mob_reset function. Add code for us-sensors
 *  12-Jul-94   Add mob_output_open() and ..close() routines.
 *  14-Jul-94   Removing of timepoint in mob_free forgotten.
 *  08-Oct-94   Sensor-data-functions added
 *
 ********************************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "mob_private.h"
#include "error.h"
#include "basic.h"
#include "mob_limits.h"
#include "command.h"


void mob_output_open(mob m)
{
    if (m->com_out_open) m->com_out_open(m->com_uservar);
}

void mob_output_close(mob m)
{
    if (m->com_out_close) m->com_out_close(m->com_uservar);
}

void mob_output(mob m, char *out)
{
    if (m->com_out) m->com_out(m->com_uservar, out);
}

static void mob_command(scheduler *s, timepoint *tp, double t)
{
    mob m = tp->workspace;
    static char com[MAX_COMMAND_LENGTH+1];

    m->t_of_next_cmd = t + m->dt_command;

    if (m->com_in) {
        com[0] = 0;
	com[MAX_COMMAND_LENGTH] = 0;
	m->com_in(com, m->com_uservar);
	if (0!=com[0]) {
	    mob_do_command(m, t, com);
	}
    }

    if (t > m->t_of_next_cmd)
	error("mob: mob_command() - t of next cmd is lower than actual time");

    scheduler_point_add(s, tp, m->t_of_next_cmd);
}


static void mob_update_us_sensors(scheduler *s, timepoint *tp, double t)
{
    mob m = tp->workspace;
    int i, j, k, index, old_count;
    double atime, temp;
    us_sensor_description *sensor;


    /* us-sensor read
     */
    if (m->read_node>=0 && m->read_node< MAX_NR_NODES) 
    for (k=0; k<MAX_NR_SENSORS_PER_NODE; k++) {
	j = node_adr(m->read_node, k);
	i = m->sensor_nr[j];
#if 0
	printf("mob_update_us_sensor() Read sensor with node-adr %d, and nr %d\n",
	       j, i);
#endif
	if (i<0) continue;
	sensor = &(m->us_sensor[i]);
/*
    for (i=0; i<MAX_NR_US_SENSORS; i++) {
	sensor = &(m->us_sensor[i]);
*/
	old_count = sensor->activation_count;
	j = sensor->activation_count - 1;
	atime = t;
	index = -1;

	while (j>=0 && (temp=sensor->activation_time[j]) <= t) {

	    /* determine earliest sensor activation time
	     */
	    if (temp >= m->sensor_invoke_time && temp <= atime) {
		atime = temp;
		index = j;
	    }
	    
	    j--;
            sensor->activation_count -= 1;
            sensor->free_activations += 1;
        }


        /* compute TOF
         */

	atime -= m->sensor_invoke_time;

        if (drand() < m->us_dropout_probability) 
	    sensor->tof = 10.0 * m->dt_us_scan * drand();
	else 
	    sensor->tof = atime * (m->us_rand_range*drand() + m->us_rand_start);

#ifdef DEBUG_DETAIL
	fprintf(stderr, "TOF of sensor %d is %fs\n", i, sensor->tof);
#endif
	/* select sensor-data
	 */

#if 0
	printf("sensor activation index =%d\n", index);
#endif
	
	if (index >= 0 && sensor->scan_mode != USSM_INACTIVE &&
	    sensor->sensor_cp[index] != 0 && m->sd_select) {
	    m->sd_select(sensor->sensor_cp[index]);
	    sensor->sensor_cp[index] = 0;
	}
	
	/* free obsolent sensor-data
	 */
#if 0
	printf("sensor %d: old-count=%d, act-count=%d\n", 
	       i, old_count, sensor->activation_count);
#endif
	if (m->sd_free) {
	    for (j=sensor->activation_count; j<old_count; j++)
		if (sensor->sensor_cp[j]) m->sd_free(sensor->sensor_cp[j]);
	}
    }


    if (m->us_mode == SM_AUTOMATIC) {
	m->read_node = m->read_node + 1;
	if (m->read_node == 6 /* MAX_NR_NODES */) {
	    m->read_node = 0;
	    m->us_read_complete = TRUE;
	}
    }
    else if (m->invoke_us_read) m->us_read_complete = TRUE;


    /* invoke transmission of the us-signal of the current node.
     */
    if (m->invoke_us_read && m->us_scan) {
	for (i=0; i<MAX_NR_SENSORS_PER_NODE; i++) {
	    j = m->sensor_nr[node_adr(m->read_node, i)];
#if 0
	    printf(" mob_update_us_sensors: try to invoke Sensor nr %d\n", j);
#endif
	    if (j>=0 && m->us_sensor[j].scan_mode != USSM_INACTIVE) {
#if 0
		printf(" mob_update_us_sensors: Invoke Sensor nr %d\n", j);
#endif
		m->us_scan(m, j, m->us_uservar);
	    }
	}
    }

#if 0
    if (m->invoke_us_read && m->us_scan) {
	m->us_read_complete = TRUE;

	for (i=0; i<MAX_NR_US_SENSORS; i++) {
	    if (m->us_sensor[i].scan_mode != USSM_INACTIVE) {
		m->us_scan(m, i, m->us_uservar);
	    }
	}
    }
#endif

    m->sensor_invoke_time = t;
    scheduler_point_add(s, tp, t + m->dt_us_scan);
}




void mob_command_delay(mob m, double t)
{
    m->t_of_next_cmd = t;
}


void mob_set_io(mob m,
		   void (*com_in)(char *, void *uservar),
		   void (*com_out)(void *uservar, char *),
		   void *uservar)
{
    m->com_in = com_in;
    m->com_out = com_out;
    m->com_uservar = uservar;
}


void mob_set_io2(mob m,
		   void (*com_in)(char *, void *uservar),
		   void (*com_out)(void *uservar, char *),
		   void (*com_out_open)(void *uservar),
		   void (*com_out_close)(void *uservar),
		   void *uservar)
{
    m->com_in = com_in;
    m->com_out = com_out;
    m->com_out_open = com_out_open;
    m->com_out_close = com_out_close;
    m->com_uservar = uservar;
}

#define set_us(sensor,node,adr,dir) \
	m->us_sensor[sensor-1].direction = dir; \
	m->sensor_node_adr[sensor-1] = node_adr(node-1,adr-1); \
	m->sensor_nr[node_adr(node-1,adr-1)] = sensor-1;

static void mob_init_us_sensors(mob m)
{
    int i;

    m->us_mode  = SM_VOID;
    m->invoke_us_read = FALSE;
    m->us_read_complete = FALSE;
    m->node_invoke = 0;
    m->read_node = 0;

    for (i=0; i<MAX_NR_US_SENSORS; i++) {
	m->us_sensor[i].scan_mode = USSM_INACTIVE;
	m->us_sensor[i].direction = 0;
	m->us_sensor[i].emission_angle = 300;
	m->us_sensor[i].tof = 0;

        m->sensor_node_adr[i] = -1;

	m->us_sensor[i].activation_time = 0;
	m->us_sensor[i].sensor_cp = 0;
	m->us_sensor[i].activation_count = 0;
	m->us_sensor[i].free_activations = 0;
    }

    for (i=0; i<MAX_NODE_ADR; i++)
	m->sensor_nr[i] = -1;

    set_us( 1, 6, 2, USD_FRONT);
    set_us( 2, 5, 2, USD_FRONT);
    set_us( 3, 4, 2, USD_FRONT|USD_RIGHT);
    set_us( 4, 3, 2, USD_FRONT|USD_RIGHT);
    set_us( 5, 2, 2, USD_FRONT|USD_RIGHT);
    set_us( 6, 5, 3, USD_RIGHT);
    set_us( 7, 6, 3, USD_RIGHT);
    set_us( 8, 1, 4, USD_RIGHT);
    set_us( 9, 2, 4, USD_RIGHT|USD_BACK);
    set_us(10, 3, 4, USD_RIGHT|USD_BACK);
    set_us(11, 4, 4, USD_RIGHT|USD_BACK);
    set_us(12, 5, 4, USD_BACK);
    set_us(13, 6, 4, USD_BACK);
    set_us(14, 1, 1, USD_BACK);
    set_us(15, 2, 1, USD_LEFT|USD_BACK);
    set_us(16, 3, 1, USD_LEFT|USD_BACK);
    set_us(17, 4, 1, USD_LEFT|USD_BACK);
    set_us(18, 5, 1, USD_LEFT);
    set_us(19, 6, 1, USD_LEFT);
    set_us(20, 1, 2, USD_LEFT);
    set_us(21, 4, 3, USD_LEFT|USD_FRONT);
    set_us(22, 3, 3, USD_LEFT|USD_FRONT);
    set_us(23, 2, 3, USD_LEFT|USD_FRONT);
    set_us(24, 1, 3, USD_FRONT);
}


mob mob_new2(scheduler *s, mob_parameters par)
{
    mob m;
    vehicle_parameters vp;

    m = (mob)malloc(sizeof(adt_mob));
    if (m) {
	vp.a = par.a;
	vp.b = par.b;
	vp.c = par.c;
	vp.d = par.d;
        m->vehicle = vehicle_new(s, &vp);
	m->parameters = par;
	m->bumper_on = FALSE;

        m->com_in = 0;
        m->com_out = 0;
        m->com_out_open = 0;
        m->com_out_close = 0;
        m->scheduler = s;
	m->dt_command = 0.1;
	m->command_tp.task = 0;
	m->command_tp.workspace = m;
	m->command_tp.free = 0;
	m->command_tp.function = mob_command;
	scheduler_point_add(s, &(m->command_tp), scheduler_time(s)+1e-3);

	m->motion_server = MS_NONE;

        m->odom_on = FALSE;

	mob_init_us_sensors(m);
	m->dt_us_scan = 0.1;
	m->us_uservar = 0;
	m->us_scan = 0;
	m->us_sensor_tp.task = 0;
	m->us_sensor_tp.workspace = m;
	m->us_sensor_tp.free = 0;
	m->us_sensor_tp.function = mob_update_us_sensors;
	m->sd_free = 0;
	m->sd_select = 0;
	scheduler_point_add(s, &(m->us_sensor_tp), scheduler_time(s) + m->dt_us_scan);


        m->us_rand_range = 0.0;
        m->us_rand_start = 1.0;
	m->us_dropout_probability = 0.0;
    }
    else error("mob: mob_new() - can't alloc memory for adt mob");

    return m;
}

mob mob_new(scheduler *s)
{
    mob_parameters par = {
	0.272, 0.508, 0.225, 0.045
    };

    return mob_new2(s, par);
}

static void mob_free_us_sensors(mob m)
{
    int i, j;
    us_sensor_description *sensor;

    for (i=0; i<MAX_NR_US_SENSORS; i++) {
	sensor = &(m->us_sensor[i]);

	if (sensor->activation_time)  {
	    free(sensor->activation_time);
	    sensor->activation_time = 0;
	}
	if (sensor->sensor_cp) {
	    if (m->sd_free) {
		for (j=0; j<sensor->activation_count; j++)
		    if (sensor->sensor_cp[j]) m->sd_free(sensor->sensor_cp[j]);
	    }
	    free(sensor->sensor_cp);
	    sensor->sensor_cp = 0;
	}
	sensor->activation_count = 0;
	sensor->free_activations = 0;
    }
}



void mob_free(mob m)
{

    scheduler_point_unlink(m->scheduler, &(m->command_tp));
    scheduler_point_unlink(m->scheduler, &(m->us_sensor_tp));
    vehicle_free(m->vehicle);
    free_motion_server(m);
    mob_free_us_sensors(m);
    free(m);
}


int motion_type(mob m)
{
    switch(vehicle_motion_type(m->vehicle)) {
    case VMT_HALT:  return MT_HALT;
    case VMT_STOP:  return MT_STOP;
    case VMT_ACC:   return MT_ACCEL;
    case VMT_DEC:   return MT_DECEL;
    case VMT_CONST: return MT_CONST;
    case VMT_KILL:  return MT_KILL;
    default:
	error("mob: motion_type() - unknown vehicle motion type");
	break;
    }
    return MT_STOP;
}



void change_motion_server(mob m, int ms)
{
    if (motion_type(m)!=MT_STOP) motion_kill(m);

    switch(m->motion_server) {
    case MS_NONE:
    case MS_MOTN:
	break;
    case MS_MOTV:
	scheduler_point_unlink(m->scheduler, &(m->motion.motv.change_tp));
	queue_free(m->motion.motv.motions);
	break;
    default:
	error("mob: change_motion_server() - illegal init motion server!");
	break;
    }

    m->motion_server = ms;

    switch(m->motion_server) {
    case MS_NONE:
	break;
    case MS_MOTN:
	m->motion.motn.linear_motion = -1;
	break;
    case MS_MOTV:
	m->motion.motv.motions = queue_new(sizeof(motv_motion));
	break;
    default:
	error("mob: change_motion_server() - motion server not supported!");
	break;
    }
}

void free_motion_server(mob m)
{
    change_motion_server(m, MS_NONE);
}


void mob_get_motion_state(mob m, mob_state *real, mob_state *intern)
{
    vehicle_state vreal, vintern;

    vehicle_get_state(m->vehicle, &vreal, &vintern);
    if (real) {
	real->x = vreal.x;
	real->y = vreal.y;
	real->angle = vreal.angle;
	real->v = vreal.v;
	real->w = vreal.w;
	real->a = vreal.a;
	real->u = vreal.u;
	real->angle_fl = vreal.angle_fl;
	real->angle_fr = vreal.angle_fr;
    }
    if (intern) {
	intern->x = vintern.x;
	intern->y = vintern.y;
	intern->angle = vintern.angle;
	intern->v = vintern.v;
	intern->w = vintern.w;
	intern->a = vintern.a;
	intern->u = vintern.u;
	intern->angle_fl = vintern.angle_fl;
	intern->angle_fr = vintern.angle_fr;
    }
}

void mob_set_motion_state(mob m, mob_state *real,
			  mob_state *intern)
{
    vehicle_state vreal, vintern;

    if (real) {
	vreal.x = real->x;
	vreal.y = real->y;
	vreal.angle = real->angle;
	vreal.v = real->v;
	vreal.w = real->w;
	vreal.a = real->a;
	vreal.u = real->u;
	vreal.angle_fl = real->angle_fl;
	vreal.angle_fr = real->angle_fr;
    }
    if (intern) {
	vintern.x = intern->x;
	vintern.y = intern->y;
	vintern.angle = intern->angle;
	vintern.v = intern->v;
	vintern.w = intern->w;
	vintern.a = intern->a;
	vintern.u = intern->u;
	vintern.angle_fl = intern->angle_fl;
	vintern.angle_fr = intern->angle_fr;
    }

    vehicle_set_state(m->vehicle,
		      ((real) ? &vreal : 0),
		      ((intern) ? &vintern : 0));
}


void mob_get_intern_state(mob m, mob_motion_state *ms)
{
    vehicle_state real, intern;

    vehicle_get_state(m->vehicle, &real, &intern);

    ms->x = intern.x;
    ms->y = intern.y;
    ms->angle = intern.angle;
    ms->v = intern.v;
    ms->w = intern.w;
    ms->a = intern.a;
    ms->u = intern.u;
}

void motion_add(mob m, double t, double a, double u)
{
    vehicle_add_motion(m->vehicle, a, u, t);
}

void motion_kill(mob m)
{
    vehicle_kill_motion(m->vehicle, MAX_LINEAR_ACC, MAX_ANGULAR_ACC);
}

double motion_stop_time(mob m)
{
    return vehicle_get_stop_time(m->vehicle);
}

void empty_motion_queue(mob m)
{
    vehicle_empty_motion_queue(m->vehicle);
}


void mob_set_vel_disturbance(mob m, mob_disturbance md)
{
     vel_disturbance vd;

     vd.v_low  = md.v_low;
     vd.v_high = md.v_high;
     vd.w_low  = md.w_low;
     vd.w_high = md.w_high;

     vehicle_set_vel_disturbance(m->vehicle, vd);
}

void mob_halt(mob m)
{
    vehicle_halt(m->vehicle);
    change_motion_server(m, MS_NONE);

    if (m->bumper_on) mob_output(m, "... HALT with bumper on ...");
    else mob_output(m, "... HALT with bumper OFF! ...");
}

void mob_resume(mob m)
{
    vehicle_resume(m->vehicle);

    mob_output(m, "... RESUME ...");
}


void    mob_reset(mob m)
{
    mob_halt(m);
    mob_resume(m);
    m->dt_command = 0.1;
    m->odom_on = FALSE;

    scheduler_point_remove(m->scheduler, &(m->us_sensor_tp));
    scheduler_point_remove(m->scheduler, &(m->command_tp));


    scheduler_point_add(m->scheduler, &(m->command_tp),
			scheduler_time(m->scheduler) + 1e-7);

    scheduler_point_add(m->scheduler, &(m->us_sensor_tp), 
			scheduler_time(m->scheduler) + m->dt_us_scan);

    mob_free_us_sensors(m);

    mob_init_us_sensors(m);

    mob_output(m, "... RESET ...");
}

void mob_us_sensor_activation(mob m, int sensor_nr, double time,
			      void *uscp)
{
    us_sensor_description *sensor;
    int i;


#if 0
    printf("sensor_activation with time %f, uscp=%p\n", time, uscp);
#endif
    if (sensor_nr <0 || sensor_nr >=MAX_NR_US_SENSORS)
	error("mob.c: mob_us_sensor_activation() - illegal sensor number");

    sensor = &(m->us_sensor[sensor_nr]);

    if (m->sensor_node_adr[sensor_nr] >= 0) {

	/* sensor installed
	 */

	if (sensor->free_activations<=0) {

	    sensor->free_activations = 10;

	    if (sensor->activation_count>0) {
		sensor->activation_time = realloc(sensor->activation_time,
			       (sensor->activation_count + sensor->free_activations)
			       * sizeof(double));
		sensor->sensor_cp = realloc(sensor->sensor_cp,
			       (sensor->activation_count + sensor->free_activations)
			       * sizeof(void *));
	    }
	    else {
		sensor->activation_time =
			malloc(sensor->free_activations * sizeof(double));
		sensor->sensor_cp = malloc(
			       sensor->free_activations
			       * sizeof(void *));
	    }

            if (sensor->activation_time == 0 || sensor->sensor_cp == 0)
		error("mob.c: mob_us_sensor_activation() - can't alloc memory for"
		      "us-sensor activation times");
        }

        i = sensor->activation_count;
	while ((i>0)  && (time > sensor->activation_time[i-1])) {
	    sensor->activation_time[i] = sensor->activation_time[i-1];
	    sensor->sensor_cp[i] = sensor->sensor_cp[i-1];
            i--;
        }

#ifdef DEBUG_DETAIL
        fprintf(stderr, "index=%d for us-sensor activation at time %f of sensor %d\n",
	        i, time, sensor_nr);
#endif
	sensor->activation_time[i] = time;
	sensor->sensor_cp[i] = uscp; 

	sensor->free_activations = sensor->free_activations - 1;
        sensor->activation_count = sensor->activation_count + 1;
    }
}

void mob_set_us_sensor_function(mob m, void (*us_scan)(mob,int,void *), void *uservar)
{
    m->us_uservar = uservar;
    m->us_scan    = us_scan;
}

void    mob_set_us_sensor_data_function(mob m, void (*free)(), 
					void (*select)())
{
    m->sd_free = free;
    m->sd_select = select;
}


