/*********************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: mob.h
 *  TYPE		: module definition
 *  Time-stamp: <94/11/13 15:34:06 stolz>
 *********************************************************************
 *  MODULE              : mob
 *
 *  FUNCTION
 *
 *	This module provide a mobile robot, a "mob", which understand
 *	the Robuter-Commands of the French RobuterII.
 *	To emulate the timing, the scheduler-module is needed.
 *
 *  EXAMPLE
 *	#include "mob.h"
 *
 *	void com_in(char *input, void *data)
 *	{
 *	    scanf("%s", &input);
 *	    if (input[0] == '!') {
 *	        input[0] = '\0';
 *	        mob_reset((mob)data);
 *	    }
 *	}
 *      void com_out(void *data, char *output)
 *      {
 *	    mob m = (mob)data;
 *          mob_state ms;
 *
 *	    mob_get_motion_state(m, &ms, 0);
 *	    printf("motion-type:%d, x/y/angle=%f/%f/%f\n", 
 *		   motion_type(m), ms.x, ms.y, ms.angle);
 *	    printf("OUTPUT> %s\n", output);
 *	}
 *	void main()
 *	{
 *	    scheduler s = scheduler_new();
 *	    mob aramis = mob_new(s);
 *	
 *	    mob_set_io(aramis, com_in, com_out, mob);
 *	
 *	    scheduler_sync(s, 60.0);  -- drive 1min.
 *	
 *	    mob_free(aramis);
 *	    scheduler_free(s);
 *	}
 *	
 *  NOTE
 *	see also in conmobs.c for an simple usage-example.
 *
 *  HISTORY
 *  20-Apr-94   Creation date.
 *  19-May-94	Add mob_stop() function.
 *
 **********************************************************************/
#ifndef MOB_H
#define MOB_H

#ifndef SCHEDULER_H
#include "scheduler.h"
#endif


/* maximum length of a zero-terminated command string. (incl. '\0')
 */
#define	MAX_COMMAND_LENGTH		300


/* abstract datatype mob
 */
typedef struct adt_mob*			mob;


/* the motion/position-state of a mob
 */
typedef struct {
    double	x, y, angle;		/* position */
    double      v, w;                   /* velocity */
    double	a, u;                   /* acceleration */
    double	angle_fl, angle_fr;	/* position of front-wheels */
} mob_state;


/* the geometric parameters of the mob
 * (see also in vehicle.h)
 */
typedef struct {
    double a, b, c, d;
} mob_parameters;


/* the random disturbance of the mob-velocity
 */
typedef struct {
    double v_low, v_high;
    double w_low, w_high;
} mob_disturbance;


/* the types of motion
 * normaly only the following types are visible:
 * MT_STOP            - mob is stopped
 * MT_ACCEL, MT_DECEL - mob is accelerating or decelerating
 * MT_KILL            - the current motion is killed to MT_STOP
 * MT_HALT            - mob is halted by intern or extern reasons.
 */
enum MotionTypes {
    MT_STOP, MT_START, MT_ACCEL, MT_CONST, MT_CHANGE, MT_DECEL, MT_END, 
    MT_KILL, MT_HALT
};


/* create or free a mob 
 */
mob     mob_new(scheduler *);
mob     mob_new2(scheduler *, mob_parameters);
void    mob_free(mob);


/* set the input/output-functions
 */
void 	mob_set_io(mob m,
			   void (*com_in)(char *, void *uservar),
			   void (*com_out)(void *uservar, char *),

			   void *uservar);

/* extended version also alow the notification of
 * an output-open/close.
 */
void mob_set_io2(mob m,
		   void (*com_in)(char *, void *uservar),
		   void (*com_out)(void *uservar, char *),
		   void (*com_out_open)(void *uservar),
		   void (*com_out_close)(void *uservar),
		   void *uservar);


/* request or set the real/intern current position/velocity
 * use a zero-pointer if the parameter doesn't bother you.
 */
void    mob_get_motion_state(mob, mob_state *real,
					  mob_state *intern);
void    mob_set_motion_state(mob, mob_state *real,
					  mob_state *intern);

/* velocity disturbance which influence the real mob 
 * position and velocity. Default values are 1,1,1,1.
 * So there is no difference between real and intern position.
 */
void	mob_set_vel_disturbance(mob, mob_disturbance);


/* stop the mob and do a soft-reset of the command-interpreter
 */
void    mob_reset(mob);


/* halt the mob and lock the interpreter so 
 * that only a resume or reset can unlock the command-interpreter.
 */
void	mob_halt(mob);
void 	mob_resume(mob);


/* request current motion-type
 */
int	motion_type(mob m);


/* activate a ultrasonic sensor at a specific time.
 * this will not automatically generate a sensor-value.
 * Only if this activation is selected from the 
 * us-sensor-measure-algorithm, the time is used
 * to compute the distance to the obstacle.
 * For this aim, a pointer to a sensor-data is also 
 * stored with the sensor-time.
 * also stores some sensor-function for further processing
 * by extern functions.
 */
void	mob_us_sensor_activation(mob m, int sensor_nr, double time, 
				 void *sensor_data);

/* To set the function which is called for an us-sensor-measuring of
 * an specific sensor-nr use the following routine.
 * If the mob invoke a sensor-measing, the function us-scan
 * will be called with the mob, sensor-nr und uservar as parameters.
 */
void	mob_set_us_sensor_function(mob m, void (*us_scan)(mob,int,void *), 
				   void *uservar);

/* The specified function select is called with the sensor-data
 * as parameter if the sensor-activation is selected, if this 
 * activation don't lead to a selection, then free is called 
 * with the sensor-data as parameter.
 */
void    mob_set_us_sensor_data_function(mob m, void (*free)(), 
					void (*select)());


#endif
