/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: mob_private.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 16:56:16 stolz>
 **************************************************************************
 *  MODULE              : mob
 *
 *  FUNCTION
 *	This header file provide the mob-datatype and
 *	exports private functions of the mob-module to related
 *      modules like command or com_motv, etc.
 *
 *
 *  NOTE
 *
 *  HISTORY
 *  20-Apr-94   Creation date.
 *  16-Jun-94   Add definitions for us-sensors.
 *
 *************************************************************************/
#ifndef MOB_PRIVATE_H
#define MOB_PRIVATE_H

#ifndef MOB_H
#include "mob.h"
#endif

#ifndef VEHICLE_H
#include "vehicle.h"
#endif

#ifndef QUEUE_H
#include "queue.h"
#endif

#ifndef BASIC_H
#include "basic.h"
#endif


/* Motion-Server of the Robuter
 */
enum ServerEnum {MS_NONE, MS_MOTN, MS_MOTV, MS_MOTP};


/*
 * defs for basic MOTN motion
 */
typedef struct {
    double	x_start;	/* information for distance/angle requests */
    double	y_start;
    double	angle_start;
    /* -1->no distance, 1->linear motion, 0->angular motion
     */
    int 	linear_motion;
} motn_info;


/*
 * defs for MOTV motion
 */
#define MOTV_QUEUE_SIZE   10

enum Motv_Motion_Type {MVT_VELOCITY, MVT_DISTANCE,
		       MVT_ANGLE, MVT_TIME, MVT_INFINITE};

typedef struct {
    int type;
    double v, w, a, u, time, distance, angle;

    double ta, tc;
    double v_max, w_max;

    double t_start, x_start, y_start, angle_start;
} motv_motion;

typedef struct {
    queue	motions;
    motv_motion last_motion;
    timepoint   change_tp;
}  motv_info;



/* us-sensor read-modes
 */
enum sensor_mode {SM_VOID, SM_REQUEST, SM_AUTOMATIC};


/* definitions for the us-sensors
 */
#define MAX_NR_US_SENSORS	24

#define MAX_NR_NODES		9
#define MAX_NR_SENSORS_PER_NODE	6
#define MAX_NODE_ADR		(MAX_NR_SENSORS_PER_NODE * MAX_NR_NODES)

#define node_adr(node,adr)	(MAX_NR_SENSORS_PER_NODE*(node)+(adr))

enum us_direction {
	USD_FRONT=1, USD_RIGHT=2, USD_BACK=4, USD_LEFT=8
};

enum us_scan_mode { USSM_LOW, USSM_HIGH, USSM_INACTIVE };



typedef struct
{
    char scan_mode;
    char direction;
    int  emission_angle;		/* in 1/10 degrees */
    double   tof;			/* us-read result in secs */
    double   *activation_time;          /* sorted dynamic array!, from high to low */
    void     **sensor_cp;               /* corresponding sensor-data */

    int      activation_count;
    int	     free_activations;

} us_sensor_description;


/* definition o an mob.
 */
typedef struct adt_mob adt_mob;
struct adt_mob {
    vehicle	vehicle;
    mob_parameters parameters;
    boolean	bumper_on;

    /* command input/output
     */
    double	dt_command;
    double	t_of_next_cmd;
    void	(*com_in)(char *, void *uservar);
    void	(*com_out)(void *uservar, char *);
    void	(*com_out_open)(void *uservar);
    void	(*com_out_close)(void *uservar);
    void 	*com_uservar;
    scheduler	*scheduler;
    timepoint	command_tp;

    int         motion_server;
    union {
	motn_info motn;
	motv_info motv;
    } 		motion;

    boolean	odom_on;
    double      odom_offset_x;
    double      odom_offset_y;
    double      odom_offset_angle;

    int		us_mode;       		/* sensor mode */
    boolean     invoke_us_read;		/* invoke a sensor read */
    boolean	us_read_complete;	/* ->sensor-results are valid */
    double	sensor_invoke_time;
    int         node_invoke;

    int         read_node;		/* node of sensor-read */
    unsigned long sensor_selection;	/* sensor mask of sensor-read */

    us_sensor_description
		us_sensor[MAX_NR_US_SENSORS];
    signed char	sensor_node_adr[MAX_NR_US_SENSORS]; /* sensor_nr -> node+adr */
		/* sensor_node_adr <0 -> sensor not installed */
    signed char	sensor_nr[MAX_NODE_ADR];	    /* node+adr -> sensor_nr */
		/* sensor_nr <0 -> sensor not installed */

    double	dt_us_scan;
    void	(*us_scan)(mob m, int sensor_nr, void *us_uservar);
    void 	*us_uservar;
    timepoint	us_sensor_tp;

    void        (*sd_free)(void *);
    void        (*sd_select)(void *);

    double	us_rand_range;
    double	us_rand_start;
    double	us_dropout_probability;
};



typedef struct {			/* intern motion state */
    double x, y, angle, v, w, a, u;
} mob_motion_state;


/* exported private functions 
 * for output:
 */
void	mob_output(mob m, char *out);
void	mob_output_open(mob m);
void	mob_output_close(mob m);
void	mob_command_delay(mob m, double t);

/* to request the intern position/velocity
 */
void	mob_get_intern_state(mob, mob_motion_state *);

/* motion-server related:
 */
#define motion_server(m) (m->motion_server)
void	change_motion_server(mob m, int ms);
void	free_motion_server(mob m);

/* motion related functions
 * (a motion is a movement of the mob with constant acc during a
 *  specific time-intevall)
 * The motions are stored in a queue!
 * NOTE: this is a basic-motion an has nothing to do with
 *       MOTV or MOTN.
 */
void	motion_add(mob m, double t, double a, double u);
void	motion_kill(mob m);
double	motion_stop_time(mob m); 
void	empty_motion_queue(mob m);



#endif
