/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: mob_world.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:56:25 stolz>
 **************************************************************************
 *  MODULE              : mob_world
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in mob_world.h!
 *
 *  HISTORY
 *  19-May-94   Creation date
 *  30-May-94   Fix a really damned bug. (malloc(mw_object))!
 *  06-Jun-94   Add some mob-specific functions.
 *  13-Jun-94   simplify object creation, change object-structure,
 *              implementation of most routines
 *  16-Jun-94   write mob-code in mwo_add_to_space().
 *  17-Jun-94   write collision-test routines.
 *  19-Jun-94   Split mob_world into mob_world+mw_object+mw_mob.
 *  21-Jun-94   Add actuator-concept.
 *  12-Jul-94   Add remote-nrs for remote-controll objects.
 *  12-Jul-94   Add pvm-server support.
 *  14-Sep-94   Add real-us-computation. (with triangle-projections)
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "mob_world_private.h"
#include "mw_object.h"
#include "mwo_mob.h"
#include "error.h"
#include "strerror.h"
#include "pvm_server.h"

char *mob_world_io_error_string = 0;


int mob_world_remote_nr()
{
    static int act_remote_nr = 0;

    act_remote_nr++;

    return act_remote_nr;
}

void mob_world_set_remote(mob_world mw, int nr, mw_object obj)
{
    if (nr >= mw->min_remote_nr) {

	if (mw->remote_control == 0) {
	    mw->max_remote_nr = mw->min_remote_nr + 50;
	    mw->remote_control = malloc(50 * sizeof(mw_object));
	}

	if (nr >= mw->max_remote_nr) {
	    mw->max_remote_nr += 50;
	    mw->remote_control = realloc(mw->remote_control, 
					(mw->max_remote_nr - mw->min_remote_nr)
					* sizeof(mw_object));
	}
	if (mw->remote_control == 0) {
	    error("mob_world: mob_world_set_remote() - can't alloc memory for remote-index-array");
	}
	    
	mw->remote_control[nr - mw->min_remote_nr] = obj;
    }
}


mw_object mob_world_get_remote(mob_world mw, int nr)
{
    if ((nr>= mw->min_remote_nr) && (nr < mw->max_remote_nr) && 
	mw->remote_control)
	return mw->remote_control[nr - mw->min_remote_nr];
    else return 0;
}


static void mob_world_free_us_results(mob_world mw)
{
    us_sensor_check_point *sr;

    while (queue_first(&sr, mw->us_sensor_results)) {
	if (sr) uscp_free(sr);
    }
}

void mob_world_store_us_result(mob_world mw, us_sensor_check_point *usscp)
{
    if (usscp) {
	if (mw->us_sensor_protocol) {
	    usscp->t = scheduler_time(mw->scheduler);
	    usscp->detailed = TRUE;
	    queue_add(mw->us_sensor_results, &usscp);
	} else uscp_free(usscp);
    }
}

boolean usscp_undetail(void *x)
{
    us_sensor_check_point *usscp = *(us_sensor_check_point **)x;
    boolean b = usscp->detailed;

    usscp->detailed = FALSE;

    return b;
}

static void mob_world_store_mob_checkpoints(mob_world mw)
{
    mw_object obj;
    mob_check_point cp;
    double t = scheduler_time(mw->scheduler);

    for (obj=list_first(mw->mobs); obj; obj=list_next(mw->mobs)) {
	cp.mob = obj;
	cp.t = t;
	mob_get_state(obj, &(cp.state));
	queue_add(mw->mob_check_points, &cp);
    }
}

void mob_world_sync_actuators(mob_world mw)
{
    mw_object obj;
    mw_object controller;

    for (obj=list_first(mw->actuators); obj; obj=list_next(mw->actuators)) {
	if (obj->actuator.nr >=0) {
	    controller = mw_object_controller(obj);
	    if (controller) mw_object_req_actuator(obj, controller);
	}
    }
}

/*
 * private functions
 */
static void mob_collision_test(scheduler *s, timepoint *tp, double t)
{
    mob_world mw = tp->workspace;
    mw_object obj;

#ifdef DEBUG_DETAIL
    fprintf(stderr, "mob_collision_test() at time %f\n", t);
#endif

    mob_world_sync_actuators(mw);

    for (obj=list_first(mw->mobs); obj; obj=list_next(mw->mobs)) {
#ifdef DEBUG_DETAIL
	fprintf(stderr, "Mob Collision test with mob%p, type=%d\n",
		obj, mob_motion_type(obj));
#endif
	if ((mob_motion_type(obj) != MOB_HALTED) && mw_object_collision(obj)) {

#ifdef DEBUG_DETAIL
	    fprintf(stderr, "Collision! Halt mob %p.\n", obj);
#endif
	    
	    mob_emergency_halt(obj);
	}
    }

    scheduler_point_add(s, tp, t + mw->dt_collision_test);
}


/**************************************************************************
 * world oriented functions
 */

mob_world mob_world_new()
{
    mob_world	mw;
    vector3     space_ref = {-10.0, -10.0, -0.05};
    vector3     space_box = {20.0, 20.0, 4.0};
    int         space_dim[3] = {10, 10, 8};

    ps_open();

    mw = malloc(sizeof(adt_mob_world));
    if (mw) {

        mw->scheduler = &(mw->scheduler_struct);
	mw->scheduler->workspace = mw;

	scheduler_new(mw->scheduler, 0.0);

	mw->dt_collision_test = 0.1;			/* 0.01; */
	mw->tp_collision_test.task = 0;
	mw->tp_collision_test.workspace = mw;
	mw->tp_collision_test.free = 0;
	mw->tp_collision_test.function = mob_collision_test;
	scheduler_point_add(mw->scheduler, &(mw->tp_collision_test),
			    scheduler_time(mw->scheduler));

        mw->mobs = list_new();
	mw->objects = list_new();
	mw->actuators = list_new();
	mw->us_sections = list_new();

	mw->mob_check_point_protocol = FALSE;
	mw->us_sensor_protocol = FALSE;
	mw->us_sensor_results = queue_new(sizeof(us_sensor_check_point *));
	mw->mob_check_points = queue_new(sizeof(mob_check_point));
/*
	mw->space = space_new();
*/
	mw->space = space_new2(space_ref, space_box, space_dim);

	mw->us_air_damping = 1.0;   /* m^-1 */
	mw->air_speed = 343.6;		/* m/s */

	mw->max_reflexion_count = 1;
	mw->real_us = FALSE;

	mw->remote_control = 0;
	mw->min_remote_nr  = mob_world_remote_nr();
	mw->max_remote_nr  = mw->min_remote_nr;
    }
    else error("mob_world: mob_world_new() - can't alloc memory for adt mob_world");

    return mw;
}


void mob_world_clear(mob_world mw)
{
    mw_object obj;

    for (obj=list_first(mw->objects); obj; obj=list_first(mw->objects))
	mw_object_free(obj);

    list_remove_all(mw->objects);
    list_remove_all(mw->mobs);
    list_remove_all(mw->actuators);

    list_remove_all(mw->us_sections);

    mob_world_free_us_results(mw);
    queue_free(mw->mob_check_points);
    mw->mob_check_points = queue_new(sizeof(mob_check_point));

    if (mw->remote_control) free(mw->remote_control);
    mw->remote_control = 0;
    mw->min_remote_nr  = mob_world_remote_nr();
    mw->max_remote_nr  = mw->min_remote_nr;
}

void mob_world_free(mob_world mw)
{
    mob_world_clear(mw);

    space_free(mw->space);
    list_free(mw->mobs);
    list_free(mw->objects);
    list_free(mw->actuators);
    list_free(mw->us_sections);
    mob_world_free_us_results(mw);
    queue_free(mw->us_sensor_results);
    queue_free(mw->mob_check_points);
    scheduler_free(mw->scheduler);

    free(mw);
}


void mob_world_simulate_to(mob_world mw, double time)
{
    double t;

    ps_recv(mw); 

    t = scheduler_time(mw->scheduler);

    if (t >= time)
	error("mob_world: mob_world_simulate_to()"
	      " - time-parameter smaller than actual time");

    if (mw->us_sensor_protocol)
	queue_last_to_first(mw->us_sensor_results, usscp_undetail);

    scheduler_sync(mw->scheduler, time);

    mob_world_sync_actuators(mw);

    if (mw->mob_check_point_protocol)
	mob_world_store_mob_checkpoints(mw);
}


double	mob_world_get_time(mob_world mw)
{
    return scheduler_time(mw->scheduler);
}


void mob_world_reset(mob_world mw)
{
    mw_object mob;

    scheduler_reset(mw->scheduler, 0.0);

    scheduler_point_add(mw->scheduler, &(mw->tp_collision_test),
			scheduler_time(mw->scheduler));

    mob = list_first(mw->mobs);
    while (mob) {
	mob_soft_reset(mob);
	mob = list_next(mw->mobs);
    }
}

mob_world mob_world_load(char *filename)
{
    mob_world mw;
    FILE *handle;
    boolean b;

    mob_world_io_error_string = 0;

    handle = fopen(filename, "r");
    if (handle == 0) {
	mob_world_io_error_string = "can't open file for reading";
	return 0;
    }

    mw = mob_world_new();
    b = mw_insert(handle, mw);
    if (b) mob_world_io_error_string = mw_file_error_string;

    fclose(handle);

    if (b) {
	mob_world_free(mw);
	return 0;
    }
    else {
	return mw;
    }
}


boolean	mob_world_save(mob_world mw, char *filename)
{
    FILE *handle;
    boolean b;

    mob_world_io_error_string = 0;

    handle = fopen(filename, "w");
    if (handle == 0) {
	mob_world_io_error_string = "can't open file for writing";

	return TRUE;
    }

    b = mob_world_write(handle, mw);
    if (b) mob_world_io_error_string = mw_file_error_string;

    fclose(handle);

    return b;
}



boolean mob_world_insert_objects(mob_world mw, char *filename)
{
    FILE *handle;
    boolean b;

    mob_world_io_error_string = 0;

    handle = fopen(filename, "r");
    if (handle == 0) {
	mob_world_io_error_string = "can't open file for reading";
	return TRUE;
    }

    b = mw_insert(handle, mw);
    if (b) mob_world_io_error_string = mw_file_error_string;

    fclose(handle);

    return b;
}


boolean	mob_world_is_empty(mob_world mw)
{
    return list_is_empty(mw->objects);
}

list	mob_world_objects(mob_world mw)
{
    return mw->objects;
}

void mob_world_set(mob_world mw, int select, double value)
{
    switch(select) {
    case MWS_AIR_SPEED:
	mw->air_speed = value;
	break;
    case MWS_US_AIR_DAMPING:
	mw->us_air_damping = value;
	break;
    case MWS_COLLISION_STEP:
	mw->dt_collision_test = max(value, 1.0e-6);
	break;
    case MWS_MAX_REFLEXION_COUNT:
	mw->max_reflexion_count = max((int)value, 0);
	break;
    default:
	warning("mob_world: mob_world_set() - unknown select code");
    }
}

double	mob_world_get(mob_world mw, int select)
{
    switch(select) {
    case MWS_AIR_SPEED:
	return mw->air_speed;
	break;
    case MWS_US_AIR_DAMPING:
	return mw->us_air_damping;
	break;
    case MWS_COLLISION_STEP:
	return mw->dt_collision_test;
	break;
    case MWS_MAX_REFLEXION_COUNT:
	return (double)(mw->max_reflexion_count);
	break;
    default:
	warning("mob_world: mob_world_get() - unknown select code");
    }
    return 0.0;
}

void mob_world_setb(mob_world mw, int select, boolean value)
{
    switch(select) {
    case MWB_MOB_PROTOCOL:
	if (value) {
	    if (! mw->mob_check_point_protocol) {
		mob_world_store_mob_checkpoints(mw);
	    }
	}
	else {
	    if (mw->mob_check_point_protocol) {
		queue_free(mw->mob_check_points);
		mw->mob_check_points = queue_new(sizeof(mob_check_point));
	    }
	}
	mw->mob_check_point_protocol = value;
	break;

    case MWB_US_SENSOR_PROTOCOL:
	if ((!value) && mw->us_sensor_protocol) {
	    mob_world_free_us_results(mw);
	}
	mw->us_sensor_protocol = value;
	break;

    case MWB_REAL_US:
	mw->real_us = value;
	break;

    default:
	warning("mob_world: mob_world_setb() - unknown select code");
    }
}


boolean mob_world_getb(mob_world mw, int select)
{
    switch(select) {
    case MWB_MOB_PROTOCOL:
	return mw->mob_check_point_protocol;
	break;
    case MWB_US_SENSOR_PROTOCOL:
	return mw->us_sensor_protocol;
	break;
    case MWB_REAL_US:
	return mw->real_us;
    default:
	warning("mob_world: mob_world_getb() - unknown select code");
    }
    return FALSE;
}
