/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: mobs_interface.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/03/03 15:17:01 stolz>
 **************************************************************************
 *  MODULE              : mobs_interface
 *
 *  IMPORT		: pvm3, stdio, stdlib, string, time
 *
 *  IMPLEMENTATION
 *
 *	This module use PVM to communicate with the mobile robot
 *	simulator(MOBS). Its aim is to make a connection to the
 *	simulator an assign to an object with a specific name.
 *      Commands can be send to the object. The reply-string
 *	Is the return-message of this object.
 *
 *	The communication scheme looks like:
 *
 *	INTERFACE  (msgtag)			  (msgtag)    MOBS
 *
 *	| mi_open()                                             |
 *	|      MID_OPEN -- name,interface_tid -->               |
 *      |               <--------- id ----------- MID_ACCEPT    |t_sim=0
 * -----|                                                       |
 *  |                                                           |
 *  dt_cmd                                                      |
 *  |                                                           |
 * -----|                                                       |
 *	| mi_send()                                             |
 *	|	MI_SEND	---- id, dt_cmd, cmd --->               |-----
 *      |                                                           |
 *      |                                                          dt_sim
 *      |                                                           |
 * 	|		<---- dt_sim, reply ----- MID_REPLY     |-----
 * -----|                                                       |
 *  |                                                           |
 *  dt_cmd                                                      .
 *  |                                                           .
 * -----|                                                       .
 * 	| mi_send()                                             .
 *	|	MI_SEND	---- id, dt_cmd, cmd --->
 *	|		<---- dt_sim, reply ----- MID_REPLY
 * -----|
 *  |
 *  dt_cmd
 *  |
 * -----|
 *	| mi_get_camera_frame
 *	|	MI_SEND	---- id, dt_cmd, camera_nr --->         |-----
 *      |                    width,height,color                     |
 *      |                                                          dt_sim
 *      |                                                           |
 * 	|     <---- dt_sim, frame_size, frame ----- MID_REPLY   |-----
 *      |   (frame is an array of uchar with the length frame_size)
 *      |   (the frame come's not necessarily from the MOBS!)
 * -----|
 *  |
 *  dt_cmd
 *  |
 * -----|
 *	| mi_close
 *	|      MI_CLOSE ------ id, dt_cmd ------>
 *      |
 *      |
 *    time
 *
 *  NOTE
 *      The module specification and definition is given in
 *	mobs_interface.h!
 *
 *	Don't define DEBUG_DETAIL for compiling mobs_interface.c if you
 *	wan't get debugging information on stdout.
 *
 *  HISTORY
 *  06-Jul-94   Creation date.
 *  22-Sep-94   mi_open returns now 0 if the object don't exist.
 *  29-Sep-94   Add get_frame() function.
 *  03-Mar-95   Use gettimeofday() instead clock() to calculate the delta-t
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>

#include "pvm3.h"

#include "mobs_interface.h"
#include "mobs_interface_def.h"

/*
#define DEBUG_DETAIL
*/

/* local defines for the simulator-server
 */
#define SERVERNAME  "mobs"
#define HOSTNAME    ""



/* private definition of the adt mobs_obj
 */
struct mobs_obj_adt {
    int      mobs_tid;
    int      interface_tid;

    char    *name;
    long     id;

    struct timeval last_recv_time;
    double   sim_dt;
};
typedef struct mobs_obj_adt mobs_obj_adt;




/* local variable for the management of reply-strings
 * of any size. Only used by mi_send().
 */
static char *reply_string = 0;
static char reply_length = 0;


/* defines for timing-computations
 */
#define after(t1,t2) ((t1)->tv_sec > (t2)->tv_sec || ( (t1)->tv_sec == (t2)->tv_sec && (t1)->tv_usec > (t2)->tv_usec ))
#define diff(t1,t2)  (((double)(t1)->tv_sec - (t2)->tv_sec) + (double)((t1)->tv_usec - (t2)->tv_usec)*1e-6)




static void clone_string(char **sp, char *string)
/**************************************************************************
 * local routine to copy a string.
 */
{
    int len;

    if ((len=strlen(string))>0) {
	*sp = malloc(len+1);
	if (*sp == 0) {
	    fprintf(stderr, "mobs_interface: clone_string() - can't alloc memory\n");
	    exit(10);
	}
	strcpy(*sp, string);
    }
    else *sp = 0;
}


mobs_obj mi_open(char *objectname, int mobs_tid)
/**************************************************************************
 */
{
    mobs_obj mo = 0;
    int mytid;

    /* allocate mobs_obj-structure
     * and set some default-values
     */
    mo = malloc(sizeof(mobs_obj_adt));
    if (mo==0) return 0;

    mo->sim_dt = 0.0;
    clone_string(&(mo->name), objectname);

    /* check own tid-number.
     */
    mo->interface_tid = pvm_mytid();
    if (mo->interface_tid < 0) {
	fprintf(stderr, "mobs_interface: mi_open() - can't request my pvm-tid\n");
	goto error;
    }

    /* use given mobs_tid or use parent-tid as simulator.
     */
    if (mobs_tid == -1) mobs_tid = pvm_parent();

    if (mobs_tid >= 0) {
	mo->mobs_tid = mobs_tid;
	if (pvm_pstat(mobs_tid) != PvmOk) {
	    fprintf(stderr, "mobs_interface: mi_open() - mobs-tid %d is not valid\n", mobs_tid);
            goto error;
	}
    }
    else {
	fprintf(stderr, "mobs_interface: mi_open() - no legal simulator-tid\n");
	goto error;
    }


    /* sending object-name and tid-number to the simulator
     */
#ifdef DEBUG_DETAIL
    printf("mi_open() - send obj-name '%s' and interface-tid %d\n",
	   mo->name, mo->interface_tid);
#endif
    pvm_initsend(PvmDataDefault);
    pvm_pkint(&(mo->interface_tid), 1, 1);
    pvm_pkstr(mo->name);
    pvm_send(mo->mobs_tid, MID_OPEN);

    /* receiving acception and object-id for further references.
     */
#ifdef DEBUG_DETAIL
    printf("mi_open() - wait for acception\n");
#endif

    pvm_recv(-1, MID_ACCEPT);
    pvm_upklong(&(mo->id), 1, 1);

#ifdef DEBUG_DETAIL
    printf("mi_open() - given. id = %ld\n", mo->id);
#endif

    if (mo->id == 0) goto error;

    /* store actual clock for the delta-time computation.
     */
    gettimeofday(&(mo->last_recv_time), (struct timezone*)0);

    return mo;

error:
    free(mo);
    return 0;
}


void mi_close(mobs_obj mo)
/**************************************************************************
 */
{
    if (mo) {
	struct timeval now;
	double dt;

	/* compute time difference between last receive and now.
	 */
	gettimeofday(&now, (struct timezone*)0);
	if (after(&now, &(mo->last_recv_time)))
	    dt = diff(&now, &(mo->last_recv_time));
	else dt = 0.0;

	dt += mo->sim_dt;

	/* send id and dt to the simulator for closing the line
	 */
#ifdef DEBUG_DETAIL
	printf("mi_close() - send id %ld and dt %f\n to sim_tid %d",
		mo->id, dt, mo->mobs_tid);
#endif
	pvm_initsend(PvmDataDefault);
	pvm_pklong(&(mo->id), 1, 1);
	pvm_pkdouble(&dt, 1, 1);
	pvm_send(mo->mobs_tid, MID_CLOSE);

	if (mo->name) free(mo->name);
	free(mo);
    }
}


void mi_send(mobs_obj mo, char *cmd, char *reply, int max_reply_size)
/**************************************************************************
 */
{
    struct timeval now;
    double dt;
    int len;

    if (mo) {

	/* compute time difference between last receive and now.
	 */
	gettimeofday(&now, (struct timezone*)0);
	if (after(&now, &(mo->last_recv_time)))
	    dt = diff(&now, &(mo->last_recv_time));
	else dt = 0.0;

	dt += mo->sim_dt;

	if (cmd==0) {
	    cmd = "";
	    len = 0;
	}
	else len = strlen(cmd);

	/* send id, dt and cmd-string to the simulator
	 */
#ifdef DEBUG_DETAIL
	printf("mi_send() - now=%ld, last=%ld\n, dt=%fs\n",
	       now, mo->last_recv_time, dt);

	printf("mi_send() - send id %ld, dt %f, cmd-len %d, cmd '%s' to mobs-tid %d\n",
	       mo->id, dt, len, cmd, mo->mobs_tid);
#endif
	pvm_initsend(PvmDataDefault);
	pvm_pklong(&(mo->id), 1, 1);
	pvm_pkdouble(&dt, 1, 1);
	pvm_pkint(&len, 1, 1);
	if (len>0) pvm_pkstr(cmd);
	pvm_send(mo->mobs_tid, MID_SEND);


	/* receive sim-dt
	 */
#ifdef DEBUG_DETAIL
	printf("mi_send() - receive reply\n");
#endif
	pvm_recv(-1, MID_REPLY);
	pvm_upkdouble(&(mo->sim_dt), 1, 1);
	pvm_upkint(&len, 1, 1);

	if (len<0) len = 0;

	if (len >= reply_length) {
            if (reply_string) free(reply_string);
	    reply_string = malloc(2 * len + 10);
	    if (reply_string==0) {
		fprintf(stderr, "mobs_interface: mi_send() - can't alloc memory for reply-string");
		exit(20);
	    }
	}
	if (len>0) pvm_upkstr(reply_string); else reply_string[0] = '\0';

#ifdef DEBUG_DETAIL
	printf("mi_send() - reply is sim_dt = %fs, reply-len %d, reply '%s'\n", 
		mo->sim_dt, len, reply_string);

	printf("mi_send() - done.\n");
#endif
	if (reply && max_reply_size>0)
	    strncpy(reply, reply_string, max_reply_size);

	gettimeofday(&(mo->last_recv_time), (struct timezone*)0);
    }
}


long mi_get_camera_frame(unsigned char *frame, mobs_obj mo, 
			 int camera_nr, int width, int height, int color)
/**************************************************************************
 */
{
    if (mo && frame && width>0 && height>0) {
	struct timeval now;
	double dt;
	long frame_size;

	/* compute time difference between last receive and now.
	 */
	gettimeofday(&now, (struct timezone*)0);
	if (after(&now, &(mo->last_recv_time)))
	    dt = diff(&now, &(mo->last_recv_time));
	else dt = 0.0;

	dt += mo->sim_dt;

#ifdef DEBUG_DETAIL
	printf("mi_get_camera_frame() - now=%ld, last=%lds\n, dt=%fs\n",
	       now, mo->last_recv_time, dt);
	
	printf("mi_get_camera_frame() - camera %d, width %d, height %d, color %d\n",
	       camera_nr, width, height, color);
#endif

	pvm_initsend(PvmDataDefault);
	pvm_pklong(&(mo->id), 1, 1);
	pvm_pkdouble(&dt, 1, 1);
	pvm_pkint(&camera_nr, 1, 1);
	pvm_pkint(&width, 1, 1);
	pvm_pkint(&height, 1, 1);
	pvm_pkint(&color, 1, 1);
	pvm_send(mo->mobs_tid, MID_REQ_FRAME);
	

	pvm_recv(-1, MID_FRAME);
	pvm_upkdouble(&(mo->sim_dt), 1, 1);
	pvm_upklong(&frame_size, 1, 1);

#ifdef DEBUG_DETAIL
	printf("mi_get_camera_frame() - frame has sim_dt = %fs, frame-size %ld\n",
		mo->sim_dt, frame_size);
#endif

	if (frame_size>0) {
	    pvm_upkbyte(frame, frame_size, 1);
	}

	gettimeofday(&(mo->last_recv_time), (struct timezone*)0);
	
	return frame_size;
    }
    else return 0;
}

