/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: mobs_interface.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 16:57:05 stolz>
 **************************************************************************
 *  MODULE              : mobs_interface
 *
 *  FUNCTION
 *
 *	This module provides the pvm-routines for connection to "MOBS",
 *	the Mobile rOBot Simulator.
 *
 *      First, to connect to an named-object of the simulator, call
 *	mi_open(object_name, simulator_tid). The simulator_tid should
 *	be a valid tid-number of an running MOBS-simulator or -1 for
 *      using the parent-tid for the simulator-process. Use the
 *	return-value of this function for further references to this
 *	object. A 0 as return-value indicates an error. Serious errors
 *      report also an error-message on stderr!
 *
 *      To send a command, in string-form, to the opened object the
 *	function mi_send(obj, cmd-string, reply, max_reply_len) is
 *	available. You get the reply-message in the <reply>-array.
 *	The maximum length of the reply which is copied into <reply>
 *	is max_reply_length without the 0-Termination of the string.
 *	(so make your array 1 element longer!)
 *
 *      Instead of to send a command, its also possible to request 
 *	a camera-frame. This is done by mi_get_camera_frame().
 *	It's up to the user to provide sufficent space for the
 *	frame. This is [width*height] for an grayscale-frame which
 *	is requested by color=0, and [width*height*3] for an color
 *	frame (parameter color=1). The Functions returns the length
 *	of the grabed frame. A null indicates a male-function. For 
 *      example due to a missed camera or the failure to render a 
 *	image. The Pixel-order in the frame is from top->buttom
 *	and left->right. The color-frame-pixel consists of the 
 *	elements red, green and blue. Possible values range from
 *	0(dark) to 255(bright).
 *
 *	To close the connection to an MOBS-object, simply call
 *	mi_close(obj). And don't forget this!
 *
 *
 *  EXAMPLE
 *
 *	mobs_obj obj;
 *	unsigned char frame[64*64];
 *
 *	obj = mi_open("portos", -1);
 *	if (obj) {
 *	    mi_send(obj, "MOTV ON", 0, 0);
 *	    mi_send(obj, "MOTV LA V+ D=2500", 0, 0);
 *          if (mi_get_camera_frame(frame, mobs_obj, 1, 64, 64, 0)==0) {
 *	        printf("No camera#1 mounted on the robot\n");
 *	    }
 *	    mi_close(obj);
 *      }
 *	else printf("Can't open portos\n");
 *
 *
 *  NOTE
 *      Link this module with -lpvm3. This headerfile can also be
 *      included from C++ sources.
 *	Don't define DEBUG for compiling mobs_interface.c if you
 *	wan't get debugging information on stdout.
 *
 *  HISTORY
 *  06-Jul-94   Creation date.
 *  11-Jul-94   Change intern Protocol to the simulator. Write remarks.
 *  29-Sep-94   Add get_frame() function.
 *
 **************************************************************************/
#ifndef MOBS_INTERFACE_H
#define MOBS_INTERFACE_H


#ifdef __cplusplus
extern "C" {
#endif

/************************************************************************* 
 * abstract datatype mobs_obj
 */
typedef struct mobs_obj_adt *mobs_obj;


/*************************************************************************
 * functions which works on mobs_obj
 */

mobs_obj mi_open(char *objectname, int mobs_tid);

void     mi_send(mobs_obj object, char *cmd, char *reply, 
		 int max_reply_size);

long     mi_get_camera_frame(unsigned char *frame, mobs_obj mo, 
			     int camera_nr, int width, int height, 
			     int color);

void     mi_close(mobs_obj);


#ifdef __cplusplus
}
#endif

#endif



