/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: mw_file.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:57:41 stolz>
 **************************************************************************
 *  MODULE              : mw_file
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in mw_file.h!
 *
 *  HISTORY
 *  25-Jun-94   Creation date
 *  29-Jun-94   Finish Read/Write routines (yet), ToDo: mob-geometry!
 *              Use mw_test_io for a quick test.
 *  12-Jul-94   Write orientations as Quaternion-values q0 q1 q2 q3
 *  14-Jul-94   Eleminate Error in mw_insert() function. 
 *  14-Sep-94   Add r/w of max_reflexion count and camera-defs.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>

#include "app2d.h"
#include "mw_file.h"
#include "rev.h"
#include "error.h"
#include "strerror.h"



/*********************************************************************
 * the Mobs File Format:
 * (in EBNF)
 *
 * token-types: 
 * { } " keywords, <int>, <double>, <string>, <vector3>, <quaternion>
 * 

# Remark to end of line!

<MOF-file> = [<world>|<object>]

<world> =  world { 
             [dt_collision_test <double>]
             [air_speed <double>]
             [us_air_damping <double>]
             [objects {
             [<object>]+
             }]
           }

<object> = (group|triangle|rectangle|cuboid) {
             [<obstacle>|<mob>|<sensor>|<camera>]
           [name "<string>"]
           [position <vector3>]
           [orientation <quaternion>]
           [us_reflexion_damping <double>]
           [marked (true|false)]
           [<appearance>]
	   [visualization "<string>"]
           [actuator <nr>]
           [<members>|					# group
            edges <vector3> <vector3> <vector3>|	# triangle
           xy_size <double> <double>			# rectangle
           edges_length <vector3>]			# cuboid
}

<members> = members { [<object>]+ }

<appearance> = appearance {
                 nr_points <int>
                 [code_length <int>
                 points {
                   [vector]+
                 }
                 code [<int>]+]
               }

<obstacle> = obstacle

<mob> = mob {
          geometry <a:double> <b:double> <c:double> <d:double> 
	           <r1:double> <r2:double>
        }

<sensor> = sensor {
             nr <int>
           }
 
<camera> = camera {
             nr <int>
             angle <float>
             focus_distance <float>
             aspect_ration <float>  
           }

<vector3> = <double> <double> <double>

<quaternion> = <double> <double> <double> <double>

keyword = (a..z|A..Z) [a..z|A..Z|_]+

 *********************************************************************
 */
 

static int indent_pos = 0;

#define write_error      {mw_file_error(fh);return TRUE;}

#define write_char(c)    {if(fputc((c),fh)!=(c))write_error}
#define write_string(s)  {if(fprintf(fh,s)<0)write_error}
#define write_double(d)  {if(fprintf(fh," %f", d)<0)write_error}
#define write_int(i)     {if(fprintf(fh," %d", i)<0)write_error}
#define write_v3(v)      {if(fprintf(fh," %f %f %f", v[0], v[1],v[2])<0)write_error} 
#define write_indent     {if(indent(fh))write_error} 
#define write_nl         {if(fputc('\n',fh)!='\n')write_error}
#define write_open_i     {if(indent_open(fh))write_error}
#define write_close_i    {if(indent_close(fh))write_error}
#define write_boolean(b) {if(fprintf(fh,((b)?" true":" false"))<0)write_error}
#define write_2d(app2d)  {if((app2d)){if(mw_write_app2d(fh,app2d))write_error}} 
#define write_q(q);      {if(quaternion_write(fh, q))write_error}
#define write_obj(obj)   {if(mw_object_write(fh,obj))write_error}

char *mw_file_error_string = 0;


static void mw_file_error(FILE *fh)
{
    indent_pos = 0;

    mw_file_error_string = strerror(ferror(fh));
    if (mw_file_error_string==0)
	mw_file_error_string = "unkown error?";
}

boolean quaternion_write(FILE *fh, quaternion q)
{
    vector3 axis;
/*
    double angle;

    q_get_av(&angle, &axis, q);

    write_double(angle*180/PI);
    write_v3(axis);
*/
    write_double(q[0]);
    v3_set(axis, q[1], q[2], q[3]);
    write_v3(axis);

    return FALSE;
}

boolean indent(FILE *fh)
{
    int i;

    write_nl;

    for (i=indent_pos; i>0; i--)
	write_char(' ');

    return FALSE;
}

boolean indent_open(FILE *fh)
{
    indent_pos++;

    write_char(' ');
    write_char('{');

    return FALSE;
}

boolean indent_close(FILE *fh)
{
    if (indent_pos>0) indent_pos--;

    write_indent;
    write_char('}');

    return FALSE;
}


boolean mw_write_app2d(FILE *fh, app_2d *app)
{
    int i;

    if (app==0) return FALSE;

    write_indent; write_string("appearance"); write_open_i;

    write_indent; write_string("nr_points"); write_int(app->nr_points);
    if (app->nr_points>0) {
	int len;

	write_indent; write_string("code_length"); 
	len = app2d_code_length(app);
	write_int(len);

	write_indent; write_string("points"); write_open_i;
	for (i=0; i< app->nr_points; i++) {
	    write_indent; write_v3(app->point[i]);
        }
        write_close_i;

	write_indent; write_string("code");
	i=0;
	while (i < len) {
	    write_int(app->code[i]);
	    i++;
        }
    }

    write_close_i;
 
    return FALSE;
}

boolean mw_object_write(FILE *fh, mw_object obj)
{
    char *name;
    
    
    mw_file_error_string = 0;

    write_indent; 

    switch (obj->type) {
    case MWO_GROUP: write_string("group"); break;
    case MWO_TRIANGLE: write_string("triangle"); break;
    case MWO_RECTANGLE:	write_string("rectangle"); break;
    case MWO_CUBOID:	write_string("cuboid"); break;
    default:
	warning("mw_file: mw_object_write() - can't write unknown object type");
	return FALSE;
	break;
    }
    write_open_i;

    write_indent;
    switch (obj->function) {
    case MWF_OBSTACLE:	
	write_string("obstacle");
	break;
	
    case MWF_MOB:
	write_string("mob"); write_open_i;
	write_indent; write_string("geometry");

	write_double(obj->func.mob.a);
	write_double(obj->func.mob.b);
	write_double(obj->func.mob.c);
	write_double(obj->func.mob.d);
	write_double(obj->func.mob.r1);
	write_double(obj->func.mob.r2);
	write_close_i;
	break;

    case MWF_SENSOR:
	write_string("sensor"); write_open_i;
	write_indent; write_string("nr"); write_int(obj->func.sensor.sensor_nr);
	write_close_i;
	break;

    case MWF_CAMERA:
	write_string("camera"); write_open_i;
	write_indent; write_string("nr"); write_int(obj->func.camera.camera_nr);
	write_indent; write_string("angle");
	write_double(obj->func.camera.angle);
	write_indent; write_string("focus_distance");
	write_double(obj->func.camera.focus_distance);
	write_indent; write_string("aspect_ratio");
	write_double(obj->func.camera.aspect_ratio);
	write_close_i;
	break;

    default:
	error("mw_file: mw_object_write() - unknown object function");
	return FALSE;
	break;
    }
	name = mw_object_get_name(obj);
	if (name) {
	    write_indent; write_string("name ");
    	write_char('"'); write_string(name); write_char('"'); 
    }

    write_indent; write_string("position");
    write_v3(obj->position);

    write_indent; write_string("orientation");
    write_q(obj->orientation);

    write_indent; write_string("color");
    write_double(obj->r);
    write_double(obj->g);
    write_double(obj->b);

    write_indent; write_string("us_reflexion_damping");
    write_double(obj->us_reflexion_damping);

    write_indent; write_string("marked");
    write_boolean(obj->marked);

    write_2d(obj->appearance_2d);

    if (obj->appearance_3d) {
	write_indent; write_string("visualization ");
    	write_char('"'); write_string((char *)(obj->appearance_3d));
	write_char('"'); 
    }

    if (obj->actuator.nr >= 0) {
	write_indent; write_string("actuator"); write_int(obj->actuator.nr);
    }

    switch (obj->type) {
    case MWO_GROUP:
        if (list_is_not_empty(obj->obj.group.members)) {
	    mw_object o;
	    
	    write_indent; write_string("members"); write_open_i;

	    o = list_first(obj->obj.group.members);
	    while (o) {
	        write_obj(o);
		o = list_next(obj->obj.group.members);
	    }

	    write_close_i;
	}
	break;
    case MWO_TRIANGLE:
	write_indent; write_string("edges"); write_open_i;
	write_indent; write_v3(obj->obj.triangle.edges[0]);
	write_indent; write_v3(obj->obj.triangle.edges[1]);
	write_indent; write_v3(obj->obj.triangle.edges[2]);
	write_close_i;
	break;

    case MWO_RECTANGLE:
	write_indent; write_string("xy_size");
	write_double(obj->obj.rectangle.x_size);
	write_double(obj->obj.rectangle.y_size);
	break;

    case MWO_CUBOID:
	write_indent; write_string("edges_length"); 
	write_v3(obj->obj.cuboid.edges_length);
	break;
    default:
	break;
    }

    write_close_i;

    write_nl;

    return FALSE;
}

/* liefert TRUE bei einem Fehler!
 */
boolean mob_world_write(FILE *fh, mob_world world)
{
    indent_pos = 0;
    mw_file_error_string = 0;

    write_string("# written by module mw_file: ");
    write_string(project_id_string);
    write_string(", ");
    write_string(project_version_string);
    write_nl;

    write_string("world");
    write_open_i;

    write_indent; write_string("dt_collision_test"); 
    write_double(world->dt_collision_test);

    write_indent; write_string("air_speed");
    write_double(world->air_speed);

    write_indent; write_string("us_air_damping");
    write_double(world->us_air_damping);

    write_indent; write_string("max_reflexion_count");
    write_int(world->max_reflexion_count);

    if (!mob_world_is_empty(world)) {
	mw_object obj;

	write_indent; write_string("objects");
	write_open_i;
	
	for (obj=list_first(world->objects); obj; obj=list_next(world->objects))
	    write_obj(obj);

	write_close_i;
	
    }

    write_close_i;
    write_nl;

    return FALSE;
}




enum tokentypes {
  TOKEN_EOF, TOKEN_STRING, TOKEN_NUM, 
  TOKEN_OPEN_I, TOKEN_CLOSE_I, TOKEN_QUOTED_STRING
};

static long line_nr = 0;

#define MAX_TOKEN_LEN	100

static char token_string[MAX_TOKEN_LEN+1];
static int token = TOKEN_EOF;
static boolean token_unread = FALSE;
static long token_int;
static double token_double;


/* boolean expression
 */
#define token_is_close_i (token==TOKEN_CLOSE_I)
#define token_is_open_i  (token==TOKEN_OPEN_I)
#define token_is_eof     (token==TOKEN_EOF)
#define token_is_string  (token==TOKEN_STRING)
#define token_is_qstring (token==TOKEN_QUOTED_STRING)
#define token_is_int     (token==TOKEN_NUM)
#define token_is_double  (token==TOKEN_NUM)


#define read_error(txt)	{read_error_txt(txt);goto exception;}

#define read_token	{if(token_read(fh))goto exception;}
#define unread_token	token_unread=TRUE;

#define read_open_i 	{read_token;if(!token_is_open_i)read_error("{ expected");}
#define read_close_i	{read_token;if(!token_is_close_i)read_error("} expected");}
#define read_object(mw,head)	if(mw_insert_object(fh,mw,head))goto exception;
#define read_world(mw)	if(mw_insert_world(fh,mw))goto exception;
#define read_int        {read_token;if(!token_is_int)read_error("integer value expected");}
#define read_double	{read_token;if(!token_is_double)read_error("float value expected");}
#define read_quoted_string {read_token;if(!token_is_qstring)read_error("quotated string expected");}
#define read_vector(v)     {if(read_v3(fh,v))goto exception;}
#define read_quaternion(q) {if(read_q(fh,q))goto exception;}
#define read_boolean(b)    {if(read_b(fh,b))goto exception;}
#define read_app2d(a)      {if(read_2d(fh,a))goto exception;}
#define read_c(c)          {c=fgetc(fh);if(c=='\n')line_nr++;}


static void read_error_txt(char *txt)
{
    if (txt==0) txt = "unknown read error?";

    token_string[0] = '\0';
    sprintf(&token_string[0], "line %ld: %s", line_nr, txt);

    mw_file_error_string = token_string;
/*
    mw_file_error_string = txt;
*/
}

static boolean token_read(FILE *fh)
{
    int i = 0;
    int c;
    boolean alpha, num, punkt;

    if (token_unread) {
	token_unread = FALSE;
	return FALSE;
    }

    /* read over space and remarks
     */
    token = TOKEN_EOF;
    read_c(c);
    while(1) {
	while (isspace(c)) {
	    read_c(c);
	}
	
	if (c=='#') {
	    read_c(c);
	    while (c!=EOF) {
		if (c=='\n') break;
		read_c(c);
	    }
	}
	else break;
    }

    switch(c) {
    case '{': 
#ifdef DEBUG_DETAIL
	printf("token({)\n", token_string);
#endif	
	token = TOKEN_OPEN_I; 
	return FALSE;
    case '}': 
	token = TOKEN_CLOSE_I; 
#ifdef DEBUG_DETAIL
	printf("token(})\n", token_string);
#endif	
	return FALSE;
    case EOF:
	token = TOKEN_EOF;
	return FALSE;
    default: 
	break;
    }
    
    i = 0;
    if (c=='"') {
	token = TOKEN_QUOTED_STRING;

	read_c(c);
	while (c!='"') {
	    if (c==EOF) read_error("quoted string end with EOF");
	    token_string[i] = c;
	    i++;
	    if (i>=MAX_TOKEN_LEN)
		read_error("quoted string to large");
	    read_c(c);
	}
	token_string[i] = '\0';

#ifdef DEBUG_DETAIL
	printf("token(quoted string) %s\n", token_string);
#endif	

	return FALSE;
    }

    alpha = FALSE;
    num = FALSE;
    punkt = FALSE;

    while(1) {
	if (isdigit(c)) num = TRUE;
        else if ((isalpha(c)||c=='_') && !num) {alpha = TRUE; c = tolower(c);}
	else if (c=='.' && !punkt) punkt = TRUE;
	else if (c=='-' || c=='+') {
	    if (i==0 || token_string[i-1]!='e') {}
	    else read_error("syntax error in int/float expression");
	}
	else if (isspace(c)) break;
	else if (c==EOF) break;
	else read_error("illegal token");
	
        token_string[i] = c;
	i++;
	if (i>=MAX_TOKEN_LEN) read_error("string to large");
	read_c(c);
    }

    token_string[i] = '\0';

    if (alpha) {
	token = TOKEN_STRING;
#ifdef DEBUG_DETAIL
	printf("token(string) %s\n", token_string);
#endif	
    }
    else if (num && !punkt) {
	token = TOKEN_NUM;
	token_int = atol(token_string);
	token_double = token_int;

#ifdef DEBUG_DETAIL
	printf("token(int) %ld\n", token_int);
#endif	
    }
    else if (num && punkt) {
	token = TOKEN_NUM;
	token_double = atof(token_string);
	token_int = token_double;
#ifdef DEBUG_DETAIL
	printf("token(double) %f\n", token_double);
#endif	
    }
    else read_error("syntax error - can't identify token as string/int/float or  \"{}");

    if (c==EOF && !feof(fh)) read_error(strerror(ferror(fh)));

    return FALSE;

exception:
    return TRUE;
}

boolean string_cmp(const char *string)
{
    int i;

    if (!token_is_string) return FALSE; 

    i = 0;
    while (token_string[i] == string[i]) {
	if (string[i]==0) return TRUE;
	i++;
    }
    return FALSE;
}


static boolean read_v3(FILE *fh, vector3 *v)
{
    read_token;
    if (!token_is_double) read_error("3-vector expected");
    (*v)[0] = token_double;

    read_token;
    if (!token_is_double) read_error("3-vector expected");
    (*v)[1] = token_double;

    read_token;
    if (!token_is_double) read_error("3-vector expected");
    (*v)[2] = token_double;

    return FALSE;

 exception:
    return TRUE;
}



static boolean read_q(FILE *fh, quaternion *q)
{
    double angle;
    vector3 axis;

    read_token;
    if (!token_is_double) read_error("angle in degrees expected");
    angle = token_double;

    read_vector(&axis);
/*
    q_set_av(q, angle*PI/180.0, axis);
*/
    (*q)[0] = angle;
    (*q)[1] = axis[0];
    (*q)[2] = axis[1];
    (*q)[3] = axis[2];

    return FALSE;

 exception:
    return TRUE;
}


static boolean read_b(FILE *fh, boolean *b)
{
    read_token;
    if (token_is_string) {
	if (string_cmp("true")) *b = TRUE;
	else if (string_cmp("false")) *b = FALSE;
	else read_error("boolean value (true or false) expected.");
    }
    else if (token_is_int) {
	*b = (token_int>0);
    }
    else read_error("boolean value (true or false) expected.");

    return FALSE;

 exception:
    return TRUE;
}


static boolean read_2d(FILE *fh, app_2d **app)
{
    read_token;
    if (string_cmp("appearance")) {
	int nr_points = 0;
	int code_length = 0;

	read_open_i;
	read_token;
	if (!string_cmp("nr_points")) read_error("keyword 'nr_points' expected.");
	read_int;
	nr_points = max(nr_points, token_int);

	read_token;
	if (string_cmp("code_length")) {
	    read_int;
	    code_length = max(code_length, token_int);
	    if (nr_points==0) code_length = 0;
	}
	else unread_token;

	*app = app2d_new(nr_points, code_length);
	if (nr_points>0) {
	    int i;
	    vector3 v;

	    read_token;
	    if (!string_cmp("points"))
		read_error("keyword 'points' expected.");

	    read_open_i;

	    for (i=0; i<nr_points; i++) {
		read_vector(&v);
		if ((*app)->point) {
		    v3_let((*app)->point[i], v);
		}
	    }
	    
	    read_close_i;

	    read_token;
	    if (!string_cmp("code"))
		read_error("keyword 'code' expected.");

	    for (i=0; i<code_length; i++) {
		read_int;
		if ((*app)->code) (*app)->code[i] = token_int;
	    }
	}

	read_close_i;
    }
    else read_error("keyword 'appearance' expected.");

    return FALSE;

 exception:
    return TRUE;
}


static boolean mw_insert_object(FILE *fh, mob_world mw, mw_object head)
{
    int type, function;
    mw_object obj = 0;
    vector3   position;
    quaternion orientation;

    v3_let(position, v3_zero);
    q_let(orientation, q_zero);

    /* determine object type
     */
    read_token;
    if (string_cmp("group")) type = MWO_GROUP;
    else if (string_cmp("triangle")) type = MWO_TRIANGLE;
    else if (string_cmp("rectangle")) type = MWO_RECTANGLE;
    else if (string_cmp("cuboid")) type = MWO_CUBOID;
    else read_error("illegal object specifier");

    read_open_i;

    /* determine object function and read object-parameters
     */
    read_token;
    function = MWF_OBSTACLE;
    if (string_cmp("obstacle")) function = MWF_OBSTACLE;
    else if (string_cmp("sensor")) function = MWF_SENSOR;
    else if (string_cmp("camera")) function = MWF_CAMERA;
    else if (string_cmp("mob")) function = MWF_MOB;
    else unread_token;

    obj = mw_object_new(mw, type, function);
    if (head) mw_object_add_to_group(head, obj);

    switch(function) {
    case MWF_MOB:
	read_open_i;
	read_token;
	if (string_cmp("geometry")) {
	    double a, b, c, d, r1, r2;
	    read_double; a = token_double;
	    read_double; b = token_double;
	    read_double; c = token_double;
	    read_double; d = token_double;
	    read_double; r1 = token_double;
	    read_double; r2 = token_double;
	}
	else read_error("mob parameters expected");
	read_close_i;
	break;
    case MWF_SENSOR:
	read_open_i;
	read_token;
	if (string_cmp("nr")) {
	    read_int;
	    mw_object_set_sensor(obj, (int)token_int);
	}
	else read_error("sensor parameters expected");
	read_close_i;
	break;
    case MWF_CAMERA:
	read_open_i;
	while(1) {
	    read_token;
	    if (token_is_close_i) break;
            else if (string_cmp("nr")) {
	        read_int;
		obj->func.camera.camera_nr = token_int;
		mw_object_set_sensor(obj, (int)token_int);
            }
	    else if (string_cmp("angle")) {
	        read_double;
		obj->func.camera.angle = token_int;
            }
            else if (string_cmp("focus_distance")) {
	        read_double;
		obj->func.camera.focus_distance  = token_int;
            }
            else if (string_cmp("aspect_ratio")) {
	        read_double;
		obj->func.camera.aspect_ratio = token_int;
            }
	    else read_error("camera parameters expected");
	}
	break;
    default:
	break;
    }

    while(1) {
	read_token;
	if (string_cmp("name")) {
	    read_quoted_string;
	    if (token_string[0]) mw_object_set_name(obj, token_string);
	}
	else if (string_cmp("position")) {
	    read_vector(&position);
	}
	else if (string_cmp("orientation")) {
	    read_quaternion(&orientation);
	}
	else if (string_cmp("color")) {
	    read_double;
	    obj->r = token_double;
	    read_double;
	    obj->g = token_double;
	    read_double;
	    obj->b = token_double;
	}
	else if (string_cmp("us_reflexion_damping")) {
	    read_double;
	    obj->us_reflexion_damping = token_double;
	}
	else if (string_cmp("marked")) {
	    boolean b;

	    read_boolean(&b);
	    if (b) (void)mw_object_mark(obj, b);
	}
	else if (string_cmp("appearance")) {
	    unread_token;
	    read_app2d(&(obj->appearance_2d));
	}
	else if (string_cmp("visualization")) {
	    read_quoted_string;
	    if (token_string[0]) {
		mw_object_set_3d(obj, token_string);
	    }
	}
	else if (string_cmp("actuator")) {
	    read_int;
	    mw_object_set_actuator_nr(obj, token_int);
	}
	else if (type==MWO_GROUP && string_cmp("members")) {
	    read_open_i
	    while(1) {
		read_token;
		if (token_is_close_i) break;
		unread_token;
		read_object(mw, obj);
	    }
	}
	else if (type==MWO_TRIANGLE && string_cmp("edges")) {
	    vector3 edges[3], a, b, normal;
	    double d;

	    read_open_i;
	    read_vector(&edges[0]);
	    read_vector(&edges[1]);
	    read_vector(&edges[2]);
	    read_close_i;

	    v3_sub(a, edges[1], edges[0]);
	    v3_sub(b, edges[2], edges[0]);
	    v3_mul(normal, a, b);
	    d = v3_sum(normal);
	    if (fabs(d)<=1e-6)
		read_error("triangle size is to smaller than 1e-6m^2");

	    mwo_triangle_set(obj, edges);
	}
	else if (type==MWO_RECTANGLE && string_cmp("xy_size")) {
	    double x_size, y_size;

	    read_double;
	    x_size = token_double;
	    read_double;
	    y_size = token_double;

	    if (x_size<=1e-6 && y_size<=1e-6)
		read_error("x/y-size of the rectangle is below 1e-6m");

	    mwo_rectangle_set(obj, x_size, y_size);
	}
	else if (type==MWO_CUBOID && string_cmp("edges_length")) {
	    vector3 v;

	    read_vector(&v);
	    if (v[0]<=1e-6 || v[1]<=1e-6 || v[2]<=1e-6)
		read_error("edges-length of the cuboid is below 1e-6m");

	    mwo_cuboid_set(obj, v);
	}
	else if (token_is_close_i) break;
	else read_error("unknown object parameter");
    }

    mw_object_set_position(obj, position, orientation);

    return FALSE;

exception:
    if (obj) mw_object_free(obj);
    return TRUE;
}


static boolean mw_insert_world(FILE *fh, mob_world mw)
{
    read_token;
    if (!string_cmp("world")) 
	read_error("keyword 'world' expected");
    read_open_i;

    while(1) {
	read_token;
	if (token_is_close_i) break;

    	if (string_cmp("dt_collision_test")) {
	    read_double;
            if (token_double<=1e-6)
	        read_error("value too small, only accept values >1e-6");	
	    mw->dt_collision_test = token_double;
        }
    	else if (string_cmp("air_speed")) {
	    read_double;
	    mw->air_speed = token_double;
        }
    	else if (string_cmp("us_air_damping")) {
	    read_double;
	    mw->us_air_damping = token_double;
        }
	else if (string_cmp("max_reflexion_count")) {
            read_int;
            mw->max_reflexion_count = token_int;
	}
    	else if (string_cmp("objects")) {
	    read_open_i;
	    while(1) {
		read_object(mw, 0);
		read_token;
		if (token_is_close_i) break;
	 	else unread_token;
            } 
        }
	else read_error("unknown keyword or missing }");
    }

    return FALSE;

exception:
    return TRUE;
}


boolean mw_insert(FILE *fh, mob_world mw)
{
    mw_file_error_string = 0;
    line_nr = 1;

    read_token;
    if (string_cmp("world")) {
	unread_token;
	read_world(mw);
    }
    else {
	while(!token_is_eof) {
	    unread_token;
	    read_object(mw, 0);
	    read_token;
	}
    }
    return FALSE;

exception:
    return TRUE;
}
