/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: mw_object.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:57:59 stolz>
 **************************************************************************
 *  MODULE              : mw_object
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in mw_object.h!
 *
 *  HISTORY
 *  19-May-94   Creation date
 *  30-May-94   Fix a really damned bug. (malloc(mw_object))!
 *  06-Jun-94   Add some mob-specific functions.
 *  13-Jun-94   simplify object creation, change object-structure,
 *              implementation of most routines
 *  16-Jun-94   write mob-code in mwo_add_to_space().
 *  17-Jun-94   write collision-test routines.
 *  19-Jun-94   Split mob_world into mob_world+mw_object+mw_mob.
 *  04-Jul-94   Add hom_matrix,.. to mw_object-structure (for optimization)
 *  12-Jul-94   Add mw_object_by_name() and remote()-functions.
 *  14-Jul-94   Insert setting of mob-us_sensor-routine in mw_object_new().
 *  02-Sep-94   Add extern-function-call mechanism.
 *  05-Sep-94   Add id/user_data access-routines.
 *  29-Sep-94   Remove obsolent routines, cleanup for better sensor-usage
 *  01-Oct-94   Add triangle-connections for the cuboid.
 *  08-Oct-94   Triangle scaling per side-length + vector(degree)
 *  07-Nov-94   Change defaukt air and reflexion damping
 *  11-Nov-94   Fix bug in mw_object_ungroup which cause a segment-fault
 *              at the ivd_server-prozess.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "mob_world_private.h"
#include "mw_object.h"
#include "mwo_mob.h"
#include "error.h"
#include "strerror.h"
#include "pvm_server.h"

/* function prototypes
 */
static void mw_object_inform_actuator(mw_object obj, mw_object controller);


/* exported variables
 */
char *mw_object_io_error_string = 0;



/* local variables
 */
static void (*extern_object_new)(mw_object) = 0;
static void (*extern_object_free)(mw_object obj) = 0;
static void (*extern_object_change)(mw_object obj) = 0;

static boolean actuator_req = FALSE;



void mw_object_set_extern(void (*new)(mw_object),
			  void (*free)(mw_object), 
			  void (*change)(mw_object))
{
    extern_object_new = new;
    extern_object_free = free;
    extern_object_change = change;
}


int mw_object_id(mw_object obj)
{
    return obj->id;
}

void mw_object_set_id(mw_object obj, int id)
{
    obj->id = id;
}

void mw_object_set_user_data(mw_object obj, void *user_data)
{
    obj->user_data = user_data;
}

void *mw_object_user_data(mw_object obj)
{
    return obj->user_data;
}




/**************************************************************************
 * picking
 */

mw_object mw_object_pick(mob_world mw, vector3 start, vector3 direction)
{
    triangle t;
    mw_object obj;
    double dist;

    t = triangle_pick(&dist, mw->space, start, direction);

    if (t) {
	obj = (mw_object)triangle_get_user_data(t);
	while (obj->group_head)
	    obj = obj->group_head;

	return obj;
    }
    else return 0;
}


/**************************************************************************
 * object oriented functions
 */

static int cuboid_side0_connect[12] = {
	10, 4, 1, 7, 9, 3, 11, 5, 0, 6, 2, 8
};
static int cuboid_side2_connect[12] = {
	8, 2, 10, 5, 1, 7, 9, 3, 11, 4, 0, 6
};


static void mwo_add_to_space(mw_object obj)
{
    triangle tri;

    /* compute homogene-transformation-matrix for
     * the computation of the absolute edges.
     */
    hm_set_v3q(&(obj->htm), obj->position, obj->orientation);

    if (obj->group_head) {
	hm_mul(&(obj->htm), obj->htm, obj->group_head->htm);
    }

    switch(obj->type) {
    case MWO_GROUP:
	{
	    mw_object o;

	    o=list_first(obj->obj.group.members);
	    while (o) {
		mwo_add_to_space(o);
		o=list_next(obj->obj.group.members);
	    }
	}
	break;

    case MWO_TRIANGLE:
	hm_vmul(&(obj->abs_edge[0]), obj->htm, obj->obj.triangle.edges[0]);
	hm_vmul(&(obj->abs_edge[1]), obj->htm, obj->obj.triangle.edges[1]);
	hm_vmul(&(obj->abs_edge[2]), obj->htm, obj->obj.triangle.edges[2]);
	tri = triangle_new(obj->world->space, obj->abs_edge);
	list_add(obj->triangles, tri);
	triangle_set_user_data(tri, obj);
	break;

    case MWO_RECTANGLE:
        {
	    double x2, y2;
	    vector3 v;

	    x2 = 0.5 * obj->obj.rectangle.x_size;
	    y2 = 0.5 * obj->obj.rectangle.y_size;

	    v3_set(v, -x2, -y2, 0.0);
	    hm_vmul(&(obj->abs_edge[0]), obj->htm, v);
	    v3_set(v, -x2, +y2, 0.0);
	    hm_vmul(&(obj->abs_edge[1]), obj->htm, v);
	    v3_set(v, +x2, +y2, 0.0);
	    hm_vmul(&(obj->abs_edge[2]), obj->htm, v);
	    v3_set(v, +x2, -y2, 0.0);
	    hm_vmul(&(obj->abs_edge[3]), obj->htm, v);
	    
	    tri = triangle_new(obj->world->space, obj->abs_edge);
	    list_add(obj->triangles, tri);
	    triangle_set_user_data(tri, obj);

	    v3_let(v, obj->abs_edge[1]);
	    v3_let(obj->abs_edge[1], obj->abs_edge[3]);
	    tri = triangle_new(obj->world->space, obj->abs_edge);
	    list_add(obj->triangles, tri);
	    triangle_set_user_data(tri, obj);
	    v3_let(obj->abs_edge[1], v);

	}
	break;

    case MWO_CUBOID:
        {
	    int i, a=0, b, c, j=0, k=0;
	    double x2, y2, z2;
	    vector3 edge[3];
	    vector3 v;
	    triangle tri[12];

	    x2 = 0.5 * obj->obj.cuboid.edges_length[0];
	    y2 = 0.5 * obj->obj.cuboid.edges_length[1];
	    z2 = 0.5 * obj->obj.cuboid.edges_length[2];

/*
  zyx
0 ---
1 --+
2 -+-
3 -++
4 +--
5 +-+
6 ++-
7 +++
*/
            for (i=0; i<8; i++) {
		v3_set(v, ((i&1) ? x2 : -x2), ((i&2) ? y2 : -y2),
		       ((i&4) ? z2 : -z2));

		hm_vmul(&(obj->abs_edge[i]), obj->htm, v);
            }

/* triangles:
i=0: a=0, b=1, c=2
i=1: a=3, b=2, c=1
i=2: a=1, b=3, c=5
i=3: a=7, b=5, c=3
i=4: a=2, b=6, c=3
i=5: a=7, b=3, c=6
i=6: a=4, b=5, c=6
i=7: a=7, b=6, c=5
i=8: a=0, b=2, c=4
i=9: a=6, b=4, c=2
i=10: a=1, b=5, c=0
i=11: a=4, b=0, c=5
 */
	    for (i=0; i<12; i++) {

		if (i&1) {
		    a = (a & ~j) | ((a+j)&j);
		    a = (a & ~k) | ((a+k)&k);
		} else {
		    a = (1<<((i>>1)%4))>>1;
		    j = 1 << ((i>>1)%3);
		    k = 1 << (((i>>1)+1)%3);
		}

		b = (a & ~j) | ((a+j)&j);
		c = (a & ~k) | ((a+k)&k);

		v3_let(edge[0], obj->abs_edge[a]);
		v3_let(edge[1], obj->abs_edge[b]);
		v3_let(edge[2], obj->abs_edge[c]);
	        tri[i] = triangle_new(obj->world->space, edge);
		list_add(obj->triangles, tri[i]);
	        triangle_set_user_data(tri[i], obj);
	    }

	    for (i=0; i<12; i++) {
		triangle_set_connections(tri[i], 
				tri[cuboid_side0_connect[i]], 2,
				0, -1,
				tri[cuboid_side2_connect[i]], 0);

				}

        }
	break;

    default:
	error("mw_object: mwo_add_to_space() - unknown object type");
	break;
    }
}


static void mwo_remove_from_space(mw_object obj)
{
    triangle t;

    for (t=list_first(obj->triangles); t; t=list_next(obj->triangles))
	triangle_free(t);

    list_remove_all(obj->triangles);

    if (obj->type == MWO_GROUP) {
	mw_object o;
	
	o = list_first(obj->obj.group.members);
	while (o) {
	    mwo_remove_from_space(o);
	    o = list_next(obj->obj.group.members);
	}
    }
}



void mw_object_triangles(list *triangles, mw_object obj)
{
    triangle t;

    if (obj) {
	for (t=list_first(obj->triangles); t; t=list_next(obj->triangles))
	    list_add(*triangles, t);
	
	if (obj->type == MWO_GROUP) {
	    mw_object o;
	    
	    o = list_first(obj->obj.group.members);
	    while(o) {
		mw_object_triangles(triangles, o);
		o = list_next(obj->obj.group.members);
	    }
	}
    }
}

boolean mw_object_collision(mw_object obj)
{
    triangle t;
    list triangles;
    boolean collision = FALSE;

    triangles = list_new();
    mw_object_triangles(&triangles, obj);
    
    for (t=list_first(triangles); t; t=list_next(triangles)) {
	triangle_set_group_nr(t, 1);
    }

    for (t=list_first(triangles); t; t=list_next(triangles)) {
	if ((collision=triangle_collision(t))) break;
    }

    for (t=list_first(triangles); t; t=list_next(triangles)) {
	triangle_set_group_nr(t, 0);
    }

    list_free(triangles);

    return collision;
}



boolean mw_object_highlight(mw_object obj, boolean highlight)
{
    boolean b;

    b = obj->highlighted;
    obj->highlighted = highlight;

    if (obj->type == MWO_GROUP) {
	mw_object temp;

        temp = list_first(obj->obj.group.members);
        while (temp) {
            (void)mw_object_highlight(temp, highlight);
	    temp = list_next(obj->obj.group.members);
	}
    }

    return b;
}


boolean mw_object_mark(mw_object obj, boolean mark)
{
    boolean b;

    b = obj->marked;
    obj->marked = mark;

    if (obj->type == MWO_GROUP) {
	mw_object temp;

        temp = list_first(obj->obj.group.members);
        while (temp) {
            (void)mw_object_mark(temp, mark);
	    temp = list_first(obj->obj.group.members);
	}
    }

    return b;
}


mw_object mw_object_group(mob_world mw, list objects, int function)
{
    mw_object head, obj;
    double temp;

    head = mw_object_new(mw, MWO_GROUP, function);

    /* build the sum of the member-positions
     */
    v3_let(head->position, v3_zero);
    q_let(head->orientation, q_zero);

    obj = list_first(objects);
    while (obj) {

	if (mw != obj->world)
	    error("mw_object: mw_object_group() - list containts at least "
		  "one object from another world");

	v3_inc(head->position, obj->position);
	list_add(head->obj.group.members, obj);
	

	obj = list_next(objects);
    }

    /* don't create a goup if it has no members!
     */
    if (list_is_empty(head->obj.group.members)) {
	mw_object_free(head);
	return 0;
    }

    /* calculate the position of the group-head as
     * average of the position-sum.
     * Substract this from the members to get the
     * relative coordinate from the head to the members.
     */
    temp = 1.0 / (double)list_size(head->obj.group.members);
    v3_scalar(head->position, head->position, temp);

    obj=list_first(head->obj.group.members);
    while (obj) {
	v3_dec(obj->position, head->position);

	list_remove(mw->objects, obj);
	obj->group_head = head;

	obj = list_next(head->obj.group.members);
    }


    /* call extern functions
     */
    if (extern_object_change) {
	extern_object_change(head);
	obj=list_first(head->obj.group.members);
	while (obj) {
	    extern_object_change(obj);
	    obj = list_next(head->obj.group.members);
	}
    }
    

    /* unhighlight all members of this group!
     */
    (void)mw_object_highlight(head, FALSE);

    return head;
}


void mw_object_add_to_group(mw_object group, mw_object obj)
{
    quaternion q;

    if (group->type != MWO_GROUP)
        error("mw_object: mw_object_add_to_group() - 1.parameter is not a group");

    if (obj->group_head != 0)
        error("mw_object: mw_object_add_to_group() - obj already member of a group");

    if (obj->world != group->world)
        error("mw_object: mw_object_add_to_group() - obj and group from different worlds");

    /* hang object out from the world-objects into the member list
     * of the group
     */
    list_remove(obj->world->objects, obj);
    list_add(group->obj.group.members, obj);
    obj->group_head = group;

    /* calculate the relative position and orientation from
     * the group-origin to the object
     */
    v3_dec(obj->position, group->position);
    q_neg(q, group->orientation);
    q_mul(&(obj->orientation), q, obj->orientation);


    /* call extern functions */
    if (extern_object_change)  {
	extern_object_change(obj);
	extern_object_change(group);
    }
}


void mw_object_ungroup(mw_object obj)
{
    mw_object o;
    hom_matrix t;

    if (obj->type != MWO_GROUP || obj->group_head)
	error("mw_object: mw_object_ungroup() - object not a group"
	      " or not the highest group-head");

    hm_set_v3q(&t, obj->position, obj->orientation);

    (void)mw_object_highlight(obj, FALSE);

    o = list_first(obj->obj.group.members);
    while (o) {
	hm_vmul(&(o->position), t, o->position);
	q_mul(&(o->orientation), o->orientation, obj->orientation);

	o->group_head = 0;

	list_add(o->world->objects, o);
	if (o->function == MWF_MOB) {
	    list_add(o->world->mobs, o);
        }

	/* call extern functions
	 */
	if (extern_object_change) 
	    extern_object_change(o);
	
	o = list_next(obj->obj.group.members);
    }

    list_remove_all(obj->obj.group.members);

    mw_object_free(obj);
}




mw_object mw_object_new(mob_world mw, int type, int function)
{
    mw_object	obj;

    obj = malloc(sizeof(adt_mw_object));
    if (obj==0)
	error("mw_object: mw_object_new() - can't alloc memory for adt mw_object");

    obj->type        = type;
    obj->function    = function;
    obj->world       = mw;
    obj->triangles   = list_new();
    obj->highlighted = FALSE;
    obj->marked      = FALSE;
    obj->group_head  = 0;
    obj->name        = 0;
    obj->appearance_2d = 0;
    obj->appearance_3d = 0;
    obj->actuator.nr  = -1;
    obj->sensors      = 0;
    obj->r = 1.0;
    obj->g = 1.0;
    obj->b = 1.0;
    obj->abs_edge = 0;
    hm_let(&(obj->htm), hm_zero);
    obj->remote_nr = -1;
    obj->remote_tid = -1;
    obj->remote_time = 0.0;
    obj->remote_nr_ios = 0;

    v3_let(obj->position, v3_zero);
    q_let(obj->orientation, q_zero);
    obj->us_reflexion_damping = 1e-5;  /* -5dB */

    switch(type) {
    case MWO_GROUP:
	obj->obj.group.members = list_new();
	break;

    case MWO_CUBOID:
	v3_set(obj->obj.cuboid.edges_length, 1.0, 1.0, 1.0);
	obj->abs_edge = malloc(8*sizeof(vector3));
	if (obj->abs_edge == 0)
	    error("mw_object: mw_object_new() - can't alloc memory for abs-edges");
	mwo_add_to_space(obj);
	break;

    case MWO_RECTANGLE:
	obj->obj.rectangle.x_size = 1.0;
	obj->obj.rectangle.y_size = 1.0;
	obj->abs_edge = malloc(4*sizeof(vector3));
	if (obj->abs_edge == 0)
	    error("mw_object: mw_object_new() - can't alloc memory for abs-edges");
	mwo_add_to_space(obj);
	break;

    case MWO_TRIANGLE:
        v3_set(obj->obj.triangle.edges[0], -0.25, -0.25, 0.0);
        v3_set(obj->obj.triangle.edges[1], -0.25, +0.75, 0.0);
        v3_set(obj->obj.triangle.edges[2], +0.75, -0.25, 0.0);
	obj->abs_edge = malloc(3*sizeof(vector3));
	if (obj->abs_edge == 0)
	    error("mw_object: mw_object_new() - can't alloc memory for abs-edges");
	mwo_add_to_space(obj);
	break;

    default:
	error("mw_object() mw_object_new() - unknown object type");
	break;
    }
    list_add(mw->objects, obj);

    switch(function) {
    case MWF_MOB:
	obj->func.mob.a = 0.272;
	obj->func.mob.b = 0.508;
	obj->func.mob.c = 0.225;
	obj->func.mob.d = 0.045;
	obj->func.mob.r1 = 0.118;
	obj->func.mob.r2 = 0.061;

	obj->func.mob.mob = mob_new(mw->scheduler);
/*	mob_set_io(obj->func.mob.mob, mwo_mob_input, mwo_mob_output, obj); */
	mob_set_io2(obj->func.mob.mob, mwo_mob_input, mwo_mob_output, 
		   mwo_mob_open_output, mwo_mob_close_output, obj);
	mob_set_us_sensor_function(obj->func.mob.mob, 
				   mwo_mob_us_sensor_scan,
				   obj);
	mob_set_us_sensor_data_function(obj->func.mob.mob, uscp_free, uscp_select);

	obj->func.mob.io_buffer = text_new();
	obj->func.mob.input_queue = queue_new(sizeof(io_marker));
	obj->func.mob.output_queue = queue_new(sizeof(io_marker));

	obj->func.mob.reply = text_new();

	list_add(mw->mobs, obj);
	break;

    case MWF_SENSOR:
	obj->func.sensor.sensor_nr = 0;
	obj->func.sensor.angle = 0.2618;	/* 15Degree in RAD */
	obj->func.sensor.radius = 0.02; /* 2cm */
	obj->func.sensor.threshold = -42.0;
        v3_set(obj->func.sensor.focus, 0.0, 0.0,
		- 0.5 * obj->func.sensor.radius / tan(0.5 * obj->func.sensor.angle));
	break;
    case MWF_CAMERA:
	obj->func.camera.camera_nr = CAMERA_OFFSET;
	obj->func.camera.angle = 30.0;	/* degree */
	obj->func.camera.focus_distance = 1.5;
	obj->func.camera.aspect_ratio = 1.0;
	break;
    case MWF_OBSTACLE:
	break;

    default:
	error("mw_object.c: mw_object_new() - unknown object function");
	break;
    }
    
    /* call extern functions
     */
    if (extern_object_new) 
	extern_object_new(obj);

    return obj;
}


void mw_object_free(mw_object obj)
{
    if (obj->group_head) {
	list_remove(obj->group_head->obj.group.members, obj);
    }

    mwo_remove_from_space(obj);

    switch(obj->type) {
    case MWO_GROUP:
	{
	    mw_object m;

            m = list_first(obj->obj.group.members);
	    while (m) {
		m->group_head = 0;
                mw_object_free(m);
		m = list_next(obj->obj.group.members);
	    }
	    list_free(obj->obj.group.members);
	}

    case MWO_CUBOID:
    case MWO_RECTANGLE:
    case MWO_TRIANGLE:
	break;

    default:
	error("mw_object() mw_object_free() - unknown object type");
    }
    list_remove(obj->world->objects, obj);


    switch(obj->function) {
    case MWF_MOB:
	queue_free(obj->func.mob.input_queue);
	queue_free(obj->func.mob.output_queue);
	text_free(obj->func.mob.io_buffer);
	text_free(obj->func.mob.reply);
	list_remove(obj->world->mobs, obj);
	mob_free(obj->func.mob.mob);
	break;

    case MWF_SENSOR:
    case MWF_CAMERA:
	{
	    mw_object controller = mw_object_controller(obj);

            if (controller)
		controller->sensors[obj->func.sensor.sensor_nr] = 0;
	}
	break;

    default:
	break;
    }


/* obsolent, dynamic binding between sensor->controller
    if (obj->sensors) {
	    int i;

	    for (i=0; i<MAX_NR_SENSORS; i++) {
		if (obj->sensors[i])
                    (obj->sensors[i])->func.sensor.controller = 0;
            }
    }
*/

    if (obj->appearance_2d) app2d_free(obj->appearance_2d);
    if (obj->abs_edge) free(obj->abs_edge);

    mob_world_set_remote(obj->world, obj->remote_nr, 0);
    if (obj->actuator.nr>=0) mw_object_set_actuator_nr(obj, -1);

    list_free(obj->triangles);

    /* call extern functions
     */
    if (extern_object_free)
	extern_object_free(obj);

    if (obj->sensors) free(obj->sensors);
    if (obj->appearance_3d) free(obj->appearance_3d);

    mw_object_kill_remote(obj);

    free(obj);
}


int mw_object_type(mw_object obj)
{
    return obj->type;
}

int mw_object_function(mw_object obj)
{
    return obj->function;
}

void mw_object_position(vector3 *position, quaternion *orientation, mw_object obj)
{
    if (position) {
	v3_let(*position, obj->position);
    }

    if (orientation) {
	q_let(*orientation, obj->orientation);
    }
}



void mw_object_set_position(mw_object obj, vector3 position, quaternion orientation)
{
    if (is_zero(obj->position[0] - position[0]) &&
	is_zero(obj->position[1] - position[1]) &&
	is_zero(obj->position[2] - position[2]) &&
	is_zero(obj->orientation[0] - orientation[0]) &&
	is_zero(obj->orientation[1] - orientation[1]) &&
	is_zero(obj->orientation[2] - orientation[2]) &&
	is_zero(obj->orientation[3] - orientation[3])) return;
	
    mwo_remove_from_space(obj);

    v3_let(obj->position, position);
    q_let(obj->orientation, orientation);

    mwo_add_to_space(obj);

    if (obj->actuator.nr>=0) {
	mw_object controller;

	controller = mw_object_controller(obj);
	if (!actuator_req && controller)
	    mw_object_inform_actuator(obj, controller); 
    }

    /* call extern function
     */
    if (extern_object_change)
	extern_object_change(obj);
}



void mw_object_move(mw_object obj, vector3 position, quaternion orientation)
{
    vector3 new_pos;
    quaternion new_orient;

    v3_add(new_pos, obj->position, position);
    q_mul(&new_orient, obj->orientation, orientation);

    mw_object_set_position(obj, new_pos, new_orient);
}


void mw_object_scale(mw_object obj, vector3 scale)
{
    scale[0] = fabs(scale[0]);
    scale[1] = fabs(scale[1]);
    scale[2] = fabs(scale[2]);

    mwo_remove_from_space(obj);

    switch(obj->type) {
    case MWO_GROUP:
/*
	{
	    mw_object m;
            vector3   nscale;

	    v3_qrot(&nscale, scale, obj->orientation);

	    m=list_first(obj->obj.group.members);
	    while(m) {
		mw_object_scale(m, nscale);
		m = list_next(obj->obj.group.members);
            }
        }
*/
	break;

    case MWO_CUBOID:
	if (is_zero(scale[0]) || is_zero(scale[1]) || is_zero(scale[2]))
	    error("mw_object: mw_object_scale() - scale factor zero?");

	v3_let(obj->obj.cuboid.edges_length, scale);
	break;

    case MWO_RECTANGLE:
	if (is_zero(scale[0]) || is_zero(scale[1]))
	error("mw_object: mw_object_scale() - scale factor zero?");
	
	obj->obj.rectangle.x_size = scale[0];
	obj->obj.rectangle.y_size = scale[1];
	break;

    case MWO_TRIANGLE:
	if (is_zero(scale[0]) || is_zero(scale[1]) || is_zero(scale[2])) 
	    error("mw_object: mw_object_scale() - scale factor zero?");
	{
	    double ax, bx, by, centerx, centery;

	    ax = scale[0];
	    bx = scale[1]*cos(scale[2]*PI/180.0); 
	    by = scale[1]*sin(scale[2]*PI/180.0);
	    centerx = (ax + bx) / 3.0;
	    centery = by / 3.0;
	    v3_set(obj->obj.triangle.edges[0], -centerx, -centery, 0.0);
	    v3_set(obj->obj.triangle.edges[1], ax-centerx, -centery, 0.0);
	    v3_set(obj->obj.triangle.edges[2], bx-centerx, by-centery, 0.0);
	}
        break;

    default:
	error("mw_object: mw_object_scale() - unknown object type");
        break;
    }

    mwo_add_to_space(obj);

    /* call extern function
     */
    if (extern_object_change)
	extern_object_change(obj);
}

void mw_object_get_scale(vector3 *scale, mw_object obj)
{
    switch(obj->type) {
    case MWO_GROUP:
	v3_let(*scale, v3_zero);
	break;

    case MWO_CUBOID:
	v3_let(*scale, obj->obj.cuboid.edges_length);
	break;

    case MWO_RECTANGLE:
	v3_set(*scale, obj->obj.rectangle.x_size,  
	       obj->obj.rectangle.y_size, 0.0);
	break;

    case MWO_TRIANGLE:
	{
	    double a, b, angle;
	    vector3 va, vb;

	    v3_sub(va, obj->obj.triangle.edges[1], obj->obj.triangle.edges[0]);
	    v3_sub(vb, obj->obj.triangle.edges[2], obj->obj.triangle.edges[0]);
	    a = v3_sum(va);
	    b = v3_sum(vb);
	    angle = acos(v3_product(va, vb) / (a*b));
	    angle *= 180.0 / PI;
	    v3_set(*scale, a, b, angle);
	}
        break;

    default:
	error("mw_object: mw_object_get_scale() - unknown object type");
        break;
    }
}

void mw_object_set_name(mw_object obj, char *name)
{
    if (obj->name) free(obj->name);
    obj->name = 0;

    if (name!=0 && *name != '\0') {
	obj->name = malloc(strlen(name)+1);
	if (obj->name) {
	    strcpy(obj->name, name);
	}
	else error("mw_object: mob_set_name() - can't alloc memory for the name");
    }

    /* call extern function
     */
    if (extern_object_change)
	extern_object_change(obj);
}

char* mw_object_get_name(mw_object obj)
{
    return obj->name;
}

mw_object	mw_object_new2(mob_world mw, int type, int function,
			       vector3 position, quaternion orientation,
			       vector3 scale)
{
    mw_object obj;

    obj = mw_object_new(mw, type, function);
    mw_object_scale(obj, scale);
    mw_object_set_position(obj, position, orientation);

    return obj;
}


mw_object	mw_object_new3(mob_world mw, int type, int function,
			       vector3 position, quaternion orientation)
{
    mw_object obj;

    obj = mw_object_new(mw, type, function);
    mw_object_set_position(obj, position, orientation);

    return obj;
}

void mw_object_get_abs_v3(vector3 *abs, mw_object obj, vector3 rel)
{
    hm_vmul(abs, obj->htm, rel);
}

void mw_object_get_abs_q(quaternion *abs, mw_object obj, quaternion rel)
{
    q_let(*abs, rel);

    while (obj) {
	q_mul(abs, *abs, obj->orientation);
	obj = obj->group_head;
    }
}



mw_object mw_object_controller(mw_object obj)
{
/*    if (obj->actuator.nr <0) return 0;
*/
    while (obj) {
	if (is_controller(obj->function)) break;
	obj = obj->group_head;
    }
    return obj;
}

static void mw_object_inform_actuator(mw_object obj, mw_object controller)
{
    if (controller) switch(controller->function) {
    case MWF_MOB:
	mwo_mob_inform_actuator(obj, controller);
	break;
    default:
	break;
    }
}

void mw_object_req_actuator(mw_object obj, mw_object controller)
/* request the position/orientation of the mw_object obj from the
 * pair controller+actuator_nr.
 */
{
    actuator_req = TRUE;

    if (controller) switch(controller->function) {
    case MWF_MOB:
	mwo_mob_req_actuator(obj, controller);
	break;
    default:
	break;
    }

    actuator_req = FALSE;
}


void mw_object_set_2d(mw_object obj, void *app)
{
    obj->appearance_2d = app;
}

void mw_object_set_3d(mw_object obj, void *app)
{
    if (obj->appearance_3d) free(obj->appearance_3d);
    obj->appearance_3d = 0;

    if (app && (*(char *)app) != 0) {
	obj->appearance_3d = malloc(strlen((char *)app)+1);
	if (obj->appearance_3d) {
	    strcpy((char *)(obj->appearance_3d), (char *)app);
	}
	else
	    error("mw_object: mw_object_set_3d()"
		  " - can't alloc memory for 3d-app-string");
    }

    if (extern_object_change)
	extern_object_change(obj);
}

void *mw_object_get_3d(mw_object obj)
{
    return obj->appearance_3d;
}

void mw_object_set_actuator_nr(mw_object obj, int nr)
{
    if (obj->actuator.nr>=0 && nr<0)
	list_remove(obj->world->actuators, obj);

    if (obj->actuator.nr<0 && nr>=0) 
	list_add(obj->world->actuators, obj);

    obj->actuator.nr = nr;
}

boolean	mw_objects_save(list objects, char *filename)
{
    FILE *handle;
    mw_object obj;
    boolean b;

    mob_world_io_error_string = 0;

    handle = fopen(filename, "w");
    if (handle == 0) {
	mw_object_io_error_string = strerror(ferror(handle));

	return TRUE;
    }

    obj = list_first(objects);
    while (obj) {
	b = mw_object_write(handle, obj);
	if (b) {
	    mw_object_io_error_string = mw_file_error_string;
	    break;
	}
	obj = list_next(objects);
    }

    fclose(handle);

    return b;
}


void mwo_cuboid_set(mw_object obj, vector3 edges_length)
{
    if (obj->type != MWO_CUBOID)
	error("mw_object: mwo_cuboid_set() - object not a cuboid");

    mwo_remove_from_space(obj);

    v3_let(obj->obj.cuboid.edges_length, edges_length);

    mwo_add_to_space(obj);


    /* call extern function
     */
    if (extern_object_change)
	extern_object_change(obj);
}


void mwo_triangle_set(mw_object obj, vector3 *edges)
{
    if (obj->type != MWO_TRIANGLE)
	error("mw_object: mwo_triangle_set() - object not a triangle");
    mwo_remove_from_space(obj);

    v3_let(obj->obj.triangle.edges[0], edges[0]);
    v3_let(obj->obj.triangle.edges[1], edges[1]);
    v3_let(obj->obj.triangle.edges[2], edges[2]);

    mwo_add_to_space(obj);

    /* call extern function
     */
    if (extern_object_change)
	extern_object_change(obj);
}

void mwo_rectangle_set(mw_object obj, double x_size, double y_size)
{
    if (obj->type != MWO_RECTANGLE)
	error("mw_object: mwo_rectangle_set() - object not a rectangle");
    mwo_remove_from_space(obj);

    obj->obj.rectangle.x_size = x_size;
    obj->obj.rectangle.y_size = y_size;

    mwo_add_to_space(obj);

    /* call extern function
     */
    if (extern_object_change)
	extern_object_change(obj);
}


void mw_object_get_rel_v3(vector3 *rel, mw_object obj, vector3 abs)
{
    v3_let(*rel, abs);

}

void mwo_cuboid_get(vector3 *edges_length, mw_object obj)
{
    if (obj->type != MWO_CUBOID)
	error("mw_object: mwo_cuboid_get() - object not a cuboid");

    v3_let(*edges_length, obj->obj.cuboid.edges_length);
}

void mwo_triangle_get(vector3 *edges, mw_object obj)
{
    if (obj->type != MWO_TRIANGLE)
	error("mw_object: mwo_triangle_get() - object not a triangle");

    v3_let(edges[0], obj->obj.triangle.edges[0]);
    v3_let(edges[1], obj->obj.triangle.edges[1]);
    v3_let(edges[2], obj->obj.triangle.edges[2]);
}

void mwo_rectangle_get(double *x_size, double *y_size, mw_object obj)
{
    if (obj->type != MWO_RECTANGLE)
	error("mw_object: mwo_rectangle_get() - object not a rectangle");

    *x_size = obj->obj.rectangle.x_size;
    *y_size = obj->obj.rectangle.y_size;
}


void mw_object_set_color(mw_object obj, float r, float g, float b)
{
    if (r<=0.0) obj->r = 0.0;
    else if (r>=1.0) obj->r = 1.0;
    else obj->r = r;

    if (g<=0.0) obj->g = 0.0;
    else if (g>=1.0) obj->g = 1.0;
    else obj->g = g;

    if (b<=0.0) obj->b = 0.0;
    else if (b>=1.0) obj->b = 1.0;
    else obj->b = b;

    /* call extern function
     */
    if (extern_object_change)
	extern_object_change(obj);
}

void mw_object_get_color(float *r, float *g, float *b, mw_object obj)
{
    *r = obj->r;
    *g = obj->g;
    *b = obj->b;
}


mw_object mw_object_head(mw_object obj)
{
    if (obj) return obj->group_head;
    else return 0;
}


mw_object	mw_object_top_head(mw_object obj)
{
    if (obj==0) return 0;

    while (obj->group_head) 
	obj = obj->group_head;

    return obj;
}


void mw_object_members(list *l, mw_object obj, boolean all)
{
    if (obj==0) return;

    if (obj->type==MWO_GROUP) {
	mw_object o;

	o = list_first(obj->obj.group.members);
	while (o) {
	    if (o->type == MWO_GROUP) {
		if (all) mw_object_members(l, o, all);
	    }
	    else list_add(*l, o);
	    o = list_next(obj->obj.group.members);
	}
    }
    else list_add(*l, obj);
}



void	mw_object_set_sensor(mw_object sensor, int nr)
{
    mw_object controller = sensor;

    if (!is_sensor(sensor->function))
	error("mw_object: mw_object_set_sensor() - 1.parameter is not a sensor");
    
    while (controller && !is_controller(controller->function)) {
	controller = controller->group_head;
    }

    if (controller==0)
	error("mw_object: mw_object_set_sensor() - sensor not member of a controller");

    if (nr<0 || nr>=MAX_NR_SENSORS) {
	warning("mw_object: mw_object_set_sensor() - sensor number out of range, skipping!");
	return;
    }

    if (controller->sensors == 0) {
	int i;

	controller->sensors = malloc(sizeof(mw_object)*MAX_NR_SENSORS);
	if (controller->sensors == 0)
	    error("mw_object: mw_object_set_sensor() - can't alloc memory for sensor-array");

	for (i=0; i<MAX_NR_SENSORS; i++)
	    controller->sensors[i] = 0;
    }

    controller->sensors[nr] = sensor;
    switch(sensor->function) {
    case MWF_CAMERA:
	sensor->func.camera.camera_nr   = nr;
	break;
    default:
	sensor->func.sensor.sensor_nr   = nr;
    }
}

int	mw_object_get_sensor_nr(mw_object obj)
{
    if (obj->function == MWF_SENSOR)
	return obj->func.sensor.sensor_nr;
    else if (obj->function == MWF_CAMERA)
	return obj->func.camera.camera_nr;
    else return -1;
}


mw_object mw_object_get_sensor(mw_object controller, int nr)
{
    if (controller->sensors && nr>=0 && nr<MAX_NR_SENSORS)
	return controller->sensors[nr];
    else	
	return 0;
}

void	mw_object_set_us_reflexion_damping(mw_object obj, double value)
{
    obj->us_reflexion_damping = value;
}

double	mw_object_get_us_reflexion_damping(mw_object obj)
{
    return obj->us_reflexion_damping;
}


int mw_object_get_remote_tid(mw_object obj)
{
    if (obj) return obj->remote_tid;
    else return -1;
}

void mw_object_set_remote(mw_object obj, int tid)
{
    if (obj) {
	mw_object_kill_remote(obj);

	obj->remote_time = scheduler_time(obj->world->scheduler);
	obj->remote_nr = mob_world_remote_nr();
	obj->remote_tid = tid;
	
	if (tid==-1) mob_world_set_remote(obj->world, obj->remote_nr, 0);
	else mob_world_set_remote(obj->world, obj->remote_nr, obj);
    }
}

void mw_object_kill_remote(mw_object obj)
{
    if (obj->remote_tid!=-1) {
	ps_kill(obj);
	mob_world_set_remote(obj->world, obj->remote_nr, 0);
	obj->remote_tid = -1;
	obj->remote_time = 0.0;
    }
}

static mw_object mwo_by_name(list objs, char *name)
{
    mw_object obj;

    obj = list_first(objs);
    while (obj) {
	if ((obj->name) && (strcmp(name, obj->name)==0)) return obj;

	if ((obj->type == MWO_GROUP) &&
	    (obj = mwo_by_name(obj->obj.group.members, name))) {
	    return obj;
	}

	obj = list_next(objs);
    }

    return 0;
}

mw_object mw_object_by_name(mob_world mw, char *name)
{
    if (mw==0 || name==0 || *name=='\0') 
	return 0;

    return mwo_by_name(mw->objects, name);
}

void mw_object_send_cmd(mw_object obj, double t, char *cmd)
{
    if (obj->function == MWF_MOB)
	mob_send_cmd(obj, t, cmd);

}
