/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: mw_object.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 16:58:07 stolz>
 **************************************************************************
 *  MODULE              : mw_object
 *
 *  FUNCTION
 *	Functions for to generate, delete and manipulate a
 *	hierarchival mobs-objects are provided.
 *
 *  NOTE
 *
 *  HISTORY
 *  21-Apr-94   Creation date.
 *  06-Jun-94   Add some mob-specific functions
 *  13-Jun-94   Write code for add/remove objects from space.
 *  16-Jun-94   Add mob_world_is_empty() and mob_soft_reset()
 *  19-Jun-94   Split mob_world into mob_world+mw_object+mw_mob.
 *
 **************************************************************************/
#ifndef MW_OBJECT_H
#define MW_OBJECT_H

#ifndef MOB_WORLD_H
#include "mob_world.h"
#endif

#ifndef BASIC_H
#include "basic.h"
#endif

#ifndef VECTOR_H
#include "vector.h"
#endif

#ifndef LIST_H
#include "list.h"
#endif


/**************************************************************************
 * abstract datatypes
 */
typedef struct adt_mw_object*	mw_object;


/**************************************************************************
 * object specifica (type and function(ality))
 */
enum enum_mw_object {
    MWO_TRIANGLE=123, MWO_RECTANGLE, MWO_CUBOID, MWO_GROUP
};

enum enum_mw_func {
    MWF_MOB=456, MWF_SENSOR, MWF_CAMERA, MWF_OBSTACLE
};

#define is_sensor(function) ((function)==MWF_SENSOR||(function)==MWF_CAMERA)
#define is_controller(function) ((function)==MWF_MOB)
#define is_obstacle(function) ((function)==MWF_OBSTACLE)




/**************************************************************************
 * object oriented functions
 */

/* object creating/freeing
 */
mw_object	mw_object_new(mob_world, int type, int function);
mw_object	mw_object_new2(mob_world, int type, int function,
			       vector3 position, quaternion orientation,
			       vector3 scale);
mw_object	mw_object_new3(mob_world, int type, int function,
			       vector3 position, quaternion orientation);
void            mw_object_free(mw_object);

/* object type
 */
int		mw_object_type(mw_object);
int		mw_object_function(mw_object);

/* change position and orientation
 */
void		mw_object_position(vector3 *position, quaternion *orientation,
				   mw_object);
void		mw_object_set_position(mw_object, vector3 position,
				       quaternion orientation);
void    	mw_object_move(mw_object, vector3 position,
			       quaternion orientation);
void		mw_object_get_abs_v3(vector3 *abs, mw_object, vector3 rel);
void		mw_object_get_rel_v3(vector3 *rel, mw_object, vector3 abs);
void            mw_object_get_abs_q(quaternion *abs, mw_object obj, quaternion rel);

/* picking
 */
mw_object	mw_object_pick(mob_world, vector3 start, vector3 direction);

/* grouping
 */
mw_object	mw_object_group(mob_world, list objects, int function);
void		mw_object_ungroup(mw_object);
void		mw_object_add_to_group(mw_object group, mw_object obj);

/* hierarchy information
 */
mw_object	mw_object_head(mw_object);
mw_object	mw_object_top_head(mw_object);
void		mw_object_members(list *l, mw_object, boolean all);

/* marking and highlighting
 */
boolean		mw_object_mark(mw_object, boolean mark);
boolean		mw_object_highlight(mw_object, boolean highlight);

/* object name
 */
void		mw_object_set_name(mw_object, char *name);
char*		mw_object_get_name(mw_object);

/* saving
 */
extern char *mw_object_io_error_string;
boolean		mw_objects_save(list objects, char *filename);

/* collision test
 */
boolean		mw_object_collision(mw_object obj);

/* actuator
 */
void		mw_object_set_actuator_nr(mw_object obj, int nr);
mw_object	mw_object_controller(mw_object obj);

/* appearance and geometry and color
 */
void		mw_object_set_2d(mw_object obj, void *);
void            mw_object_set_3d(mw_object obj, void *);
void           *mw_object_get_3d(mw_object obj);
void		mw_object_scale(mw_object, vector3);
void		mw_object_get_scale(vector3 *, mw_object);

void		mwo_cuboid_set(mw_object, vector3 edges_length);
void		mwo_triangle_set(mw_object, vector3 *edges);
void		mwo_rectangle_set(mw_object, double x_size, double y_size);

void		mwo_cuboid_get(vector3 *edges_length, mw_object);
void		mwo_triangle_get(vector3 *edges, mw_object);
void		mwo_rectangle_get(double *x_size, double *y_size, mw_object);

void		mw_object_set_color(mw_object, float r, float g, float b);
void		mw_object_get_color(float *r, float *g, float *b, mw_object);


/* sensor (camera & us-sensor)
 */
void		mw_object_set_sensor(mw_object sensor, int nr);
int		mw_object_get_sensor_nr(mw_object obj);
mw_object	mw_object_get_sensor(mw_object controller, int nr);



/* us parameters
 */
void		mw_object_set_us_reflexion_damping(mw_object, double);
double		mw_object_get_us_reflexion_damping(mw_object);

void		mw_object_req_actuator(mw_object obj, mw_object controller);

mw_object	mw_object_by_name(mob_world mw, char *name);
void		mw_object_set_remote(mw_object obj, int tid);
int		mw_object_get_remote_tid(mw_object obj);
void            mw_object_kill_remote(mw_object obj);

void		mw_object_send_cmd(mw_object, double t, char *cmd);

void mw_object_set_extern(void (*new)(mw_object),
			  void (*free)(mw_object), 
			  void (*change)(mw_object));
int mw_object_id(mw_object obj);
void mw_object_set_id(mw_object obj, int id);
void mw_object_set_user_data(mw_object obj, void *user_data);
void *mw_object_user_data(mw_object obj);
void mw_object_triangles(list *triangles, mw_object obj);



#endif
