/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: mob_world.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/03/02 18:48:36 stolz>
 **************************************************************************
 *  MODULE              : mob_world
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in mob_world.h!
 *
 *  HISTORY
 *  19-May-94   Creation date
 *  30-May-94   Fix a really damned bug. (malloc(mw_object))!
 *  06-Jun-94   Add some mob-specific functions.
 *  13-Jun-94   simplify object creation, change object-structure,
 *              implementation of most routines
 *  16-Jun-94   write mob-code in mwo_add_to_space().
 *  17-Jun-94   write collision-test routines.
 *  19-Jun-94   Add pvm_server-support.
 *  20-Jul-94   Take only inputs which are <act_time. (in mwo_mob_input)
 *  14-Sep-94   Add routines for us-computation with triangle-projections.
 *  02-Mar-95   Remove bug which results in an position-set of the mob by
 *              any other actuator.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "app2d.h"

#include "mwo_mob.h"
#include "mob_world_private.h"
#include "error.h"
#include "pvm_server.h"



/* code for mob-us_sensor_scan
 */
void mwo_mob_us_sensor_scan(mob m, int sensor_nr, void *uservar)
{
    mw_object obj = uservar;
    mw_object sensor;

#ifdef DEBUG_DETAIL
    fprintf(stderr, "us_sensor_scan(mob=%p, sensor_nr=%d, uservar=%p)\n",
                    m, sensor_nr, uservar);
#endif

    if (uservar==0 || obj->function != MWF_MOB) {
	warning("mwo_mob: us_sensor_scan() - uservar is not a mob");
    }

    sensor = mw_object_get_sensor(obj, sensor_nr);
#if 0
    printf("init sensor scan for sensor (%p), nr=%d\n", sensor, sensor_nr);
#endif

    if (sensor) us_sensor_scan(sensor);
}

/**************************************************************************
 * mob-specific functions, public
 */
void mob_send_command(mw_object mob, char *command)
{
    mob_send_cmd(mob, scheduler_time(mob->world->scheduler), command);
}

void mob_send_cmd(mw_object mob, double t, char *command)
{
    long com_len;
    io_marker i_marker;
    static char com_prefix[60];
    

    if (mob->function != MWF_MOB)
	error("mwo_mob: mob_send_command() - object is not a mob!");

    com_len = strlen(command);

    if (com_len > MAX_COMMAND_LENGTH)
	error("mwo_mob: mob_send_command() - max command length exceeded");

    com_prefix[0] = '\0';
    sprintf(&com_prefix[0], "\n<t=%.3fs: ", t);
    text_add(mob->func.mob.io_buffer, com_prefix);

    i_marker.time = t;
    i_marker.start = text_marker(mob->func.mob.io_buffer);
    i_marker.end = text_add2(mob->func.mob.io_buffer, command, com_len);

    queue_add(mob->func.mob.input_queue, &i_marker);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "SEND CMD '%s'\n", command);
#endif
}


void mwo_mob_input(char *com, void *uservar)
{
    mw_object  mob = uservar;
    io_marker i_marker;
    double t = scheduler_time(mob->world->scheduler);
    
    ps_sync_obj(mob, t);

    *com = '\0';
    if (queue_look_first(&i_marker, mob->func.mob.input_queue) &&
	i_marker.time <= t) {
	
	queue_first(&i_marker, mob->func.mob.input_queue);
	text_get(com, mob->func.mob.io_buffer, i_marker.start, i_marker.end);
	
#ifdef DEBUG_DETAIL
	fprintf(stderr, "DO CMD '%s'\n", com);
#endif
	    
    }
}


void mwo_mob_open_output(void *uservar)
{
    mw_object  mob = uservar;

    text_clear(mob->func.mob.reply);
}

void mwo_mob_close_output(void *uservar)
{
    mw_object  mob = uservar;
    char *output;
    long len;

    text_add2(mob->func.mob.reply, "", (long)1);
    text_ref(&output, &len, mob->func.mob.reply, TEXT_START, TEXT_END);
    ps_reply(mob, output);
}

void mwo_mob_output(void *uservar, char *out)
{
    mw_object  mob = uservar;
    io_marker o_marker;
    static char out_prefix[30];

    if (out) {

	if (mob->function != MWF_MOB)
	    error("mwo_mob: mob_send_command() - object is not a mob!");

	sprintf(out_prefix, "\n>t=%.3fs: ", scheduler_time(mob->world->scheduler));
        text_add(mob->func.mob.io_buffer, out_prefix);

	o_marker.time = scheduler_time(mob->world->scheduler);
	o_marker.start = text_marker(mob->func.mob.io_buffer);
	o_marker.end = text_add(mob->func.mob.io_buffer, out);
	queue_add(mob->func.mob.output_queue, &o_marker);

        text_add(mob->func.mob.reply, out);

#ifdef DEBUG_DETAIL
	fprintf(stderr, "OUTPUT> '%s'\n", out);
#endif

    }
}

int mob_motion_type(mw_object obj)
{
   switch(motion_type(obj->func.mob.mob)) {
    case MT_STOP:
	return MOB_STOPPED;
	break;
    case MT_HALT:
	return MOB_HALTED;
	break;
    default:
	return MOB_MOVING;
	break;
    }
}


void mob_get_state(mw_object obj, mwo_mob_state *state)
{
    mob_state real, intern;


    if (obj->function != MWF_MOB)
	error("mwo_mob: mob_get_state() - obj is not a mob!");

    mob_get_motion_state(obj->func.mob.mob, &real, &intern);

    state->x          = intern.x;
    state->y          = intern.y;
    state->angle      = intern.angle;
    state->v          = intern.v;
    state->w          = intern.w;
    state->a          = intern.a;
    state->u          = intern.u;
    state->x_real     = real.x;
    state->y_real     = real.y;
    state->angle_real = real.angle;

    switch(motion_type(obj->func.mob.mob)) {
    case MT_STOP:
	state->type = MOB_STOPPED;
	break;
    case MT_HALT:
	state->type = MOB_HALTED;
	break;
    default:
	state->type = MOB_MOVING;
	break;
    }
}

void mob_get_io(mw_object obj, char **text, int *len)
{
    long l;

    if (obj->function != MWF_MOB)
	error("mwo_mob: mob_get_io() - obj is not a mob!");

    text_ref(text, &l, obj->func.mob.io_buffer, TEXT_START, TEXT_END);

    *len = l;
}


void mob_clear_io(mw_object obj)
{
    if (obj->function != MWF_MOB)
	error("mwo_mob: mob_clear_io() - obj is not a mob!");

    text_clear(obj->func.mob.io_buffer);
}

void            mob_emergency_halt(mw_object mob)
{
    if (mob->function != MWF_MOB)
	error("mwo_mob: mob_emergency_halt() - obj is not a mob!");

    mob_halt(mob->func.mob.mob);
}

void mob_emergency_resume(mw_object mob)
{
    if (mob->function != MWF_MOB)
	error("mwo_mob: mob_resume() - obj is not a mob!");

    mob_resume(mob->func.mob.mob);
}




void mob_soft_reset(mw_object mob)
{
    if (mob->function != MWF_MOB)
	error("mwo_mob: mob_soft_reset() - obj is not a mob!");

    queue_free(mob->func.mob.input_queue);
    queue_free(mob->func.mob.output_queue);
    mob->func.mob.input_queue = queue_new(sizeof(io_marker));
    mob->func.mob.output_queue = queue_new(sizeof(io_marker));
    text_clear(mob->func.mob.io_buffer);

    mob_reset(mob->func.mob.mob);
}




static vector3 box_offset =   { 0.2284, 0.000, 0.255 };
static vector3 box_size =     { 0.950,  0.550, 0.500 };
static vector3 fl_wheel_pos = { 0.508,  0.255, 0.000 };
static vector3 fr_wheel_pos = { 0.508, -0.255, 0.000 };

static vector3 us_sensor_size = {0.030, 0.030, 1.0};
static vector3 us_sensor_normal = {0.0, 0.0, 1.0};
static vector3 us_sensor_pos[24] = {
    {+0.7410,  0.0000, +0.3130},
    {+0.7410, -0.2170, +0.3130},
    {+0.7235, -0.2630, +0.3130},
    {+0.6935, -0.3040, +0.3130},
    {+0.6575, -0.3340, +0.3130},
    {+0.5675, -0.3400, +0.3130},
    {+0.2285, -0.3400, +0.3130},
    {-0.1105, -0.3400, +0.3130},
    {-0.2005, -0.3340, +0.3130},
    {-0.2365, -0.3040, +0.3130},
    {-0.2665, -0.2630, +0.3130},
    {-0.2840, -0.2170, +0.3130},
    {-0.2840,  0.0000, +0.3130},
    {-0.2840, +0.2170, +0.3130},
    {-0.2665, +0.2630, +0.3130},
    {-0.2365, +0.3040, +0.3130},
    {-0.2005, +0.3340, +0.3130},
    {-0.1105, +0.3400, +0.3130},
    {+0.2285, +0.3400, +0.3130},
    {+0.5675, +0.3400, +0.3130},
    {+0.6575, +0.3340, +0.3130},
    {+0.6935, +0.3040, +0.3130},
    {+0.7235, +0.2630, +0.3130},
    {+0.7410, +0.2170, +0.3130}
};
static double us_sensor_angle[24] = {
	0.0,
	0.0,
	-30.0,
	-45.0,
	-60.0,
	-90.0,
	-90.0,
	-90.0,
	-120.0,
	-135.0,
	-150.0,
	-180.0,
	-180.0,
	-180.0,
	-210.0,
	-225.0,
	-240.0,
	-270.0,
	-270.0,
	-270.0,
	-300.0,
	-315.0,
	-330.0,
	0.0
};
static vector3 camera_axis = {0.0, 1.0, 0.0}; 
static double  camera_angle = 10.0*PI/180.0;
static vector3 camera_pos = {0.5, 0.0, 0.7};
static vector3 camera_size = {0.2, 0.05, 0.05};


mw_object mob_create(mob_world mw)
{
    int i;
    mw_object mob, sensor, box, wheel;
    vector3 v;
    quaternion sensor_orient, sensor_rot;

    mob = mw_object_new(mw, MWO_GROUP, MWF_MOB);
    mw_object_set_2d(mob, &def_2d_mob);
    mw_object_set_actuator_nr(mob, ACT_MOB);

    box = mw_object_new2(mw, MWO_CUBOID, MWF_OBSTACLE, box_offset, q_zero, box_size);
    mw_object_set_2d(box, &def_2d_empty);
    mw_object_add_to_group(mob, box);

    wheel = mw_object_new3(mw, MWO_GROUP, MWF_OBSTACLE, fl_wheel_pos, q_zero);
    mw_object_set_2d(wheel, &def_2d_wheel);
    mw_object_add_to_group(mob, wheel);
    mw_object_set_actuator_nr(wheel, ACT_MOB_WHEEL_FL);

    wheel = mw_object_new3(mw, MWO_GROUP, MWF_OBSTACLE, fr_wheel_pos, q_zero);
    mw_object_set_2d(wheel, &def_2d_wheel);
    mw_object_add_to_group(mob, wheel);
    mw_object_set_actuator_nr(wheel, ACT_MOB_WHEEL_FR);

    v3_set(v, 0.0, 1.0, 0.0);
    q_set_av(&sensor_rot, PI/2, v);

    for (i=0; i<24; i++) {
	q_set_av(&sensor_orient, us_sensor_angle[i]*PI/180.0, us_sensor_normal);

	q_mul(&sensor_orient, sensor_rot, sensor_orient);
	sensor = mw_object_new2(mw, MWO_RECTANGLE, MWF_SENSOR,
				us_sensor_pos[i], sensor_orient, us_sensor_size);
	mw_object_set_2d(sensor, &def_2d_us_sensor);

	mw_object_add_to_group(mob, sensor);
	mw_object_set_sensor(sensor, i);
    }

    /* camera
     */

    q_set_av(&sensor_orient, camera_angle, camera_axis);
    sensor = mw_object_new2(mw, MWO_CUBOID, MWF_CAMERA,
				camera_pos, sensor_orient, camera_size);
    mw_object_set_2d(sensor, &def_2d_camera);
    
    mw_object_add_to_group(mob, sensor);
    mw_object_set_sensor(sensor, CAMERA_OFFSET);

    return mob;
}


void mwo_mob_inform_actuator(mw_object obj, mw_object mob)
{
    mob_state real, intern;
    vector3 v;

    if (mob->function != MWF_MOB)
	error("mwo_mob: mob_mob_inform_actuator() - controller is not a mob!");

#ifdef DEBUG_DETAIL
    fprintf(stderr, "mwo_mob_inform_actuator()\n");
#endif

    switch(obj->actuator.nr) {
    case ACT_MOB: 
	mob_get_motion_state(mob->func.mob.mob, &real, &intern);

	real.x = obj->position[0];
	real.y = obj->position[1];
	q_get_av(&real.angle, &v, obj->orientation);
	
	mob_set_motion_state(mob->func.mob.mob, &real, &real);
	break;
    default:
	break;
    }
}

void mwo_mob_req_actuator(mw_object obj, mw_object controller)
{
    mob_state real;
    vector3 v;
    quaternion q;

#ifdef DEBUG_DETAIL
    fprintf(stderr, "mwo_mob_req_actuator() - obj=%p, act_nr=%d, controller=%p\n",
	    obj, obj->actuator.nr, controller);
#endif

    mob_get_motion_state(controller->func.mob.mob, &real, 0);

    switch(obj->actuator.nr) {
    case ACT_MOB: 
	v3_set(v, 0.0, 0.0, 1.0);
	q_set_av(&q, real.angle, v);
	v[0] = real.x;
	v[1] = real.y;
	v[2] = obj->position[2];
	mw_object_set_position(obj, v, q);
	break;

    case ACT_MOB_WHEEL_FL:
	v3_set(v, 0.0, 0.0, 1.0);
	q_set_av(&q, real.angle_fl, v);
	mw_object_set_position(obj, obj->position, q);
	break;

    case ACT_MOB_WHEEL_FR:
	v3_set(v, 0.0, 0.0, 1.0);
	q_set_av(&q, real.angle_fr, v);
	mw_object_set_position(obj, obj->position, q);
	break;

    default:
	break;
    }
}
