/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: parray.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:58:51 stolz>
 **************************************************************************
 *  MODULE              : parray
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in parray.h!
 *
 *  HISTORY
 *  29-Aug-94   Creation date
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "parray.h"
#include "error.h"

typedef struct adt_parray adt_parray;
typedef struct adt_parray {
    int size;
    int max_size;
    pointer *elements;
};

parray parray_new(int elem_size)
{
    parray pa = malloc(sizeof(adt_parray));

    if (pa) {
	pa->size = 0;
	pa->max_size = 0;
	pa->elements = 0;
    }
    else error("parray: parray_new() - can't alloc memory for adt parray");

    return pa;
}


void parray_free(parray pa)
{
    if (pa) {
	if (pa->elements) free(pa->elements);
	free(pa);
    }
}

void parray_clear(parray pa)
{
    pa->size = 0;
}


int parray_size(parray pa)
{
    return pa->size;
}


void parray_resize(parray pa, int index)
{
    if (index >= pa->max_size) {

	pa->max_size = index + 32 + pa->max_size >> 2;

	if (pa->elements)
	    pa->elements = (pointer *)realloc(pa->elements, pa->max_size * sizeof(pointer));
	else
	    pa->elements = (pointer *)malloc(sizeof(pointer) * pa->max_size);

	if (pa->elements==0)
	    error("parray: parray_resize() - can't alloc memory for elements");
    }
}

void parray_set(parray pa, int index, pointer elem)
{
    if (index>=0) {
        if (index >= pa->max_size) parray_resize(pa, index);

	pa->elements[index] = elem;
    }
    else error("parray: parray_set() - index < 0");
}


pointer parray_get(parray pa, int index)
{
    if (index>=0) {

        if (index >= pa->size)
	    error("parray: parray_get() - access on a unitialized element");

	return pa->elements[index];

    }
    else error("parray: parray_get() - index < 0");
}

void parray_copy(parray pa, int target, int source)
{
    if (source < 0 || source >=pa->size)
	error("parray: parray_copy() - source-index out of range");

    if (target<0)
	error("parray: parray_copy() - target-index < 0");

    if (target >= pa->max_size) parray_resize(pa, target);

    pa->elements[target] = pa->elements[source];
}

void parray_swap(parray pa, int a, int b)
{
    pointer h;

    if (a < 0 || a >=pa->size)
	error("parray: parray_swap() - 1.index out of range");

    if (b < 0 || b >=pa->size)
	error("parray: parray_copy() - 2.index out of range");

    h = pa->elements[b];
    pa->elements[b] = pa->elements[a];
    pa->elements[a] = h;
}

pointer parray_exchange(parray pa, int index, pointer p)
{
    pointer h;

    if (index < 0 || index >= pa->size)
	error("parray: parray_copy() - index out of range");

    h = pa->elements[index];
    pa->elements[index] = p;

    return h;
}


pointer* parray_array_ref(parray pa)
{
    return pa->elements;
}


void parray_for_all(parray pa, void (*f)(pointer))
{
    int i;

    if (f)
	for (i=0; i<pa->size; i++)
	    f(&(pa->elements[i]));
}


void parray_print(parray pa)
{
    int i;

    if (pa) {
	printf("parray {\nsize = %ld\nmax_size = %ld\nelements = ",
	       pa->size, pa->max_size);
	for (i=0; i<pa->size; i++) printf(" %p", pa->elements[i]);
	printf("\n}\n");
    }
}
