/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: pdlist.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:59:04 stolz>
 **************************************************************************
 *  MODULE              : pdlist
 *  IMPORT		: error, basic
 *
 *  IMPLEMENTATION
 *
 *	The priority list ist implemented as double-linked list.
 *
 *  NOTE
 *      The module specification and definition is given in pdlist.h!
 *
 *  HISTORY
 *  28-May-94   Creation date. Implementation and validation.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "pdlist.h"
#include "error.h"


/**************************************************************************
 * structure to store the element and priority-information
 */
typedef struct pdlist_elem pdlist_node;
struct pdlist_elem {
    double	priority;
    pelem	element;
    pdlist_node*	lower;
    pdlist_node*	higher;
};


/**************************************************************************
 * definition of the abstract datatype pdlist
 */
typedef struct adt_pdlist	adt_pdlist;
struct adt_pdlist {
    long size;
    pdlist_node	*highest, *lowest, *current;
};



pdlist	pdlist_new()
/**************************************************************************
 */
{
    pdlist l;

    l = (pdlist)malloc(sizeof(adt_pdlist));
    if (l) {
	l->size    = 0;
	l->highest = 0;
	l->lowest  = 0;
	l->current = 0;
    }
    else error("pdlist: pdlist_new() - can't alloc memory for adt pdlist");

    return l;

}


void    pdlist_free(pdlist l)
/**************************************************************************
 */
{
    pdlist_node *n, *t;

    for (n=l->highest; n; n=t) {
        t = n->lower;
	free(n);
    }

    free(l);
}


void	pdlist_add(pdlist l, double pri, pelem e)
/**************************************************************************
 */
{
    pdlist_node *n;
    pdlist_node *np;

    if (l->size == LONG_MAX)
	error("pdlist: pdlist_add() - max number of elements reached");

    n = (pdlist_node *)malloc(sizeof(pdlist_node));

    if (n==0) error("pdlist: pdlist_add() - can't alloc memory for a pdlist-node");

    n->priority = pri;
    n->element = e;
    l->size = l->size + 1;

    /* search first element which has a lower or same priority
     */
    for (np=l->highest; np && (np->priority > pri); np=np->lower)
	;

    /* initialize the node-pointers
     */
    n->lower = np;
    n->higher = (np ? np->higher : l->lowest);

    /* set the pointers of the higher and lower elements
     */
    if (n->lower) n->lower->higher = n;
    if (n->higher) n->higher->lower = n;

    /* set the highest/lowest pointer of the pdlist.
     */
    if (np==l->highest) l->highest = n;
    if (np==0) l->lowest  = n;
}


void	pdlist_clear(pdlist l)
/**************************************************************************
 */
{
    pdlist_node *n, *t;

    for (n=l->highest; n; n=t) {
        t = n->lower;
	free(n);
    }

    l->size = 0;
    l->highest = 0;
    l->lowest  = 0;
}


void	pdlist_remove(pdlist l, pelem e)
/**************************************************************************
 * FUNCTION
 *   Remove all nodes with the element e!
 */
{
    pdlist_node *n, *t;

    for (n=l->highest; n; n=t) {
        t = n->lower;
	if (n->element == e) {

	    if (n->higher) {
		n->higher->lower = n->lower;
	    } else l->highest = n->lower;

	    if (n->lower) {
		n->lower->higher = n->higher;
	    } else l->lowest = n->higher;

            l->size = l->size - 1;
            free(n);
	}
    }
}

void	pdlist_remove_pri(pdlist l, double pri)
/**************************************************************************
 */
{
    pdlist_node *n, *t;

    for (n=l->highest; n; n=t) {
        t = n->lower;
	if (n->priority == pri) {

	    if (n->higher) {
		n->higher->lower = n->lower;
	    } else l->highest = n->lower;

	    if (n->lower) {
		n->lower->higher = n->higher;
	    } else l->lowest = n->higher;

            l->size = l->size - 1;
            free(n);
	}
    }
}


pelem	pdlist_highest(double *pri, pdlist l)
/**************************************************************************
 */
{
    l->current = l->highest;

    if (l->current) {
	if (pri) *pri = l->current->priority;
        return l->current->element;
    }
    else return 0;
}


pelem	pdlist_lower(double *pri, pdlist l)
/**************************************************************************
 */
{
    if (l->current) l->current = l->current->lower;

    if (l->current) {
	if (pri) *pri = l->current->priority;
        return l->current->element;
    }
    else return 0;
}


pelem	pdlist_lowest(double *pri, pdlist l)
/**************************************************************************
 */
{
    l->current = l->lowest;

    if (l->current) {
	if (pri) *pri = l->current->priority;
        return l->current->element;
    }
    else return 0;
}


pelem	pdlist_higher(double *pri, pdlist l)
/**************************************************************************
 */
{
    if (l->current) l->current = l->current->higher;

    if (l->current) {
	if (pri) *pri = l->current->priority;
        return l->current->element;
    }
    else return 0;
}


boolean	pdlist_highest_pri(double *pri, pdlist l)
/**************************************************************************
 */
{
     if (l->highest) {

	if (pri) *pri = l->highest->priority;

	return TRUE;
     }
     else return FALSE;
}


boolean	pdlist_lowest_pri(double *pri, pdlist l)
/**************************************************************************
 */
{
     if (l->lowest) {

	if (pri) *pri = l->lowest->priority;

	return TRUE;
     }
     else return FALSE;
}


boolean	pdlist_in(pdlist l, pelem e)
/**************************************************************************
 */
{
    pdlist_node *n;

    for (n=l->highest; n; n=n->lower) {

	if (n->element == e) return TRUE;

    }
    return FALSE;
}


boolean	pdlist_pri_in(pdlist l, double pri)
/**************************************************************************
 */
{
    pdlist_node *n;

    for (n=l->highest; n && (n->priority >= pri); n=n->lower) {

	if (n->priority == pri) return TRUE;

    }
    return FALSE;
}


long	pdlist_nr_elem(pdlist l, pelem e)
/**************************************************************************
 */
{
    pdlist_node *n;
    long count = 0;

    for (n=l->highest; n; n=n->lower)
	if (n->element == e) count++;

    return count;
}


long	pdlist_nr_pri(pdlist l, double pri)
/**************************************************************************
 */
{
    pdlist_node *n;
    long count = 0;

    for (n=l->highest; n && (n->priority>=pri); n=n->lower)
	if (n->priority == pri) count++;

    return count;
}


boolean	pdlist_is_empty(pdlist l)
/**************************************************************************
 */
{
    return (l->size == 0);
}


boolean	pdlist_is_not_empty(pdlist l)
/**************************************************************************
 */
{
    return (l->size != 0);
}


long	pdlist_size(pdlist l)
/**************************************************************************
 */
{
    return l->size;
}


