/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: pdlist.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 16:59:09 stolz>
 **************************************************************************
 *  MODULE              : pdlist
 *  IMPORT              : basic
 *
 *  FUNCTION
 *	This module provide a priority-list as abstract datatype(adt). The
 *	elements which can be stored in the pdlist are "void-pointers".
 *      The type of the priority is "double". The list can contain more
 *      than one objects with the same priority or elements. But in
 *	this case the order can not be given.
 *
 *      To initialize the pdlist, call pdlist_new(). Now you can add or
 *	remove elements with pdlist_add(), pdlist_remove(). Note that
 *	pdlist_remove() deletes all objects with the same element! Also
 *	do pdlist_remove_pri() delete all objects with the same priority.
 *
 *	It gives 4 functions to traverse through a priority-list.
 *	ATTENTION, they should never be mixed with add/remove calls!
 *	The routines are pdlist_highest(), pdlist_lowest(), pdlist_lower()
 *	and pdlist_higher(). They return the element and the priority
 *	of the element. 0 indicates that no element was found. If you
 *	don't want to know the priority, call this functions with zero
 *      as priority-pointer. Please look also in the example for
 *	the usage of this functions!
 *
 *	For some usefull information of the pdlist, use pdlist_is_empty(),
 *	pdlist_is_not_empty() or pdlist_size().
 *
 *	The highest or lowest priority can be requested by a call of
 *	pdlist_highest_pri() and pdlist_lowest_pri(). The values the
 *	var-parameter pri are only valid if the functions return TRUE.
 *
 *	To found out if a element or a priority exists in a list use
 *	pdlist_in() and pdlist_pri_in(). The get the count of founded
 *	objects with the same elements or priorities, use pdlist_nr_elem()
 *	or pdlist_nr_pri().
 *
 *	To clear all objects use the pdlist_clear() procedure. But to
 *	free the adt pdlist, use pdlist_free(). This frees all the elements
 *	and the adt.
 *
 *  EXAMPLE
 *
 *	#include "pdlist.h"
 *
 *	void main()
 *      {
 *          pdlist l;
 *	    char *text1 = "Hello!";
 *	    char *text2 = "How are you?";
 *	    char *text3 = "Thanks, Fine.";
 *	    char *e;
 *	    double pri;
 *
 *
 *          l = pdlist_new();
 *
 *          pdlist_add(l, 2.0, text1);
 *          pdlist_add(l, -10.0, text3);
 *          pdlist_add(l, 5.0, text1);
 *          pdlist_add(l, 1.0, text2);
 *          pdlist_add(l, 0.0, text2);
 *
 *
 *	    pdlist_remove_pri(l, 0);	// remove last add of text2
 *
 *          for (e=pdlist_highest(0, l); e; e=pdlist_lower(0, l))
 *		fputs(e);
 *
 *          // this should print out:
 *          // > Hello! Hello! How are you? Thanks Fine.
 *
 *
 *	    pdlist_remove(l, text1);	// remove both hellos
 *
 *	    for (e=pdlist_lowest(&pri, l); e; e=pdlist_higher(&pri, l))
 *		printf("%ld : %s\n", pri, e);
 *
 *          // now, you should get:
 *          // > -10 : Thanks Fine.
 *	    // > 1 : How are you?
 *
 *
 *          if (pdlist_lowest_pri(&pri, l))
 *		printf("lowest pri = %f\n", pri);
 *          // output >lowest pri = -10.000
 *
 *
 *	    printf("elements in list: %ld", pdlist_size(l));  // 2, other votes?
 *
 *          pdlist_free(l);
 *	}
 *
 *  NOTE
 *
 *  HISTORY
 *  28-May-94   Creation date. Pass module-test.
 *
 **************************************************************************/

#ifndef PDLIST_H
#define PDLIST_H


#ifndef BASIC_H
#include "basic.h"
#endif

/**************************************************************************
 * defines the abstract datatype pdlist, and the element type.
 */

typedef struct adt_pdlist*	pdlist;
typedef	void*			pelem;

/**************************************************************************
 * init and free a priority-list
 */
pdlist	pdlist_new();
void    pdlist_free(pdlist l);

/**************************************************************************
 * add a element to the pdlist
 */
void	pdlist_add(pdlist l, double pri, pelem e);

/**************************************************************************
 * clear the pdlist, remove elements e, remove all elements
 * with a given priority.
 */
void	pdlist_clear(pdlist l);
void	pdlist_remove(pdlist l, pelem e);
void	pdlist_remove_pri(pdlist l, double pri);

/**************************************************************************
 * traverse from highes to lowest priority
 * contents of pri only valid if returnvalue!=0 and pri!=0
 */
pelem	pdlist_highest(double *pri, pdlist l);
pelem	pdlist_lower(double *pri, pdlist l);

/**************************************************************************
 * traverse from lowest to highest priority
 * contents of pri only valid if returnvalue!=0 and pri!=0
 */
pelem	pdlist_lowest(double *pri, pdlist l);
pelem	pdlist_higher(double *pri, pdlist l);

/**************************************************************************
 * request lowest and highest priority
 * pri will be set if returnvalue is true and pri!=0.
 */
boolean pdlist_highest_pri(double *pri, pdlist l);
boolean pdlist_lowest_pri(double *pri, pdlist l);

/**************************************************************************
 * aquire if a element or a priority exists.
 * returns TRUE if this it exists.
 */
boolean	pdlist_in(pdlist l, pelem e);
boolean	pdlist_pri_in(pdlist l, double pri);

/**************************************************************************
 * Request how many of them exists.
 */
long	pdlist_nr_elem(pdlist l, pelem e);
long	pdlist_nr_pri(pdlist l, double pri);

/**************************************************************************
 * request the size of the pdlist.
 */
boolean	pdlist_is_empty(pdlist l);
boolean	pdlist_is_not_empty(pdlist l);
long	pdlist_size(pdlist l);

#endif
