/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: plist.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:59:17 stolz>
 **************************************************************************
 *  MODULE              : plist
 *  IMPORT		: error, basic
 *
 *  IMPLEMENTATION
 *
 *	The priority list ist implemented as double-linked list.
 *
 *  NOTE
 *      The module specification and definition is given in plist.h!
 *      Define MODULE_TEST to generate a test-routine inclusive a
 *	main()-block.
 *	Define NDEBUG to get rid of debugging-output. This also
 *	disables the plist_print()-call of the module-test.
 *
 *  HISTORY
 *  28-May-94   Creation date. Implementation and validation.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "plist.h"
#include "error.h"


/**************************************************************************
 * structure to store the element and priority-information
 */
typedef struct plist_elem plist_node;
struct plist_elem {
    long	priority;
    pelem	element;
    plist_node*	lower;
    plist_node*	higher;
};


/**************************************************************************
 * definition of the abstract datatype plist
 */
typedef struct adt_plist	adt_plist;
struct adt_plist {
    long size;
    plist_node	*highest, *lowest, *current;
};



plist	plist_new()
/**************************************************************************
 */
{
    plist l;

    l = (plist)malloc(sizeof(adt_plist));
    if (l) {
	l->size    = 0;
	l->highest = 0;
	l->lowest  = 0;
	l->current = 0;
    }
    else error("plist: plist_new() - can't alloc memory for adt plist");

    return l;

}


void    plist_free(plist l)
/**************************************************************************
 */
{
    plist_node *n, *t;

    for (n=l->highest; n; n=t) {
        t = n->lower;
	free(n);
    }

    free(l);
}


void	plist_add(plist l, long pri, pelem e)
/**************************************************************************
 */
{
    plist_node *n;
    plist_node *np;

    if (l->size == LONG_MAX)
	error("plist: plist_add() - max number of elements reached");

    n = (plist_node *)malloc(sizeof(plist_node));

    if (n==0) error("plist: plist_add() - can't alloc memory for a plist-node");

    n->priority = pri;
    n->element = e;
    l->size = l->size + 1;

    /* search first element which has a lower or same priority
     */
    for (np=l->highest; np && (np->priority > pri); np=np->lower)
	;

    /* initialize the node-pointers
     */
    n->lower = np;
    n->higher = (np ? np->higher : l->lowest);

    /* set the pointers of the higher and lower elements
     */
    if (n->lower) n->lower->higher = n;
    if (n->higher) n->higher->lower = n;

    /* set the highest/lowest pointer of the plist.
     */
    if (np==l->highest) l->highest = n;
    if (np==0) l->lowest  = n;
}


void	plist_clear(plist l)
/**************************************************************************
 */
{
    plist_node *n, *t;

    for (n=l->highest; n; n=t) {
        t = n->lower;
	free(n);
    }

    l->size = 0;
    l->highest = 0;
    l->lowest  = 0;
}


void	plist_remove(plist l, pelem e)
/**************************************************************************
 * FUNCTION
 *   Remove all nodes with the element e!
 */
{
    plist_node *n, *t;

    for (n=l->highest; n; n=t) {
        t = n->lower;
	if (n->element == e) {

	    if (n->higher) {
		n->higher->lower = n->lower;
	    } else l->highest = n->lower;

	    if (n->lower) {
		n->lower->higher = n->higher;
	    } else l->lowest = n->higher;

            l->size = l->size - 1;
            free(n);
	}
    }
}

void	plist_remove_pri(plist l, long pri)
/**************************************************************************
 */
{
    plist_node *n, *t;

    for (n=l->highest; n; n=t) {
        t = n->lower;
	if (n->priority == pri) {

	    if (n->higher) {
		n->higher->lower = n->lower;
	    } else l->highest = n->lower;

	    if (n->lower) {
		n->lower->higher = n->higher;
	    } else l->lowest = n->higher;

            l->size = l->size - 1;
            free(n);
	}
    }
}


pelem	plist_highest(long *pri, plist l)
/**************************************************************************
 */
{
    l->current = l->highest;

    if (l->current) {
	if (pri) *pri = l->current->priority;
        return l->current->element;
    }
    else return 0;
}


pelem	plist_lower(long *pri, plist l)
/**************************************************************************
 */
{
    if (l->current) l->current = l->current->lower;

    if (l->current) {
	if (pri) *pri = l->current->priority;
        return l->current->element;
    }
    else return 0;
}


pelem	plist_lowest(long *pri, plist l)
/**************************************************************************
 */
{
    l->current = l->lowest;

    if (l->current) {
	if (pri) *pri = l->current->priority;
        return l->current->element;
    }
    else return 0;
}


pelem	plist_higher(long *pri, plist l)
/**************************************************************************
 */
{
    if (l->current) l->current = l->current->higher;

    if (l->current) {
	if (pri) *pri = l->current->priority;
        return l->current->element;
    }
    else return 0;
}


boolean	plist_highest_pri(long *pri, plist l)
/**************************************************************************
 */
{
     if (l->highest) {

	if (pri) *pri = l->highest->priority;

	return TRUE;
     }
     else return FALSE;
}


boolean	plist_lowest_pri(long *pri, plist l)
/**************************************************************************
 */
{
     if (l->lowest) {

	if (pri) *pri = l->lowest->priority;

	return TRUE;
     }
     else return FALSE;
}


boolean	plist_in(plist l, pelem e)
/**************************************************************************
 */
{
    plist_node *n;

    for (n=l->highest; n; n=n->lower) {

	if (n->element == e) return TRUE;

    }
    return FALSE;
}


boolean	plist_pri_in(plist l, long pri)
/**************************************************************************
 */
{
    plist_node *n;

    for (n=l->highest; n && (n->priority >= pri); n=n->lower) {

	if (n->priority == pri) return TRUE;

    }
    return FALSE;
}


long	plist_nr_elem(plist l, pelem e)
/**************************************************************************
 */
{
    plist_node *n;
    long count = 0;

    for (n=l->highest; n; n=n->lower)
	if (n->element == e) count++;

    return count;
}


long	plist_nr_pri(plist l, long pri)
/**************************************************************************
 */
{
    plist_node *n;
    long count = 0;

    for (n=l->highest; n && (n->priority>=pri); n=n->lower)
	if (n->priority == pri) count++;

    return count;
}


boolean	plist_is_empty(plist l)
/**************************************************************************
 */
{
    return (l->size == 0);
}


boolean	plist_is_not_empty(plist l)
/**************************************************************************
 */
{
    return (l->size != 0);
}


long	plist_size(plist l)
/**************************************************************************
 */
{
    return l->size;
}



/****************************************************************************
 * MODULE-TEST-routines
 */

#ifdef MODULE_TEST

#define N       8


/* test data
 */
static long test_pri[N] = {
	10, -1, 4, 5, 3, 3, 10, -2
};

static pelem test_elem[N] = {
	(pelem)1, (pelem)2, (pelem)3, (pelem)2, (pelem)4, (pelem)1, (pelem)2, (pelem)5
};



void plist_print(plist l)
{
    plist_node *n;

    fprintf(stderr, "PLIST(%ld) %s", l->size, ((l->size != 0) ? " =" : " = empty"));

    for (n=l->highest; n; n=n->lower)
	fprintf(stderr, " (%ld, %p)", n->priority, n->element);

    fprintf(stderr, "\n");
}


int plist_test()
{
    int failed, errors = 0;
    char index;

    plist l;
    long pri = 0;
    pelem e;
    int i;
    boolean b;

    index = 'a';
    fprintf(stderr, "PLIST-TEST from %s %s\n"
            "I. generate an empty list via plist_new(..)\n",
	    __DATE__, __TIME__);

    l = plist_new();

    if (failed = (!plist_is_empty(l))) errors++;
    fprintf(stderr, "  %c) test function plist_is_empty() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (plist_is_not_empty(l))) errors++;
    fprintf(stderr, "  %c) test function plist_is_not_empty() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (0 != plist_size(l))) errors++;
    fprintf(stderr, "  %c) test function plist_size() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (0!=plist_highest(0, l))) errors++;
    fprintf(stderr, "  %c) use function plist_highest() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (0!=plist_lowest(0, l))) errors++;
    fprintf(stderr, "  %c) use function plist_lowest() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (0!=plist_lower(0, l))) errors++;
    fprintf(stderr, "  %c) use function plist_lower() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (0!=plist_higher(0, l))) errors++;
    fprintf(stderr, "  %c) use function plist_higher() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (plist_highest_pri(0, l))) errors++;
    fprintf(stderr, "  %c) test function plist_highest_pri() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (plist_lowest_pri(0, l))) errors++;
    fprintf(stderr, "  %c) test function plist_lowest_pri() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));


    fprintf(stderr, "II. Add %d items to the empty plist\n", N);
    index = 'a';
    for (i=0; i<N; i++) {
	plist_add(l, test_pri[i], test_elem[i]);
	fprintf(stderr, "     Adding Element %p with priority %ld\n",
		test_elem[i], test_pri[i]);
    }

#ifdef DEBUG_DETAIL
    fprintf(stderr, "  Print out contents of the plist:\n");
    plist_print(l);
#endif


    if (failed = (plist_is_empty(l))) errors++;
    fprintf(stderr, "  %c) test function plist_is_empty() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (!plist_is_not_empty(l))) errors++;
    fprintf(stderr, "  %c) test function plist_is_not_empty() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (N != plist_size(l))) errors++;
    fprintf(stderr, "  %c) test function plist_size() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = ((pelem)2!=plist_highest(0, l))) errors++;
    fprintf(stderr, "  %c) use function plist_highest() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = ((pelem)5!=plist_lowest(0, l))) errors++;
    fprintf(stderr, "  %c) use function plist_lowest() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = ((pelem)2!=plist_higher(0, l))) errors++;
    fprintf(stderr, "  %c) use function plist_higher() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = ((pelem)5!=plist_lower(0, l))) errors++;
    fprintf(stderr, "  %c) use function plist_lower() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (0!=plist_lower(0, l))) errors++;
    fprintf(stderr, "  %c) use function plist_lower() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    failed = !(b=plist_highest_pri(&pri, l));
    if (failed) errors++;
    else failed = (pri != 10);
    fprintf(stderr, "  %c) plist_highest_pri() returns %s, %ld : %s\n", index++,
                    (b ? "true" : "false"), (b ? pri : 0),
			(failed ? "FAILED" : "PASSED"));

    failed = !(b=plist_lowest_pri(&pri, l));
    if (failed) errors++;
    else failed = (pri != -2);
    fprintf(stderr, "  %c) plist_lowest_pri() returns %s, %ld : %s\n", index++,
                    (b ? "true" : "false"), (b ? pri : 0),
			(failed ? "FAILED" : "PASSED"));

    pri = plist_nr_elem(l, (pelem)2);
    failed = (pri != 3); if (failed) errors++;
    fprintf(stderr, "  %c) request plist_nr_elem(2) returns %ld : %s\n", index++,
            pri,
            (failed ? "FAILED" : "PASSED"));

    pri = plist_nr_pri(l, 10);
    failed = (pri != 2); if (failed) errors++;
    fprintf(stderr, "  %c) request plist_nr_pri(2) returns %ld : %s\n", index++,
            pri,
            (failed ? "FAILED" : "PASSED"));


    fprintf(stderr, "III. remove elements 2 and priorities -10\n"
		    "     with plist_remove() and plist_remove_pri()\n");
    index = 'a';
    plist_remove_pri(l, 10);
    plist_remove(l, (pelem)2);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "  Print out contents of the plist:\n");
    plist_print(l);
#endif

    if (failed = (N-4 != (pri=plist_size(l)))) errors++;
    fprintf(stderr, "  %c) plist_size() returns %ld: %s\n", index++, pri,
			(failed ? "FAILED" : "PASSED"));

    pri = 0;
    e = plist_highest(&pri, l);
    failed = ((pelem)3 != e) || (pri != 4);
    if (failed) errors++;
    fprintf(stderr, "  %c) plist_highest() returns %ld, %p: %s\n", index++,
		        pri, e,
			(failed ? "FAILED" : "PASSED"));

    pri = 0;
    e = plist_lowest(&pri, l);
    failed = ((pelem)5 != e) || (pri != -2);
    if (failed) errors++;
    fprintf(stderr, "  %c) plist_lowest() returns %ld, %p: %s\n", index++,
		        pri, e,
			(failed ? "FAILED" : "PASSED"));

    b = plist_in(l, (pelem)4);
    failed = !b;
    if (failed) errors++;
    fprintf(stderr, "  %c) plist_in(4) : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    b = plist_pri_in(l, 3);
    failed = !b;
    if (failed) errors++;
    fprintf(stderr, "  %c) plist_pri_in(3) : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    fprintf(stderr, "IV.  empty list with plist_clear()\n");
    index = 'a';
    plist_clear(l);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "  Print out contents of the plist:\n");
    plist_print(l);
#endif

    if (failed = (0 != plist_size(l))) errors++;
    fprintf(stderr, "  %c) test function plist_size() : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));


    plist_free(l);


    fprintf(stderr, "\nTotal amount of errors: %d. Test %s!\n\n",
			errors, errors ? "FAILED" : "PASSED");


    return errors;
}

int main(int argc, char **argv)
{
    return plist_test();
}

#endif
