/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: plist.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 16:59:22 stolz>
 **************************************************************************
 *  MODULE              : plist
 *  IMPORT              : basic
 *
 *  FUNCTION
 *	This module provide a priority-list as abstract datatype(adt). The
 *	elements which can be stored in the plist are "void-pointers".
 *      The type of the priority is "long". The list can contain more
 *      than one objects with the same priority or elements. But in
 *	this case the order can not be given.
 *
 *      To initialize the plist, call plist_new(). Now you can add or
 *	remove elements with plist_add(), plist_remove(). Note that
 *	plist_remove() deletes all objects with the same element! Also
 *	do plist_remove_pri() delete all objects with the same priority.
 *
 *	It gives 4 functions to traverse through a priority-list.
 *	ATTENTION, they should never be mixed with add/remove calls!
 *	The routines are plist_highest(), plist_lowest(), plist_lower()
 *	and plist_higher(). They return the element and the priority
 *	of the element. 0 indicates that no element was found. If you
 *	don't want to know the priority, call this functions with zero
 *      as priority-pointer. Please look also in the example for
 *	the usage of this functions!
 *
 *	For some usefull information of the plist, use plist_is_empty(),
 *	plist_is_not_empty() or plist_size().
 *
 *	The highest or lowest priority can be requested by a call of
 *	plist_highest_pri() and plist_lowest_pri(). The values the
 *	var-parameter pri are only valid if the functions return TRUE.
 *
 *	To found out if a element or a priority exists in a list use
 *	plist_in() and plist_pri_in(). The get the count of founded
 *	objects with the same elements or priorities, use plist_nr_elem()
 *	or plist_nr_pri().
 *
 *	To clear all objects use the plist_clear() procedure. But to
 *	free the adt plist, use plist_free(). This frees all the elements
 *	and the adt.
 *
 *  EXAMPLE
 *
 *	#include "plist.h"
 *
 *	void main()
 *      {
 *          plist l;
 *	    char *text1 = "Hello!";
 *	    char *text2 = "How are you?";
 *	    char *text3 = "Thanks, Fine.";
 *	    char *e;
 *	    long pri;
 *
 *
 *          l = plist_new();
 *
 *          plist_add(l, 2, text1);
 *          plist_add(l, -10, text3);
 *          plist_add(l, 5, text1);
 *          plist_add(l, 1, text2);
 *          plist_add(l, 0, text2);
 *
 *
 *	    plist_remove_pri(l, 0);	// remove last add of text2
 *
 *          for (e=plist_highest(0, l); e; e=plist_lower(0, l))
 *		fputs(e);
 *
 *          // this should print out:
 *          // > Hello! Hello! How are you? Thanks Fine.
 *
 *
 *	    plist_remove(l, text1);	// remove both hellos
 *
 *	    for (e=plist_lowest(&pri, l); e; e=plist_higher(&pri, l))
 *		printf("%ld : %s\n", pri, e);
 *
 *          // now, you should get:
 *          // > -10 : Thanks Fine.
 *	    // > 1 : How are you?
 *
 *
 *          if (plist_lowest_pri(&pri, l))
 *		printf("lowest pri = %ld\n", pri);
 *          // output >lowest pri = -10
 *
 *
 *	    printf("elements in list: %ld", plist_size(l));  // 2, other votes?
 *
 *          plist_free(l);
 *	}
 *
 *  NOTE
 *
 *  HISTORY
 *  28-May-94   Creation date. Pass module-test.
 *
 **************************************************************************/

#ifndef PLIST_H
#define PLIST_H


#ifndef BASIC_H
#include "basic.h"
#endif

/**************************************************************************
 * defines the abstract datatype plist, and the element type.
 */

typedef struct adt_plist*	plist;
typedef	void*			pelem;

/**************************************************************************
 * init and free a priority-list
 */
plist	plist_new();
void    plist_free(plist l);

/**************************************************************************
 * add a element to the plist
 */
void	plist_add(plist l, long pri, pelem e);

/**************************************************************************
 * clear the plist, remove elements e, remove all elements
 * with a given priority.
 */
void	plist_clear(plist l);
void	plist_remove(plist l, pelem e);
void	plist_remove_pri(plist l, long pri);

/**************************************************************************
 * traverse from highes to lowest priority
 * contents of pri only valid if returnvalue!=0 and pri!=0
 */
pelem	plist_highest(long *pri, plist l);
pelem	plist_lower(long *pri, plist l);

/**************************************************************************
 * traverse from lowest to highest priority
 * contents of pri only valid if returnvalue!=0 and pri!=0
 */
pelem	plist_lowest(long *pri, plist l);
pelem	plist_higher(long *pri, plist l);

/**************************************************************************
 * request lowest and highest priority
 * pri will be set if returnvalue is true and pri!=0.
 */
boolean plist_highest_pri(long *pri, plist l);
boolean plist_lowest_pri(long *pri, plist l);

/**************************************************************************
 * aquire if a element or a priority exists.
 * returns TRUE if this it exists.
 */
boolean	plist_in(plist l, pelem e);
boolean	plist_pri_in(plist l, long pri);

/**************************************************************************
 * Request how many of them exists.
 */
long	plist_nr_elem(plist l, pelem e);
long	plist_nr_pri(plist l, long pri);

/**************************************************************************
 * request the size of the plist.
 */
boolean	plist_is_empty(plist l);
boolean	plist_is_not_empty(plist l);
long	plist_size(plist l);

#endif
