/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: polygon.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 16:59:28 stolz>
 **************************************************************************
 *  MODULE              : polygon
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in polygon.h!
 *
 *  HISTORY
 *  08-Aug-94   Creation date
 *  11-Aug-94   Passed module test.
 *  15-Aug-94   Split Index-part from the triangulation. ->ipolygon3...
 *  23-Sep-94   use point[j] instead point[p[j]] in the point-in-triangle
 *              test. Fixed it!
 *
 **************************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "polygon.h"
#include "error.h"





void convex_polygon3_triangulation(
  vector3 *point,
  int nr_points,
  void (*f)(vector3 *))
/**************************************************************************
 */
{
    vector3 points[3];
    int i;

    if (nr_points<3 || f==0) return;

    v3_let(points[0], point[0]);
    nr_points--;
    for (i=1; i<nr_points; i++) {
	v3_let(points[1], point[i]);
	v3_let(points[2], point[i+1]);
	f(points);
    }
}



void convex_ipolygon3_triangulation(
  vector3 *point,
  int nr_points,
  int *index,
  void (*f)(vector3 *))
/**************************************************************************
 */
{
    vector3 points[3];
    int i;

    if (nr_points<3 || f==0) return;

    v3_let(points[0], point[index[0]]);
    nr_points--;
    for (i=1; i<nr_points; i++) {
	v3_let(points[1], point[index[i]]);
	v3_let(points[2], point[index[i+1]]);
	f(points);
    }
}



int ipolygon3_triangulation(
  vector3 *point,
  int nr_points,
  int *index,
  vector3 normal,
  void (*f)(vector3 *))
/**************************************************************************
 */
{
    vector3 points[3];
    int i, j, k, l;
    vector3 p_rel, face_dir, face_2, face_3, t;
    static int *p = 0;		/* index-array to point[] */
    static int p_size = 0;	/* size of p */
    double d;
    int nr_t = 0;

#if DEBUG_DETAIL
    printf("p3_triangulation() - split a convex/concave polygon into triangles.\n");
#endif

    if (nr_points<3 || f==0) return 0;

    /* alloc memory for index-array
     */
    if (p_size < nr_points) {
        p_size = nr_points;

	if (p) p = (int *)realloc(p, p_size * sizeof(int));
	else p = (int *)malloc(p_size * sizeof(int));

	if (p==0) error("polygon: polygon3_triangulation() - can't alloc memory");
    }

    /* initialize index-array
     */
    if (index) {
        for (i=0; i<nr_points; i++) p[i] = index[i];
    } else {
        for (i=0; i<nr_points; i++) p[i] = i;
    }

    /* remove a triangle from the polygon-face
     * until it has only 3 points
     */
    if (nr_points>3) for (i=0; i<nr_points; i++) {


	j = (i+1) % nr_points;
        k = (i+2) % nr_points;

#if DEBUG_DETAIL
	printf("p3_triangulation: can triangle (%d %d %d) splitted from poly?\n",
		i, j, k);
#endif

        v3_sub(p_rel, point[p[k]], point[p[j]]);
        v3_sub(t, point[p[j]], point[p[i]]);
	v3_mul(face_dir, t, normal);
	d = v3_product(face_dir, p_rel) /* not necessary: */ / v3_sum(face_dir); 
	
   
#if DEBUG_DETAIL
	printf("p3_triangulation: inside offset = %f\n", d);
#endif
        if ( d > 0.0) {

#if DEBUG_DETAIL
	    printf("p3_triangulation: sides %d->%d->%d are convex, good!\n", i, j, k);
#endif

	    v3_sub(t, point[p[k]], point[p[j]]);
	    v3_mul(face_2, t, normal);
            v3_sub(t, point[p[i]], point[p[k]]);
	    v3_mul(face_3, t, normal);

	    /* lines i->j->k are konvex,
             * check if a point lies in the triangle (i,j,k)
	     */
	    for (l = (k+1)%nr_points; l != i; l=(l+1)%nr_points) {

#if DEBUG_DETAIL
		printf("p3_triangulation: lies point %d in the triangle?\n", l);
#endif

                /* point l in the side of the plane i->j ?
		 */
                v3_sub(p_rel, point[p[l]], point[p[j]]);
		if (v3_product(p_rel, face_dir) >= 0.0) { /* yes */

#if DEBUG_DETAIL
		    printf("p3_triangulation: point lies inside plain (%d %d)\n", i, j);
#endif

                    /* point l in the side of the plane j->k ?
		     */
		    v3_sub(p_rel, point[p[l]], point[p[k]]);

		    if (v3_product(p_rel, face_2) >= 0.0) { /* yes */

#if DEBUG_DETAIL
		        printf("p3_triangulation: point lies inside plain (%d %d)\n", j, k);
#endif

	                /* point l in the side of the plane k->i ?
			 */
		        v3_sub(p_rel, point[p[l]], point[p[i]]);

		        if (v3_product(p_rel, face_3) >= 0.0) {

#if DEBUG_DETAIL
		            printf("p3_triangulation: point lies inside plain (%d %d)\n", k, i);
			    printf("p3_triangulation: Yes, the point lies in the triangle!\n");
#endif
			    break; /* yes */
			}
		    }
		}
#if DEBUG_DETAIL
		printf("p3_triangulation: no\n");
#endif
            }
	    if (l==i) {
#if DEBUG_DETAIL
		printf("p3_triangulation: ok, build triangle (%d %d %d), remove point %d\n",
		       i, j, k, j);
#endif
		/* no other point lies in triangle,
		 * build triangle (i,j,k)
                 */
		v3_let(points[0], point[p[i]]);
		v3_let(points[1], point[p[j]]);
		v3_let(points[2], point[p[k]]);
		f(points);
		nr_t ++;

                /* remove point j from the index-array
		 */
		if (k>j) for (;k<nr_points; k++) p[k-1] = p[k];
		nr_points--;
                i = -1;
                if (nr_points == 3) break;
            }
        }
    }

    if (nr_points!=3) {
	return -1;
#if DEBUG_DETAIL
	printf("p3_triangulation: !! bad polygon !!\n normale = {%f, %f, %f}\n polygon-idx =",
		v3_elems(normal));
	for (i=0; i<nr_points; i++)
	    printf(" %d", p[i]);
	
	printf("\n polygon-points =");
	for (i=0; i<nr_points; i++)
	    printf(" {%f,%f,%f}", v3_elems(point[p[i]]));
#endif
	warning("polygon: polygon3_triangulation() - incomplete triangulation");
	return -1;
    }

    v3_let(points[0], point[p[0]]);
    v3_let(points[1], point[p[1]]);
    v3_let(points[2], point[p[2]]);
    f(points);
    nr_t++;

    return nr_t;
}



int polygon3_triangulation(
  vector3 *point,
  int nr_points,
  vector3 normal,
  void (*f)(vector3 *))
/**************************************************************************
 */
{
    return ipolygon3_triangulation(point, nr_points, 0, normal, f);
}
