/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: polygon.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 16:59:33 stolz>
 **************************************************************************
 *  MODULE              : polygon
 *
 *  FUNCTION
 *
 *  NOTE
 *	Compile with NDEBUG defined if you don't want any debugging
 *	output!
 *
 *  HISTORY
 *  08-Aug-94   Creation date.
 *  11-Aug-94   Passed module test.
 *  15-Aug-94   Add index-polygon3-triangulation routines.
 *
 **************************************************************************/
#ifndef POLYGON_H
#define POLYGON_H

#ifndef VECTOR_H
#include "vector.h"
#endif

void convex_polygon3_triangulation(
  vector3 *point,
  int nr_points,
  void (*f)(vector3 *));
/****************************************************************************
 * NAME
 *	convex_polygon3_triangulation(...)	split convex polygon into
 *					        triangles.
 *
 * FUNCTION
 *	A convex polygon face will be triangulated. For each triangle
 *	which can be splitted from the polygon face, the function f
 *	will be called. The Parameter of the function f are the array
 *	to the 3 edges of the triangle.
 *
 *	If <f> is NULL or <nr_points> is smaller than 3, the function
 *	do nothing!
 *
 * INPUT
 *	point           array of the polygon-points
 *	nr_points	number of points of this polygon. the last point
 *			is connected to the first!
 *	f		function which will be called for an triangle-
 *			creation.
 *
 * OUTPUT
 *	NONE.
 *
 * RESULTS
 *      None. Only the function <f> will be called for every triangle.
 *
 * NOTE
 *	This routine works only proper for a 
 *      convex non-overlapping polygon!
 *
 * EXAMPLE
 *
 *	vector3 simple_poly[5] = {
 *		-25.0,   5.0,   0.0,
 *		-10.0,  15.0,   0.0,
 *		 10.0,  10.0,   0.0,
 *		  0.0, -10.0,   0.0,
 *		-15.0, -10.0,   0.0
 *	};
 *
 *      void print_triangle(vector3 *t)
 *	{
 *          printf("triangle (%f %f %f) - (%f %f %f) - (%f %f %f)\n",
 *		    v3_elems(t[0]), v3_elems(t[1]), v3_elems(t[2]);
 *
 *	}
 *
 *
 *	{
 *	...
 *
 *          convex_polygon3_triangulation(simple_poly, 5, print_triangle);
 *	}
 *
 *
 ****************************************************************************
 *
 */


int polygon3_triangulation(
  vector3 *point,
  int nr_points,
  vector3 normal,
  void (*f)(vector3 *));
/****************************************************************************
 * NAME
 *	polygon3_triangulation(...)	split a polygon into triangles.		        triangles.
 *
 * FUNCTION
 *	A non-overlapping polygon face will be triangulated. The polygon
 *	face can be convex or concave.  For each triangle
 *	which can be splitted from the polygon face, the function f
 *	will be called. The Parameter of the function f are the array
 *	to the 3 edges of the triangle.
 *
 *	If <f> is NULL or <nr_points> is smaller than 3, the function
 *	do nothing!
 *
 * INPUT
 *	point           array of the polygon-points
 *	nr_points	number of points of this polygon. the last point
 *			is connected to the first!
 *	normal		normal-face of the polygon to determine inside/
 *			outside. the inside lies on the right if you
 *			go on the polygon-points from 0 ... nr_points-1.
 *			(use the right-hand-rule:
 *				thumb : (point[i]->point[i+1])
 *				index-finger: normal
 *                              middle-finger: inside of the polygon!)
 *	f		function which will be called for an triangle-
 *			creation.
 *
 * OUTPUT
 *	An warning on stderr if the algorithm cannot complete the
 *	triangulation.
 *
 * RESULTS
 *      The number of created triangles will be returned. A -1 indicates
 *      a incomplete triangulation (caused by a overlapping polygon or
 *      wrong normal-direction).
 *      For every created triangle, the function <f> will be called.
 *
 * NOTE
 *	This routine works only proper for a
 *      convex or cancave non-overlapping polygon!
 *      If you run this function with a wrong normal, you will get
 *	a warning on stderr and a -1 result,
 *      but probably only after some calls of <f> with wrong triangles.
 *
 * EXAMPLE
 *
 *	vector3 simple_poly[5] = {
 *		-25.0,   5.0,   0.0,
 *		-10.0,  -5.0,   0.0,
 *		 10.0,  10.0,   0.0,
 *		  0.0, -10.0,   0.0,
 *		-15.0, -10.0,   0.0
 *	};
 *
 *      void print_triangle(vector3 *t)
 *	{
 *          printf("triangle (%f %f %f) - (%f %f %f) - (%f %f %f)\n",
 *		    v3_elems(t[0]), v3_elems(t[1]), v3_elems(t[2]));
 *
 *	}
 *
 *
 *	{
 *	    vector3 normal = {0.0, 0.0, 1.0};
 *	...
 *
 *          polygon3_triangulation(simple_poly, 5, normal, print_triangle);
 *	}
 *
 *
 ****************************************************************************
 *
 */


void convex_ipolygon3_triangulation(
  vector3 *point,
  int nr_points,
  int *index,
  void (*f)(vector3 *));
/****************************************************************************
 * FUNCTION
 *	same as convex_polygon3_triangulation(), but the polygon is defined
 *	as an index-array to the absolute points in the point-parameter.
 */


int ipolygon3_triangulation(
  vector3 *point,
  int nr_points,
  int *index,
  vector3 normal,
  void (*f)(vector3 *));
/****************************************************************************
 * FUNCTION
 *	same as polygon3_triangulation(), but the polygon is defined
 *	as an index-array to the absolute points in the point-parameter.
 */



#endif
