/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: pvm_server.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/03/10 19:36:35 stolz>
 **************************************************************************
 *  MODULE              : pvm_serve
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in pvm_server.h!
 *
 *  HISTORY
 *  09-Jul-94   Creation date.
 *  10-Mar-95   change sync-mode for a better performance.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>    /* for sleep() */

#ifndef NO_PVM
#include "pvm3.h"
#endif

#include "pvm_server.h"
#include "mob_world.h"
#include "mw_object.h"
#include "mob_world_private.h"
#include "mobs_interface_def.h"
#include "ivd.h"
#include "error.h"
#include "basic.h"

static int ps_tid = -1;
static boolean ps_sync_mode = FALSE;

/*
#define DEBUG_DETAIL 1
*/


void ps_synchron(boolean mode)
{
    ps_sync_mode = mode;
}


void ps_open()
{
#ifndef NO_PVM
    ps_tid = pvm_mytid();
    if (ps_tid<0) {
	warning("PVM not started! ");
	ps_tid = -1;
    }
#endif
#ifdef DEBUG_DETAIL
    printf("pvm_server: my tid is %d\n", ps_tid);
#endif
}

int  ps_server_tid()
{
    return ps_tid;
}


void ps_close(mob_world mw)
{
    if (ps_tid>=0) {
    }
    ps_tid = -1;
}

void ps_sync_obj(mw_object obj, double t)
{
    int nr_tries = 4;

#ifndef NO_PVM
#ifdef DEBUG_DETAIL
    printf("ps_sync_obj(obj=%p, t=%fs)\n", obj, t);
#endif

    if (obj && obj->remote_tid>0 && obj->remote_nr_ios==0
	&& obj->remote_time < t) {
	    
#ifdef DEBUG_DETAIL
	printf("ps_sync_obj: recv for object %p\n", obj);
#endif
	
	ps_recv(obj->world);
#ifdef DEBUG_DETAIL
	printf("ps_sync_obj: stat remote tid %d\n", obj->remote_tid);
#endif

	if (pvm_pstat(obj->remote_tid) != PvmOk) {
	    mw_object_set_remote(obj, -1);
	    obj->remote_nr_ios = 0;
	    return;
	}

	if (!ps_sync_mode) return;

	while (obj->remote_time < t && nr_tries>0) {

#ifdef DEBUG_DETAIL
	    printf("ps_sync_obj: sleep for one second\n");
#endif
	    sleep(1);

#ifdef DEBUG_DETAIL
	    printf("ps_sync_obj: recv for object %p\n", obj);
#endif
	    ps_recv(obj->world);
	    nr_tries--;
	}
	if (obj->remote_time < t) {
#ifdef DEBUG_DETAIL
	    printf("ps_sync_obj:  Time-out! Client %d don't respond to obj '%s'\n",
		    obj->remote_tid, obj->name);
	    printf("ps_sync_obj: Currently t=%fs, remote_t=%fs, remote_nr_ios=%d\n",
		   t, obj->remote_time, obj->remote_nr_ios);
#endif
	    if (pvm_pstat(obj->remote_tid) != PvmOk) {
		printf("ps_sync_obj: Client %d not reachable via pvm, close connection!\n",
		       obj->remote_tid);
		
		mw_object_set_remote(obj, -1);
		obj->remote_nr_ios = 0;
	    }
	}
    }
#endif
}

void ps_recv(mob_world mw)
{
    int bufid, bytes, msgtag, client_tid;
    long id;		/* object remote-nr */
    double dt;
    int slen;		/* string-len */
    mw_object obj;
    static char *buf = 0;
    static int buf_size = 0;

#ifndef NO_PVM
    if (mw==0 || ps_tid<0) return;
    if (buf==0) {
	buf = malloc(301 * sizeof(char));
	buf_size = 300;
	if (buf == 0) 
	    error("pvm_server: ps_recv() - can't alloc memory for string-buffer");
    }
    
    while(pvm_probe(-1, -1)) {
	
	bufid = pvm_recv(-1,-1);
	pvm_bufinfo(bufid, &bytes, &msgtag, &client_tid);

	switch(msgtag) {
	case MID_SEND:
	    pvm_upklong(&id, 1, 1);
	    pvm_upkdouble(&dt, 1, 1);
	    pvm_upkint(&slen, 1, 1);
	    if (slen > buf_size) {
		free(buf);
		slen = slen + buf_size;
		buf = malloc((slen+1) * sizeof(char));
		if (buf == 0) 
		    error("pvm_server: ps_recv() - can't alloc memory for string-buffer");
	    }
	    if (slen>0) pvm_upkstr(buf); else buf[0] = '\0';

#ifdef DEBUG_DETAIL
	    printf("Receiving from client %d the cmd '%s' for id=%ld with dt=%f\n",
		       client_tid, buf, id, dt);
#endif	    
	    obj = mob_world_get_remote(mw, id);
	    if (obj) {
		obj->remote_time += dt;
#ifdef DEBUG_DETAIL
	        printf("Send cmd to obj %p at time %fs\n", 
			obj, obj->remote_time);
#endif	    
		mw_object_send_cmd(obj, obj->remote_time, buf);

	        obj->remote_nr_ios = obj->remote_nr_ios + 1;

	    } else {
		char *error_reply = "This object don't understand messages";
		int len;

		len = strlen(error_reply);
		dt = 0.0;
#ifdef DEBUG_DETAIL
		printf("Send error-reply to client %d\n", client_tid);
#endif
		pvm_initsend(PvmDataDefault);
		pvm_pkdouble(&dt, 1, 1);
		pvm_pkint(&len, 1, 1);
		if (len>0) pvm_pkstr(error_reply);
		pvm_send(client_tid, MID_REPLY);
	    }
	    break;


	case MID_OPEN:
	    pvm_upkint(&client_tid, 1, 1);
	    pvm_upkstr(buf);
#ifdef DEBUG_DETAIL
	    printf("Received from client %d the open request for '%s'\n",
		    client_tid, buf);

#endif	

	    obj = mw_object_by_name(mw, buf);
	    if (obj==0 || mw_object_get_remote_tid(obj) != -1) {
#ifdef DEBUG_DETAIL
		printf("Sending accept with zero-id...\n");
#endif
		id = 0;
		pvm_initsend(PvmDataDefault);
	        pvm_pklong(&id, 1, 1);
	        pvm_send(client_tid, MID_ACCEPT);
	    }
	    else {
	        mw_object_set_remote(obj, client_tid);

	        obj->remote_nr_ios = 0;
		id = (long)(obj->remote_nr);
#ifdef DEBUG_DETAIL
		printf("Sending accept with id %ld\n", id);
#endif
		pvm_initsend(PvmDataDefault);
	        pvm_pklong(&id, 1, 1);
	        pvm_send(client_tid, MID_ACCEPT);
	    }
	    break;

	case MID_CLOSE:
	    pvm_upklong(&id, 1, 1);
	    pvm_upkdouble(&dt, 1, 1);
#ifdef DEBUG_DETAIL
	    printf("Received from client %d the close request for id=%ld, dt = %f\n",
		    client_tid, id, dt);
#endif
	    obj = mob_world_get_remote(mw, (int)id);
	    if (obj) {
		mw_object_set_remote(obj, -1);
		obj->remote_nr_ios = 0;
	    }
	    break;

	case MID_REQ_FRAME:
	    {
		int camera_nr, width, height, color;
		mw_object camera = 0;

		pvm_upklong(&id, 1, 1);
		pvm_upkdouble(&dt, 1, 1);
		pvm_upkint(&camera_nr, 1, 1);
		pvm_upkint(&width, 1, 1);
		pvm_upkint(&height, 1, 1);
		pvm_upkint(&color, 1, 1);
		
		camera_nr += CAMERA_OFFSET;
#ifdef DEBUG_DETAIL
		printf("ps_sync: Received from client %d the frame request with camera=%d, width=%d, height=%d, color=%d, for id=%ld, dt = %f\n",
		       client_tid, camera_nr, width, height, color, id, dt);
#endif

		dt = 0.0;

		obj = mob_world_get_remote(mw, id);
		if (obj)
		    camera = mw_object_get_sensor(obj, camera_nr);

		if (camera && mw_object_function(camera)==MWF_CAMERA) {
#ifdef DEBUG_DETAIL
		    printf("ps_sync: send request to ivd-server\n");
#endif
		    ivd_req_camera_frame(camera, client_tid, 
					 width, height, color);
		}
		else {
#ifdef DEBUG_DETAIL
		    printf("ps_sync: send zero-frame to tid %d\n", client_tid);
#endif
		    ivd_send_zero_frame(client_tid);
		}

	    }
	    break;

	default:
#ifdef DEBUG_DETAIL
	    printf("Unknown message with tag %d.\n", msgtag);
#endif
	    break;
	}
    }
#endif
}


void ps_reply(mw_object obj, char *reply)
{
    int len;
    double dt, t;

#ifndef NO_PVM
    if (ps_tid<0 || obj==0 || obj->remote_tid<0 || obj->remote_nr_ios<=0) return;

    obj->remote_nr_ios = obj->remote_nr_ios - 1;

    len = strlen(reply);

    t = scheduler_time(obj->world->scheduler);
    dt = t - obj->remote_time;

#ifdef DEBUG_DETAIL
    printf("Send reply to client %d with t=%fs, remote_t=%f -> dt=%f, reply='%s'\n",
	    obj->remote_tid, t, obj->remote_time, dt, reply);
#endif

    pvm_initsend(PvmDataDefault);
    pvm_pkdouble(&dt, 1, 1);
    pvm_pkint(&len, 1, 1);
    if (len>0) pvm_pkstr(reply);
    pvm_send(obj->remote_tid, MID_REPLY);

    obj->remote_time = t;

#endif    
}

void ps_kill(mw_object obj)
{
#ifndef NO_PVM
    if (obj) {
	if (obj->remote_tid>=0) {
	    pvm_kill(obj->remote_tid);
	}
    }
#endif
}
