/********************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: queue.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 17:00:07 stolz>
 ********************************************************************************
 *  MODULE              : queue
 *  IMPORT		: error, basic
 *
 *  IMPLEMENTATION
 *	The queue is implemented as dynamic array. If the array is
 *	full the size of the array will be doubled and the old
 *	contents will be copied into the new array. The size of the
 *	array will be 1/4 when it its only less then 1/4 occupied.
 *	It will never be smaller than the INIT_QUEUE_BLOCK_SIZE.
 *      This is all done in the resize_queue_buf() procedure.
 *
 *	The size of the elements is passed via a integers. So be
 *	careful on a system with 16bit-integers?!
 *
 *  NOTE
 *      The specification and definition is given in queue.h!
 *      If the Macro MODULE_TEST is specified, a test-routine
 *	is included with a own main()-routine. Link with module
 *	error.o to performe a test run for "verification" of
 *	this module.
 *
 *  HISTORY
 *  19-Apr-94   comments added
 *  19-Apr-94   modul-test passed!
 *  20-Apr-94   copy-mechanism in resize_queue_buf() checked. extend mod-test
 *  04-Jul-94   Add traversal mechanism. Not Tested!
 *  20-Jul-94   queue_look_first/last() added.
 *
 ********************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "queue.h"
#include "error.h"

#define INIT_QUEUE_BLOCK_SIZE	10


/****************************************************************************
 */
typedef struct adt_queue adt_queue;
struct adt_queue {
    int elem_size;
    int size;
    int	first;
    int block_size;
    char *elements;
};


#define queue_buf_to_small(q) (q->size == q->block_size)
#define queue_buf_to_large(q) ((q->block_size >> 2) > q->size + INIT_QUEUE_BLOCK_SIZE)



queue queue_new(int elem_size)
/****************************************************************************
 */
{
    queue q;

    if (elem_size <= 0) error("queue: queue_new() - called with elem_size <=0");

    q = (queue)malloc(sizeof(adt_queue));
    if (q) {
	q->elem_size = elem_size;
	q->size = 0;
	q->first = 0;
	q->block_size = 0;
	q->elements = 0;
    }
    else error("queue: queue_new() - can't alloc memory for adt queue");

    return q;
}

void queue_free(queue q)
/****************************************************************************
 */
{
    free(q->elements);
    free((void *)q);
}


static void resize_queue_buf(queue q)
/****************************************************************************
 */
{
    char *elems;
    int old_bsize, cpysize;

    old_bsize = q->block_size;

    if (queue_buf_to_small(q)) {	/* queue-buf is full */

        /* double capacity of queue-block
	 */
        if (0 == q->block_size) q->block_size = INIT_QUEUE_BLOCK_SIZE;
	else q->block_size *= 2;

    } else if (queue_buf_to_large(q)) {

	/* shrinkt buffer to 1/4 of old size
	 */
	q->block_size >>= 2;
	q->block_size = max(q->block_size, INIT_QUEUE_BLOCK_SIZE);
    }

    if (old_bsize != q->block_size) {

	/* allocate new block
         */
        elems = malloc(q->block_size * q->elem_size);
	if (0==elems)
	    error("queue: resize_queue_buf() - can't alloc memory for queue-block");

	/* copy contents from old to new block
	 * from old[first]..old[block_end] -> new[0]
         * from old[0]..old[first-1] -> new[block_end-first]
	 * so the first elem in the new block is nr.0
	 */

	if (q->size>0) {
	    cpysize = old_bsize - q->first;
	    cpysize = min(cpysize, q->size);

	    if (cpysize >0)
 	        memcpy(elems,
		       &(q->elements[q->elem_size * q->first]),
		       q->elem_size * cpysize);

            cpysize = q->size - cpysize;

            if (cpysize >0)
		memcpy(&elems[q->elem_size * (old_bsize - q->first)],
		       q->elements,
                       q->elem_size * cpysize);

            q->first = 0;
	}

	if (q->elements) free(q->elements);

	q->elements = elems;
    }
}



void queue_add(queue q, element e)
/****************************************************************************
 */
{
    if (queue_buf_to_small(q)) resize_queue_buf(q);

    /* copy element into queue-block
     */
    memcpy(&(q->elements[q->elem_size * ((q->first+q->size) % q->block_size)]),
	   e, q->elem_size);

    q->size ++;
}

boolean	queue_first(element e, queue q) /* FIFO */
/****************************************************************************
 */
{
    if (0==q->size) return FALSE;
    else {
	memcpy(e, &(q->elements[q->elem_size * q->first]), q->elem_size);
	q->first = (q->first + 1) % q->block_size;
	q->size --;
	if (queue_buf_to_large(q)) resize_queue_buf(q);

	return TRUE;
    }
}

boolean queue_last(element e, queue q)  /* FILA */
/****************************************************************************
 */
{
    if (0==q->size) return FALSE;
    else {
	memcpy(e, &(q->elements[q->elem_size * (q->first + q->size - 1)]),
	       q->elem_size);
	q->size --;
	if (queue_buf_to_large(q)) resize_queue_buf(q);

	return TRUE;
    }
}

boolean queue_is_empty(queue q)
/****************************************************************************
 */
{
    return (q->size == 0);
}

boolean queue_is_not_empty(queue q)
/****************************************************************************
 */
{
    return (q->size != 0);
}

int queue_size(queue q)
/****************************************************************************
 */
{
    return q->size;
}

void	queue_first_to_last(queue q, boolean (*f)(element))
/****************************************************************************
 */
{
    int i;

    if (f==0) return;

    for (i=0; i<q->size; i++) {
	if (!f(&(q->elements
		 [q->elem_size * ((q->first + i)% (q->block_size))])))
	    break;
    }
}

void	queue_last_to_first(queue q, boolean (*f)(element))
/****************************************************************************
 */
{
    int i;

    if (f==0) return;

    for (i=q->size - 1; i>=0; i--) {
	if (!f(&(q->elements
		 [q->elem_size * ((q->first + i)%(q->block_size))])))
	    break;
    }
}


boolean queue_look_first(element e, queue q)
/****************************************************************************
 */
{
    if (0==q->size) return FALSE;
    else {
	memcpy(e, &(q->elements[q->elem_size * q->first]), q->elem_size);

	return TRUE;
    }
}


boolean queue_look_last(element e,queue q)
/****************************************************************************
 */
{
    if (0==q->size) return FALSE;
    else {
	memcpy(e, &(q->elements[q->elem_size * (q->first + q->size - 1)]),
	       q->elem_size);

	return TRUE;
    }
}


/****************************************************************************
 * MODULE-TEST-routines
 */

#ifdef MODULE_TEST

struct q_elem {
    int		nr;
    char	text[11];
    int		nr2;
};

char *texts[32] = {
    "Alpha", "Beta", "Gamma",
};

static struct q_elem queue_data[12] = {
	{1, "Januar    ", 31},
	{2, "Februar   ", 28},
	{3, "Maerz     ", 31},
	{4, "April     ", 30},
	{5, "Mai       ", 31},
	{6, "Juni      ", 30},
	{7, "Juli      ", 31},
	{8, "August    ", 31},
	{9, "September ", 30},
	{10,"Oktober   ", 31},
	{11,"November  ", 30},
	{12,"Dezember  ", 31}
};


int queue_test()
{
    int failed, errors = 0;
    char index;
    queue q;
    struct q_elem e;
    int i;

    index = 'a';
    fprintf(stderr, "QUEUE-TEST from %s %s\n"
            "    %c) generate an empty list via queue_new(..)\n",
	    __DATE__, __TIME__, index++);

    q = queue_new(sizeof(struct q_elem));

    if (failed = (!queue_is_empty(q))) errors++;
    fprintf(stderr, "    %c) test function queue_is_empty(..) on empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (queue_is_not_empty(q))) errors++;
    fprintf(stderr, "    %c) test function queue_is_not_empty(..) on empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (0 != queue_size(q))) errors++;
    fprintf(stderr, "    %c) test function queue_size(..) on empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = queue_first(&e, q)) errors++;
    fprintf(stderr, "    %c) use function queue_first(..) on empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = queue_last(&e, q)) errors++;
    fprintf(stderr, "    %c) use function queue_last(..) on empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));



    fprintf(stderr, "  Add 12 items to the empty list\n");
    for (i=0; i<12; i++) {
        fprintf(stderr, "      elem: {%d, %10s, %d}\n", 
		queue_data[i].nr, queue_data[i].text, queue_data[i].nr2);
	queue_add(q, &queue_data[i]);
    }



    if (failed = (queue_is_empty(q))) errors++;
    fprintf(stderr, "    %c) test function queue_is_empty(..) on not empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (!queue_is_not_empty(q))) errors++;
    fprintf(stderr, "    %c) test function queue_is_not_empty(..) on not empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (12 != queue_size(q))) errors++;
    fprintf(stderr, "    %c) test function queue_size(..) on not empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (queue_first(&e, q) && 0==memcmp(&e, &queue_data[0], sizeof(struct q_elem))) {
    }
    else {
	failed = TRUE;
	errors++;
    }
    fprintf(stderr, "    %c) use function queue_first(..) on not empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (queue_last(&e, q) && 0==memcmp(&e, &queue_data[11], sizeof(struct q_elem))) {
    }
    else {
	failed = TRUE;
	errors++;
    }
    fprintf(stderr, "    %c) use function queue_last(..) on not empty list: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if (failed = (10 != queue_size(q))) errors++;
    fprintf(stderr, "    %c) test function queue_size(..) after removing of 2 elems: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));


    for (i=1; i<11; i++) {
	if (queue_first(&e, q) && 0==memcmp(&e, &queue_data[i], sizeof(struct q_elem))) {
            fprintf(stderr, "      elem: {%d, %10s, %d}\n", e.nr, e.text, e.nr2);
	}
	else {
	    failed = TRUE;
	    errors++;
        }
	fprintf(stderr, "    %c.%d) call of queue_first(..) : %s\n", index, i,
			(failed ? "FAILED" : "PASSED"));

	if (failed = ((10-i) != queue_size(q))) errors++;
	fprintf(stderr, "    %c.%d) call of queue_size(..): %s\n", index, i,
			(failed ? "FAILED" : "PASSED"));
    }
    index++;



    fprintf(stderr, "    %c) Freeing mem via queue_free(..),\n"
		    "    and create a new one with queue_new()\n", index++);
    queue_free(q);
    q = queue_new(sizeof(struct q_elem));

    fprintf(stderr, "    Add 100 items to the empty list\n");
    for (i=0; i<100; i++) {
	queue_add(q, &queue_data[i%12]);
    }

    fprintf(stderr, "    Obtain elements with queue_first()\n"
		    "    and check for queue_size()\n");

    failed = FALSE;
    for (i=100; i>=0; i--) {
	if (queue_size(q)!=i) {
	    failed = TRUE;
	    errors++;
	    fprintf(stderr, "    %d. queue_size() call FAILED (!=%d)\n", 101-i, i);
	}

	if (queue_first(&e, q) && 0==memcmp(&e, &queue_data[(100-i)%12], sizeof(struct q_elem))) {
	}
	else {
            if (i!=0){
	        failed = TRUE;
	        errors++;
		fprintf(stderr, "    %d. queue_first() call FAILED\n", 101-i);
	    }
        }
    }
    fprintf(stderr, "    test for adding/removing 100 elements: %s\n",
			(failed ? "FAILED" : "PASSED"));



    fprintf(stderr, "Total amount of errors: %d. Test %s!\n\n",
			errors, errors ? "FAILED" : "PASSED");


    return errors;
}

int main(int argc, char **argv)
{
    return queue_test();
}

#endif
