/********************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: queue.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 17:00:14 stolz>
 ********************************************************************************
 *  MODULE              : queue
 *  IMPORT		: basic
 *
 *  FUNCTION
 *
 *	This module make a abstract datatype "queue" available. The
 *	elements of this queue must have the same size and structure
 *	but they can be of any size. The size of the queue is "un-
 *	limited". The information of the elements is copied byte
 *	for byte to the queue. So be aware if your elements contains
 *	pointers!
 *
 *	The queue is initialized via the queue_new() call. The
 *	parameter is the size of one element in bytes.
 *      Now elements can be added via the queue_add() call. Note
 *	that the type element is defined as pointer! All functions
 *	which requires a element as parameter mean a pointer to
 *	this element.
 *
 *	Its up to the user to use the queue as fifo-buffer or stack.
 *	Use queue_first() to get the first element which was added
 *	to the queue. Or use queue_last() for an usage as stack.
 *      The element-parameter is only be valid if these functions
 *	returns TRUE.
 *
 *	The number of elements which are stored in the queue can be
 *      inquired be the queue_size() function. To get the information
 *	if the queue is empty or not use queue_is_empty() and
 *	queue_is_not_empty().
 *
 *	The queue and his contents is freed by a call of queue_free().
 *
 *  EXAMPLE
 *
 *	#include "queue.h"
 *
 *	struct elem {
 *	    int nr;
 *	    char sig;
 *      };
 *
 *
 *      void main()
 *      {
 *	    int i;
 *	    queue q;
 *          struct elem e;
 *
 *	    q = queue_new(sizeof(struct elem));
 *
 *	    // insert 4 elements
 *	    for (i=0; i<4; i++) {
 *		e.nr = i+1;
 *		e.sig = 'a' + i;
 *              queue_add(q, &e);
 *	    }
 *
 *          if (queue_is_empty(q)) printf("error, empty queue?\n");
 *
 *          printf("%d elements in my queue.\n", queue_size(q));
 *
 *          // the next statements should print out nr.1, sig=a
 *	    if (queue_first(&e, q)) {
 *              printf("elem with nr.%d, sig=%c\n", e.nr, e.sig);
 *	    }
 *          else printf("hoppla, error something is wrong!\n");
 *
 *          // the next statements should print out nr.4, sig=d
 *	    if (queue_last(&e, q)) {
 *              printf("elem with nr.%d, sig=%c\n", e.nr, e.sig);
 *	    }
 *          else printf("Is there a bug in this module?\n");
 *
 *	    queue_free(q);
 *      }
 *
 *
 *  NOTE
 *	The size of the elements is passed via a integers. So be
 *	careful on a system with 16bit-integers?!
 *
 *  HISTORY
 *  19-Apr-94   Creation date.
 *  19-Apr-94	Module-test passed.
 *  04-Jul-94   Add traversal mechanism. Not Tested!
 *  20-Jul-94   queue_look_first/last() added.
 *
 ********************************************************************************/
#ifndef QUEUE_H
#define QUEUE_H

#ifndef BASIC_H
#include "basic.h"
#endif

/****************************************************************************
 * abstract datatype queue is hidden behind a opac pointer.
 */

typedef struct adt_queue*	queue;
typedef void *			element;

/****************************************************************************
 * new and free a queue.
 */
queue	queue_new(int elem_size);
void	queue_free(queue);

/****************************************************************************
 * add/remove element
 */
void    queue_add(queue, element);
boolean	queue_first(element, queue);
boolean	queue_last(element, queue);

/****************************************************************************
 * look at first/last element
 */
boolean queue_look_first(element, queue);
boolean queue_look_last(element, queue);

/****************************************************************************
 * queue characteristics
 */
boolean queue_is_empty(queue);
boolean queue_is_not_empty(queue);
int	queue_size(queue);

/****************************************************************************
 * queue traversal
 */
void	queue_first_to_last(queue, boolean (*f)(element));
void	queue_last_to_first(queue, boolean (*f)(element));

#endif
