#ifndef ROB_H
#define ROB_H

/* general constants */

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif


#define ROBOCMDLENGTH      256
#define ROBOMSGLENGTH      256
#define MAXEXECLENGTH      20
#define MAXHOSTNAMELENGTH  256
#define MAXROBNAMELENGTH  256
#define MAXDEVICETYPELENGTH  256

#define ROBOREPLYLENGTH    256  /* obsolent use ROBMSGLEN */

/* for rob_pvm_start */

#define RobViaSimulator    403  /* all rob_commands are routed to simulator */
#define RobViaRS232        402  /* all rob_commands are routed to RS232 immediately*/
#define RobViaServer       401  /* all rob_commands are routed to a robotserver */


/* rob error codes */

#define RobOk              0	/* okay */
#define RobPvmFail         -1	/* pvm not started / stopped */
#define RobCameraFail      1    /* camera not available */
#define RobFileFail        2   /* can't save image-file */

#define CAMERA_DEF_WIDTH   128
#define CAMERA_DEF_HEIGHT  128
#define CAMERA_DEF_COLOR   0

/* general tty behavior */

#define Activated 1
#define NotActivated 0

#define RobDebugDefault    NotActivated	/* no debugging */
#define RobTTYDefault      Activated	/* tty I-O */



/* Interface functions */

#ifdef __cplusplus
extern "C" {
#endif



typedef struct {
    char	*rob_name;

    int		start_simulator;  /* true or false */
    int		simulator_tid;

    /* arguments without succesfully parsed rob-args
     */
    int		argc;
    char	**argv;
} rob_args;



rob_args *rob_parse_args(int argc, char **argv);
/****************************************************************************
 * FUNCTION
 *	Parse the command line args (and set the equivalent options of
 *	the rob-library). The parse understand the following arguments:
 *
 *	-r[ob] <name>           -> set rob_name in the args.
 *	-s[im]                  -> simulator-switch (no effect)
 *	-simtid	<tid-number>	-> connect to spec simulator-tid
 *
 * OUTPUT
 *	If <name> or <tid-number> is missing, or the function can't alloc
 *	memory, it prints out an error-text on stderr and perfom an exit(10).
 *
 * RESULTS
 *	The routine returns a pointer to the rob_args. There you find
 *	the rob-name, ..., and also the argument array without
 *	the successfully parsed arguments.
 *
 * NOTE
 *
 * EXAMPLE
 *
 *      main(int argc, char **argv)
 *      {
 *	    char *rob_name = "portos";
 *          int  rob;
 *	    rob_args *args;
 *
 *          args = rob_parse_args(argc, argv);
 *          if (args->rob_name) rob_name = args->rob_name;
 *
 *	    rob_pvm_start();
 *          rob = rob_init(rob_name);
 *	    ...
 *
 ****************************************************************************
 *
 */

/* the following two functions are obsolent, use rob_start() and 
 * rob_halt() instead!
 */
int rob_pvm_start();
int rob_pvm_halt();

int rob_start(int ttyflag);
int rob_halt();

int rob_init(char* robname);
int rob_exit(int rob);

int rob_command(int rob,char* command,char* reply);

int rob_camera_frame(unsigned char *frame, int rob, int camera_nr,
		     int width, int height, int color);
/****************************************************************************
 * FUNCTION
 *	Requests a camera-frame from a robot. 
 *
 * PARAMETERS
 *	frame		pointer to a array of byte-values(0-255) 
 *      rob             robot
 *      camera_nr       nr of requested camera-device (0,1,...)
 *      width, height   requested image-size
 *      color           0 for grayscale and 1 for color-images
 *      
 *                    
 * RESULTS
 *	The routine returns RobOk if the camera exists and a rendered 
 *      images will stored in the frame-array (1 byte for gray,
 *      3 bytes (rgb) for color-images, order: left->right, top->bottom.)
 *      On an error, you get CameraFail.
 *
 * EXAMPLE
 *
 *      unsigned char frame[CAMERA_DEF_WIDTH][CAMERA_DEF_HEIGHT];
 * 
 *      if (RobOk != rob_camera_frame(unsigned char *)frame, rob, 0,
 *                                  CAMERA_DEF_WIDTH, CAMERA_DEF_HEIGHT, 0)
 *          printf("*** error: no line to the camera\n");      
 *
 *
 ****************************************************************************
 *
 */

int rob_save_frame(char *fname, 
		   unsigned char *frame, int width, int height, int color);
/****************************************************************************
 * FUNCTION
 *	Saves the frame into a file in pgm- or ppm-format
 *
 * PARAMETERS
 *      fname           filename of the output-file
 *	frame		pointer to a array of byte-values(0-255) 
 *      width, height   requested image-size
 *      color           0 for grayscale and 1 for color-images
 *      
 *                    
 * RESULTS
 *	The routine returns RobOk for an successful write or
 *      RobFileFail for an error.
 *
 * EXAMPLE
 *
 *      unsigned char frame[3*CAMERA_DEF_WIDTH*CAMERA_DEF_HEIGHT];
 * 
 *      rob_save_frame("frame.ppm", frame, 
 *			   CAMERA_DEF_WIDTH, CAMERA_DEF_HEIGHT, 1);
 *
 ****************************************************************************
 *
 */


#ifdef __cplusplus
}
#endif

#endif
