/********************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: scheduler.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 17:01:05 stolz>
 ********************************************************************************
 *  MODULE              : scheduler
 *  IMPORT              : math.h (math library)
 *
 *  IMPLEMENTATION
 *      tasks ist is implemented as a list. timepoints as a rising sorted list
 *      with "time" as sort criteria.
 *
 *  NOTE
 *      The module specification and definition is given in scheduler.h!
 *	Also the function specification insist in scheduler.h.
 *      The test-routine is only available if NO_TEST_CODE is undefined.
 *      When BUILD_MODULE_TEST is defined in addition, a main()-routine is
 *      available which invokes the module-test.
 *
 *  HISTORY
 *  03-Mar-94   Creation date
 *  07-Mar-94   Adding comments
 *  08-Mar-94   Test-Routine sucessfully passed!
 *  24-Mar-94   last added Timepoints with same time is sorted in at the end
 *
 ********************************************************************************/

#include <math.h>

#ifndef NO_TEST_CODE
#include <stdio.h>
#endif

#include "scheduler.h"

#define SMALLEST_ZERO   1e-8

#define DOUBLE_EQUAL(x,y) (fabs((x)-(y))<=SMALLEST_ZERO)
#define DOUBLE_NOT_EQUAL(x,y) (fabs((x)-(y))>SMALLEST_ZERO)



void scheduler_new(scheduler *s, double time)
/****************************************************************************
 */
{
    s->time = time;
    s->lastsynctime = time;
    s->tasks = 0;
    s->points = 0;
}


double scheduler_time(scheduler *s)
/****************************************************************************
 */
{
  return s->time;
}

void scheduler_sync(scheduler *s, double time)
/****************************************************************************
 */
{
    task *t;
    timepoint *point;

    if (time < s->time) {
	scheduler_reset(s, time);
	return ;
    }

    for (point=s->points; point && (point->time <= time); point=s->points) {
	s->points = point->next;

	if (DOUBLE_NOT_EQUAL(s->lastsynctime, point->time)) {
	    for (t=s->tasks; t; t=t->next)
		if (t->sync) t->sync(s, t, point->time, s->lastsynctime);
	    s->lastsynctime = point->time;
	}
	s->time = point->time;
	if (point->function) point->function(s, point, point->time);

	if (point->free) point->free(s, point);
    }

    /* synchronise workspace to time if necessary */
    if (DOUBLE_NOT_EQUAL(s->lastsynctime, time)) {
	for (t=s->tasks; t; t=t->next)
	    if (t->sync) t->sync(s, t, time, s->lastsynctime);
    }
    s->lastsynctime = time;
    s->time = time;
}

void scheduler_task_add(scheduler *s, task *t)
/****************************************************************************
 */
{
    t->next = s->tasks;
    s->tasks = t;
}

void scheduler_task_remove(scheduler *s, task *t)
/****************************************************************************
 */
{
    task **tp;
    timepoint **p, *point;

    for (p=&(s->points); *p; ) {
	point = *p;

	if (point->task == t) {
	    *p = point->next;
	    p = &(point->next);

	    if (point->free) point->free(s, point);
	}
	else p = &(point->next);
    }

    for (tp = &(s->tasks); *tp && (*tp)!=t; tp = &((*tp)->next))
	;

    if (*tp) {
	*tp = t->next;
	if (t->free) t->free(s, t);
    }
}

void scheduler_point_add(scheduler *s, timepoint *p, double time)
/****************************************************************************
 */
{
    timepoint **tp;

    if (time < (s->time)) return;

    p->time = time;

    for (tp = &(s->points); *tp && (*tp)->time <= p->time; tp = &((*tp)->next))
	;

    p->next = *tp;
    *tp = p;
}

void scheduler_point_unlink(scheduler *s, timepoint *p)
/****************************************************************************
 */
{
    timepoint **tp;

    for (tp = &(s->points); *tp && (*tp)!=p; tp = &((*tp)->next))
	;

    if (*tp) {
      *tp = p->next;
    }
}

void scheduler_point_remove(scheduler *s, timepoint *p)
/****************************************************************************
 */
{
    timepoint **tp;

    for (tp = &(s->points); *tp && (*tp)!=p; tp = &((*tp)->next))
	;

    if (*tp) {
      *tp = p->next;
    }

    if (p->free) p->free(s, p);
}

void scheduler_free(scheduler *s)
/****************************************************************************
 */
{
    task *stask, *t;
    timepoint *point, *p;

    for (p=s->points; p; ) {
	point = p;
	p = p->next;

	if (point->free) point->free(s, point);
    }
    s->points = 0;

    for (t=s->tasks; t; ) {
	stask = t;
	t = t->next;

	if (stask->free) stask->free(s, stask);
    }
    s->tasks = 0;

    s->time = 0.0;
    s->lastsynctime = 0.0;
}

void scheduler_reset(scheduler *s, double time)
/****************************************************************************
 */
{
    task *stask, *t;
    timepoint *point, *p;

    for (p=s->points; p; ) {
	point = p;
	p = p->next;

	if (point->free) point->free(s, point);
    }
    s->points = 0;

    for (t=s->tasks; t; ) {
	stask = t;
	t = t->next;

	if (stask->reset) stask->reset(s, stask, time);
    }

    s->time = time;
    s->lastsynctime = time;
}



#ifndef NO_TEST_CODE
/****************************************************************************
 * TEST-CODE
 */

static struct statistic {
    int t_sync[3];
    int t_free[3];
    int t_reset[3];
    int p_free[4];
    int p_func[4];
    double time;
    int errors;
} ws;

static scheduler sched;
static task tasks[3];
static timepoint points[4];

static void task_free(scheduler *s, task *t)
{
    int tnr;
    int errold = ws.errors;

    tnr = (int)(t->workspace);
    fprintf(stderr, "      > task#%d - free()\n", tnr);
    if (&tasks[tnr] != t) ws.errors++;
    if (s != &sched) ws.errors++;
    ws.t_free[tnr]++;

    if (ws.errors != errold) {
	fprintf(stderr, "      > #parameter-errors-total:%d\n", ws.errors);
    }
}

static void task_sync(scheduler *s, task *t, double time, double lastsync)
{
    int tnr;
    int errold = ws.errors;

    tnr = (int)(t->workspace);
    fprintf(stderr, "      > task#%d - sync() to time %f\n", tnr, time);
    if (&tasks[tnr] != t) ws.errors++;
    if (s != &sched) ws.errors++;
    ws.t_sync[tnr]++;

    if (ws.errors != errold) {
	fprintf(stderr, "      > #parameter-errors-total:%d\n", ws.errors);
    }
}

static void task_reset(scheduler *s, task *t, double time)
{
    int tnr;
    int errold = ws.errors;

    tnr = (int)(t->workspace);
    fprintf(stderr, "      > task#%d - reset() to time %f\n", tnr, time);
    if (&tasks[tnr] != t) ws.errors++;
    if (s != &sched) ws.errors++;
    if (time != ws.time) ws.errors++;
    ws.t_reset[tnr]++;

    if (ws.errors != errold) {
	fprintf(stderr, "      > #parameter-errors-total:%d\n", ws.errors);
    }
}

static void point_func(scheduler *s, timepoint *tp, double time)
{
    int pnr;
    int errold = ws.errors;

    pnr = (int)(tp->workspace);
    fprintf(stderr, "      > point#%d - function() at time %f\n", pnr, time);
    if (&points[pnr] != tp) ws.errors++;
    if (s != &sched) ws.errors++;

    ws.p_func[pnr]++;
  
    if (ws.errors != errold) {
	fprintf(stderr, "      > #parameter-errors-total:%d\n", ws.errors);
    }

    if (pnr!=3) {
	time += 1.3;
    }
    else {
	 if (ws.p_func[pnr]>3) return;
    }

    
    fprintf(stderr, "      > point#%d - adding same point to time %f\n", pnr, time);
    scheduler_point_add(s, tp, time);
}

static void point_free(scheduler *s, timepoint *tp)
{
    int pnr;
    int errold = ws.errors;

    pnr = (int)(tp->workspace);
    fprintf(stderr, "      > point#%d - free()\n", pnr);
    if (&points[pnr] != tp) ws.errors++;
    if (s != &sched) ws.errors++;

    ws.p_free[pnr]++;

    if (ws.errors != errold) {
	fprintf(stderr, "      > #parameter-errors-total:%d\n", ws.errors);
    }
}


static void reset_counters()
{
    int i;

    for (i=0; i<3; i++) {
	ws.t_sync[i] = 0;
	ws.t_free[i] = 0;
	ws.t_reset[i] = 0;
    }
    for (i=0; i<4; i++) {
	ws.p_func[i] = 0;
	ws.p_free[i] = 0;
    }
}


int scheduler_test()
/****************************************************************************
 */
{
    int failed, errors = 0;
    char index;
    int i;


    index = 'a';
    fprintf(stderr, "SCHEDULER-TEST\n"
          "    Initialize scheduler via scheduler_new(..) to time 0.0\n");

    sched.workspace = &ws;
    scheduler_new(&sched, 0.0);

    if ((failed = (&ws != sched.workspace))) errors++;
    fprintf(stderr, "    %c) test of workspace pointer after calling: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    ws.errors = 0;

    for (i=0; i<3; i++) {
	ws.t_sync[i] = 0;
	ws.t_free[i] = 0;
	ws.t_reset[i] = 0;
	tasks[i].workspace = (void *)i;
        if (i==2) {
	    tasks[i].sync = 0;
	    tasks[i].free = 0;
	    tasks[i].reset = 0;
        } else {
	    tasks[i].sync = task_sync;
	    tasks[i].free = task_free;
	    tasks[i].reset = task_reset;
        }
    }

    for (i=0; i<4; i++) {
	ws.p_func[i] = 0;
	ws.p_free[i] = 0;
        points[i].workspace = (void *)i;
	points[i].task = &tasks[((i>=2) ? 2 : i)];
	points[i].function = point_func;
	points[i].free = point_free;
    }
    points[3].function = 0;
    points[3].free = 0;

    fprintf(stderr, "    Adding two tasks via scheduler_task_add(..)\n");

    scheduler_task_add(&sched, &tasks[0]);
    scheduler_task_add(&sched, &tasks[1]);

    ws.time = 1.0;
    scheduler_sync(&sched, 1.0);

    if ((failed = (ws.t_sync[0]!=1 || ws.t_free[0]!=0 || ws.t_reset[0]!=0 ||
                  ws.t_sync[1]!=1 || ws.t_free[1]!=0 || ws.t_reset[1]!=0 )))
	errors++;
    fprintf(stderr, "    %c) synchronization to time 1.0: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    reset_counters();

    scheduler_point_add(&sched, &points[0], 1.0 );
    scheduler_point_add(&sched, &points[1], 1.2 );
    scheduler_point_add(&sched, &points[2], 1.5 );
    scheduler_point_add(&sched, &points[3], 1.5 );

    fprintf(stderr, "    Adding four points at times 1.0, 1.2, 1.5, 1.5\n");
    fprintf(stderr, "    Point#3 without a free() or function()!\n");
  
    ws.time = 2.0;
    scheduler_sync(&sched, 2.0);

    if ((failed = (ws.t_sync[0]!=3 || ws.t_free[0]!=0 || ws.t_reset[0]!=0 ||
                  ws.t_sync[1]!=3 || ws.t_free[1]!=0 || ws.t_reset[1]!=0 ||
		  ws.p_func[0]!=1 || ws.p_free[0]!=1 ||
		  ws.p_func[1]!=1 || ws.p_free[1]!=1 ||
		  ws.p_func[2]!=1 || ws.p_free[2]!=1 ||
		  ws.p_func[3]!=0 || ws.p_free[3]!=0 )))
	errors++;
    fprintf(stderr, "    %c) t->sync,p->func and p->free to time 2.0: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    reset_counters();

    fprintf(stderr, "    Adding point to time 2.6\n");
    fprintf(stderr, "    ->should be the points: 2.3, 2.5, 2.6, 2.8!\n");
    fprintf(stderr, "    point at 2.6 should self-add at same time!\n");

    points[3].function = point_func;
    points[3].free = 0;
    scheduler_point_add(&sched, &points[3], 2.6 );

    fprintf(stderr, "    Now sync to time 3.0\n");
    ws.time = 3.0;
    scheduler_sync(&sched, 3.0);

    if ((failed = (ws.t_sync[0]!=5 || ws.t_free[0]!=0 || ws.t_reset[0]!=0 ||
		   ws.t_sync[1]!=5 || ws.t_free[1]!=0 || ws.t_reset[1]!=0 ||
		   ws.p_func[0]!=1 || ws.p_free[0]!=1 ||
		   ws.p_func[1]!=1 || ws.p_free[1]!=1 ||
		   ws.p_func[2]!=1 || ws.p_free[2]!=1 ||
		   ws.p_func[3]!=4 || ws.p_free[3]!=0 )))
	errors++;
    fprintf(stderr, "    %c) check of _point_add() in main+function(): %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    reset_counters();



    fprintf(stderr, "    Removing task#1 (and his point)\n");

    scheduler_task_remove(&sched, &tasks[1]);

    fprintf(stderr, "    Sync to time 4.0  -> points 3.6, [4.1]\n");

    ws.time = 4.0;
    scheduler_sync(&sched, 4.0);

    if ((failed = (ws.t_sync[0]!=2 || ws.t_free[0]!=0 || ws.t_reset[0]!=0 ||
		   ws.t_sync[1]!=0 || ws.t_free[1]!=1 || ws.t_reset[1]!=0 ||
		   ws.p_func[0]!=1 || ws.p_free[0]!=1 ||
		   ws.p_func[1]!=0 || ws.p_free[1]!=1 ||
		   ws.p_func[2]!=0 || ws.p_free[2]!=0 ||
		   ws.p_func[3]!=0 || ws.p_free[3]!=0 )))
	errors++;
    fprintf(stderr, "    %c) check if task#1 and his point were deleted: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    reset_counters();

    

    fprintf(stderr, "    Adding task#1 and #2 and reseting to time 1.1\n");
    fprintf(stderr, "    Task#2 has no reset(),sync() or free() functions!\n");
    fprintf(stderr, "    This should free point#0(4.9) and #2(4.1)\n");

    scheduler_task_add(&sched, &tasks[2]);
    scheduler_task_add(&sched, &tasks[1]);

    ws.time = 1.1;
    scheduler_reset(&sched, 1.1);

    if ((failed = (ws.t_sync[0]!=0 || ws.t_free[0]!=0 || ws.t_reset[0]!=1 ||
		   ws.t_sync[1]!=0 || ws.t_free[1]!=0 || ws.t_reset[1]!=1 ||
		   ws.t_sync[2]!=0 || ws.t_free[2]!=0 || ws.t_reset[2]!=0 ||
		   ws.p_func[0]!=0 || ws.p_free[0]!=1 ||
		   ws.p_func[1]!=0 || ws.p_free[1]!=0 ||
		   ws.p_func[2]!=0 || ws.p_free[2]!=1 ||
		   ws.p_func[3]!=0 || ws.p_free[3]!=0 )))
	errors++;
    fprintf(stderr, "    %c) scheduler_reset() function: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    reset_counters();

    


    fprintf(stderr, "    Sync scheduler to time 0.9 (before actual time!)\n");
    fprintf(stderr, "    This should cause a reset.\n");

    ws.time = 0.9;
    scheduler_sync(&sched, 0.9);

    if ((failed = (ws.t_sync[0]!=0 || ws.t_free[0]!=0 || ws.t_reset[0]!=1 ||
		   ws.t_sync[1]!=0 || ws.t_free[1]!=0 || ws.t_reset[1]!=1 ||
		   ws.t_sync[2]!=0 || ws.t_free[2]!=0 || ws.t_reset[2]!=0 ||
		   ws.p_func[0]!=0 || ws.p_free[0]!=0 ||
		   ws.p_func[1]!=0 || ws.p_free[1]!=0 ||
		   ws.p_func[2]!=0 || ws.p_free[2]!=0 ||
		   ws.p_func[3]!=0 || ws.p_free[3]!=0 )))
	errors++;
    fprintf(stderr, "    %c) scheduler_sync() with reset()-invokation: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    reset_counters();

    


    fprintf(stderr, "    Adding 4 points to the scheduler(#0 before time, #3 without free())\n");
    fprintf(stderr, "    Freeing the scheduler->free points+tasks in this order!\n");

    points[3].function = 0;
    points[3].free = 0;
    scheduler_point_add(&sched, &points[0], 0.7 );
    scheduler_point_add(&sched, &points[1], 1.2 );
    scheduler_point_add(&sched, &points[2], 1.5 );
    scheduler_point_add(&sched, &points[3], 1.5 );

    scheduler_free(&sched);

    if ((failed = (ws.t_sync[0]!=0 || ws.t_free[0]!=1 || ws.t_reset[0]!=0 ||
		   ws.t_sync[1]!=0 || ws.t_free[1]!=1 || ws.t_reset[1]!=0 ||
		   ws.t_sync[2]!=0 || ws.t_free[2]!=0 || ws.t_reset[2]!=0 ||
		   ws.p_func[0]!=0 || ws.p_free[0]!=0 ||
		   ws.p_func[1]!=0 || ws.p_free[1]!=1 ||
		   ws.p_func[2]!=0 || ws.p_free[2]!=1 ||
		   ws.p_func[3]!=0 || ws.p_free[3]!=0 )))
	errors++;
    fprintf(stderr, "    %c) check scheduler_free(): %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    reset_counters();


    fprintf(stderr, "\nReported parameter-errors: %d - %s\n",
			ws.errors, (ws.errors ? "FAILED" : "PASSED"));

    errors += ws.errors;

    fprintf(stderr, "Total amount of errors: %d. Test %s!\n\n",
			errors, errors ? "FAILED" : "PASSED");


    return errors;
}

#ifdef BUILD_MODULE_TEST
int main()
{
    return scheduler_test();
}
#endif

#endif
