/********************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: scheduler.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 17:01:13 stolz>
 ********************************************************************************
 *  MODULE              : scheduler
 *
 *  FUNCTION
 *	This module provide structures and functions for a time-oriented
 * 	synchronised scheduling of more than on tasks.
 *
 *	Many tasks works on a big "problem". This problem is time-dependet.
 *	Some of this task has to do something on a special "timepoints". But
 *      the whole data of the problem should be consistent at this points.
 *
 *      The scheduler now transform the problem from the actual time t0 to a
 *      time t1 in the future. During [t0,t1] some tasks want to "change" the
 *      data (global "workspace"). Therefore, tasks can add "functions" that
 *      should called at this discret "timepoints". Every task has a "sync"
 *	function to update the "workspace" to a new time. The scheduler
 *      guaranty that the "workspace" is synchronised at time t1. Also the
 *      functions get only synchronised data!
 *
 *      So the scheduler hold the global "workspace" of a problem and stores
 *      the "tasks" and functions of time-"points".
 *	Functions to initialise ("new"), or deallocate("free") or "reset" the
 *      scheduler to a new time where provided.
 *
 *      Therefore must have a "task" in addition to his local "workspace"
 *      and "sync"-function also functions to "free" and "reset".
 *      Functions to add and remove tasks are specified. Also timepoints of
 *      a removed task get "free"d.
 *
 *      A "timepoint" defines a "time" at which a user-"function" should called.
 *      Also a "free"-function is necessary to provide a consistent freeing
 *      of timepoints.
 *
 *  NOTES
 *	Don't set elements in the private part of a structure!
 *      Nearly in all structures, the public part should be filled in completly.
 *	Don't feed the same tasks or timepoints to different schedulers!
 *	Don't use any of the provided function within a called sync(), free(),
 *	reset() or function() routinen if its not explicit alloweded!
 *	This header file defines SCHEDULER_H.
 *	The module test-routine is only available if NO_TEST_CODE is undefined.
 *
 *  HISTORY
 *  07-Mar-94   Creation date, Adding of remarks
 *  08-Mar-94   Adding the lastsync-parameter to task->sync.
 *  24-Mar-94   Adding _time(), _point_unlink(), _point_remove().
 *
 ********************************************************************************/
#ifndef SCHEDULER_H
#define SCHEDULER_H


/****************************************************************************
 * exported types
 */
typedef struct scheduler scheduler;
typedef struct task task;
typedef struct timepoint timepoint;


/****************************************************************************
 * task-structure with local workspace, sync, reset and free-functions
 * The elements sync, reset and free must be set to 0 if no function
 * should be called!
 * The workspace element is for user specific data and is not used by
 * the scheduler.
 */
struct task {
    void *workspace;
    void (*sync)(scheduler *, task *, double time, double lastsync);
    void (*reset)(scheduler *, task *, double time);
    void (*free)(scheduler *, task *);
/* private */
    task *next;
};


/****************************************************************************
 * Structure for an timepoint.
 * Set the task-pointer appropriated to your task.
 * Also set function or free to 0 if you don't use them.
 */
struct timepoint {
    void *workspace;
    task *task;
    void (*function)(scheduler *s, timepoint *tp, double time);
    void (*free)(scheduler *s, timepoint *tp);
/* private */
    double time;
    timepoint *next;
};


/****************************************************************************
 * The scheduler-structure with his global workspace.
 * use scheduler_new(..) to initialize this structure!
 */
struct scheduler {
    void *workspace;
/* private */
    double time;
    double lastsynctime;
    task *tasks;
    timepoint *points;
};


void scheduler_new(scheduler *s, double time);
/****************************************************************************
 * NAME
 *	scheduler_new(...)		scheduler initialize
 *
 * FUNCTION
 *	Initialize the private part of the scheduler structure "s".
 *      It's not allowed to perform any scheduler-function before a this!
 * 	The start-time of the scheduler is set to "time". The list of "tasks"
 *      and "timepoint"s is empty.
 *
 * EXAMPLE
 *	scheduler my_s;
 *
 *	my_s.workspace = <pointer to user-specific data>;
 *
 *	scheduler_new(&my_s, 0.0);
 *
 * SEE ALSO
 *	scheduler_free(), scheduler_reset(), scheduler_sync(),
 *	scheduler_task_add(), scheduler_point_add()
 *
 ****************************************************************************
 *
 */

double scheduler_time(scheduler *s);
/****************************************************************************
 * NAME
 *	double scheduler_time(...)	     get current time of scheduler
 *
 * FUNCTION
 *      Returns the current time of the scheduler. 
 *
 * NOTE
 *	This function can be called even in task->sync(),task->free(),
 *      task->reset(), timepoint->function(),timepoint->free(). 
 *
 ****************************************************************************
 *
 */

void scheduler_sync(scheduler *s, double time);
/****************************************************************************
 * NAME
 *	scheduler_sync(...)		data synchronisation at time
 *
 * FUNCTION
 *	Synchronise the scheduler and his user-data to new time.
 *      Call all functions of timepoints which lies between the old and
 *      the new time. Free them after calling!
 *      Before a timepoint-function is called. The workspace is synchronised
 *	to this time via task->sync().
 *      After doing all timepoints, the workspace is synchronised to the
 *      target time, specified by the scheduler_sync-option.
 *	If the specified time is below the current scheduler time, a
 *	scheduler_reset() with this time is performed.
 *
 * EXAMPLE
 *	scheduler my_s;
 *
 *	... set workspace ...
 *	scheduler_new(&my_s, 1.0);
 *      ... add some tasks and timepoints ...
 *      scheduler_sync(&my_s, 1.1);             - advance to time 1.1
 *      ...
 *      scheduler_sync(&my_s, 1.2);             - sync to time 1.2
 *
 * NOTES
 *      It's possible to add the same timepoint in the "function" or the
 *      "free" routine with a call of scheduler_add_point(..).
 *
 * SEE ALSO
 *	scheduler_free(), scheduler_reset(), scheduler_sync(), add_task()
 *	scheduler_point_add()
 *
 ****************************************************************************
 *
 */

void scheduler_task_add(scheduler *s, task *t);
/****************************************************************************
 * NAME
 *	scheduler_task_add(...)		add task to scheduler
 *
 * FUNCTION
 *      Add a task to the scheduler's list of tasks. The public part of
 * 	the task-structure must be initialized!
 *      Synchronization, Free-ing and Resets are then also performed on
 *	this task. Don't add the same task more than once to a scheduler
 *	nor feed the structure to different schedulers!
 *
 * EXAMPLE
 *      void my_sync(scheduler *s, task *t, double time, double lastsync);
 *
 *	scheduler my_s;
 *      problem_data data;
 *	task	my_task;
 *
 *	my_s.workspace = data;
 *	scheduler_new(&my_s, 0.0);
 *
 *	my_task.workspace = data;
 *	my_task.sync  = my_sync;
 *	my_task.free  = 0;
 *	my_task.reset = 0;
 *	scheduler_add_task(&my_s, &my_task);
 *
 *	scheduler_sync(&my_s, 2.0);
 *
 *	scheduler_remove_task(&my_s, &my_task);
 *
 * NOTES
 *      Don't use this function within a called sync(), free(), reset()
 *	or function().
 *
 * SEE ALSO
 *	scheduler_task_remove(), scheduler_point_add()
 *
 ****************************************************************************
 *
 */

void scheduler_task_remove(scheduler *s, task *t);
/****************************************************************************
 * NAME
 *	scheduler_task_remove(...)	remove task+his points from scheduler
 *
 * FUNCTION
 *	Removes a task from a schedulers task-list. The scheduler performs
 *	a task->free() for this task. But before doing this all timepoints
 *	with the same task-pointer get freed via a point->free() call.
 *
 * EXAMPLE
 *	see example of scheduler_task_add()
 *
 * NOTES
 *      Don't use this function within a called sync(), free(), reset()
 *	or function().
 *
 * SEE ALSO
 *	scheduler_task_add(), scheduler_free()
 *
 ****************************************************************************
 *
 */

void scheduler_point_add(scheduler *s, timepoint *p, double time);
/****************************************************************************
 * NAME
 *	scheduler_point_add(...)        add timepoint to scheduler
 *
 * FUNCTION
 *      Add a timepoint to the schedulers list of timepoints. The specified
 *	function() of the timepoint-structure will be called if the time is
 *	reached. The free()-routine get invoked to free the timepoint. This
 *	is done p.e. in scheduler_task_remove(), scheduler_free() or after the
 *	function()-calling. free() of the timepoint is called before the
 *	free() of the corresponding task! Only timepoints are added which
 *	lies in the future. This means timepoint.time >= scheduler.time.
 *
 *	The timepoints are stored in order so that the correct calling
 *	order is guarantied. The calling order of timepoints at the same
 *	time is undefined. All timepoint->function()'s at the a time t1
 *	will also be called of an scheduler_sync(s, t1) invokation.
 *
 *	This function call also called within a scheduler_task_add(),
 *      timepoint->function(), task->reset() or task->sync().
 *
 *      Before the scheduler calls the timepoint->function(), he decouple the
 *	structure. So it's possible to add the same timepoint in the
 *	timepoint->function().
 *
 * EXAMPLE
 *
 *	my_function(scheduler *s, timepoint *tp, time *t)
 *	{
 *          printf("function call at time %f\n", t);
 *
 *          scheduler_point_add(s, tp, t+0.01);
 *      }
 *
 *	my_tp_free(scheduler *s, timepoint *tp)
 *	{
 *	    printf("tp freed!\n");
 *	    free(tp);
 *      }
 *
 *	main()
 *	{
 *	    scheduler my_s;
 *	    task my_t;
 *	    timepoint my_tp1;
 *	    timepoint *my_tp2;
 *
 *
 *	    scheduler_new(&my_s, 0.0);
 *	... scheduler and task initialization ...
 *
 *          my_tp1.workspace = 0;
 *          my_tp1.function  = my_function;
 *          my_tp1.free = 0;
 *          scheduler_point_add(&my_s, &my_tp1, 1.0);
 *
 *	    my_tp2 = malloc(sizeof(timepoint));
 *          my_tp2->workspace = 0;
 *          my_tp2->function  = my_function;
 *          my_tp2->free = my_free;
 *          scheduler_point_add(&my_s, my_tp2, 1.3);
 *
 *	    scheduler_sync(&my_s, 1.5);
 *
 *	    scheduler_free(&my_s);
 *	}
 *
 *	-> OUTPUT:
 *          function call at time 1.0
 *          function call at time 1.1
 *          function call at time 1.2
 *          function call at time 1.3
 *          tp freed!
 *          function call at time 1.3
 *          function call at time 1.4
 *          function call at time 1.5
 *
 * NOTES
 *	This is the only function which can be call within a
 *	scheduler_task_add(), timepoint->function(), task->reset()
 *	or task->sync(). The last three will be invoked from the scheduler
 *	automatically.
 *
 * SEE ALSO
 *	scheduler_free(), scheduler_reset(), scheduler_sync(),
 *	scheduler_task_add(), scheduler_point_add(),
 *      scheduler_point_unlink(), scheduler_point_remove()
 *
 ****************************************************************************
 *
 */

void scheduler_point_unlink(scheduler *s, timepoint *p);
/****************************************************************************
 * NAME
 *	scheduler_point_unlink(...)        unlink timepoint from scheduler
 *
 * FUNCTION
 *      Removes a timepoint from the schedulers list of timepoints. The
 *      timepoint will not be freed. Its possible now to add the unlinked
 *      point to a different time vie scheduler_point_add().
 *
 * EXAMPLE
 *
 *	my_function(scheduler *s, timepoint *tp, time *t)
 *      {
 *          printf("function call at time %f\n", t);
 *          scheduler_point_add(s, tp, t+0.01);
 *      }
 *
 *	main()
 *	{
 *	    scheduler my_s;
 *	    task my_t;
 *	    timepoint my_tp;
 *
 *	    scheduler_new(&my_s, 0.0);
 *	... scheduler and task initialization ...
 *
 *          my_tp.workspace = 0;
 *          my_tp.function  = my_function;
 *          my_tp.free = 0;
 *          scheduler_point_add(&my_s, &my_tp, 1.0);
 *
 *	    scheduler_point_remove(&my_s, &my_tp);
 *   
 *          scheduler_point_add(&my_s, &my_tp, 2.0);
 *          ...
 *	}
 *
 * NOTES
 *	This function can also called within a scheduler_task_add()
 *      timepoint->function(), task->reset() or task->sync(). 
 *
 * SEE ALSO
 *	scheduler_point_add(), scheduler_point_removet()
 *
 ****************************************************************************
 *
 */

void scheduler_point_remove(scheduler *s, timepoint *p);
/****************************************************************************
 * NAME
 *	scheduler_point_add(...)        add timepoint to scheduler
 *
 * FUNCTION
 *      Add a timepoint to the schedulers list of timepoints. The specified
 *	function() of the timepoint-structure will be called if the time is
 *	reached. The free()-routine get invoked to free the timepoint. This
 *	is done p.e. in scheduler_task_remove(), scheduler_free() or after the
 *	function()-calling. free() of the timepoint is called before the
 *	free() of the corresponding task! Only timepoints are added which
 *	lies in the future. This means timepoint.time >= scheduler.time.
 *
 *	The timepoints are stored in order so that the correct calling
 *	order is guarantied. The calling order of timepoints at the same
 *	time is undefined. All timepoint->function()'s at the a time t1
 *	will also be called of an scheduler_sync(s, t1) invokation.
 *
 *	This function call also called within a scheduler_task_add(),
 *      timepoint->function(), task->reset() or task->sync().
 *
 *      Before the scheduler calls the timepoint->function(), he decouple the
 *	structure. So it's possible to add the same timepoint in the
 *	timepoint->function().
 *
 * EXAMPLE
 *
 *	my_function(scheduler *s, timepoint *tp, time *t)
 *	{
 *          printf("function call at time %f\n", t);
 *
 *          scheduler_add_function(s, tp, t+0.01);
 *      }
 *
 *	my_tp_free(scheduler *s, timepoint *tp)
 *	{
 *	    printf("tp freed!\n");
 *	    free(tp);
 *      }
 *
 *	main()
 *	{
 *	    scheduler my_s;
 *	    task my_t;
 *	    timepoint my_tp1;
 *	    timepoint *my_tp2;
 *
 *
 *	    scheduler_new(&my_s, 0.0);
 *	... scheduler and task initialization ...
 *
 *          my_tp1.workspace = 0;
 *          my_tp1.function  = my_function;
 *          my_tp1.free = 0;
 *          scheduler_add_point(&my_s, &my_tp1, 1.0);
 *
 *	    my_tp2 = malloc(sizeof(timepoint));
 *          my_tp2->workspace = 0;
 *          my_tp2->function  = my_function;
 *          my_tp2->free = my_free;
 *          scheduler_add_point(&my_s, my_tp2, 1.3);
 *
 *	    scheduler_sync(&my_s, 1.5);
 *
 *	    scheduler_free(&my_s);
 *	}
 *
 *	-> OUTPUT:
 *          function call at time 1.0
 *          function call at time 1.1
 *          function call at time 1.2
 *          function call at time 1.3
 *          tp freed!
 *          function call at time 1.3
 *          function call at time 1.4
 *          function call at time 1.5
 *
 * NOTES
 *	This is the only function which can be call within a
 *	scheduler_task_add(), timepoint->function(), task->reset()
 *	or task->sync(). The last three will be invoked from the scheduler
 *	automatically.
 *
 * SEE ALSO
 *	scheduler_free(), scheduler_reset(), scheduler_sync(),
 *	scheduler_task_add(), scheduler_point_add()
 *
 ****************************************************************************
 *
 */




void scheduler_free(scheduler *s);
/****************************************************************************
 * NAME
 *	scheduler_free(...)		free scheduler
 *
 * FUNCTION
 *	Free all timepoints and tasks of a scheduler.
 *	A freed scheduler could be reused after a scheduler_new().
 *
 * EXAMPLE
 *
 *	scheduler my_s;
 *
 *	scheduler_new(&my_s, 0.0);
 *      ... do something ...
 *      scheduler_free(&my_s);
 *
 * SEE ALSO
 *	scheduler_free(), scheduler_reset(), scheduler_sync(),
 *	scheduler_task_add(), scheduler_point_add()
 *
 ****************************************************************************
 *
 */

void scheduler_reset(scheduler *s, double time);
/****************************************************************************
 * NAME
 *	scheduler_reset(...)		resets scheduler to spec. time
 *
 * FUNCTION
 *	This function is necessary to begin the "problem-solving" again.
 *	First, all timepoints get free'd. Than for each task a task->reset()
 *      call is invoked to the specified "time". So the "workspace" can
 *	reseted to this time. This don't mean that the tasks would be
 *	removed! To remove all tasks use the scheduler_free() function.
 *
 *	After this function the scheduler should be consistent. The
 *	scheduler time is now the reset-time!
 *
 * EXAMPLE
 *
 *	void my_sync(scheduler *s, task *ta, double t, double t_old)
 *	{
 *          double d;
 *
 *	    d = t * t + 10.0;
 *	    *((double *)ta->workspace) = d;
 *	    printf("sync data to %lf\n", d);
 *
 *	}
 *
 *      void my_reset(scheduler *s, task *ta)
 *	{
 *	    double *dp;
 *
 *	    dp = (double *)ta->workspace;
 *          *dp = 10.0;
 *	    printf("reset data to 10.0\n");
 *	}
 *
 *	scheduler my_s;
 *	task my_t;
 *	double	data;
 *
 *      my_s.workspace = &data;
 *      scheduler_new(&my_s, 0.0);
 *
 *      my_t.workspace = &data;
 *	my_t.free = 0;
 *	my_t.sync = my_sync;
 * 	my_t.reset = my_reset;
 *	scheduler_task_add(&my_s, &my_t);
 *
 *	scheduler_reset(&my_s, 0.0);
 *
 *	scheduler_sync(&my_s, 1.0);
 *	scheduler_sync(&my_s, 2.0);
 *	scheduler_sync(&my_s, 3.0);
 *
 *	scheduler_reset(&my_s, 0.0);
 *
 *	-> OUTPUT:
 *          reset data to 10.0
 *          sync data to 11.0
 *          sync data to 14.0
 *          sync data to 19.0
 *          reset data to 10.0
 *
 * SEE ALSO
 *	scheduler_free(), scheduler_reset(), scheduler_sync(),
 *	scheduler_task_add(), scheduler_point_add<()
 *
 ****************************************************************************
 *
 */

#ifndef NO_TEST_CODE
int scheduler_test();
/****************************************************************************
 * NAME
 *	scheduler_test(...)	test functionality of scheduler module
 *
 * FUNCTION
 *	This routine performs a represantive boundle of test-cases to 
 *	see if the specified functionality is provided! This routine is
 *	only necessary during the module-development.
 *
 * OUTPUT
 *	The "informativ" text should be printed out in stderr.
 *
 * RESULTS
 *	return number of founded mail-functions.
 *	A 0 means that the scheduler-functions runs without error through the
 *	test.
 *
 * NOTE
 *	This routine is only availabe if NO_TEST_CODE is undefined.
 * EXAMPLE
 *	if (scheduler_test()) printf("error, don't use scheduler-routines!\n");
 *
 ****************************************************************************
 *
 */
#endif

#endif
