/********************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: space.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 17:01:23 stolz>
 ********************************************************************************
 *  MODULE              : space
 *  IMPORT		: list, error, basic, vector
 *
 *  IMPLEMENTATION
 *
 *	The space is organized as a grid with fixed dimensions. Each
 *	triangle is member of one or more grids or of the outside.
 *      This reduce the effort for collision-tests etc. The grid, also
 *	named as partition, is perpendicular. All partition build
 *	the "box".
 *
 *
 *  NOTE
 *      The module specification and definition is given in space.h.
 *	Define DEBUG_DETAIL to get the debugging output!!.
 *
 *
 *  HISTORY
 *  02-May-94   Creation date
 *  10-May-94   add partition clipping
 *  10-May-94	module test passed, space_new() and triangle_add() works fine
 *  13-May-94   triangle/line piercing succesfull. test passed.
 *  14-May-94	error in triangle piercing detected. some test cases added.
 *  15-May-94	triangle piercing now bug-free! WHOW!
 *  16-May-94   split space-module into space,triangle,clipping
 *  27-May-94   implement and succ. test triangle_pick().
 *  28-May-94   problem in next_part_index() with diagonals fixed.
 *  22-Jun-94   Introduce a group-number for an triangle ->coll-test!
 *  04-Jul-94   Change Collision-Test. Add space_new2().
 *  14-Aug-94   Add triangle-projection support.
 *  03-Oct-94   Deactivate the distance-criteria for the tp-triangle-selection.
 *
 ********************************************************************************/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <float.h>

#include "space_private.h"
#include "error.h"
#include "basic.h"
#include "pdlist.h"


#define MIN_PROJ_DISTANCE 1e-2    /* 0.01m */

bounding_box box_new(vector3 position, vector3 edges,
		     int dim_x, int dim_y, int dim_z)
/****************************************************************************
 * FUNCTION
 * 	Generate a bounding box at the specified position, edge and
 *	dimension. Also allocate and initialize the partitions.
 *
 * PREREQUISITES
 *	dim_x, dim_y, dim_y > 0 (take care of your memory!)
 *	edges > 0
 */
{
    bounding_box b;
    int x, y, z;
    vector3 p, e;

    if (dim_x <=0 || dim_y <= 0 || dim_z <= 0)
	error("space: box_new() - dimension argument<=0");

    if (!v3_is_positive(edges))
	error("space: box_new() vector edges are not positive");

    b = malloc(sizeof(adt_bounding_box));
    if (b) {

	v3_let(b->position, position);
	v3_let(b->edges, edges);
	v3_add(b->end_pos, position, edges);

	b->dim[0] = dim_x;
	b->dim[1] = dim_y;
	b->dim[2] = dim_z;

	b->nr_partitions = dim_x * dim_y * dim_z;

	b->nr_triangles = 0;
    }
    else error("space: box_new() - can't alloc memory for adt bounding_box");

    e[0] = edges[0] / (double)dim_x;
    e[1] = edges[1] / (double)dim_y;
    e[2] = edges[2] / (double)dim_z;

    v3_let(b->partition_edges, e);

    b->partitions = malloc(sizeof(partition)*b->nr_partitions);
    if (b->partitions) {
	for (x=0;x<dim_x; x++)
	    for (y=0;y<dim_y; y++)
		for (z=0;z<dim_z; z++) {
		    p[0] = position[0]+ (double)x/(double)dim_x * edges[0];
		    p[1] = position[1]+ (double)y/(double)dim_y * edges[1];
		    p[2] = position[2]+ (double)z/(double)dim_z * edges[2];
		    v3_let(part(b,x,y,z).position, p);
		    v3_let(part(b,x,y,z).edges, e);
		    v3_add(part(b,x,y,z).max_edge, p, e);
		    part(b,x,y,z).triangles = list_new();
		}
    }
    else error("space: box_new() - can't alloc memory for partitions");

    b->outside_triangles = list_new();

    return b;
}


void box_free(bounding_box b)
/****************************************************************************
 * frees the box and deallocate the partitions.
 */
{
    long l;

    for (l=b->nr_partitions-1; l>=0; l--)
	list_free(b->partitions[l].triangles);

    free(b->partitions);
    list_free(b->outside_triangles);
    free(b);
}



space space_new()
/****************************************************************************
 * initialize a space structure inclusive the bounding-box.
 */
{
    space s;
    vector3 edges;

    v3_set(edges, 1.0, 1.0, 1.0);

    s = (space)malloc(sizeof(adt_space));
    if (s) {
	s->box = box_new(v3_zero, edges, 2, 2, 2);
	s->triangles = list_new();
    }
    else error("space: space_new() - can't alloc memory for adt space");

    return s;
}


space space_new2(vector3 ref, vector3 edges, int dim[3])
/****************************************************************************
 * initialize a space structure inclusive the bounding-box.
 */
{
    space s;

    if (dim[0]<=0 || dim[1]<=0 || dim[2]<=0 ||
	edges[0]<ZERO_MAX || edges[1]<ZERO_MAX || edges[2]<ZERO_MAX)
	error("space: space_new2() - dimension or edges lower than 0");

    s = (space)malloc(sizeof(adt_space));
    if (s) {
	s->box = box_new(ref, edges, dim[0], dim[1], dim[2]);
	s->triangles = list_new();
    }
    else error("space: space_new() - can't alloc memory for adt space");

    return s;
}


void space_free(space s)
/****************************************************************************
 * frees the space and all of his triangles
 */
{
    triangle t;

    box_free(s->box);

    for (t=list_first(s->triangles); t; t=list_next(s->triangles)) {
	free(t);
    }
    list_free(s->triangles);
}




static void triangle_boundaries(int *min, int *max, bounding_box b,
			  vector3 *abs_edges)
/****************************************************************************
 * FUNCTION
 *	The boundries of a sub-box in which lies the triangles
 *      are computed. This is done by computing the min/max-values
 *	of the triangle-edges and transform them into partition
 *	indices.
 *	part_edges represent the
 *
 * RESULTS
 *    min[0..2], max[0..2]
 *    the triangle probably lies in the partitions
 *    0<=min[i]<max[i]<=dim[i]
 */
{
    double min_coord, max_coord;
    int i;


    for (i=0; i<3; i++) {


	/* compute min and max coordinate of the triangle
	 */
	min_coord = abs_edges[0][i];
	max_coord = min_coord;

	if (abs_edges[1][i] < min_coord) min_coord = abs_edges[1][i];
	else if (abs_edges[1][i] > max_coord) max_coord = abs_edges[1][i];

	if (abs_edges[2][i] < min_coord) min_coord = abs_edges[2][i];
	else if (abs_edges[2][i] > max_coord) max_coord = abs_edges[2][i];

        /* transform this into a valid partition-index
	 */
	min_coord -= b->position[i];
	max_coord -= b->position[i];

	min[i] = (int)(min_coord / b->partition_edges[i]);
	min[i] = max(0, min[i]);
	max[i] = (int)(max_coord / b->partition_edges[i]) + 1;
	max[i] = min(b->dim[i], max[i]);
    }
}


boolean triangle_in_partition(triangle t, vector3 position,
			      vector3 max_edge)
/****************************************************************************
 * FUNCTION
 *	check if a triangle lies in the space, which is spaned of the
 *	partition.
 *
 * RESULTS
 *      returns TRUE if the triangle lies in the partition or
 *      cut the partition-boundries.
 */
{
    vector3 *abs_edges;

    abs_edges = triangle3_abs_edges(t->triangle);

    /* fast test by checking if one edge of the triangle lies in
     * the partition
     */
    if (v3_is_in_range(position, abs_edges[0], max_edge) ||
        v3_is_in_range(position, abs_edges[1], max_edge) ||
        v3_is_in_range(position, abs_edges[2], max_edge)) {

	return TRUE;

    } else {
	list triangles, clipped;
	int i;

	/* clip the triangle at the boundries of the box
         * if after clipping to all 3 dimension a clipped
         * triangle is returned, the triangle is member of
	 * this partition
	 */
	triangles = list_new();
	list_add(triangles, t->triangle);

        clipped = triangle3_clip(triangles, 0, position[0], max_edge[0]);

	list_free(triangles);
        triangles = clipped;
	if (list_size(triangles)==0) {
            triangle3_list_free(triangles);
	    return FALSE;
	}

        clipped = triangle3_clip(triangles, 1, position[1], max_edge[1]);

	triangle3_list_free(triangles);
        triangles = clipped;
	if (list_size(triangles)==0) {
            triangle3_list_free(triangles);
	    return FALSE;
	}

        clipped = triangle3_clip(triangles, 2, position[2], max_edge[2]);
	triangle3_list_free(triangles);
	i = list_size(clipped);
	triangle3_list_free(clipped);

	return (i>0) ? TRUE : FALSE;
    }
}


boolean triangle_out_of_box(triangle t, bounding_box b)
{
    vector3 *abs_edges;

    abs_edges = triangle3_abs_edges(t->triangle);

    if (v3_out_of_range(b->position, abs_edges[0], b->end_pos) ||
        v3_out_of_range(b->position, abs_edges[1], b->end_pos) ||
        v3_out_of_range(b->position, abs_edges[2], b->end_pos)) {

	return TRUE;
    }
    else {
	return FALSE;
    }
}



void box_add_triangle(bounding_box b, triangle t)
/****************************************************************************
 * FUNCTION
 *	Add a triangle to a box.
 *	This means the insertion to the triangle-lists of
 *	the correct partitions.
 *
 */
{
    int x,y,z;
    int min[3],max[3];
    partition *p;

    b->nr_triangles = b->nr_triangles + 1;

    triangle_boundaries(min, max, b, triangle3_abs_edges(t->triangle));

    for (x=min[0];x<max[0];x++)
	for (y=min[1];y<max[1];y++)
	    for (z=min[2];z<max[2];z++) {

		p = &(part(b,x,y,z));

#ifdef DEBUG_DETAIL
    fprintf(stderr, "Is triangle %p in partition [%d][%d][%d]? ",
	    t, x, y, z);
#endif

		if (triangle_in_partition(t, p->position, p->max_edge)) {
		    list_add(t->member_boxes, p->triangles);
		    list_add(p->triangles, t);
#ifdef DEBUG_DETAIL
		    fprintf(stderr, "yes.\n");
		} else {
		    fprintf(stderr, "no.\n");
#endif
		}
	    }

    /* add triangle to outside list if it is necessary
     */
    if (triangle_out_of_box(t, b)) {

	list_add(b->outside_triangles, t);
	list_add(t->member_boxes, b->outside_triangles);
    }
}



void box_remove_triangle(bounding_box b, triangle t)
/****************************************************************************
 * FUNCTION
 *      Remove the triangle from a box.
 *	This means the removing of the triangle of all triangle-list
 *	in which the triangle is a member.
 */
{
    list l;

    /* decrement element counter
     */
    b->nr_triangles = b->nr_triangles - 1;

    /* remove boxes from partitions (also from the outside)
     */
    l = list_first(t->member_boxes);
    while(l) {
	list_remove(l, t);
	l = list_next(t->member_boxes);
    }
}


triangle3 triangle_triangle3(triangle t)
{
    return t->triangle;
}


triangle triangle_new(space s, vector3 *edges)
/****************************************************************************
 * FUNCTION
 *      Generate a triangle with 3 edges.
 *	This routine computes all structure-elements out of the
 *	three edges.
 *	The triangle is also includes into the space s. s=0 denotes a free
 *	triangle, which is not a member of an space.
 *
 * RESULT
 *	a pointer to a triangle structure.
 */
{
    triangle t;

    t = malloc(sizeof(adt_triangle));
    if (t) {
        t->space = s;

	t->triangle = triangle3_new(edges);
	if (t->triangle == 0)
	    error("space: triangle_new() - triangle is too small!");

        t->user_data = 0;
	t->group_nr  = 0;
	t->center[0] = (edges[0][0] + edges[1][0] + edges[2][0])/3.0;
	t->center[1] = (edges[0][1] + edges[1][1] + edges[2][1])/3.0;
	t->center[2] = (edges[0][2] + edges[1][2] + edges[2][2])/3.0;

	if (s) {
	    t->member_boxes = list_new();

            list_add(s->triangles, t);

	    box_add_triangle(s->box, t);
	}
    }
    else error("space: triangle_new() - can't alloc memory for adt triangle");
    return t;
}


void triangle_move(triangle t, vector3* edges)
/****************************************************************************
 * FUNCTION
 *	Change the triangle t to the new edges.
 *	This will rearrange the position in the bounding-box.
 *
 */
{
    if (t->space) box_remove_triangle(t->space->box, t);

    triangle3_free(t->triangle);
    t->triangle = triangle3_new(edges);
    if (t->triangle == 0) 
	error("space: triangle_move() - triangle is too small!");

    if (t->space) box_add_triangle(t->space->box, t);
}

void triangle_free(triangle t)
/****************************************************************************
 * FUNCTION
 *      Removes a triangle from the space and deallocate the
 *      structure.
 */
{
    if (t->space) {
	box_remove_triangle(t->space->box, t);

    	list_remove(t->space->triangles, t);
        list_free(t->member_boxes);
    }

    triangle3_free(t->triangle);
    free(t);
}


boolean	triangle_collision(triangle t)
/****************************************************************************
 * FUNCTION
 *	Check if a triangle which is member of an space
 *      is piercing a other triangle of this space. 
 *	The Routine checks only triangles with a different
 *	group-nr, or against all if the group-nr is 0.
 *
 * RESULT
 *	Returns TRUE for piercing.
 */
{
    triangle t2;
    list l;
    list tl;

    if (t->space == 0)
	error("space: triangle_collision() - triangle is not member of a space");

    /* generate a list of unique triangles
     */
    tl = list_new();
    for (l=list_first(t->member_boxes); l; l=list_next(t->member_boxes)) {

	t2 = list_first(l);
	while (t2) {
            if ((t!=t2) && (t->group_nr!=t2->group_nr || 
		 (t->group_nr | t2->group_nr)==0)) {

		list_remove(tl, t2);
		list_add(tl, t2);
	    }
            t2=list_next(l);
	}
    }
    t2 = list_first(tl);
    while (t2) {
	if (triangle3_piercing(t->triangle, t2->triangle)) {
	    list_free(tl);
	    return TRUE;
	}
	t2=list_next(tl);
    }
    list_free(tl);

/* old routine:
    for (l=list_first(t->member_boxes); l; l=list_next(t->member_boxes)) {

	t2 = list_first(l);
	while (t2) {
            if (t->group_nr!=t2->group_nr && t!=t2 &&  
		triangle3_piercing(t->triangle, t2->triangle))
		return TRUE;
            t2=list_next(l);
	}
    }
*/
    return FALSE;
}

vector3* triangle_edges(triangle t)
{
    return triangle3_abs_edges(t->triangle);
}

void triangles_do(space s, void (*doit)(triangle t, vector3 *edges))
{
    triangle t;

    for (t=list_first(s->triangles); t; t=list_next(s->triangles)) {
	doit(t, triangle3_abs_edges(t->triangle));
    }
}


static void partition_index(int *index, bounding_box b, vector3 position)
/* Compute the partition-indices of the point "position"
 * returns an index[0] < 0 for out-of-box
 */
{
    if (index) {
	index[0] = -1;
	index[1] = -1;
	index[2] = -1;
    }
    else error("space: partition_index() - index-array-pointer is zero!");

    if (v3_is_in_range(b->position, position, b->end_pos)) {
	index[0] = (int)((position[0] - b->position[0])/b->partition_edges[0]);
	index[1] = (int)((position[1] - b->position[1])/b->partition_edges[1]);
	index[2] = (int)((position[2] - b->position[2])/b->partition_edges[2]);

	if (index[0] <0 || index[0] >= b->dim[0])
	   error("space: partition_index() - x-index out of range!");
	if (index[1] <0 || index[1] >= b->dim[0])
	   error("space: partition_index() - y-index out of range!");
	if (index[2] <0 || index[2] >= b->dim[0])
	   error("space: partition_index() - z-index out of range!");
    }

#ifdef DEBUG_DETAIL
    fprintf(stderr, "part-index of the point is [%d,%d,%d]\n", index[0], index[1], index[2]);
#endif
}


static void first_partition_index(vector3 *next, int *index,
				  bounding_box b,
				  vector3 position, vector3 direction)
{
    int i;
    double d, alpha=0.0;
    vector3 v;

    index[0] = -1;
    index[1] = -1;
    index[2] = -1;

    for (i=0; i<3; i++) {

	if (is_not_zero(direction[i])) {

	    d = (b->position[i] - position[i]) / direction[i];
#ifdef DEBUG_DETAIL
            fprintf(stderr, "factor[%d]=%f to reach box->position[%d]\n",
		    i, d, i);
#endif
	    if (d>0.0) {
		d *= (1.0 + 10 * ZERO_MAX);
	        v3_addscal(v, position, d, direction);

	        if (v3_is_in_range(b->position, v, b->end_pos)
                    && (alpha==0.0 || d < alpha)) {
                    alpha = d;
                    v3_let(*next, v);
	        }
            }


	    d = (b->end_pos[i] - position[i]) / direction[i];
#ifdef DEBUG_DETAIL
            fprintf(stderr, "factor[%d]=%f to reach b->end_pos[%d]\n",
		    i, d, i);
#endif
	    if (d>0.0) {
		d *= (1.0 + 10 * ZERO_MAX);
	        v3_addscal(v, position, d, direction);

	        if (v3_is_in_range(b->position, v, b->end_pos)
                    && (alpha==0.0 || d < alpha)) {
                    alpha = d;
                    v3_let(*next, v);
	        }
            }
	}
    }

#ifdef DEBUG_DETAIL
    fprintf(stderr, "alpha to reach bounding-box: %f\n", alpha);
    fprintf(stderr, "point in box: (%1.2f %1.2f %1.2f)\n", 
		    (*next)[0], (*next)[1], (*next)[2]);
#endif

    if (alpha!=0.0)
	partition_index(index, b, *next);
}


static void next_partition_index(vector3 *next, int *index, bounding_box b,
			     vector3 position, vector3 direction)

{
    double alpha = 0.0;
    double d;
    int index_diff[3] = {0, 0, 0};
    int i;
    partition *p;

#ifdef DEBUG_DETAIL
    fprintf(stderr, "next_part_index() curr. index=[%d,%d,%d]\n",
		    index[0], index[1], index[2]);
#endif

    if (index[0] < 0 || index[0] >= b->dim[0])
	error("space: next_partition_index() - index out of range");

    /* get the partition of this point
     */
    p = &part(b, index[0], index[1], index[2]);

    /* compute for all 3 dimension the piercing point through
     * the partition-borders and remeber the smalles factor.
     * store also the index which should be inc/dec to reach the
     * new partition.
     */
    for (i=0; i<3; i++) {

	if (is_not_zero(direction[i])) {

	    d = (p->position[i] - position[i]) / direction[i];
#ifdef DEBUG_DETAIL
            fprintf(stderr, "factor[%d]=%f to reach p->position[%d]\n",
		    i, d, i);
#endif
	    if (d > ZERO_MAX) {
		if (alpha==0.0) {
		    alpha = d;
		    index_diff[0] = 0;
		    index_diff[1] = 0;
		    index_diff[2] = 0;
		    index_diff[i] = -1;
                }
		else if (is_zero(d-alpha)) {	/* line goes trough an edge */
		    alpha = d;
		    index_diff[i] = -1;
		}
		else if (d < alpha) {
		    alpha = d;
		    index_diff[0] = 0;
		    index_diff[1] = 0;
		    index_diff[2] = 0;
		    index_diff[i] = -1;
		}
	    }

	    d = (p->max_edge[i] - position[i]) / direction[i];
#ifdef DEBUG_DETAIL
            fprintf(stderr, "factor[%d]=%f to reach p->max_edge[%d]\n",
		    i, d, i);
#endif
	    if (d > ZERO_MAX) {
		if (alpha==0.0) {
		    alpha = d;
		    index_diff[0] = 0;
		    index_diff[1] = 0;
		    index_diff[2] = 0;
		    index_diff[i] = 1;
                }
		else if (is_zero(d-alpha)) {	/* line goes trough an edge */
		    alpha = d;
		    index_diff[i] = +1;
		}
		else if (d < alpha) {
		    alpha = d;
		    index_diff[0] = 0;
		    index_diff[1] = 0;
		    index_diff[2] = 0;
		    index_diff[i] = 1;
		}
	    }
	}
    }

#ifdef DEBUG_DETAIL
    fprintf(stderr, "alpha=%f, index_diff=[%d,%d,%d]\n", alpha,
	    index_diff[0], index_diff[1], index_diff[2]);
#endif

    if (alpha != 0.0) {

	for (i=0; i<3; i++) {
            index[i] += index_diff[i];

            if (index[i]<0 || index[i] >= b->dim[i]) {	/* out of box */
                index[0] = -1;
                index[1] = -1;
	        index[2] = -1;

                return;
	    }
	}

	v3_addscal(*next, position, alpha, direction);

    } else
	error("space: next_partition_index() - no new index found!");
}



triangle triangle_pick(double *distance, space s, vector3 position, vector3 direction)
/****************************************************************************
 */
{
    triangle first_triangle = 0;
    boolean outside_not_tested = TRUE;
    vector3 start;
    triangle t;
    double d;
    int    index[3];
    boolean  t_found = FALSE;
    list   l;

    /* a list with unique triangles would be better.
     * so triangles will be test more than once if they a member
     * of different partitions.
     */

    v3_let(start, position);	/* remember starting-point */

    *distance = 1.0e9;

    if (v3_is_in_range(s->box->position, position, s->box->end_pos)) {

        /* starting point "position" is in the bounding box
	 */
	partition_index(index, s->box, start);

    } else {

        /* starting point "position" is not in the bounding box
	 */

	/* compute the distance to the outside-triangles
	 */
	l =  s->box->outside_triangles;

	for (t=list_first(l); t; t=list_next(l)) {
	    if (triangle3_distance(&d, position, direction, t->triangle)) {
		if (d >= 0.0 && d < *distance) {
		    *distance = d;
                    first_triangle = t;
		}
	    }
	}

	outside_not_tested = FALSE;

	index[0] = -1;
	index[1] = -1;
	index[2] = -1;

	first_partition_index(&start, index, s->box, start, direction);
    }

    while (index[0] >= 0) {
	l =  part(s->box, index[0], index[1], index[2]).triangles;

	/* compute distance to the triangles in the partition(index[])
	 */

	for (t=list_first(l); t; t=list_next(l)) {
	    if (triangle3_distance(&d, position, direction, t->triangle)) {
		t_found = TRUE;
		if (d >= 0.0 && d < *distance) {
		    *distance = d;
                    first_triangle = t;
		}
	    }
	}

	/* triangle with minimal distance found.
	 */
	if (t_found && first_triangle) {
	    *distance = *distance * v3_sum(direction);
	    return first_triangle;
	}

	next_partition_index(&start, index, s->box, start, direction);
    }


    if (outside_not_tested) {

	l =  s->box->outside_triangles;

	for (t=list_first(l); t; t=list_next(l)) {
	    if (triangle3_distance(&d, position, direction, t->triangle)) {
		if (d >= 0.0 && d < *distance) {
		    *distance = d;
                    first_triangle = t;
		}
	    }
	}
    }

    *distance = *distance * v3_sum(direction);
    return first_triangle;

/*
    return list_first(s->triangles);
 */
}



void triangle_set_user_data(triangle t, void *userdata)
/****************************************************************************
 */
{
    t->user_data = userdata;
}

void*		triangle_get_user_data(triangle t)
/****************************************************************************
 */
{
    return t->user_data;
}

void triangle_set_group_nr(triangle t, int nr)
/****************************************************************************
 */
{
    t->group_nr = nr;
}

int triangle_get_group_nr(triangle t)
/****************************************************************************
 */
{
    return t->group_nr;
}



void triangle_print(triangle t)
/****************************************************************************
 */
{
    bounding_box b;
    vector3	*abs_edges, *rel_edges;
    double	*rel_length;

    abs_edges = triangle3_abs_edges(t->triangle);
    rel_edges = triangle3_rel_edges(t->triangle);
    rel_length = triangle3_edges_length(t->triangle);

    printf("TRIANGLE(%p) with edges:\n"
		    "(%1.2f %1.2f %1.2f) (%1.2f %1.2f %1.2f) (%1.2f %1.2f %1.2f)\n", t,
            abs_edges[0][0], abs_edges[0][1], abs_edges[0][2],
            abs_edges[1][0], abs_edges[1][1], abs_edges[1][2],
            abs_edges[2][0], abs_edges[2][1], abs_edges[2][2]);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "rel_edges: (%1.2f %1.2f %1.2f) (%1.2f %1.2f %1.2f) (%1.2f %1.2f %1.2f)\n"
                    "rel_length: %1.2f %1.2f %1.2f\n",
            rel_edges[0][0], rel_edges[0][1], rel_edges[0][2],
            rel_edges[1][0], rel_edges[1][1], rel_edges[1][2],
            rel_edges[2][0], rel_edges[2][1], rel_edges[2][2],
            rel_length[0], rel_length[1], rel_length[2]);
#endif


    printf("User data = %p\n", t->user_data);

    if (t->space) {
        printf("Is member of boxes:");
	for (b=list_first(t->member_boxes); b; b=list_next(t->member_boxes))
            printf(" %p", b);
	printf("\n");
    }
}

void triangles_print(list tl)
/****************************************************************************
 */
{
    triangle t;

    for (t=list_first(tl); t; t=list_next(tl))
	triangle_print(t);
}


void box_print(bounding_box b)
/****************************************************************************
 */
{
    int x, y, z;
    triangle t;

    printf("BOX(%p)\nat position (%f %f %f)\nWith edges (%f %f %f)\n", b,
	    b->position[0], b->position[1], b->position[2],
            b->edges[0], b->edges[1], b->edges[2]);

    printf("partition edges (%f %f %f)\n",
            b->partition_edges[0], b->partition_edges[1], b->partition_edges[2]);

    printf("Dimension: %dx%dx%d (%ld)\n",
	    b->dim[0], b->dim[1], b->dim[2], b->nr_partitions);

    printf("Total number of triangles: %ld.\n", b->nr_triangles);

    printf("Partitions:\n");

    for (x=0; x<b->dim[0]; x++)
	for (y=0; y<b->dim[1]; y++)
	    for (z=0; z<b->dim[2]; z++) {
		printf("[%d][%d][%d] with triangles:", x, y, z);
		t=list_first(part(b,x,y,z).triangles);
		while (t) {
		    printf(" %p", t);
/*
		    triangle_print(t);
*/
		    t=list_next(part(b,x,y,z).triangles);
		}
		printf("\n");
	    }

    printf("Outside triangles:");
    t=list_first(b->outside_triangles);
    while (t) {
	printf(" %p", t);
	t=list_next(b->outside_triangles);
    }
    printf("\n");
}


void space_print(space s)
/****************************************************************************
 */
{
    printf("SPACE with:\n");
    triangles_print(s->triangles);
    printf("\nAnd:\n");
    box_print(s->box);
}





/*
#define DEBUG_DETAIL 1
*/



triangle_projection	tp_new(vector3 *edges, vector3 focus)
{
    t3p  p = t3p_new(edges, focus);
    triangle_projection tp = 0;

    if (p) {
	tp = malloc(sizeof(adt_tp));
	if (tp==0) error("space: tp_new() - can't alloc memory for adt.");

        tp->ptriangle = p;
	tp->root = 0;
	tp->mirror_on = 0;
    }

    return tp;
}


triangle_projection	tp_new2(t3p p)
{
    triangle_projection tp = 0;

    if (p) {
	tp = malloc(sizeof(adt_tp));
	if (tp==0) error("space: tp_new2() - can't alloc memory for adt.");

        tp->ptriangle = p;
	tp->root = 0;
	tp->mirror_on = 0;
    }

    return tp;
}



void			tp_free(triangle_projection tp)
{
    if (tp) {
	t3p_free(tp->ptriangle);
	free(tp);
    }
}


vector3*		tp_focus(triangle_projection tp)
{
    return t3p_focus(tp->ptriangle);
}


vector3*		tp_edges(triangle_projection tp)
{
    return triangle3_abs_edges(t3p_triangle3(tp->ptriangle));
}

/*
static void point_boundaries(int *min, int *max, bounding_box b,
			  vector3 point)
{
    int i;

    for (i=0; i<3; i++) {
 	min[i] = (int)((point[i] - b->position[i]) / b->partition_edges[i]);
        max[i] = min[i] + 1;
    }
}


static void limit_boundaries(int *min, int *max, bounding_box b)
{
    int i;

    for (i=0; i<3; i++) {
	min[i] = max(0, min[i]);
        max[i] = min(b->dim[i], max[i]);
    }
}


static void index_add_boundaries(int *min, int *max, int index, int value)
{
    if (value<min[index]) min[index] = value;
    else if (value>=max[index]) max[index] = value + 1;
}


static void point_add_boundaries(int *min, int *max, bounding_box b,
			  vector3 point)
{
    int i, p;

    for (i=0; i<3; i++) {
	p = (int)((point[i] - b->position[i]) / b->partition_edges[i]);
	if (p<min[i]) min[i] = p;
	else if (p>=max[i]) max[i] = p + 1;
    }
}
*/


void project_triangle(list tp_cut, list tp_rest,
					space s, triangle_projection tp)
{

    int i, j;

    vector3 p1, *focus, *edges, focus_index;
    double d;
    triangle_projection tp2, tp_orginal;
    pdlist triangles = pdlist_new();
    list  rest3 = list_new();
    list  cut3 = list_new();
    triangle t;
    t3p p3;

    list rest_t[2];
    vector3 center;
    double focus_index_length;
    
    if (s==0 || tp==0) return;

    focus = tp_focus(tp);
    edges = tp_edges(tp);
    center[0] = (edges[0][0] + edges[1][0] + edges[2][0])/3.0;
    center[1] = (edges[0][1] + edges[1][1] + edges[2][1])/3.0;
    center[2] = (edges[0][2] + edges[1][2] + edges[2][2])/3.0;
    v3_sub(focus_index, center, *focus);
    focus_index_length = v3_sum(focus_index);


#ifdef DEBUG_DETAIL
    printf("project_triangle() - tp = %p\n", tp);
    printf("project_triangle: tp-center={%f, %f, %f}\n", v3_elems(center));
    printf("project_triangle: tp-focus-index={%f, %f, %f}, length=%f\n", v3_elems(focus_index), focus_index_length);
#endif
    /* sort the triangles in of the outside into the distance-sorted list.
     */
    t = list_first(s->triangles);
    while (t) {
	v3_sub(p1, t->center, center);
	d = v3_product(p1, focus_index) / focus_index_length;
	
	/* if (d>0.0) */ {
#ifdef DEBUG_DETAIL
	    printf("project_triangle: proj_distance = %f of triangle 0x%p (t-center %f %f %f)\n",
		   d, t, v3_elems(t->center));
#endif
	    pdlist_add(triangles, d, t);
	}
	t = list_next(s->triangles);
    }
    
#ifdef DEBUG_DETAIL
    printf("project_triangle: check against %ld triangles\n", 
	   pdlist_size(triangles));
#endif

    /* project the projection-triangle to nearest triangle
     */
    rest_t[0] = list_new();
    rest_t[1] = list_new();
    i = 0;
    j = 1;
    list_add(rest_t[i], tp);
    tp_orginal = tp;
    
    t = pdlist_lowest(&d, triangles);
    while(t) {

#ifdef DEBUG_DETAIL
	printf("project_triangle: Check against triangle %p, distance %f\n",  t, d);
	triangle_print(t);
#endif

	tp = list_first(rest_t[i]);
	while (tp) {
            triangle3_projection(cut3, rest3, tp->ptriangle, t->triangle);

#ifdef DEBUG_DETAIL
            printf("project_triangle: t3p_projection between tp %p and triangle %p\n", tp, t);
	    printf("project_triangle: cut3 has %d, rest3 has %d members\n",
		   list_size(cut3), list_size(rest3));
#endif
	    /* insert the rest of the tp in the rest-list
             * and remove the projection-triangle from the list.
             */
	    if (list_is_empty(cut3)) {
#if 0
                printf("project_triangle: Add tp %p to the rest-list\n", tp);
#endif
		list_add(rest_t[j], tp);
	    } else {
	        p3 = list_first(rest3);
	        while(p3) {
		    tp2 = tp_new2(p3);
                    if (tp2) {
			tp2->root = tp;
#if 0
			printf("project_triangle: Add tp %p to the rest-list\n", tp2);
			tp_print(tp2);
#endif
			list_add(rest_t[j], tp2);
		    }
		    p3 = list_next(rest3);
                }
	    }
	    
	    
	    /* insert the projection of the tp in the projection-list
             */
            if (tp_cut) {
		
	        p3 = list_first(cut3);
	        while(p3) {
		    tp2 = tp_new2(p3);
		    if (tp2) {
			tp2->root = tp;
			tp2->mirror_on = t;
			list_add(tp_cut, tp2);
		    }
		    p3 = list_next(cut3);
                }
	    } else list_for_all(cut3, t3p_free);
	    
	    
	    /* clear the t3p-list
             * and get the next rest-triangle for projection
             */
            list_remove_all(cut3);
            list_remove_all(rest3);
	    
	    tp = list_next(rest_t[i]);
	}
	
        list_remove_all(rest_t[i]);
	j = i;
	i = 1 - i;
	
	t = pdlist_higher(&d, triangles);
    }

    
#ifdef DEBUG_DETAIL
    printf("project_triangle: result is %d rest-triangles and ? cuts\n", 
	   list_size(rest_t[i]));
#endif

    if (tp_rest) {
	tp2 = list_first(rest_t[i]);
	while (tp2) {
	    if (tp2 != tp_orginal) list_add(tp_rest, tp2);
	    tp2 = list_next(rest_t[i]);
	}
    } else {
	tp2 = list_first(rest_t[i]);
	while (tp2) {
	    if (tp2 != tp_orginal) tp_free(tp2);
	    tp2 = list_next(rest_t[i]);
	}
    }

    list_free(rest_t[0]);
    list_free(rest_t[1]);

    pdlist_free(triangles);
    list_free(rest3);
    list_free(cut3);
}


void tp_print(triangle_projection tp)
{
    if (tp) {
	printf("TRIANGLE_PROJECTION(%p) {\n", tp);
	t3p_print(tp->ptriangle);
	printf(" root = %p\n mirror on = %p\n",
		tp->root, tp->mirror_on);
	printf("}\n");
    }
    else
	printf("TRIANGLE_PROJECTION = NIL");
}


triangle_projection	tp_root(triangle_projection tp)
{
    return tp->root;
}


triangle		tp_mirror_of(triangle_projection tp)
{
    return tp->mirror_on;
}

t3p                     tp_t3p(triangle_projection tp)
{
    return tp->ptriangle;
}

void		triangle_set_connections(triangle t,
                                         triangle t0, int side0,
                                         triangle t1, int side1,
                                         triangle t2, int side2)
{
    triangle3 tc0, tc1, tc2;

    if (t) {
	if (t0) tc0 = t0->triangle; else tc0 = 0;
        if (t1) tc1 = t1->triangle; else tc1 = 0;
        if (t2) tc2 = t2->triangle; else tc2 = 0;

        triangle3_set_connections(t->triangle, tc0, side0, tc1, side1, tc2, side2);
    }
}
