/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: text.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 17:02:11 stolz>
 **************************************************************************
 *  MODULE              : text
 *  IMPORT		: error
 *
 *  IMPLEMENTATION
 *
 *	The textbuffer consist of a variable array of chars.
 *      If the size of a text which should be added by text_add...()
 * 	is greater than the free space of the array, the array is
 *	reallocated with an greater size.
 *
 *  NOTE
 *      The module specification and definition is given in text.h!
 *	Define MODULE_TEST to generate the test-code inclusive a
 *	main()-routine.
 *
 *  HISTORY
 *  24-May-94   Creation date.
 *  25-May-94   Module test passed.
 *  06-Jun-94   text_ref() add.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "text.h"
#include "error.h"


/**************************************************************************
 * define the abstract datatype textbuffer
 */
typedef struct adt_textbuffer adt_textbuffer;
struct adt_textbuffer {
    char	*text;
    long        size;
    long        free;
};



textbuffer text_new()
/**************************************************************************
 * construct a new adt textbuffer
 */
{
    textbuffer tb;

    tb = malloc(sizeof(adt_textbuffer));
    if (tb) {
	tb->free = 0;
	tb->size = 0;
	tb->text = 0;
    }
    else error("text: text_new() - can't alloc memory for adt textbuffer");

    return tb;
}


void       text_free(textbuffer tb)
/**************************************************************************
 * deconstruct the textbuffer
 */
{
    if (tb->text) free(tb->text);
    free(tb);
}


textmarker text_add2(textbuffer tb, char *text, long size)
/**************************************************************************
 * append text to the end of the buffer
 * and return the textmarker of the buffer-end.
 */
{
    long new_size;

    if (size==0) return tb->size;
    else if (size < 0)
	error("text: text_add2() - try to add a text with size <0 ?");


    if (size > tb->free) {
	if (0 == tb->text) {
	    new_size = size+size;
	    tb->text = malloc(new_size);
#ifdef DEBUG_DETAIL
	    fprintf(stderr, "malloc with %ld\n", new_size);
#endif
	}
	else {
	    new_size = (tb->size + tb->free + size)<<1 ;
	    tb->text = realloc(tb->text, new_size);
#ifdef DEBUG_DETAIL
	    fprintf(stderr, "realloc with %ld\n", new_size);
#endif
	}
	tb->free = new_size - tb->size;

#ifdef DEBUG_DETAIL
	fprintf(stderr, "text-size = %ld, free=%ld, size=%ld\n",
		tb->size, tb->free, size);
#endif
	if (0 == tb->text)
	    error("text: text_add2() - can't alloc memory for the additional text");
    }
    memcpy(&(tb->text[tb->size]), text, size);
    tb->free -= size;
    tb->size += size;

    return tb->size;
}


textmarker text_add(textbuffer tb, char *text)
/**************************************************************************
 */
{
    return text_add2(tb, text, strlen(text));
}


textmarker text_marker(textbuffer tb)
/**************************************************************************
 * return the current position in the text.
 */
{
    return tb->size;
}


void       text_get(char *target,
		    textbuffer tb,
		    textmarker start, textmarker end)
/**************************************************************************
 * copy the contents of the textbuffer between the markers start-end to
 * the array target.
 */
{
    if (start==TEXT_END)
	start = tb->size;

    if (end==TEXT_END)
	end = tb->size;

    if (start<0 || end<0 || end > tb->size)
	error("text: text_get() - start or end lies out of text");

    if (start>end)
	error("text: text_get() - start lies behind end");

    memcpy(target, &(tb->text[start]), end - start);
    target[end - start] = '\0';
}

void text_ref(char **text, long *len, 
	      textbuffer tb, textmarker start, textmarker end)
{
    if (start==TEXT_END)
	start = tb->size;

    if (end==TEXT_END)
	end = tb->size;

    if (start<0 || end<0 || end > tb->size)
	error("text: text_ref() - start or end lies out of text");

    if (start>end)
	error("text: text_ref() - start lies behind end");

    *text = &(tb->text[start]);
    *len = end - start;
}

void       text_clear(textbuffer tb)
/**************************************************************************
 */
{
    if (tb->text) {
	free(tb->text);
    }
    tb->text = 0;
    tb->size = 0;
    tb->free = 0;
}





/****************************************************************************
 * MODULE-TEST-routines
 */

#ifdef MODULE_TEST


char *text  = "This is an test-text";
char *text1 = "This";
char *text2 = " is an";
char *text3 = " test-text";


int text_test()
{
    int failed, errors = 0;
    char index;

    textbuffer tb;
    char text_result[40];

    index = 'a';
    fprintf(stderr, "TEST-TEST from %s %s\n"
            "    %c) generate an empty text via text_new(..)\n",
	    __DATE__, __TIME__, index++);

    tb = text_new();

    if ((failed = (text_marker(tb)!=0))) errors++;
    fprintf(stderr, "    %c) test function text_marker(..) on empty text: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));



    if ((failed = (text_add(tb, text1)!=4))) errors++;
    fprintf(stderr, "    %c) test function text_add(..) on empty text: %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if ((failed = (text_add(tb, text2)!=10))) errors++;
    fprintf(stderr, "    %c) test function text_add(..) : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    if ((failed = (text_add(tb, text3)!=20))) errors++;
    fprintf(stderr, "    %c) test function text_add(..) : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    text_get(text_result, tb, TEXT_START, TEXT_END);



    if ((failed = (strcmp(text, text_result)!=0))) errors++;
    fprintf(stderr, "    %c) check whole text with text_get(..) : %s\n", index++,
			(failed ? "FAILED" : "PASSED"));

    fprintf(stderr, "Text in buffer: <%s>\n", text_result);
    fprintf(stderr, "Correct Text should be: <%s>\n", text);




    fprintf(stderr, "Total amount of errors: %d. Test %s!\n\n",
			errors, errors ? "FAILED" : "PASSED");


    return errors;
}

int main(int argc, char **argv)
{
    return text_test();
}

#endif


