/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: text.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 17:02:17 stolz>
 **************************************************************************
 *  MODULE              : text
 *
 *  FUNCTION
 *
 *	This module provide a simple textbuffer, in which a character based
 *	text can be stored. This is done by providing a abstract datatype
 *      textbuffer. Additional the type textmarker is defined for referencing
 *	to a specific position in the text.
 *
 *	The textbuffer is initialized with a call of text_new(). Now you
 *	can use the textbuffer to add text with text_add() or text_add2().
 *      With this functions, the text is added at the end of the buffer.
 *	The only difference is that text_add() appends a null-terminated
 *	string and needs no therefor no size-parameter.
 *	The Marker which is returned by this functions points to the end
 *	of the text after appending. The actual textmarker can also aquired
 *	by a call of text_marker(). Its also necessary that the start-marker
 *      is before end-marker.
 *
 *	To examine the textbuffer, use the function text_get().
 *	This function copies the text between the two marks in the target-
 *	array. Its possible to use the defines TEXT_START and TEXT_END to
 *	denote to the start or end of the buffer.
 *
 *      To clear the textbuffer use text_clear().
 *
 *	The freeing of the abstract datatype textbuffer call text_free().
 *
 *  EXAMPLE
 *
 *	textbuffer tb;
 *	textmarker tm1, tm2;
 *	char text[40];
 *
 *	tb = text_new();
 *
 *	tm1 = text_add(tb, "Hello!");
 *      // the textbuffer containts now "Hello!"
 *	// the textmarker tm1 points now to the end of the text.
 *
 *	// or get the text-marker by
 *      tm1 = text_marker(tb);
 *
 *	// append "How are you?" to the text.
 *	tm2 = text_add2(tb, "How are you? Thanks fine", 12);
 *
 *	text_get(text, tb, tm1, TEXT_END);
 *	printf("%s\n", text);
 *	// the output should be "How are you?"
 *
 *	// clear only the text
 *      text_clear(tb);
 *
 *	// clear the text and free the datatype textbuffer.
 *      text_free(tb);
 *
 *
 *  HISTORY
 *  24-May-94   Creation date.
 *  25-May-94   Add notes. Pass module test.
 *  06-Jun-94   text_ref() add.
 *
 **************************************************************************/
#ifndef TEXT_H
#define TEXT_H


typedef struct adt_textbuffer*    textbuffer;
typedef long                      textmarker;


/* textmarker defines
 */
#define TEXT_START    0l        /* Start of Text */
#define TEXT_END     -1l	/* End of Text */



textbuffer text_new();
void       text_free(textbuffer);
void       text_clear(textbuffer);

textmarker text_add(textbuffer, char *);
textmarker text_add2(textbuffer, char *, long size);

void       text_get(char *target, textbuffer, textmarker start, textmarker end);

void       text_ref(char **text, long *len, 
		    textbuffer, textmarker start, textmarker end);

textmarker text_marker(textbuffer);

#endif
