/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: triangle.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 17:02:27 stolz>
 **************************************************************************
 *  MODULE              : triangle
 *  IMPORT		: list, vector, basic, error
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in triangle.h!
 *
 *  HISTORY
 *  02-May-94   Creation date
 *  13-May-94   triangle/line piercing succesfull. test passed.
 *  14-May-94	error in triangle piercing detected. some test cases added.
 *  15-May-94	triangle piercing now bug-free! WHOW!
 *  16-May-94   split space-module into space,triangle,clipping
 *  27-May-94   add triangle3_distance() and test it.
 *  15-Jun-94   triangle3_clip() bug detected, 5 clipped triangles possible!
 *  23-Jun-94   triangle3_new(0 returns now 0 if the triangle is to small!
 *  28-Jul-94   Add projection support.
 *  17-Aug-84   Split projection-routines an basic vector geometry-routines
 *              to different modules.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "triangle.h"
#include "error.h"
#include "clipping.h"
#include "basic.h"
#include "vector_geometry.h"

/* the adt triangle3
 *
 * many struct-elements are for efficient purposes.
 */
typedef struct adt_triangle3 adt_triangle3;
struct adt_triangle3 {
    vector3       abs_edges[3];	     /* absolut positon of the t-edges */
    vector3	  rel_edges[3];	     /* relativ von edge[i] to edge[(i+1)%3] */
    double        rel_length[3];     /* length of the relativ edges */
    vector3       normal;            /* triangle-normal = -rel_edge[0]x[2] */
    double        bounding_radius;   /* bounding radius from 1.edge [0] */
    triangle3	  connection_triangle[3];
    signed char	  connection_side[3];
};



boolean	line3_in_t3_plane(double *factor, vector2 *point2, vector3 *point3,
		          vector3 line_start, vector3 line_index,
			  triangle3 t)
/****************************************************************************
 * FUNCTION
 *	Compute the line-index-factor of the piercing point between the
 *	triangle t and the infinite straight line (line_start, line_index)
 *
 * RESULT
 *      Returns TRUE if the line goes through the triangle-plain.
 *      Else return FALSE.
 *      In the var-variable factor, the factor for the piercing point
 *		line_start + factor*line_index
 *	will be returned.
 */
{
    vector3 v,		/* piercing point */
            y, z;
    double d;
    double alpha, beta;


    /* length of line in direction of the triangle-normal
     */
    d = v3_product(t->normal, line_index);

#if 0
    fprintf(stderr, "line-index-length in triangle-normal-direction: %f\n", d);
#endif

    if (is_zero(d)) return FALSE; /* line parallel to triangle */

    /* Starting point of the line relative to the 1.edge
     * of the triangle
     */
    v3_sub(v, line_start, t->abs_edges[0]);

    /* v3_product() = distance of the line-starting-point
     * to the triangle-plain
     */

    /* factor d*line_index + line_start lies in triangle plain
     */
    d = - v3_product(t->normal, v) / d;

#if 0
    fprintf(stderr, "line-index-factor from line-start to triangle-plain: %f\n", d);
#endif

    if (factor) *factor = d;

    /* absolute coordinates of the piercing point
     */
    if (point3) {
	v3_addscal(*point3, line_start, d, line_index);
    }

    /* point of piercing relative from 1.edge of the triangle
     */
    v3_addscal(v, v, d, line_index);

#if 0
    fprintf(stderr, "piercing point is = (%f %f %f)\n",
	    v[0], v[1], v[2]);
#endif

    alpha = v3_product(t->rel_edges[0], v) / (t->rel_length[0] * t->rel_length[0]);
    beta = -v3_product(t->rel_edges[2], v) / (t->rel_length[2] * t->rel_length[2]);

#if 0
    fprintf(stderr, "alpha' = %f, beta' = %f\n", alpha, beta);
#endif

    v3_subscal(y, v, alpha, t->rel_edges[0]);
    v3_addscal(z, v, beta, t->rel_edges[2]);

    alpha = v3_product(z, z);
    beta = v3_product(y, y);

    d = v3_product(t->rel_edges[2], y);
    if (is_zero(d)) beta = 0.0; else beta = - beta / d;

    d = v3_product(t->rel_edges[0], z);
    if (is_zero(d)) alpha = 0.0; else alpha = alpha / d;

#if 0
    fprintf(stderr, "alpha = %f, beta = %f\n", alpha, beta);
#endif

    if (point2) {
	v2_set(*point2, alpha, beta);
    }

    return TRUE;
}



void point3_in_basis_t3(vector3 *v3, vector2 v2, triangle3 basis)
{
    v3_let(*v3, basis->abs_edges[0]);
    v3_addscal(*v3, *v3, v2[0], basis->rel_edges[0])
    v3_subscal(*v3, *v3, v2[1], basis->rel_edges[2])
}

void triangle3_of_triangle2(vector3 *t3, vector2 t2[3], triangle3 basis)
{
    point3_in_basis_t3(&(t3[0]), t2[0], basis);
    point3_in_basis_t3(&(t3[1]), t2[1], basis);
    point3_in_basis_t3(&(t3[2]), t2[2], basis);
}



list triangle3_clip(list triangles, int xyz_select, double imin, double imax)
/****************************************************************************
 * FUNCTION
 *	Clips all triangles of the list in x or y or z-coordinates at
 *	the values imin and imax. A list of clipped triangles are returned.
 *	Free this list with triangle_list_free().
 *      In the worst case, for one triangle, 2 clipped triangles would be
 *	generated.
 *
 * PREREQUISITES
 *	imin < imax
 *      0 <= xyz_select <= 2  (x=0, y=1, z=2)
 *
 * RESULTS
 *	A list of clipped triangles.
 */
{
    triangle3	t;
    int		nr_points;
    int 	i;
    vector3	points[5];
    list	clipped_triangles;
    triangle3	ct;

    clipped_triangles = list_new();

#if 0
    if (imin>imax) {
	double temp;

	warning("triangle: triangle3_clip() - min > max, exchange them!");

	temp = imin;
	imin = imax;
	imax = temp;
    }
#endif

#if 0
    fprintf(stderr, "triangle3_clip()\n");
#endif

    for (t=list_first(triangles); t; t=list_next(triangles)) {

	nr_points = 0;

	/* clip lines (edge[i]->edge[(i+1)%3]) and add also
	 * the start end end-point of the line if it lies in
	 * [imin,imax].
	 * this generates a polygon with max 4 edges.
	 */
	for (i=0; i<3; i++) {

#if 0
	    fprintf(stderr, "A) i=%d, t=%p, imin=%f, imax=%f, nr_points=%d\n",
		    i, t, imin, imax, nr_points);

	    if (i<0 || i>=3)
		warning("triangle: triangle3_clip() - i out of range!");

	    if (t==0)
		warning("triangle: triangle3_clip() - t is zero!");
#endif

	    if (in_range(imin, t->abs_edges[i][xyz_select], imax)) {
		v3_let(points[nr_points], t->abs_edges[i]);
	        nr_points++;
	    }
#if 0
	    fprintf(stderr, "B) i=%d, t=%p, imin=%f, imax=%f, nr_points=%d\n\n",
		    i, t, imin, imax, nr_points);
#endif
	    nr_points+= clip_points(&points[nr_points],
				    t->abs_edges[i], t->rel_edges[i],
				    xyz_select, imin, imax);
#if 0
	    fprintf(stderr, "C) i=%d, t=%p, imin=%f, imax=%f, nr_points=%d\n",
		    i, t, imin, imax, nr_points);
#endif

#if 1
	    if (nr_points>5)
		error("triangle.c: triangle3_clip() - more than 4 points!");
#endif

	}

	/* generate clipped triangles
	 */
	if (nr_points>=3) {
	    ct = triangle3_new(points);
	    if (ct) list_add(clipped_triangles, ct);

	    /* polygon with 4 edges? divide it
	     * into 2 triangles
             */
	    if (nr_points>=4) {
		v3_let(points[1], points[3]);
		ct = triangle3_new(points);
	    	if (ct) list_add(clipped_triangles, ct);

		if (nr_points>=5) {
		    v3_let(points[2], points[4]);
		    ct = triangle3_new(points);
		    if (ct) list_add(clipped_triangles, ct);
		}
	    }
	}
    }

    return clipped_triangles;
}




boolean	triangle3_distance(double *factor, vector3 line_start,
                                  vector3 line_index, triangle3 t)
/****************************************************************************
 * FUNCTION
 *	Compute the line-index-factor of the piercing point between the
 *	triangle t and the infinite straight line (line_start, line_index)
 *
 * RESULT
 *      Returns TRUE if the line goes through the triangle-plain.
 *      Else return FALSE.
 *      In the var-variable factor, the factor for the piercing point
 *		line_start + factor*line_index
 *	will be returned.
 */
{
    vector3 v,		/* piercing point */
            y, z;
    double alpha, beta, d;

    /* length of line in direction of the triangle-normal
     */
    d = v3_product(t->normal, line_index);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "line-index-length in triangle-normal-direction: %f\n", d);
#endif

    if (is_zero(d)) return FALSE; /* line parallel to triangle */

    /* Starting point of the line relative to the 1.edge
     * of the triangle
     */
    v3_sub(v, line_start, t->abs_edges[0]);

    /* v3_product() = distance of the line-starting-point
     * to the triangle-plain
     */

    /* factor d*line_index + line_start lies in triangle plain
     */
    d = - v3_product(t->normal, v) / d;

#ifdef DEBUG_DETAIL
    fprintf(stderr, "line-index-factor from line-start to triangle-plain: %f\n", d);
#endif

    if (factor) *factor = d;

    /* point of piercing relative from 1.edge of the triangle
     */
    v3_addscal(v, v, d, line_index);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "piercing point is = (%f %f %f)\n",
	    v[0], v[1], v[2]);
#endif

    alpha = v3_product(t->rel_edges[0], v) / (t->rel_length[0] * t->rel_length[0]);
    beta = -v3_product(t->rel_edges[2], v) / (t->rel_length[2] * t->rel_length[2]);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "alpha' = %f, beta' = %f\n", alpha, beta);
#endif

    v3_subscal(y, v, alpha, t->rel_edges[0]);
    v3_addscal(z, v, beta, t->rel_edges[2]);

    alpha = v3_product(z, z);
    beta = v3_product(y, y);

    d = v3_product(t->rel_edges[2], y);
    if (is_zero(d)) beta = 0.0; else beta = - beta / d;

    d = v3_product(t->rel_edges[0], z);
    if (is_zero(d)) alpha = 0.0; else alpha = alpha / d;

#ifdef DEBUG_DETAIL
    fprintf(stderr, "alpha = %f, beta = %f\n", alpha, beta);
#endif

    /* piercing point lies in triangle ?
     */
    if ((alpha >= 0.0) && (beta >=0.0) && ((alpha+beta) <= 1.0)) {

	return TRUE;
    }
    else
	return FALSE;
}

boolean straight_line_trough_triangle(vector3 line_start, vector3 line_index,
                                      triangle3 t)
/****************************************************************************
 * FUNCTION
 *      Check if a line (line_start->+line_index) goes through a triangle.
 *	A line which lies in the triangle-plain or is parallel to the plain
 *	is not considered.
 *
 *
 * RESULT
 *      Returns TRUE if the line goes through the triangle-plain.
 *      Else return FALSE.
 */
{
    vector3 v, y;
    double alpha, beta, d;

    /* length of line in direction of the triangle-normal
     */
    d = v3_product(t->normal, line_index);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "line-index-length in triangle-normal-direction: %f\n", d);
#endif

    if (is_zero(d)) return FALSE; /* line parallel to triangle */

    /* Starting point of the line relative to the 1.edge
     * of the triangle
     */
    v3_sub(v, line_start, t->abs_edges[0]);

    /* v3_product() = distance of the line-starting-point
     * to the triangle-plain
     */

    /* factor d*line_index + line_start lies in triangle plain
     */
    d = - v3_product(t->normal, v) / d;

#ifdef DEBUG_DETAIL
    fprintf(stderr, "line-index-factor from line-start to triangle-plain: %f\n", d);
#endif

    /* line cut the triangle-plain?
     */
    if (d<0.0 || d >1.0) return FALSE;

    /* point of piercing relative from 1.edge of the triangle
     */
    v3_addscal(v, v, d, line_index);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "piercing point is = (%f %f %f)\n",
	    v[0], v[1], v[2]);
#endif

    alpha = v3_product(t->rel_edges[0], v) / (t->rel_length[0] * t->rel_length[0]);
    beta = -v3_product(t->rel_edges[2], v) / (t->rel_length[2] * t->rel_length[2]);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "alpha' = %f, beta' = %f\n", alpha, beta);
#endif

    v3_subscal(y, v, alpha, t->rel_edges[0]);
    v3_addscal(v, v, beta, t->rel_edges[2]);

    alpha = v3_product(v, v);
    beta = v3_product(y, y);

    beta = - beta / v3_product(t->rel_edges[2], y);

    alpha = alpha / v3_product(t->rel_edges[0], v);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "alpha = %f, beta = %f\n", alpha, beta);
#endif

    /* piercing point lies in triangle ?
     */
    if ((alpha >= 0.0) && (beta >=0.0) && ((alpha+beta) <= 1.0))
	return TRUE;
    else
	return FALSE;
}


boolean triangle3_piercing(triangle3 t1, triangle3 t2)
/****************************************************************************
 * FUNCTION
 *	Check if a triangle pierce another.
 *	Returns TRUE if the triangle t1 pierce the triangle t2 or
 *	FALSE if not.
 *      If t1=t2, FALSE is also returned.
 *
 */
{
    vector3 t;

    /* fast piercing test by the bounding_radius
     */
    v3_sub(t, t1->abs_edges[0], t2->abs_edges[0]);
    if (v3_sum(t) > t1->bounding_radius + t2->bounding_radius) return FALSE;

    /* test triangle edges against triangle
     */
    if (straight_line_trough_triangle(t1->abs_edges[0], t1->rel_edges[0], t2) ||
        straight_line_trough_triangle(t1->abs_edges[1], t1->rel_edges[1], t2) ||
        straight_line_trough_triangle(t1->abs_edges[2], t1->rel_edges[2], t2) ||

        straight_line_trough_triangle(t2->abs_edges[0], t2->rel_edges[0], t1) ||
        straight_line_trough_triangle(t2->abs_edges[1], t2->rel_edges[1], t1) ||
        straight_line_trough_triangle(t2->abs_edges[2], t2->rel_edges[2], t1)) {

        return TRUE;
    }

    return FALSE;
}




triangle3 triangle3_new(vector3 *edges)
/****************************************************************************
 * FUNCTION
 *      Generate a triangle with 3 edges.
 *	This routine computes all structure-elements out of the
 *	three edges.
 *
 * RESULT
 *	a pointer to a triangle3 structure.
 *      or 0 if the triangle is to small.
 */
{
    triangle3 t;
    double d;

    t = malloc(sizeof(adt_triangle3));
    if (t) {
	v3_let(t->abs_edges[0], edges[0]);
	v3_let(t->abs_edges[1], edges[1]);
	v3_let(t->abs_edges[2], edges[2]);

	v3_sub(t->rel_edges[0], edges[1], edges[0]);
	v3_sub(t->rel_edges[1], edges[2], edges[1]);
	v3_sub(t->rel_edges[2], edges[0], edges[2]);

	t->rel_length[0] = v3_sum(t->rel_edges[0]);
	t->rel_length[1] = v3_sum(t->rel_edges[1]);
	t->rel_length[2] = v3_sum(t->rel_edges[2]);
	t->bounding_radius = max(t->rel_length[0], t->rel_length[2]);

        v3_mul(t->normal, t->rel_edges[0], t->rel_edges[2]);
        d = -v3_sum(t->normal);

        if (is_zero(t->rel_length[0]) || is_zero(t->rel_length[2]) ||
	    is_zero(1.0 - v3_product( t->rel_edges[0], t->rel_edges[2])
		    / (t->rel_length[0] * t->rel_length[2])) ||
	    is_zero(d)) { 

#ifdef ZERO_TRIANGLE_WARNING
	    printf("triangle: triangle3_new() - triangle with zero extension\n");
	    printf("t-edges: {%f, %f, %f},\n  {%f, %f, %f},\n  {%f, %f, %f}\n",
		   v3_elems(t->abs_edges[0]),
		   v3_elems(t->abs_edges[1]),
		   v3_elems(t->abs_edges[2]));
	    
	    printf("t-size: %e\n", d);

	    warning("triangle: triangle3_new() - triangle with zero extension");
#endif
	    free(t);
	    return 0;
	}
        d = 1.0 / d;
        v3_scalar(t->normal, t->normal, d);

	t->connection_triangle[0] = 0;
	t->connection_triangle[1] = 0;
	t->connection_triangle[2] = 0;
	t->connection_side[0] = -1;
	t->connection_side[1] = -1;
	t->connection_side[2] = -1;
    }
    else error("triangle: triangle3_new() - can't alloc memory for adt triangle3");

    return t;
}

vector3*	triangle3_abs_edges(triangle3 t)
{
    return &(t->abs_edges[0]);
}

vector3*	triangle3_rel_edges(triangle3 t)
{
    return &(t->rel_edges[0]);
}

double*		triangle3_edges_length(triangle3 t)
{
    return &(t->rel_length[0]);
}

void	triangle3_normal(vector3 *normal, triangle3 t)
{
    v3_let(*normal, t->normal);
}


void triangle3_free(triangle3 t)
/****************************************************************************
 * FUNCTION
 *      deallocate the triangle structure.
 */
{
    free(t);
}

void triangle3_list_free(list tl)
/****************************************************************************
 * frees all triangles in a list-structure.
 */
{
    triangle3 t;

    for (t=list_first(tl); t; t=list_next(tl)) {
	free(t);
    }

    list_free(tl);
}


void triangle3_print(triangle3 t)
{
    printf("TRIANGLE3(%p) with edges:\n"
		    "{%f, %f, %f}, {%f, %f, %f}, {%f, %f, %f}\n", t,
            v3_elems(t->abs_edges[0]),
            v3_elems(t->abs_edges[1]),
            v3_elems(t->abs_edges[2]));
/* #ifdef DEBUG_DETAIL */
#if 1
    printf("rel_edges: (%1.2f %1.2f %1.2f) (%1.2f %1.2f %1.2f) (%1.2f %1.2f %1.2f)\n"
                    "rel_length: %1.2f %1.2f %1.2f\n"
                    "normal: (%1.2f %1.2f %1.2f), b-radius: %1.2f\n",
            t->rel_edges[0][0], t->rel_edges[0][1], t->rel_edges[0][2],
            t->rel_edges[1][0], t->rel_edges[1][1], t->rel_edges[1][2],
            t->rel_edges[2][0], t->rel_edges[2][1], t->rel_edges[2][2],
            t->rel_length[0], t->rel_length[1], t->rel_length[2],
            t->normal[0], t->normal[1], t->normal[2], t->bounding_radius);

#endif
}

void triangle3_print_list(list tl)
/****************************************************************************
 */
{
    triangle3 t;

    for (t=list_first(tl); t; t=list_next(tl))
	triangle3_print(t);
}


boolean	triangle3_equal(triangle3 t1, triangle3 t2)
/****************************************************************************
 */
{
   int i, j;
   int same[3] = {-1, -1, -1};

   for (i=0; i<3; i++)
	for (j=0; j<3; j++)
	    if (v3_equal(t1->abs_edges[i], t2->abs_edges[j]))
		same[i] = j;

   if (same[0]>=0 && same[1]>=0 && same[2]>=0 &&
       same[0]!=same[1] && same[0]!=same[2] && same[1]!=same[2])

	return TRUE;
   else
        return FALSE;
}




void	triangle3_set_connections(triangle3 t,
                                  triangle3 t0, int side0,
                                  triangle3 t1, int side1,
                                  triangle3 t2, int side2)
{
    if (t) {
            t->connection_triangle[0] = t0;
            t->connection_side[0] = side0;
            t->connection_triangle[1] = t1;
            t->connection_side[1] = side1;
            t->connection_triangle[2] = t2;
            t->connection_side[2] = side2;
    }
}

void		triangle3_connection(triangle3 *ct, int *cside, triangle3 t, int side)
{
    if (side<0 || side>2 || t==0)
        error("triangle3: triangle3_connection() - wrong side or zero-triangle");

    if (ct) *ct = t->connection_triangle[side];
    if (cside) *cside = t->connection_side[side];
}

