/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: triangle_projection.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 17:02:43 stolz>
 **************************************************************************
 *  MODULE              : triangle_projection
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in
 *      triangle_projection.h!
 *
 *  HISTORY
 *  17-Aug-94   Split projection-routines from triangle module.
 *  02-Sep-94   Module-Test passed. Hurra!!
 *  19-Sep-94   Change tp-triangle-assembly.
 *  21-Sep-94   Change rest-face-algorithm. (module test fine!)
 *  22-Sep-94   Forgotten to take only projection points with +-Index!
 *              DAMNED BUG!!!
 *  03-Oct-94   Add Edge-Reflexions on Connected Triangles
 *  13-Oct-94   Another BUG: forgett to make k++ before break;->lost
 *              a point of the cut-polygon if start-point!=end-point
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "triangle_projection.h"
#include "error.h"
#include "basic.h"
#include "vector_geometry.h"
#include "polygon.h"


/*
#define DEBUG_DETAIL   1
*/


#define EPS		1.0e-3
#define EPS2            1.1e-3


#define EDGE_REFLEXIONS 1
#define PLAIN_REFLEXIONS 1
#define REST_FACES 1

/*
#undef REST_FACES
#undef PLAIN_REFLEXIONS
#undef EDGE_REFLEXIONS
*/

boolean t3p_with_edge_reflexions = TRUE;


/**************************************************************************
 * abstract datatype definition
 */
typedef struct t3p_adt t3p_adt;
struct t3p_adt {
    triangle3	triangle;
    vector3	focus;
};


t3p t3p_new(vector3 *edges, vector3 focus)
/**************************************************************************
 */
{
    t3p p;
    triangle3 t;


    t = triangle3_new(edges);
    if (t==0) return 0;

    p = malloc(sizeof(t3p_adt));
    if (p==0)
	error("triangle_projection: t3p_new() - can't alloc memory for adt t3-projection");

    p->triangle = t;
    v3_let(p->focus, focus);

    return p;
}


void t3p_free(t3p p)
/**************************************************************************
 */
{
    if (p) {
	if (p->triangle) free(p->triangle);
	free(p);
    }
}


static list *tp_cut;
static list *tp_rest;
static vector3 t1_focus;
static vector3 t2_focus;
static t3p tp;

static void init_cut_t(vector3 *p)
{
#ifdef PLAIN_REFLEXIONS
    tp = t3p_new(p, t2_focus);
#ifdef DEBUG_DETAIL
    printf("init_cut_t() - cut-triangle {%f, %f, %f}, {%f,%f, %f}, {%f, %f, %1f} with t3p=%p\n",
	   v3_elems(p[0]), v3_elems(p[1]), v3_elems(p[2]), tp);
#endif
    if (tp) list_add(*tp_cut, tp);
#endif
}

static void init_rest_t(vector3 *p)
{
#ifdef REST_FACES
    tp = t3p_new(p, t1_focus);
#ifdef DEBUG_DETAIL
    printf("init_rest_t() - rest-triangle {%f, %f, %f}, {%f,%f, %f}, {%f, %f, %1f} with t3p=%p\n",
	   v3_elems(p[0]), v3_elems(p[1]), v3_elems(p[2]), tp);
#endif
    if (tp) list_add(*tp_rest, tp);
#endif
}





/* Triangle description with intermediate points.
 * Ready to use after <A>.
 */
static int     nr_points[2];    /* [triangle] */
static int     point_nr[2][9];  /* [triangle][point] point-number to below array */

    /* all triangle edges and intermediate points.
     * note that point3 is used for the projection-triangles and
     * the rest-triangles!
     * point is in normalized-coordinates of the t2-triangle.
     * (this means t2 = (0,0) (0,1), (1,0)).
     */
static int     points;
static vector2 point[18];	  /* 0-2 edges of t1, 3-5 edges of t2, 6>= cut-p */
static vector3 point3[18];   /* absolute position */

    /* edges of t2 in the normalized coordinates of t1.
     */
static vector2 t1_point[3];

    /* Index to the point_nr of the specific triangle.
     * This means: y = ipoint[t][x] and x = point_nr[t][y].
     * Ready to use after <B>.
     */
static int     ipoint[2][18];

    /* Point belong to both triangles? TRUE / FALSE.
     * Usable after <B>.
     */
static boolean common[18]; /* point belongs to t1 and t2 */
static boolean used[18];

    /* place to hold the polygon-definition (point_nrs) of a
     * polygon for further triangulation.
     */
static int     k;  /* nr of points in the polygon. */
static int     polygon[15];

    /* index-factor of a piercing point for line<->line.
     */
static double  pindex[2];
static double  index[3][2][2];     /* idx [line-nr][triangle][which-ordered] */

    /* nr of the piercing point (index in point[])
     */
static int     ppoint[3][2][2]; /* [line-nr][triangle][piercing-point] */

    /* nr of the starting-point of the other triangle-line
     */
static int     tpoint[3][2][2]; /* [line-nr][triangle][piercing-point] */
static double  tindex[3][2][2]; /* idx on other triangle-line [line-nr][triangle] */

    /* number of piercing points of the triangle line [line-nr]->[line-nr+1].
     */
static int    cuts[2][3];       /* [triangle][line-nr] */






static void dump_tp(t3p t1, triangle3 t2)
{
#ifdef DEBUG_TP
    int i, j, l;

    printf("dump_tp() - fatal error in algoritm, dump current state.\n");
    printf("dump_tp() - START\n");

    printf("dump_tp() - projection triangle t1 and obstacle triangle t2:\n");
    printf("T1=\n");
    t3p_print(t1);
    printf("T2=\n");
    triangle3_print(t2);

    printf("\ndump_tp() - piercing points (ref from t2)\n");
    for (l=0; l<2; l++) {
	printf("triangle %d has:\n", l);
 	for (i=0; i<3; i++) {
	    printf(" %d piercing points on line %d:\n", cuts[l][i], i);
	    for (j=0; j<cuts[l][i]; j++) {
		printf("  pierce line %d(tidx=%f):"
		       " with index %f at the point[%d]= %f %f\n",
		       tpoint[i][l][j], tindex[i][l][j],
		       index[i][l][j], ppoint[i][l][j],
		       v2_elems(point[ppoint[i][l][j]]));
	    }
        }
    }

    printf("dump_tp() - list of intermediate points and edges\n");
    for (i=0; i<points; i++)
	printf(" point[%d] = %f %f%s", i, v2_elems(point[i]),
		(common[i] ? " (common)\n" : "\n"));

    printf("dump_tp() - triangle-description with intermediate points\n");
    for (l=0; l<2; l++) {
	printf(" triangle %d : ", l);
 	for (i=0; i<nr_points[l]; i++)
	    printf(" %d", point_nr[l][i]);
	printf("\n");
    }

    printf("dump_tp() - current contents of the polygon array.\n");
    printf("polygon =");
    for (i=0; i<k; i++)
	printf(" %d", polygon[i]);
    printf("\n");
    printf("dump_tp() - END\n");
#endif
}




static void make_edge_reflexion(list t3p_edges,
	triangle3 t, int side, vector3 focus, triangle3 ct, int cside,
	vector3 line_start, vector3 line_stop)
{
    vector3 efocus, line_index;
    vector3 t_normal, ct_normal;
    vector3 *t_edges, *ct_edges;
    vector3 plain_dir, edge_dir;
    vector3 edges[3];
    vector3 new_start, new_stop;
    vector3 v;
    t3p tp;
    double h;
    int center;

#ifdef DEBUG_DETAIL
    printf("make_egde_reflexion() - t=%p, side=%d, ct=%p, cside=%d\n",
           t, side, ct, cside);
    printf("make_edge_reflexion() - line_start = {%f, %f, %f}\n", v3_elems(line_start));
    printf("make_edge_reflexion() - line_stop = {%f, %f, %f}\n", v3_elems(line_stop));
#endif

    /* take middle of line as focus point.
     */
    v3_sub(v, line_start, line_stop);
    if (is_zero(v3_sum(v))) {
#ifdef DEBUG_DETAIL
	printf("make_egde_reflexion: line-start=stop!\n");
#endif
	return ;
    }

    v3_add(efocus, line_start, line_stop);
    v3_scalar(efocus, efocus, 0.5);

#ifdef DEBUG_DETAIL
    printf("make_edge_reflexion:  efocus = {%f, %f, %f}\n", v3_elems(efocus));
#endif

    /* compute vector which lies in the plane of t, is perpendiculat to the line
     * and direct into the the triangle.
     */
    v3_sub(line_index, line_stop, line_start);
    t_edges = triangle3_abs_edges(t);
    center = (side+2)%3;
    perpendicular_from_line3_to_point3(&h, &plain_dir, 0, t_edges[center],
					line_start, line_index);

#ifdef DEBUG_DETAIL
    printf("make_edge_reflexion:  plain_dir = {%f, %f, %f}\n", v3_elems(plain_dir));
#endif

    triangle3_normal(&t_normal, t);
    perpendicular_on_plain3(&h, 0, t_edges[0], t_normal, focus);
    if (h<0) v3_neg(t_normal, t_normal);

    ct_edges = triangle3_abs_edges(ct);
    triangle3_normal(&ct_normal, ct);
    perpendicular_on_plain3(&h, 0, ct_edges[0], ct_normal, focus);
    if (h<0) v3_neg(ct_normal, ct_normal);

#ifdef DEBUG_DETAIL
    printf("make_edge_reflexion:  t_normal = {%f, %f, %f}\n", v3_elems(t_normal));
    printf("make_edge_reflexion:  ct_normal = {%f, %f, %f}\n", v3_elems(ct_normal));
#endif

    v3_add(edge_dir, t_normal, ct_normal);
    if (v3_product(edge_dir, plain_dir)<0.0) {
	/* angle between edge-dir and plain-dir is greater than 90degree
	 * this is a outside-edge
	 */
	v3_sub(plain_dir, t_normal, ct_normal);
    }
    else {
	/* a inside edge
	 */
    }
    v3_scalar(plain_dir, plain_dir, 0.01);
    v3_scalar(edge_dir, edge_dir, 0.01);


#ifdef DEBUG_DETAIL
    printf("make_edge_reflexion:  plain_dir = {%f, %f, %f}\n", v3_elems(plain_dir));
    printf("make_edge_reflexion:  egde_dir = {%f, %f, %f}\n", v3_elems(edge_dir));
#endif

    h = (1.0 / v3_sum(v)) * 0.0025;
    v3_scalar(v, v, h);
    v3_add(new_start, efocus, v);
    v3_sub(new_stop, efocus, v);
    
    /* build reflexion triangles
     */
    v3_add(edges[0], new_start, edge_dir);
    v3_add(edges[1], new_stop, edge_dir);
    
    v3_add(edges[2], new_start, plain_dir);
    tp = t3p_new(edges, efocus);
    if (tp) list_add(t3p_edges, tp);
    else {
	printf("make_edge_reflexion: can't create reflexion-projection!!\n");
	printf("make_edge_reflexion: for focus: {%f, %f, %f}\n",
	       v3_elems(efocus));
	printf("make_edge_reflexion: and edges\n"
		   " {%f, %f, %f}, {%f, %f, %f}, {%f, %f, %f}\n",
	       v3_elems(edges[0]), v3_elems(edges[1]), v3_elems(edges[2]));
	warning("make_edge_reflexion: can't create reflexion-projection");
    }
    
    v3_add(edges[0], new_stop, plain_dir);
    tp = t3p_new(edges, efocus);
    if (tp) list_add(t3p_edges, tp);
    else {
	printf("make_edge_reflexion: can't create reflexion-projection!!\n");
	printf("make_edge_reflexion: for focus: {%f, %f, %f}\n",
	       v3_elems(efocus));
	printf("make_edge_reflexion: and edges\n"
	       " {%f, %f, %f}, {%f, %f, %f}, {%f, %f, %f}\n",
	       v3_elems(edges[0]), v3_elems(edges[1]), v3_elems(edges[2]));
	warning("make_edge_reflexion: can't create reflexion-projection");
    }
}




static void edge_reflexions(list t3p_edges, triangle3 t, vector3 focus)
{
    triangle3 ct;
    int i, start, stop, cside;

    for (i=0; i<3; i++) {
	start = i+3;
        stop = (i+1)%3 + 3;

#ifdef DEBUG_DETAIL
	printf("edge_reflexion: test side %d, common[%d]=%s -> common[%d]=%s, cuts=%d\n",
	       i, start, (common[start] ? "true" : "false"),
	       stop, (common[stop] ? "true" : "false"), cuts[1][i]);
#endif
	if (common[start]) {
	    if (common[stop]) ;
            else if (cuts[1][i]>0) stop = ppoint[i][1][cuts[1][i]-1];
            else continue;
	}
	else if (common[stop]) {
	    if (cuts[1][i]>0) start = ppoint[i][1][0];
            else continue;
	}
	else if (cuts[1][i]==2) {
            start = ppoint[i][1][0];
            stop = ppoint[i][1][1];
	} else continue;

#ifdef DEBUG_DETAIL
	printf("edge_reflexion: build edge reflexion from point %d to %d\n",
		start, stop);
#endif
	triangle3_connection(&ct, &cside, t, i);
	if (ct) {
	    if (start != stop)
		make_edge_reflexion(t3p_edges, t, i, focus, ct, cside,
				    point3[start], point3[stop]);
	    else {
#ifdef DEBUG_DETAIL
		printf("edge_reflexion: start-point == stop-point (=%d)!\n",
		       start);
#endif
		warning("edge_reflexion: line-start == line-stop");
	    }
	}
    }
}







void triangle3_projection(list t3p_cut, list t3p_rest,
			 t3p t1, triangle3 t2)
/**************************************************************************
 */
{
    /* temporar variables
     */
    int i, j, l, m, n, o, mf, mb, d;
    vector2 line1_index, line2_index;
    vector3 v3;
    double  h, h2;
    vector3 *t1_abs_edges;
    vector3 t1_normal;
    vector3 t2_normal;
    vector3 *t2_abs_edges;
    vector3 *t2_rel_edges;

   /********************************************************************
    */

    k=0;
    points=0;
    nr_points[0]=0;
    nr_points[1]=0;

#ifdef DEBUG_DETAIL
    printf("triangle3_projection() - project triangle t1(%p) on t2(%p) and\n"
	   "triangle3_projection() - determine rest and common face\n",
	   t1, t2);
    printf("t3p: focus of t1: {%f %f %f}\n", v3_elems(t1->focus));
#endif

    t1_abs_edges = triangle3_abs_edges(t1->triangle);
    triangle3_normal(&t1_normal, t1->triangle);
    t2_abs_edges = triangle3_abs_edges(t2);
    t2_rel_edges = triangle3_rel_edges(t2);
    triangle3_normal(&t2_normal, t2);



    /********************************************************************
     * project the edges of triangle 1 into the plain of triangle 2
     * and initialize the point[] and points with the edges of
     * t1 and t2.
     * Also initialize the number of piercing points in cuts[][] to 0.
     */
    for (i=0; i<3; i++) {

        v3_sub(v3, t1_abs_edges[i], t1->focus);

	if (!line3_in_t3_plane(&h, &point[i], &point3[i], t1_abs_edges[i], v3, t2)
	    || h <= 0.0)
	    return;

	cuts[0][i] = 0;
	cuts[1][i] = 0;

#ifdef DEBUG_DETAIL
    printf("t3p: projection of edge %d with index %f on t2 has coord {%f, %f}\n"
	   " or absolute {%f, %f, %f}\n",
	   i, h, v2_elems(point[i]), v3_elems(point3[i]));
#endif
	v3_let(point3[i+3], t2_abs_edges[i]);
    }
    v2_set(point[3], 0.0, 0.0);
    v2_set(point[4], 1.0, 0.0);
    v2_set(point[5], 0.0, 1.0);

    points = 6;


#ifdef DEBUG_DETAIL
    printf("t3p: 1. determine piercing points between t1&t2\n");
#endif

    /********************************************************************
     * Compute the piercing points of the triangles t1<->t2.
     * and store the result in cuts[][], index[][], ppoint[][][].
     * ppoint[][i][] is sorted from lowest to highest index.
     * so the line i could be devided into
     * i->ppoint[][i][0]->ppoint[][i][1]-> i+1 % ...
     * if cuts[][i] is greater than 0 or 1!
     * Also determine if the edges belongs to both triangles t1
     * and t2 and set common[] correspondingly.
     */
    for (i=0; i<3; i++) {

        /* edge i of t1 lies in t2 ?
         */
        common[i] = point2_in_triangle_eps(point[i], 0.0);

	/* edge i of t2 lies in t1 ?
	 */
	point2_in_basis_oab(&(t1_point[i]), point[i+3], point[0], point[1], point[2]);
	common[i+3] = point2_in_triangle_eps(t1_point[i], 0.0);

	for (j=0; j<3; j++) {
	    v2_sub(line1_index, point[(i+1)%3], point[i]);
	    v2_sub(line2_index, point[3+((j+1)%3)], point[3+j]);
	    if (line2_pierce_line2(&point[points], &pindex[0], &pindex[1],
				   point[i], line1_index,
				   point[j+3], line2_index) &&
		line2_at_eps(pindex[0], EPS) &&
		line2_at_eps(pindex[1], EPS)) {


		/* line i of triangle1 pierce line j of triangle2
		 * at point[points] with index pindex[0/1]
		 */

#ifdef DEBUG_DETAIL
		printf("t3p: line %d of t1 pierce with line %d of t2 at %f %f (t2-rel)\n",
				i, j, v2_elems(point[points]));

#endif

		for (l=0; l<2; l++) {

		    if (l==0) {
			k = i;
			d = j;
		    }
		    else {
			k = j;
			d = i;
                    }


		    /* sort the point in with respect of the index
                     */

		    switch(cuts[l][k]) {
		    case 0:
		        index[k][l][0]  = pindex[l];
		        ppoint[k][l][0] = points;
                        tindex[k][l][0] = pindex[1-l];
                        tpoint[k][l][0] = d;
			cuts[l][k] += 1;
			break;

		    case 1:
			if (pindex[l] < index[k][l][0]) {
			    index[k][l][1]  = index[k][l][0];
			    ppoint[k][l][1] = ppoint[k][l][0];
			    tpoint[k][l][1] = tpoint[k][l][0];
			    tindex[k][l][1] = tindex[k][l][0];

			    index[k][l][0]  = pindex[l];
			    ppoint[k][l][0] = points;
			    tpoint[k][l][0] = d;
			    tindex[k][l][0] = pindex[1-l];
			}
			else {
			    index[k][l][1]  = pindex[l];
			    ppoint[k][l][1] = points;
			    tpoint[k][l][1] = d;
			    tindex[k][l][1] = pindex[1-l];
			}
			cuts[l][k] += 1;
			break;

		    case 2:
			if (pindex[l] < index[k][l][0]) {
			    index[k][l][0]  = pindex[l];
			    ppoint[k][l][0] = points;
			    tpoint[k][l][0] = d;
			    tindex[k][l][0] = pindex[1-l];
			}
			else if (pindex[l] > index[k][l][1]) {
			    index[k][l][1]  = pindex[l];
			    ppoint[k][l][1] = points;
			    tpoint[k][l][1] = d;
			    tindex[k][l][1] = pindex[1-l];
			}
			break;

		    default:
			dump_tp(t1, t2);
			error("triangle_projection: triangle3_projection() "
			      "- more than two intermediate-points\n"
			      "  while piercing the edges of two triangles");
			break;
		    }
		}
                common[points] = TRUE;
		points++;
	    }
	}
    }

#ifdef DEBUG_DETAIL
    printf("t3p: piercing points (ref from t2)\n");
    for (l=0; l<2; l++) {
	printf("t3p: triangle %d has:\n", l);
 	for (i=0; i<3; i++) {
	    printf("t3p: %d piercing points on line %d:\n", cuts[l][i], i);
	    for (j=0; j<cuts[l][i]; j++) {
		printf("t3p: pierce line %d(tidx=%f):"
			" with index %f at the point[%d]= %f %f\n",
			tpoint[i][l][j], tindex[i][l][j],
			index[i][l][j], ppoint[i][l][j],
			v2_elems(point[ppoint[i][l][j]]));
	    }
        }
    }
#endif

    /********************************************************************
     * sort the piercing points in the edges of the triangles
     * and store the result in point_nr[][] and nr_points[] for
     * each triangle.
     */

#ifdef DEBUG_DETAIL
    printf("t3p: 2. describe triangle incl. piercing points\n");
#endif

    for (i=0; i<18; i++) {
	ipoint[0][i] = -1;
	ipoint[1][i] = -1;
	used[i] = FALSE;
    }

    for (l=0; l<2; l++) {
        k = 0;
	nr_points[l] = 0;

	for (i=0; i<3; i++) {

	    j = i + l*3;
	    m = (i+1)%3 + l*3;


	    point_nr[l][k] = j;
	    ipoint[l][j] = k++;

	    for (o=0; o<cuts[l][i]; o++) {
                if (fabs(index[i][l][o])<=EPS2) n = 0;
		else if (fabs(1.0-index[i][l][o])<=EPS2) n = 1;
		else n = 2;

		n <<=4;

                if (fabs(tindex[i][l][o])<=EPS2) ;
		else if (fabs(1.0-tindex[i][l][o])<=EPS2) n |= 1;
		else n |= 2;

#ifdef DEBUG_DETAIL
		printf("t3p: triangle:%d, line %d, cut %d, ip-code: 0x%x\n", l, i, o, n);
#endif

                switch(n) {
		case 0x20:
		    d = (tpoint[i][l][o])%3 + (1-l)*3;
		    point_nr[l][k] = d;
		    ipoint[l][d] = k++;
                    common[d] = TRUE;
#ifdef DEBUG_DETAIL
		    printf("t3p: 0x%x: set common[%d] = TRUE\n", n, d);
#endif
                    break;
		case 0x21:
		    d = (tpoint[i][l][o]+1)%3 + (1-l)*3;
                    common[d] = TRUE;
#ifdef DEBUG_DETAIL
		    printf("t3p: 0x%x: set common[%d] = TRUE\n", n, d);
#endif
                    break;
		case 0x22:
		    d = ppoint[i][l][o];
		    point_nr[l][k] = d;
		    ipoint[l][d] = k++;
                    break;
		case 0x00:
		case 0x01:
                    common[j] = TRUE;
		    d = (tpoint[i][l][o] + (n&1))%3 + (1-l)*3;
                    common[d] = TRUE;
#ifdef DEBUG_DETAIL
		    printf("t3p: 0x%x: set common[%d] = TRUE\n", n, j);
		    printf("t3p: 0x%x: set common[%d] = TRUE\n", n, d);
#endif
                    if (l==0) break;
                    ipoint[l][j] = -1;
                    k--;
		    point_nr[l][k] = d;
		    ipoint[l][d] = k++;
                    break;
		case 0x02:
                    common[j] = TRUE;
#ifdef DEBUG_DETAIL
		    printf("t3p: 0x%x: set common[%d] = TRUE\n", n, j);
#endif
		    break;
		case 0x10:
		case 0x11:
                    common[m] = TRUE;
		    d = (tpoint[i][l][o] + (n&1))%3 + (1-l)*3;
		    common[d] = TRUE;
#ifdef DEBUG_DETAIL
		    printf("t3p: 0x%x: set common[%d] = TRUE\n", n, m);
		    printf("t3p: 0x%x: set common[%d] = TRUE\n", n, d);
#endif
		    break;
		case 0x12:
                    common[m] = TRUE;
#ifdef DEBUG_DETAIL
		    printf("t3p: 0x%x: set common[%d] = TRUE\n", n, m);
#endif
		    break;
                default:
		    dump_tp(t1, t2);
                    error("triangle_projection: triangle3_projection() - "
                          "illegal intermediate point code");
                    break;
	        }
	    }
	}
	nr_points[l] = k;
    }


   /********************************************************************
    * <B>
    */


#ifdef DEBUG_DETAIL
    printf("t3p: point-list\n");
    for (i=0; i<points; i++)
	printf("t3p: point[%d] = %f %f%s", i, v2_elems(point[i]),
		(common[i] ? " (common)\n" : "\n"));

    for (l=0; l<2; l++) {
	printf("t3p: triangle %d : ", l);
 	for (i=0; i<nr_points[l]; i++)
	    printf(" %d", point_nr[l][i]);
	printf("\n");
    }
#endif

   /********************************************************************
    * <A>
    *
    */


    /********************************************************************
     * triangle t2 lies totally in t1.
     */
    if (common[3] &&  common[4] && common[5]) {

        int nearest[3];	/* t2-point from point i of t1 */

#ifdef DEBUG_DETAIL
	printf("t3p: t2 lies totally in t1, so t2 is the common face.\n");
#endif

	/* calculate the focus points
	 */
	tp_cut = &t3p_cut;
	tp_rest= &t3p_rest;
	v3_let(t1_focus, t1->focus);
	perpendicular_on_plain3(&h, &v3, t2_abs_edges[0], t2_normal, t1_focus);

	v3_dec(v3, t1_focus);
	v3_inc(v3, v3);
	v3_add(t2_focus, v3, t1_focus);

	/* cut-triangle is the same as t2.
	 */
	init_cut_t(t2_abs_edges);

	/* reflexion on the edges of t2
         */
#ifdef EDGE_REFLEXIONS
	if (t3p_with_edge_reflexions)
	    edge_reflexions(t3p_cut, t2, t1->focus);
#endif
	/* absolute coord. of the points in the t2-plane.
	 */
	for (i=3; i<points; i++) {
	    if (i>=3 && i<6) {
	    v3_let(point3[i], t2_abs_edges[i-3]);
	}
	    else {

		v3_let(v3, t2_abs_edges[0]);
		v3_addscal(v3, v3, point[i][0], t2_rel_edges[0]);
		v3_subscal(point3[i], v3, point[i][1], t2_rel_edges[2]);
	    }
	}


	/* retransform the point[s] into the t1-plane
	 */
	for (i=0; i<3; i++) {
	    v3_let(point3[i], t1_abs_edges[i]);
	}

	for (i=3; i<points; i++) {
	    v3_sub(v3, t1->focus, point3[i]);
	    (void)line3_in_t3_plane(0, 0, &point3[i],
				    point3[i], v3,
				    t1->triangle);
	}

	/* Triangle t2 lies within t1.
         * Divide the rest-area into two pieces for futher
	 * trianglulation.
	 */

	/* find nearest t2-edge from a t2-edge */
	for (i=0; i<3; i++) {
            h = v2_distance_quadrat(point[i], point[3]);
	    nearest[i] = 3;
	    for (j=4; j<6; j++) {
		h2 = v2_distance_quadrat(point[j], point[i]);
		if (h2<h) {
		    nearest[i] = j;
                    h = h2;
		}
	    }
	}

	for (i=0; i<3; i++) {
	    polygon[0] = i;
            polygon[1] = (i+1)%3;
            polygon[2] = nearest[polygon[1]];
            polygon[3] = nearest[polygon[0]];
	    k = 4;

#ifdef DEBUG_DETAIL
            printf("t3p: rest-polygon (simply t1-t2):");
            for (j=0; j<k; j++)
	        printf(" %d", polygon[j]);
            printf("\n");
#endif
	    convex_ipolygon3_triangulation(point3, 4, polygon, init_rest_t);
	}
	return ;
    }


#ifdef DEBUG_DETAIL
	printf("t3p: determine polygon-area of t1&t2\n");
#endif
    /********************************************************************
     * determine polygone-aera which lies in t1 and t2.
     * search first point which lies in t1 and t2.
     */
    /* start with point on triangle t1 */
    for (l=0; l<2; l++) {
        for (i=0; i<nr_points[l]; i++) {
	  if (common[point_nr[l][i]]) goto pfound;
	}
    }

    /* no aera found!
     */
    return;

pfound:
    d = +1; /* seek-direction */

    k = 0;      /* number of points of the polygon */
    do {
	polygon[k] = point_nr[l][i];
	used[polygon[k]] = TRUE;

#ifdef DEBUG_DETAIL
	printf("t3p: both: triangle=%d, index=%d, direction=%d, point=%d\n",
			l, i, d, polygon[k]);
#endif


	/* change triangle if the next point is not common and the
	 * current point exists on the other triangle.
	 */
	if (!common[point_nr[l][(i+d+nr_points[l])%nr_points[l]]] &&
	    ipoint[1-l][polygon[k]]>=0) {

	    j = i;
	    l = 1 - l;		        /* change triangle */
	    i = ipoint[l][polygon[k]];

	    /* select direction
             */
            mf = point_nr[l][(i+d+nr_points[l]) % nr_points[l]];
            mb = point_nr[l][(i-d+nr_points[l]) % nr_points[l]];

	    if (common[mb] && !used[mb]) d = -d;
	}
	i = (i+d+nr_points[l]) % nr_points[l];

	if (!common[point_nr[l][i]]) {
	    k++;
	    break;
	}

	k++;
	if (k==15) {
	    dump_tp(t1, t2);
	    error("triangle_projection: triangle3_projection()"
		  " - cut-polygon-size exceeded");
	}

    } while (!used[point_nr[l][i]]); /*while (polygon[0] != point_nr[l][i]);*/

#ifdef DEBUG_DETAIL
    if (polygon[0] != point_nr[l][i])
	warning("triangle_projection: triangle3_projection()"
		  " - cut-polygon don't end with 1.point");
#endif

    if (k<3) return;    /* polygon has no extension */


    /* absolute coord. of the points in the t2-plane.
     */
    for (i=0; i<points; i++) {
	if (i>=3 && i<6) {
	    v3_let(point3[i], t2_abs_edges[i-3]);
	}
	else {

	    v3_let(v3, t2_abs_edges[0]);
            v3_addscal(v3, v3, point[i][0], t2_rel_edges[0]);
            v3_subscal(point3[i], v3, point[i][1], t2_rel_edges[2]);
	}
    }

#ifdef EDGE_REFLEXIONS
    if (t3p_with_edge_reflexions)
	edge_reflexions(t3p_cut, t2, t1->focus);
#endif

#ifdef DEBUG_DETAIL
    printf("t3p: cut-polygon:");
    for (i=0; i<k; i++)
	printf(" %d", polygon[i]);
    printf("\n");
#endif

    tp_cut = &t3p_cut;
    tp_rest= &t3p_rest;
    v3_let(t1_focus, t1->focus);
    perpendicular_on_plain3(&h, &v3, t2_abs_edges[0], t2_normal, t1_focus);

    v3_dec(v3, t1_focus);
    v3_inc(v3, v3);
    v3_add(t2_focus, v3, t1_focus);
#ifdef DEBUG_DETAIL
    printf("t3p: t2-focus is: %f %f %f\n", v3_elems(t2_focus));
#endif


    /* triangulation of the cut-polygon.
     */
    convex_ipolygon3_triangulation(point3, k, polygon, init_cut_t);


    /********************************************************************
     * retransform the point[s] into the t1-plane
     * point3[] must hold the points in the t2-plane!
     */
#ifdef DEBUG_DETAIL
    printf("t3p: absolute points in the plain of t2\n");
    for (i=0; i<points; i++)
	printf("t3p: point3[%d] = {%f, %f, %f}\n",
	       i, v3_elems(point3[i]));
#endif

    for (i=0; i<3; i++) {
        v3_let(point3[i], t1_abs_edges[i]);
    }

    for (i=0; i<points; i++) {
	v3_sub(v3, t1->focus, point3[i]);
	(void)line3_in_t3_plane(0, 0, &point3[i],
				point3[i], v3,
				t1->triangle);
    }

#ifdef DEBUG_DETAIL
    printf("t3p: absolute points in the plain of t1\n");
    for (i=0; i<points; i++)
	printf("t3p: point3[%d] = {%f, %f, %f}\n",
	       i, v3_elems(point3[i]));
#endif

    /* normal of the rest-polygon-faces
     */
    v3_neg(v3, t1_normal);

#ifdef DEBUG_DETAIL
	printf("t3p: determine the rest face (t1-t2)\n");
#endif

    while(1) {

        /* search for first point of t1 which lies not in t2
	 * (a point on t1 which is not common)
	 */
	for (i=0; i<nr_points[0]; i++)
	    if (!common[point_nr[0][i]]) break;
	
	if (i==nr_points[0]) break;  /* no uncommon point found. */

	common[point_nr[0][i]] = TRUE; /* mark this point as common */

#ifdef DEBUG_DETAIL
	printf("t3p: found !common point %d on t1.\n", point_nr[0][i]);
#endif
	/* Find from i backwards the first common point
	 * and mark the uncommon points as common
	 */
	for (j=(i-1+nr_points[0])%nr_points[0];
	     ipoint[1][point_nr[0][j]]<0;
	     j=(j-1+nr_points[0])%nr_points[0]) {
	    common[point_nr[0][j]] = TRUE;
	}

	/* Find from i on forwards the first common point
	 * and mark the uncommon points as common
	 */
	for (k=(i+1)%nr_points[0];
	     ipoint[1][point_nr[0][k]]<0;
	     k=(k+1)%nr_points[0]) {
	    common[point_nr[0][k]] = TRUE; 
	}
	
#ifdef DEBUG_DETAIL
	printf("t3p: uncommon section on t1 goes from %d to %d (index++).\n", 
	       point_nr[0][j], point_nr[0][k]);
#endif
		
	/* Now find the direction to go on triangle2 to connection 
	 * between i and j without a interection with a common
	 * point to and only with common points
	 * common points 
	 * triangle.
	 */
	l = ipoint[1][point_nr[0][j]];
	m = ipoint[1][point_nr[0][k]];

	d = +1;
	n = m;
	do {
	    n = (n + d + nr_points[1])%nr_points[1];
	    o = point_nr[1][n];
	    if (n != l && (!common[o] || ipoint[0][o]>=0)) {
		d = -1;
		break;
	    }
	} while (n != l);

#ifdef DEBUG_DETAIL
	printf("t3p: index-direction from point %d on t2 is %d.\n",
	       point_nr[1][m], d);
#endif

	/* the rest polygons goes now from:
	 * t1: j ->+1 i ->+1  k   and on t2: (m+d) ->d (l-d)
	 * no build it!
	 */

	o = 0;
	for (n=i;n!=k; n=(n+1)%nr_points[0])
	    polygon[o++] = point_nr[0][n];

	for (n=m;n!=l; n=(n+d+nr_points[1])%nr_points[1])
	    polygon[o++] = point_nr[1][n];

	for (n=j;n!=i; n=(n+1)%nr_points[0])
	    polygon[o++] = point_nr[0][n];

	k = o;

#ifdef DEBUG_DETAIL
        printf("t3p: rest-polygon:");
        for (i=0; i<k; i++)
	    printf(" %d", polygon[i]);
        printf("\n");
#endif

	if (ipolygon3_triangulation(point3, k, polygon, v3, init_rest_t)<0) {

	    dump_tp(t1, t2);
/*
	    warning("triangle_projection: triangle3_projection()"
		  " - incomplete triangulation");
*/
	}
    }
}



triangle3	t3p_triangle3(t3p tp)
/**************************************************************************
 */
{
    return tp->triangle;
}



vector3*	t3p_focus(t3p tp)
/**************************************************************************
 */
{
    return &(tp->focus);
}



boolean t3p_equal(t3p a, t3p b)
/**************************************************************************
 */
{
    if (v3_equal(a->focus, b->focus) &&
	triangle3_equal(a->triangle, b->triangle))
	return TRUE;
    else
	return FALSE;
}


void t3p_print(t3p tp)
/**************************************************************************
 */
{
    vector3 *edges;
    vector3 normal;
    
    if (tp==0) {
	printf("T3P = NIL\n");
	return;
    }

    edges = triangle3_abs_edges(tp->triangle);

    printf("T3P(%p) {\n focus = {%f, %f, %f}\n"
	   " edges = {%f, %f, %f},{%f, %f, %f},{%f, %f, %f}\n",
	   tp, v3_elems(tp->focus),
	   v3_elems(edges[0]), v3_elems(edges[1]), v3_elems(edges[2]));
    triangle3_normal(&normal, tp->triangle);
    printf(" normal = {%f, %f, %f}\n}\n", 
	       v3_elems(normal));
}
