/**************************************************************************
 *  PROJECT   : DA mobile robot simulation
 *  AUTHOR(S) : Horst Stolz (HUS)
 *  FILE      : us_sensor.c
 *  TYPE      : module implementation
 *  Time-stamp: <94/12/06 18:37:43 stolz>
 **************************************************************************
 *  MODULE    : us_sensor
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in us_sensor.h.
 *
 *  HISTORY
 *  23-Mar-94   Creation date
 *  08-Oct-94   change log(I/I0) computation 
 *  15-Nov-94   Remove sensor-activation of the sensor-backside.
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "error.h"
#include "us_sensor.h"
#include "mob_world_private.h"
#include "queue.h"
#include "vector_geometry.h"

/*
#define DEBUG_DETAIL 1
*/


us_sensor_check_point *uscp_new(void *world)
{
    us_sensor_check_point *sr = (us_sensor_check_point *)malloc(sizeof(us_sensor_check_point));

    if (sr==0) error("us_sensor: uscp_new() - can't alloc memory for sensor");
    sr->path = 0;
    sr->world = world;

    return sr;
}

void uscp_free(us_sensor_check_point *uscp)
{
#if 0
    if (uscp) {
	if (uscp->path) queue_free(uscp->path);
	free(uscp);
    }
#endif
}

void uscp_select(us_sensor_check_point *uscp)
{
    if (uscp) mob_world_store_us_result(uscp->world, uscp);
}




static void activate_us_sensor(mw_object sensor, double tof, us_sensor_check_point *uscp)
{
    mw_object controller;
    

    if (sensor && sensor->function == MWF_SENSOR &&
	(controller=mw_object_controller(sensor))) {

	switch(controller->function) {
	case MWF_MOB:
#ifdef NO_MOB_ACTIVATION
	    uscp_select(uscp);
#else
	    mob_us_sensor_activation(controller->func.mob.mob,
				     sensor->func.sensor.sensor_nr, tof, uscp);
#endif
	    break;
	default:
	    break;
	}
    }
    else error("mwo_mob: activate_us_sensor() - argument 1 is not a us-sensor!");
}


/* the triangle t, which belongs to the sensor. was activated from the
 * projection tp;
 * now follow the path to the sending-sensor and determine the flying-time of the signal.
 */
static void activate_us_sensor_from_tp(mw_object sensor, triangle_projection tp,
				triangle t)
{
    us_sensor_check_point *sr;
    vector3 *edges = triangle_edges(t);
    vector3 center, new_center, index, *focus;
    vector3 normal;
    double  distance = 0.0;
    double  intensity_factor = 1.0;	/* Isending / Irecv. */
    double  index_length, cos_angle;
    double  r0 = 0.288; /* distance to focus */
    mw_object obj;

#ifdef DEBUG_DETAIL
    printf("activate_us_sensor_from_tp()\n");
#endif
    sr = uscp_new(sensor->world);

    /* Initialize the path, and the sensor-direction.
     */
    sr->path = queue_new(sizeof(vector3));
    mw_object_get_abs_v3(&(sr->direction), sensor, v3_one[2]);
    mw_object_get_abs_v3(&(sr->position), sensor, v3_zero);
    v3_dec(sr->direction, sr->position);

    /* compute center of the activation-triangle on the sensor
     */
    center[0] = (edges[0][0] + edges[1][0] + edges[2][0]) / 3.0;
    center[1] = (edges[0][1] + edges[1][1] + edges[2][1]) / 3.0;
    center[2] = (edges[0][2] + edges[1][2] + edges[2][2]) / 3.0;
    v3_let(sr->position, center);

#ifdef DEBUG_DETAIL
    printf("center = %f %f %f\n", v3_elems(center));
#endif


    /* Follow recursivly from the current tp
     * to the tp without a root.
     * Cut from the current center to the focus of the tp with the tp-plane
     * the determine the new center(on the plane) and the distance
     * from the center to the new center (for adding to the distance
     * variable).
     */
    queue_add(sr->path, center);
    while ((tp = tp_root(tp))!=0) {
        focus = tp_focus(tp);
	v3_sub(index, *focus, center);
#ifdef DEBUG_DETAIL
	printf("cut line (%f %f %f) -> (%f %f %f) with tp\n",
		v3_elems(center), v3_elems(*focus));
#endif
	if (!line3_in_t3_plane((double *)0, (vector2 *)0,
			  &new_center, center, index,
			  t3p_triangle3(tp_t3p(tp)))) {
	    warning("us_sensor: activate_us_sensor_from_tp() - line don't pierce tp");
	    continue;
	}

#ifdef DEBUG_DETAIL
	printf("new center = %f %f %f\n", v3_elems(new_center));
#endif
        v3_sub(index, new_center, center);
	index_length = v3_sum(index);

	/* consider reflexion angle from the last center to the 
	 * new center against the triangle-normal. 
	 */  
#if 1
	if (!is_zero(index_length)) {
	    triangle3_normal(&normal, triangle_triangle3(t));
	    cos_angle = fabs(v3_product(normal, index) / 
			     (index_length * v3_sum(normal)));
#ifdef DEBUG_DETAIL
	    printf("reflection-angle (degree) = %f, cos(angle)=%f\n", 
		   acos(cos_angle)*180.0/PI, cos_angle);
#endif
	    intensity_factor *= cos_angle;
	}
#endif


	t=tp_mirror_of(tp);
	if (t) {
	    /* reflection on a obstacle (or a Sensor)
	     */

	    /* consider reflexion-factor of the obstacle
	     */
	    obj = triangle_get_user_data(t);
	    intensity_factor *= obj->us_reflexion_damping;

	} else {
	    /* Source-Sensor has no mirror-triangle 
	     */
	}

        distance += index_length;
	v3_let(center, new_center);
	queue_add(sr->path, center);
    }

/*    r0 = fabs(sensor->func.sensor.focus[2]); */
    r0 = 0.5 * sensor->func.sensor.radius / tan(0.5 * sensor->func.sensor.angle); 

    /* consider Intensity-weakening the sphere-area
     */
#if 1
    intensity_factor *= (r0*r0) / (distance + r0)*(distance + r0);
#endif
    /* consider the air-damping
     */
    intensity_factor *= exp(- distance * sensor->world->us_air_damping);

#ifdef DEBUG_DETAIL
    printf("distance=%f, Irecv/Isend = %f (%fdB)\n",
		distance, intensity_factor, log10(intensity_factor));
#endif

    intensity_factor = log10(intensity_factor);

    if (distance > ZERO_MAX && 
        intensity_factor >= sensor->func.sensor.threshold) {

	double tof;

	tof = distance / sensor->world->air_speed;

	sr->result = tof;
	distance *= 0.5;
	v3_addscal(sr->end_pos, sr->position, distance, sr->direction);

#ifdef DEBUG_DETAIL
	printf("tof = %fsecs\n", tof);
#endif
	tof = +scheduler_time(sensor->world->scheduler);
	activate_us_sensor(sensor, tof, sr);
    }
    else uscp_free(sr);
}


static void real_us_sensor_scan(mw_object sensor)
{
    triangle_projection tp;
    triangle t;
    mw_object obj;
    mob_world world = sensor->world;
    list tps = list_new();
    list tp_cuts = list_new();
    list rest_tps = list_new();
    list triangles = list_new();
    int	reflexion_count = 0;
    vector3 focus;

#ifdef DEBUG_DETAIL
    printf("real_us_sensor_scan()\n");
#endif

    /* Generate for all triangles of the object sensor
     * a triangle-projection with the given focus-point
     */
    mw_object_get_abs_v3(&focus, sensor, sensor->func.sensor.focus);

    mw_object_triangles(&triangles, sensor);
    t = list_first(triangles);
    while(t) {
        tp = tp_new(triangle_edges(t), focus);
	if (tp) list_add(tps, tp);
	t = list_next(triangles);
    }
    list_free(triangles);

    /* Projection the tp-elements in the tps-list in the space
     * until there is no reflexion, or the max reflexion count is
     * exceeded. Activate the sensor if a tp meets a sensor-face.
     */
    while (!list_is_empty(tps) && (reflexion_count <= world->max_reflexion_count)) {


	tp = list_first(tps);
        while(tp) {
            project_triangle(tp_cuts, 0, world->space, tp);
	    list_add(rest_tps, tp);
	    tp = list_next(tps);
        }

#ifdef DEBUG_DETAIL
	printf("Wave #%d with %d projections -> result: %d reflexions\n",
	       reflexion_count, list_size(tps), list_size(tp_cuts));
	printf("Reflexions =\n");
/* 	list_for_all(tp_cuts, tp_print);  */
#endif

	list_remove_all(tps);

        /* Look in the cut-list for an projection onto a sensor.
         * Add the remaining elements to the tps-list for further
	 * projection.
	 */
        tp = list_first(tp_cuts);
        while(tp) {
	    if ((t = tp_mirror_of(tp))!=0 &&
		((obj=triangle_get_user_data(t)) &&
		 obj->function == MWF_SENSOR)) {
		
		vector3 *focus;
		vector3 sensor_pos;
		vector3 sensor_dir;
		vector3 focus_dir;
		double x;

		focus = tp_focus(tp);
		mw_object_get_abs_v3(&sensor_pos, obj, v3_zero);
		mw_object_get_abs_v3(&sensor_dir, obj, v3_one[2]);
		v3_dec(sensor_dir, sensor_pos);
		v3_sub(focus_dir, *focus, sensor_pos);
		x = v3_product(focus_dir, sensor_dir);
		
#ifdef DEBUG_DETAIL
		printf("sensor-pos = {%f, %f, %f}\n", v3_elems(sensor_pos));
		printf("sensor-dir = {%f, %f, %f}\n", v3_elems(sensor_dir));
		printf("focus = {%f, %f, %f}\n", v3_elems(*focus));
		printf("x = %f\n", x);
		
#endif		

		/* projection onto a sensor, so activate the sensor!
		 */
		if (/* 1 */ x<=0.0 )  { 
		    activate_us_sensor_from_tp(obj, tp, t);
		    list_add(rest_tps, tp);
		}
		else list_add(tps, tp);
	    }
	    else list_add(tps, tp);

	    tp = list_next(tp_cuts);
	}
        list_remove_all(tp_cuts);

	reflexion_count++;
    }

    /* freeing all triangle projections and
     * lists.
     */
    list_for_all(tps, tp_free);
    list_for_all(rest_tps, tp_free); 
    list_free(tps);
    list_free(tp_cuts);
    list_free(rest_tps);
}






/* straight line sensor-simulation
 */

static void us_sensor_distance_scan(mw_object sensor)
{
    double distance;
    triangle t;
    vector3 position, direction;

    mw_object_get_abs_v3(&position, sensor, v3_zero);
    mw_object_get_abs_v3(&direction, sensor, v3_one[2]);
    v3_dec(direction, position);
    v3_addscal(position, position, 1e-4, direction);

    t = triangle_pick(&distance, sensor->world->space, position,
		      direction);

#ifdef DEBUG_DETAIL
    fprintf(stderr, "sensor_pos = %f %f %f\n",
	    position[0], position[1], position[2]);
    fprintf(stderr, "sensor_dir = %f %f %f\n",
	    direction[0], direction[1], direction[2]);
    fprintf(stderr, "triangle=%p, distance = %fm\n", t, distance);
#endif

    if (t && (distance >= ZERO_MAX)) {
	us_sensor_check_point	*sr = uscp_new(sensor->world);
	double tof;

	v3_let(sr->position, position);
	v3_let(sr->direction, direction);
	tof = scheduler_time(sensor->world->scheduler);
	tof += 2 * distance / sensor->world->air_speed;

	sr->result = tof;
	v3_addscal(sr->end_pos, position, distance, sr->direction);

	activate_us_sensor(sensor, tof, sr);
    }
}

void us_sensor_scan(mw_object sensor)
{
    if (sensor->world->real_us) real_us_sensor_scan(sensor);
    else us_sensor_distance_scan(sensor);
}
