/**************************************************************************
 *  PROJECT   : DA mobile robot simulation
 *  AUTHOR(S) : Horst Stolz (HUS)
 *  FILE      : vector.c
 *  TYPE      : module implementation
 *  Time-stamp: <94/11/07 17:40:30 stolz>
 **************************************************************************
 *  MODULE              : vector
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in vector.h.
 *
 *  HISTORY
 *  23-Mar-94   Creation date
 *
 **************************************************************************/

#include <stdio.h>
#include <math.h>

#include "vector.h"
#include "basic.h"




vector3 v3_zero = {0.0, 0.0, 0.0};
vector3 v3_one[3] = {
    {1.0, 0.0, 0.0},
    {0.0, 1.0, 0.0},
    {0.0, 0.0, 1.0}
};

vector2 v2_zero = {0.0, 0.0};

quaternion q_zero = {1.0, 0.0, 0.0, 0.0};

hom_matrix hm_zero = {
    {1.0, 0.0, 0.0, 0.0},
    {0.0, 1.0, 0.0, 0.0},
    {0.0, 0.0, 1.0, 0.0},
    {0.0, 0.0, 0.0, 1.0}
};



/* rechte-Hand-drehung +angle!! */
void q_set_av(quaternion *q, double angle, vector3 axis)
{
    double d, a;

#ifdef DEBUG_DETAIL
    printf("q_set_av() - set quaternion from angle+axis\n");
    printf("q_set_av: angle = %f, axis = {%f %f %f}\n",
	   angle, v3_elems(axis));
#endif

    /* restrict angle between -PI and +PI
     */
    a = fmod(angle, PIx2);
/*
    if (a>PI) a = a - PIx2; 
    else if (a<-PI) a = PIx2 + a;
*/
    /* restrict angle between 0 and +2*PI
     */
    if (a<0.0) a = PIx2 + a;


#ifdef DEBUG_DETAIL
    printf("q_set_av: restricted angle = %f\n", a);
#endif

    /* quaternion-def: 1 = q0^2 + d^2 * (q1^2 + q2^2 + q3^2)
     * -> d^2 = (1 - q^0) / (q1^2 + q2^2 + q3^2)
     */
    d = v3_product(axis,axis);
    if (is_zero(d)) {
	q_let(*q, q_zero);
	return;
    }

    /* 1 = q0^2 + q1^2 + q2^2 + q3^2
     * or 1 = q0^2 + d^2 * ( q1^2 + q2^2 + q3^2 )
     * -> d^2 = (1 - q0^2) / ( q1^2 + q2^2 + q3^2 )
     */
    (*q)[0] = cos(a/2.0);

    d = sqrt((1.0 - (*q)[0]*(*q)[0]) / d);

    d = -d;
/*    if (a >= 0.0) d = -d; */

    (*q)[1] = axis[0] * d;
    (*q)[2] = axis[1] * d;
    (*q)[3] = axis[2] * d;

#ifdef DEBUG_DETAIL
    printf("q_set_av: q = {%f, %f, %f, %f}\n", q_elems(*q));
#endif
}

void q_mul(quaternion *x, quaternion a, quaternion b)
{
    quaternion e;

    e[0] = a[0]*b[0] - (v3_product(&a[1], &b[1]));
    v3_mul(&e[1], &a[1], &b[1]);

    e[1] += a[0]*b[1] + b[0]*a[1];
    e[2] += a[0]*b[2] + b[0]*a[2];
    e[3] += a[0]*b[3] + b[0]*a[3];

    (*x)[0] = e[0];
    (*x)[1] = e[1];
    (*x)[2] = e[2];
    (*x)[3] = e[3];
}


/* [x][y] */

void hm_mul(hom_matrix *e, hom_matrix a, hom_matrix b)
{
    hom_matrix t;

    int x,y;

    for (x=0; x<4; x++)
	for (y=0; y<4; y++)
	    t[x][y]= a[x][0]*b[0][y] + a[x][1]*b[1][y]
		+ a[x][2]*b[2][y] + a[x][3]*b[3][y];

    for (x=0; x<4; x++)
	for (y=0; y<4; y++)
	    (*e)[x][y] = t[x][y];
}



void v3_qrot(vector3 *y, vector3 x, quaternion q)
{
    double q00 = q[0] * q[0];
    double q01 = q[0] * q[1];
    double q02 = q[0] * q[2];
    double q03 = q[0] * q[3];
    double q11 = q[1] * q[1];
    double q12 = q[1] * q[2];
    double q13 = q[1] * q[3];
    double q22 = q[2] * q[2];
    double q23 = q[2] * q[3];
    double q33 = q[3] * q[3];
    double m[3][3];
    vector3 v;

    m[0][0] = 2.0 * (q00 + q11) - 1.0;
    m[1][0] = 2.0 * (q12 + q03);
    m[2][0] = 2.0 * (q13 - q02);
    m[0][1] = 2.0 * (q12 - q03);
    m[1][1] = 2.0 * (q00 + q22) - 1.0;
    m[2][1] = 2.0 * (q23 + q01);
    m[0][2] = 2.0 * (q13 + q02);
    m[1][2] = 2.0 * (q23 - q01);
    m[2][2] = 2.0 * (q00 + q33) - 1.0;

    v[0] = m[0][0] * x[0] + m[1][0] * x[1] + m[2][0] * x[2];
    v[1] = m[0][1] * x[0] + m[1][1] * x[1] + m[2][1] * x[2];
    v[2] = m[0][2] * x[0] + m[1][2] * x[1] + m[2][2] * x[2];

    v3_let(*y, v);
}



void hm_set_v3q(hom_matrix *x, vector3 v, quaternion q)
{
    double q00 = q[0] * q[0];
    double q01 = q[0] * q[1];
    double q02 = q[0] * q[2];
    double q03 = q[0] * q[3];
    double q11 = q[1] * q[1];
    double q12 = q[1] * q[2];
    double q13 = q[1] * q[3];
    double q22 = q[2] * q[2];
    double q23 = q[2] * q[3];
    double q33 = q[3] * q[3];

    (*x)[0][0] = 2.0 * (q00 + q11) - 1.0;
    (*x)[1][0] = 2.0 * (q12 + q03);
    (*x)[2][0] = 2.0 * (q13 - q02);
    (*x)[0][1] = 2.0 * (q12 - q03);
    (*x)[1][1] = 2.0 * (q00 + q22) - 1.0;
    (*x)[2][1] = 2.0 * (q23 + q01);
    (*x)[0][2] = 2.0 * (q13 + q02);
    (*x)[1][2] = 2.0 * (q23 - q01);
    (*x)[2][2] = 2.0 * (q00 + q33) - 1.0;

    (*x)[3][0] = v[0];
    (*x)[3][1] = v[1];
    (*x)[3][2] = v[2];

    (*x)[0][3] = 0.0;
    (*x)[1][3] = 0.0;
    (*x)[2][3] = 0.0;
    (*x)[3][3] = 1.0;
}

/* HC = 
 *
 *           | v
 *   Matrix  | e
 *           | k
 *   -----------
 *   0  0  0 | 1
 *
 */
void hm_vmul(vector3 *y, hom_matrix t, vector3 x)
{
    vector3 v;
    
    v[0] = t[0][0] * x[0] + t[1][0] * x[1] + t[2][0] * x[2] + t[3][0];
    v[1] = t[0][1] * x[0] + t[1][1] * x[1] + t[2][1] * x[2] + t[3][1];
    v[2] = t[0][2] * x[0] + t[1][2] * x[1] + t[2][2] * x[2] + t[3][2];

    v3_let(*y, v);
}


void q_get_av(double *angle, vector3 *axis, quaternion q)
{
#ifdef DEBUG_DETAIL
    printf("q_get_av() - get angle and axis from quaternion\n");
#endif

    if (angle) {
	if (q[0] >= 1.0) *angle = 0.0;
	else if (q[0] <= -1.0) *angle = PI*2.0; 
	else *angle = acos(q[0]) * 2.0;
#ifdef DEBUG_DETAIL
	printf("q_get_av: angle = %f\n", *angle);
#endif
    }
    
    if (axis) {
	(*axis)[0] = -q[1]; 
	(*axis)[1] = -q[2]; 
	(*axis)[2] = -q[3]; 
#ifdef DEBUG_DETAIL
	printf("q_get_av: axis = {%f, %f, %f}\n", v3_elems(*axis));
#endif
    }
}

void hm_let(hom_matrix *e, hom_matrix a)
{
    int x,y;

    for (x=0; x<4; x++)
	for (y=0; y<4; y++)
	    (*e)[x][y] = a[x][y];
}
