/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: vector_geometry.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 17:04:01 stolz>
 **************************************************************************
 *  MODULE              : vector_geometry
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in 
 *	vector_geometry.h!
 *
 *  HISTORY
 *  17-Aug-94   Split routines from triangle.c.
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "vector_geometry.h"
#include "error.h"




void point2_in_basis_oab(vector2 *point_basis, vector2 point,
		vector2 basis_o, vector2 basis_a, vector2 basis_b)
/****************************************************************************
 * FUNCTION
 *	Transform <point> into the new basis (basis_o, basis_a, basis_b).
 *
 * RESULT
 *      The result-vector will be stored in <point_basis>.
 */
{
    vector2 v, y, z;
    vector2 rel_a, rel_b;
    double alpha, beta, h;

    /* point from origin of basis
     */
    v2_sub(v, point, basis_o);

    /* relative vectors from a,b to o.
     */
    v2_sub(rel_a, basis_a, basis_o);
    v2_sub(rel_b, basis_b, basis_o);

    alpha = v2_product(rel_a, v) / v2_product(rel_a, rel_a);
    beta = v2_product(rel_b, v) / v2_product(rel_b, rel_b);

    v2_subscal(y, v, alpha, rel_a);
    v2_subscal(z, v, beta, rel_b);

    alpha = v2_product(z, z);
    beta = v2_product(y, y);

    h = v2_product(rel_b, y);
    if (is_zero(h)) beta = 0.0; else beta = beta / h;

    h = v2_product(rel_a, z);
    if (is_zero(h)) alpha = 0.0; else alpha = alpha / h;

    if (point_basis) {
	v2_set(*point_basis, alpha, beta);
#if 0
	printf("point %f %f -> %f %f\n",
		v2_elems(point), v2_elems(*point_basis));
#endif
    }
}


boolean point2_in_triangle(vector2 point)
/****************************************************************************
 * FUNCTION
 *	Determine if a point lies in the standard triangle which is build of
 *      the edges (0,0) (1,0) and (0,1).
 *
 * RESULT
 *      TRUE	point lies in the triangle or on an edge.
 *	FALSE   point lies out of the triangle.
 */
{
    if ((point[0] >= 0.0) && (point[1] >=0.0) && ((point[0]+point[1]) <= 1.0)) {

	return TRUE;
    }
    else
	return FALSE;
}


boolean point2_in_triangle_eps(vector2 point, double eps)
/****************************************************************************
 * FUNCTION
 *	Determine if a point lies in the standard triangle which is build of
 *      the edges (0,0) (1,0) and (0,1).
 *
 * RESULT
 *      TRUE	point lies in the triangle or on an edge.
 *	FALSE   point lies out of the triangle.
 */
{
    if ((point[0] >= -eps) && (point[1] >= -eps) && 
	((point[0]+point[1]) <= (1.0+eps))) {

	return TRUE;
    }
    else
	return FALSE;
}



#if 0
boolean line2_pierce_line2(vector2 *point, double *index1, double *index2,
			   vector2 line1_start, vector2 line1_index,
                           vector2 line2_start, vector2 line2_index)
/* an welchem Punkt schneiden sich die beiden Geraden?
 */
{
    vector2 v;
    double line1_index_quadrat;
    double i1, i2;

#if DEBUG_DETAIL
    printf("l2l: line1 %f %f -> %f %f\nl2l: line2 %f %f -> %f %f\n",
	    v2_elems(line1_start), v2_elems(line1_index),
	    v2_elems(line2_start), v2_elems(line2_index));
#endif

    /* starting point of line2 viewed from the starting point
     * of line1
     */
    v2_sub(v, line2_start, line1_start);

    line1_index_quadrat = v2_product(line1_index, line1_index);

    i1 = v2_product(line1_index, v) / line1_index_quadrat;

    v[0] = i1 * line1_index[0] - v[0];
    v[1] = i1 * line1_index[1] - v[1];

    /* 1/i2 is the index2 so that x=line2_start+1/i2*line2_index
     */
    i2 = v2_product(v, v);
    if (is_zero(i2)) return FALSE;

    i2 = v2_product(line2_index, v) / i2;

#if DEBUG_DETAIL
    printf("l2l: c = %f %f\n", v2_elems(v));
    printf("l2l: alpha = %f, e = %f\n", i1, i2);
#endif
    if (is_zero(i2)) return FALSE;

    i2 = 1/i2;

    i1 += i2 * v2_product(line2_index, line1_index) / line1_index_quadrat;

#if DEBUG_DETAIL
    printf("l2l: i1 = %f, i2 = %f\n", i1, i2);
#endif

    if (index1) *index1 = i1;
    if (index2) *index2 = i2;
    if (point) {
	v2_addscal(*point, line2_start, i2, line2_index);

#if DEBUG_DETAIL
        printf("l2l: point = %f %f\n", v2_elems(*point));
#endif

    }

    return TRUE;
}
#endif


boolean line2_pierce_line2(vector2 *point, double *index1, double *index2,
			   vector2 line1_start, vector2 line1_index,
                           vector2 line2_start, vector2 line2_index)
/* an welchem Punkt schneiden sich die beiden Geraden?
 */
{
    vector2 v;
    double i1, i2, d;


    d = line1_index[0]*line2_index[1] - line1_index[1]*line2_index[0];

    if (is_zero(d)) {
#if DEBUG_DETAIL
	printf("l2l: i1x*i2y-i1y*i2x == 0");
	return FALSE;
#endif
    }

    v2_sub(v, line2_start, line1_start);

    i1 = (line2_index[1]*v[0] - line2_index[0]*v[1]) / d;
    i2 = (line1_index[1]*v[0] - line1_index[0]*v[1]) / d;


#if DEBUG_DETAIL
    printf("l2l: i1 = %f, i2 = %f\n", i1, i2);
#endif

    if (index1) *index1 = i1;
    if (index2) *index2 = i2;
    if (point) {
	v2_addscal(*point, line2_start, i2, line2_index);

#if DEBUG_DETAIL
        printf("l2l: point = %f %f\n", v2_elems(*point));
#endif

    }

    return TRUE;
}


boolean line2_in(double index)
/****************************************************************************
 */
{
    if ((index > 100.0*ZERO_MAX) && (index < (1.0 - 100.0*ZERO_MAX)))
	return TRUE;
    else
	return FALSE;
}

boolean line2_in_eps(double index, double eps)
/****************************************************************************
 */
{
    if ((index >= eps) && (index <= (1.0 - eps)))
	return TRUE;
    else
	return FALSE;
}



boolean line2_at(double index)
/****************************************************************************
 */
{
    if ((index >= 0.0) && (index <= 1.0))
	return TRUE;
    else
	return FALSE;
}

boolean line2_at_eps(double index, double eps)
/****************************************************************************
 */
{
    if ((index >= -eps) && (index <= (1.0+eps)))
	return TRUE;
    else
	return FALSE;
}


void perpendicular_on_plain3(double *distance,
			     vector3 *perp_point,
                             vector3 plain_point,
                             vector3 plain_normal,
			     vector3 point)
/****************************************************************************
 * FUNCTION
 *	Determine the perpendicular distance between a point and a plain.
 *	Also computes the point of the perpendicular which lies in the
 *	plain. If the sum of the normal is zero, point=perp_point and
 *	distance=0.
 *
 * ARGUMENTS
 *	plain_point     one point on the plain
 *	plain_normal	the normal of the plain
 *	point           a point in space
 *
 * RESULTS
 *	The function fills only the results which pointer are not 0.
 *
 *	*distance	perpendicular distance between point and plain.
 *      *perp_point	point on plain.
 */
{
    double a, n, b;
    vector3 p;

    v3_sub(p, point, plain_point);
    n = v3_sum(plain_normal);
    if (is_not_zero(n)) {
        a = v3_product(plain_normal, p) / n;
	b = a / n;
    }
    else b = 0.0;

    if (distance) *distance = a;

    if (perp_point) {
	v3_subscal(*perp_point, point, b, plain_normal);
    }
}



#if 0
boolean	line3_in_plane(double *factor, vector2 *point2, vector3 *point3,
		          vector3 line_start, vector3 line_index,
			  triangle3 t)
/****************************************************************************
 * FUNCTION
 *	Compute the line-index-factor of the piercing point between the
 *	triangle t and the infinite straight line (line_start, line_index)
 *
 * RESULT
 *      Returns TRUE if the line goes through the triangle-plain.
 *      Else return FALSE.
 *      In the var-variable factor, the factor for the piercing point
 *		line_start + factor*line_index
 *	will be returned.
 */
{
    vector3 v,		/* piercing point */
            y, z;
    double d;
    double alpha, beta;


    /* length of line in direction of the triangle-normal
     */
    d = v3_product(t->normal, line_index);

#if 0
    printf("line-index-length in triangle-normal-direction: %f\n", d);
#endif

    if (is_zero(d)) return FALSE; /* line parallel to triangle */

    /* Starting point of the line relative to the 1.edge
     * of the triangle
     */
    v3_sub(v, line_start, t->abs_edges[0]);

    /* v3_product() = distance of the line-starting-point
     * to the triangle-plain
     */

    /* factor d*line_index + line_start lies in triangle plain
     */
    d = - v3_product(t->normal, v) / d;

#if 0
    printf("line-index-factor from line-start to triangle-plain: %f\n", d);
#endif

    if (factor) *factor = d;

    /* absolute coordinates of the piercing point
     */
    if (point3) {
	v3_addscal(*point3, line_start, d, line_index);
    }

    /* point of piercing relative from 1.edge of the triangle
     */
    v3_addscal(v, v, d, line_index);

#if 0
    printf("piercing point is = (%f %f %f)\n",
	    v[0], v[1], v[2]);
#endif

    alpha = v3_product(t->rel_edges[0], v) / (t->rel_length[0] * t->rel_length[0]);
    beta = -v3_product(t->rel_edges[2], v) / (t->rel_length[2] * t->rel_length[2]);

#if 0
    printf("alpha' = %f, beta' = %f\n", alpha, beta);
#endif

    v3_subscal(y, v, alpha, t->rel_edges[0]);
    v3_addscal(z, v, beta, t->rel_edges[2]);

    alpha = v3_product(z, z);
    beta = v3_product(y, y);

    d = v3_product(t->rel_edges[2], y);
    if (is_zero(d)) beta = 0.0; else beta = - beta / d;

    d = v3_product(t->rel_edges[0], z);
    if (is_zero(d)) alpha = 0.0; else alpha = alpha / d;

#if 0
    printf("alpha = %f, beta = %f\n", alpha, beta);
#endif

    if (point2) {
	v2_set(*point2, alpha, beta);
    }

    return TRUE;
}
#endif


void perpendicular_from_line3_to_point3(double *distance, vector3 *direction,
					double *meet_index,
					vector3 point,
					vector3 line_start, vector3 line_index)
{
    vector3 p;
    double m, d;

    v3_sub(p, point, line_start);
    m = v3_product(p, line_index) / v3_product(line_index, line_index);
    if (meet_index) *meet_index = m;

    v3_subscal(p, p, m, line_index);
    d = v3_sum(p);
    if (distance) *distance = d;

    if (!is_zero(d)) {
        d = 1 / d;
        v3_scalar(p, p, d);
    }
    else v3_let(p, v3_zero);

    if (direction) v3_let(*direction, p);
}

