/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: vehicle.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 17:04:18 stolz>
 **************************************************************************
 *  MODULE              : vehicle
 *  IMPORT		: basic, scheduler, error, queue, integration
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The specification and definition is given in vehicle!
 *
 *  HISTORY
 *  19-Apr-94   build module from obsolent modules robot and motion
 *  28-Apr-94   motv ad/la is running fine
 *  29-May-94	removing obsolent vehicle parameters r1 and r2.
 *              remove Runge-Kutta-Code, place it in integration.c.
 *		also add code for front-wheel-angles
 *  01-Jun-94   add front_wheel_damping for computation of the 
 *              front-wheel-angles.
 *  06-Jun-94   add vehicle_halt() and vehicle_resume() for extern halting
 *  06-Jun-94   add vehicle_set_vel_disturbance() & enable code in dstate()
 *  14-Jun-94   fix problem of random-number by using the rand() function
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>

#include "basic.h"
#include "vehicle.h"
#include "scheduler.h"
#include "queue.h"
#include "error.h"
#include "integration.h"



/**************************************************************************
 * private definition for abstract datatype vehicle
 */
typedef struct adt_vehicle adt_vehicle;
struct adt_vehicle {
	vehicle_state		real, intern;
	vehicle_parameters	parameters;
	scheduler		*scheduler;
	task			vehicle_task;
        timepoint		vehicle_tp;
	boolean                 tp_inactive;
	double			dt_integrator;
	int                     motion_type;
        queue			motions;
	double			stop_time;

	double			rel_rand_voffset, rel_rand_vstart;
	double			rel_rand_woffset, rel_rand_wstart;
};


/**************************************************************************
 * private definition for type motion
 */
typedef struct {
	double  	a, u, time;
} motion;


/**************************************************************************
 * definition for the integration-variables
 */
#define NR_IVARS	12		/* number of vars in the ivars */

typedef struct {
    double x, y, angle, angle_fl, angle_fr;
    double v, w;
    double real_x, real_y, real_angle;
} ivars;


/**************************************************************************
 * some local variables for the integration-process
 */
static vehicle act_vehicle;
static const double front_wheel_damping = 5.4321;


/**************************************************************************
 * private functions
 */


static void dstate(double *dy_double, double *y_double, double t)
/**************************************************************************
 * computes the derivate of the variables in y at the time t
 * and stores the result in dy.
 */
{
    double sina, cosa, temp;
    double v[2], n[2], r[2];
    ivars *dy = (ivars *)dy_double;
    ivars *y = (ivars *)y_double;


    /* precompute usefull terms.
     */
    sina = sin(y->angle);
    cosa = cos(y->angle);

    /* set intern acceleration and velocity
     */

    dy->v = act_vehicle->intern.a;
    dy->w = act_vehicle->intern.u;

    dy->x = y->v * cosa;
    dy->y = y->v * sina;
    dy->angle = y->w;


    /* compute derivations for real positions/angle.
     * consider some random-factors!
     */
    temp = y->v * 
	   (act_vehicle->rel_rand_voffset * drand() + act_vehicle->rel_rand_vstart);

    dy->real_x = temp * cosa;
    dy->real_y = temp * sina;

    temp = act_vehicle->rel_rand_woffset * drand() + act_vehicle->rel_rand_wstart;

    dy->real_angle = y->w * temp;

/* obsolent: without velocity disturbance!
    dy->real_x = dy->x;
    dy->real_y = dy->y;
    dy->real_angle = dy->angle;
*/

    /* compute angular velocity of the right front-wheel.
     */
    n[0] = - act_vehicle->parameters.d * cos(y->angle_fr);
    n[1] = - act_vehicle->parameters.d * sin(y->angle_fr);

    r[0] = act_vehicle->parameters.b + n[0];
    r[1] = -act_vehicle->parameters.c + n[1];

    v[0] = y->v - y->w * r[1];
    v[1] =      + y->w * r[0];

    dy->angle_fr = - (-v[0]*n[1] + v[1] * n[0])/act_vehicle->parameters.d
	* front_wheel_damping;
    

    /* compute angular acceleration of the left front-wheel.
     */
    n[0] = - act_vehicle->parameters.d * cos(y->angle_fl);
    n[1] = - act_vehicle->parameters.d * sin(y->angle_fl);

    r[0] = act_vehicle->parameters.b + n[0];
    r[1] = act_vehicle->parameters.c + n[1];

    v[0] = y->v - y->w * r[1];
    v[1] =      + y->w * r[0];

    dy->angle_fl = - (-v[0]*n[1] + v[1] * n[0])/act_vehicle->parameters.d
	* front_wheel_damping;
}



static void vehicle_task_free(scheduler *s, task *t)
{
}

static void vehicle_task_reset_to(scheduler *s, task *t, double time)
{
    vehicle v = (vehicle)t->workspace;

    v->tp_inactive = TRUE;
    scheduler_point_remove(v->scheduler, &(v->vehicle_tp));
}

static void vehicle_sync(scheduler *s, task *vt, double t, double lastsync)
{
    ivars	vars;

    act_vehicle = vt->workspace;

    /* integrate only if the vehicle is not halted
     * or stopped.
     */
    if (act_vehicle->motion_type == VMT_STOP ||
	act_vehicle->motion_type == VMT_HALT) return;

    /* integrate ivars from time lastsync to the time t
     */
    vars.x = act_vehicle->intern.x;
    vars.y = act_vehicle->intern.y;
    vars.angle = act_vehicle->intern.angle;
    vars.v = act_vehicle->intern.v;
    vars.w = act_vehicle->intern.w;
    vars.angle_fl = act_vehicle->intern.angle_fl;
    vars.angle_fr = act_vehicle->intern.angle_fr;
    vars.real_x = act_vehicle->real.x;
    vars.real_y = act_vehicle->real.y;
    vars.real_angle = act_vehicle->real.angle;

    integrator((double *)&vars, (double *)&vars, NR_IVARS, 
	       lastsync, t, act_vehicle->dt_integrator,
	       dstate);

    act_vehicle->intern.x = vars.x;
    act_vehicle->intern.y = vars.y;
    act_vehicle->intern.angle = vars.angle;
    act_vehicle->intern.angle_fl = vars.angle_fl;
    act_vehicle->intern.angle_fr = vars.angle_fr;
    act_vehicle->intern.v = vars.v;
    act_vehicle->intern.w = vars.w;

    act_vehicle->real.x = vars.real_x;
    act_vehicle->real.y = vars.real_y;
    act_vehicle->real.angle = vars.real_angle;
    act_vehicle->real.angle_fl = vars.angle_fl;
    act_vehicle->real.angle_fr = vars.angle_fr;
    act_vehicle->real.v = vars.v;
    act_vehicle->real.w = vars.w;
}



static void motion_change(scheduler *s, timepoint *tp, double t)
{
    vehicle v = tp->workspace;
    motion  m;

    if (queue_first(&m, v->motions)) {

        scheduler_point_add(s, tp, t + m.time);
	v->tp_inactive = FALSE;

	if (is_zero(m.a)) m.a = 0.0;
	if (is_zero(m.u)) m.u = 0.0;

        v->intern.a = m.a;
        v->intern.u = m.u;
	v->real.a = m.a;
        v->real.u = m.u;
    }
    else {
	v->intern.a = 0.0;
        v->intern.u = 0.0;
	v->real.a = 0.0;
        v->real.u = 0.0;


	v->tp_inactive = TRUE;
	v->stop_time = 0.0;
    }

    if (is_zero(v->intern.v)) { v->intern.v = 0.0; v->real.v = 0.0; };
    if (is_zero(v->intern.w)) { v->intern.w = 0.0; v->real.w = 0.0; };

    /* set motion type
     */
    if (v->motion_type==VMT_KILL) {
	if (v->intern.v==0.0 && v->intern.w==0.0)
	    v->motion_type = VMT_STOP;
    } else {
	if (v->intern.a==0.0 && v->intern.u==0.0) {
	    if (v->intern.v==0.0 && v->intern.w==0.0)
		v->motion_type = VMT_STOP;
	    else
		v->motion_type = VMT_CONST;
	}
	else {
	    if (v->intern.v * v->intern.a || v->intern.w * v->intern.u)
		v->motion_type = VMT_ACC;
	    else
		v->motion_type = VMT_DEC;
	}
    }
}


/**************************************************************************
 * exported functions
 */


vehicle	vehicle_new(scheduler *s, vehicle_parameters *vp)
/**************************************************************************
 */
{
    vehicle v;

    if (s==0) error("vehicle:vehicle_new() - illegal scheduler parameter(=0)");
    if (vp==0) error("vehicle:vehicle_new() - illegal vehicle parameter(=0)");
    if (vp->a <= 0.0)
	error("vehicle:vehicle_new() - vehicle parameter a <= 0.0");
    if (is_zero(vp->d))
	error("vehicle:vehicle_new() - vehicle parameter d is zero!");

    v = malloc(sizeof(adt_vehicle));
    if (v) {
        v->parameters = *vp;
        v->motions = queue_new(sizeof(motion));
	v->scheduler = s;

	v->vehicle_task.workspace = v;
	v->vehicle_task.free = vehicle_task_free;
	v->vehicle_task.reset = vehicle_task_reset_to;
	v->vehicle_task.sync = vehicle_sync;

        scheduler_task_add(s, &(v->vehicle_task));

	v->vehicle_tp.task = &(v->vehicle_task);
	v->vehicle_tp.workspace = v;
	v->vehicle_tp.free = 0;
	v->vehicle_tp.function = motion_change;
	v->tp_inactive = TRUE;
	v->stop_time = 0.0;
	v->motion_type = VMT_STOP;

	v->real.x = 0.0;
	v->real.y = 0.0;
	v->real.angle = 0.0;
	v->real.angle_fl = 0.0;
	v->real.angle_fr = 0.0;
	v->real.v = 0.0;
	v->real.w = 0.0;
	v->real.a = 0.0;
	v->real.u = 0.0;
	v->intern = v->real;

	v->dt_integrator = 0.2;

	v->rel_rand_voffset = 0.0;
	v->rel_rand_vstart  = 1.0;
	v->rel_rand_woffset = 0.0;
	v->rel_rand_wstart  = 1.0;

    } else error("vehicle:vehicle_new() - can't alloc"
		 " memory for vehicle struct");

    return v;
}


void vehicle_free(vehicle v)
/**************************************************************************
 */
{
    if (!(v->tp_inactive)) scheduler_point_remove(v->scheduler, &(v->vehicle_tp));
    v->tp_inactive = TRUE;
    scheduler_task_remove(v->scheduler, &(v->vehicle_task));
    queue_free(v->motions);
    free(v);
}


void vehicle_reset(vehicle v)
/**************************************************************************
 */
{
    motion m;

    if (!(v->tp_inactive)) scheduler_point_remove(v->scheduler, &(v->vehicle_tp));
    v->tp_inactive = TRUE;
    v->stop_time = 0.0;
    v->motion_type = VMT_STOP;

    while (queue_first(&m, v->motions))
        ;

    v->real.x = 0.0;
    v->real.y = 0.0;
    v->real.angle = 0.0;
    v->real.angle_fl = 0.0;
    v->real.angle_fr = 0.0;
    v->real.v = 0.0;
    v->real.w = 0.0;
    v->real.a = 0.0;
    v->real.u = 0.0;
    v->intern = v->real;
}


void vehicle_add_motion(vehicle v,
			double linear_acc,
			double angular_acc,
			double time)
/**************************************************************************
 */
{
    motion m;

    if (v->motion_type == VMT_HALT) {
	v->stop_time = scheduler_time(v->scheduler);
	return;
    }

    m.a = linear_acc;
    m.u = angular_acc;
    m.time = time;
    if (queue_is_empty(v->motions))
	v->stop_time = scheduler_time(v->scheduler);
    v->stop_time += time;


    queue_add(v->motions, &m);

    if (v->tp_inactive) {
	v->tp_inactive = FALSE;

	scheduler_point_add(v->scheduler, &(v->vehicle_tp),
			    scheduler_time(v->scheduler));
    }
}


void vehicle_empty_motion_queue(vehicle v)
/**************************************************************************
 */
{
    motion m;

    while (queue_first(&m, v->motions))
	;
    
    v->stop_time = 0.0;
    scheduler_point_unlink(v->scheduler, &(v->vehicle_tp));
    v->tp_inactive = TRUE;
}


void vehicle_kill_motion(vehicle v, double linear_decel,
				    double angular_decel)
/**************************************************************************
 */
{
    motion m;
    double a, u, ta, tu;

    while (queue_first(&m, v->motions))
        ;

    if (v->motion_type == VMT_KILL || v->motion_type == VMT_STOP ||
	v->motion_type == VMT_HALT) return;

    v->motion_type = VMT_KILL;
    
    if (is_zero(v->intern.v) && is_zero(v->intern.w)) {
	v->intern.a = 0.0;
	v->intern.u = 0.0;
	v->real.a = 0.0;
	v->real.u = 0.0;
	return;
    }

    a = -sign_heritage(v->intern.v, linear_decel);
    u = -sign_heritage(v->intern.w, angular_decel);

    if (is_zero(a))  ta = 0.0; else ta = fabs(v->intern.v / a);
    if (is_zero(u))  tu = 0.0; else tu = fabs(v->intern.w / u);


    
    scheduler_point_unlink(v->scheduler, &(v->vehicle_tp));
    v->tp_inactive = TRUE;

    if (ta >= tu)
        vehicle_add_motion(v, a, - v->intern.w / ta, ta);
    else
        vehicle_add_motion(v, - v->intern.v / tu, u, tu);
}


double vehicle_get_stop_time(vehicle v)
/**************************************************************************
 */
{
    return v->stop_time;
}


void vehicle_get_state(vehicle v, vehicle_state *real, vehicle_state *intern)
/**************************************************************************
 */
{
    if (real) *real = v->real;
    if (intern) *intern = v->intern;
}

void vehicle_set_state(vehicle v, vehicle_state *real, vehicle_state *intern)
/**************************************************************************
 */
{
    vehicle_reset(v);

    if (real) v->real = *real;
    if (intern) v->intern = *intern;
}


int vehicle_motion_type(vehicle v)
/**************************************************************************
 */
{
    return v->motion_type;
}


void	vehicle_halt(vehicle v)
/**************************************************************************
 */
{
    vehicle_empty_motion_queue(v);

    v->motion_type = VMT_HALT;

    v->real.a = 0.0;
    v->real.u = 0.0;
    v->real.v = 0.0;
    v->real.w = 0.0;

    v->intern.a = 0.0;
    v->intern.u = 0.0;
    v->intern.v = 0.0;
    v->intern.w = 0.0;
}


void	vehicle_resume(vehicle v)
/**************************************************************************
 */
{
    v->motion_type = VMT_STOP;
}


void vehicle_set_vel_disturbance(vehicle v, vel_disturbance rand)
/**************************************************************************
 */
{
    v->rel_rand_vstart  = rand.v_low;
    v->rel_rand_voffset = rand.v_high - rand.v_low;
    v->rel_rand_wstart  = rand.w_low;
    v->rel_rand_woffset = rand.w_high - rand.w_low;
}
