/********************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: vehicle.h
 *  TYPE		: module definition
 *  Time-stamp: <95/02/13 17:04:27 stolz>
 ********************************************************************************
 *  MODULE              : vehicle
 *
 *  FUNCTION
 *	This module provides a vehicle which is controled through constant
 *	linear and angular acceleration. The time-specific aspect is
 *	controlled via the module scheduler.
 *
 *	The vehicle has two driving wheels with radius r1. These lies on a
 *	axe with length 2 * a. The middle of the axe is the reference point
 *	of the local coordinate system. The position x and y of the vehicle
 *	describe the position of the local coordinate system to a global
 *	cartesian coordiant system. The parameter angle describe the angle
 *	from the global x-axe to the local x-axe if both lies in the same
 *	origin. The angular velocity and accelaration has the same direction
 *	as this angle.
 *
 *	The frontwheels with radius r2 are rotatable at a point with distance d 
 *      in x-direction. The angles angle_fl and angle_fr represent the angle 
 *	from the local x-axe to the distance-holder of the left and right 
 *	frontwheels.
 *
 *                  r1
 *                |---|                        r2   d
 *                                            |--|-----|
 *                WWWWXWWWW                                    left
 *		- WWWWXWWWW                   WWWXWWW--* -
 *              |     |   local coord.                 | |
 *            a |    /|\y          _w                  | | c
 *              |     |      x     \                   | |
 *     rear     -     O----->-------|------------------| -     front
 *                    |-------> v   |                  |
 *      driving       |            /                   |
 *        wheels      |                                |
 *                WWWWXWWWW                   WWWXWWW--*       right
 *    y           WWWWXWWWW
 *    |                               b
 *    |               |--------------------------------|
 *    0--- x
 *       global coord.                        frontwheels
 *
 *
 *	If the distance c is set to 0. The vehicle has only one front-wheel.
 *	Then angle_fl and angle_fr has the same value. It's not allowed to
 *	set the d to zero!
 *
 *	The motion of the vehicle is controlled by setting the linear and
 *	angular acceleration (a=dv/dt, u=dw/dt) and the time for which
 *	this values are applied. This is done with the vehicle_add_motion()
 *	procedure. All motions which are added with this functions are
 *	queued. To stop the vehicle and empty the motion-queue the
 *      procedure vehicle_kill_motion() should be called.
 *
 *      If you wan't use the vehicle_kill_motion() command, or to clear
 *      the motion queue, call vehicle_empty_motion_queue().
 *      this erase all motions in the queue. Remember that the actual
 *      velocitiies and acceleration will remain!
 *
 *      If it's necessary to halt the vehicle immediately call
 *	vehicle_halt(). Now, it's not possible to move with the
 *	vehicle_add_motion()-command. They motions will also not 
 *	stored! To disable the halt use the function vehicle_resume().
 *	Also the function vehicle_reset() clears the halt-status.
 *	The vehicle_motion_type() indicates this status with
 *	VMT_HALT.
 *
 *	The real and intern state of the vehicle will differ if the
 *	velocity-disturbance-factors are changed with a call of
 *	vehicle_set_vel_disturbance(). The parameters are range
 *	[low,..,high] for linear and angular velocity. The "real"
 *	velocity is computed by multiplying a random-value of this
 *	range with the respective velocity. This will cause a different
 *	value for the position (x,y) and the angle w!
 *	
 *
 *  NOTE
 *      Vehicle parameters r1 and r2 are obsolent. Removing them from the
 *      definition.
 *
 *  HISTORY
 *  17-Apr-94   Creation date.
 *  29-May-94   Removing obsolent parameters r1 and r2
 *  06-Jun-94   Add vehicle_halt() and vehicle_resume() for extern halting
 *  06-Jun-94   Add vehicle_set_vel_disturbance() & enable code in dstate()
 *
 ********************************************************************************/
#ifndef VEHICLE_H
#define VEHICLE_H

#ifndef SCHEDULER_H
#include "scheduler.h"
#endif


/**************************************************************************
 * Abstract datatype for a vehicle
 */
typedef struct adt_vehicle*		vehicle;


/**************************************************************************
 * motion state of the vehicle,
 * this includes position, velocity and acceleration
 */
typedef struct vehicle_state		vehicle_state;
struct vehicle_state {
    double	x, y, angle;		/* position */
    double      v, w;                   /* velocity */
    double	a, u;                   /* acceleration */
    double	angle_fl, angle_fr;	/* position of front-wheels */
};


/**************************************************************************
 * This structure describe the vehicle
 */
typedef struct vehicle_parameters	vehicle_parameters;
struct vehicle_parameters {
    double	a, b, c, d;
};


typedef struct {		/* lin/angular velocity is randomly */
    double v_low, v_high;	/* distributed in the ranges: */
    double w_low, w_high;	/* v_real = v_intern * random of [v_low .. v_high] */
} vel_disturbance;		/* w_real = w_intern * random of [w_low .. w_high] */


/**************************************************************************
 * motion types returned by vehicle_motion_type().
 *
 * VMT_STOP  = vehicle don't move (accelerations and velocities are zero)
 * VMT_ACC   = vehicle accelerate (at least linear or angular)
 * VMT_DEC   = vehicle decelerate (linear and angular)
 * VMT_KILL  = vehicle is in a kill-phase. After that it have zero vel/acc.
 * VMT_CONST = vehicle drives with constant velocities
 * VMT_HALT  = vehicle halted! (accelerations and velocities are zero)
 */
enum vehicle_motion_types {
    VMT_STOP, VMT_ACC, VMT_DEC, VMT_KILL, VMT_CONST, VMT_HALT
};



/**************************************************************************
 * datatype functions
 */
vehicle	vehicle_new(scheduler *, vehicle_parameters *);
void	vehicle_free(vehicle);
void	vehicle_reset(vehicle);

void	vehicle_add_motion(vehicle, double linear_acc, double angular_acc, double time);
void	vehicle_kill_motion(vehicle, double linear_decel, double angular_decel);
double	vehicle_get_stop_time(vehicle);

void	vehicle_get_state(vehicle, vehicle_state *real, vehicle_state *intern);
void 	vehicle_set_state(vehicle, vehicle_state *real, vehicle_state *intern);

int     vehicle_motion_type(vehicle v);
void    vehicle_empty_motion_queue(vehicle v);

void	vehicle_halt(vehicle);
void	vehicle_resume(vehicle);

void	vehicle_set_vel_disturbance(vehicle, vel_disturbance);

#endif
