/**************************************************************************
 *  PROJECT		: DA mobile robot simulation
 *  AUTHOR(S)           : Horst Stolz (HUS)
 *  FILE		: xdraw.c
 *  TYPE		: module implementation
 *  Time-stamp: <95/02/13 17:04:34 stolz>
 **************************************************************************
 *  MODULE              : xdraw
 *
 *  IMPLEMENTATION
 *
 *  NOTE
 *      The module specification and definition is given in xdraw.h!
 *
 *  HISTORY
 *  19-May-94   Creation date
 *  06-Jun-94   Add draw2d_space(), draw_get_pick_parameters().
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <X11/Intrinsic.h>


#include "xdraw.h"
#include "mob_world_private.h"
#include "error.h"
#include "basic.h"


#define get_win_pos(x,y,start,point,sinx,cosx)  \
(x)=(width>>1)+(int)(scale[0]*((start)[0]+(cosx)*(point)[0]-(sinx)*(point)[1]-center[0])); \
(y)=(height>>1)-(int)(scale[1]*((start)[1]+(cosx)*(point)[1]+(sinx)*(point)[0]-center[1]));

#define get_win_coord(x,y,point)  \
(x)=(width>>1)+(int)(scale[0]*((point)[0]-center[0])); \
(y)=(height>>1)-(int)(scale[1]*((point)[1]-center[1]));

#define get_v3_coord(point,x,y)  \
v3_set((point),center[0]+((double)((x)-(width>>1)))/scale[0],\
center[1]-((double)((y)-(height>>1)))/scale[1],0.0);


static Display *display = 0;
static Window  window = 0;
static GC      gc = 0;
static int     width, height;
static vector2 scale;
static vector2 center;
static XGCValues gcvalues;
static boolean  draw_highlighted = FALSE;
static boolean  draw_marked = FALSE;
static boolean  draw_names = FALSE;


#define MAX_NR_XEDGES 100
static XPoint xpoint[MAX_NR_XEDGES];



void draw2d_set(Display *mydisplay,
		Window mywindow,
		GC mygc,
		int mywidth,
		int myheight,
		double xscale,
		double yscale,
		vector2 mycenter)
{
    display = mydisplay;
    window  = mywindow;
    gc      = mygc;
    width = mywidth;
    height = myheight;
    v2_set(scale, xscale, yscale);
    v2_let(center, mycenter);

    if (gc==0) warning("xdraw: draw2d() GC is zero?");
}


void xy_transform(vector2 x, vector2 start, double angle, vector2 index)
{
    double sina=sin(angle);
    double cosa=cos(angle);

    x[0] = start[0]+cosa*index[0]-sina*index[1];
    x[1] = start[1]+cosa*index[1]+sina*index[0];
}


void change_gc(boolean marked, boolean highlighted)
{
    unsigned long mask = 0;

    gcvalues.line_style = ((highlighted) ? LineOnOffDash : LineSolid);
    gcvalues.line_width = ((marked) ? 2 : 0);

    if (marked != draw_marked) mask |= GCLineWidth;

    if (highlighted != draw_highlighted) mask |= GCLineStyle;

    if (mask ) XChangeGC(display, gc, mask, &gcvalues);

    draw_marked = marked;
    draw_highlighted = highlighted;
}





void draw2d_get_pick_parameters(vector3 *start, vector3 *direction, int x, int y)
{
    get_v3_coord(*start, x, y);
    (*start)[2] = 100.0;
    v3_set(*direction, 0.0, 0.0, -1.0);
}


void draw2d_striangle(triangle t, vector3 *tedges)
{
    mw_object obj;

    get_win_coord(xpoint[0].x, xpoint[0].y, tedges[0]);
    get_win_coord(xpoint[1].x, xpoint[1].y, tedges[1]);
    get_win_coord(xpoint[2].x, xpoint[2].y, tedges[2]);
    xpoint[3] = xpoint[0];

    obj = (mw_object)triangle_get_user_data(t);
    change_gc(obj->marked, obj->highlighted);

    XDrawLines(display, window, gc, xpoint, 4, CoordModeOrigin);
}

void draw2d_get_real_coord(vector3 *point, int x, int y)
{
    get_v3_coord(*point, x, y);
}

int  draw2d_distance(vector3 a, vector3 b)
{
    return (int)dround(sqrt(pow((a[0]-b[0])*scale[0],2.0) +
		 pow((a[1]-b[1])*scale[1], 2.0)));
}

void draw2d_v3_norm(vector3 *x)
{
    (*x)[2] = 0.0;
}
static void draw2d_cuboid(mw_object obj)
{
    get_win_coord(xpoint[0].x, xpoint[0].y, obj->abs_edge[4]);
    get_win_coord(xpoint[1].x, xpoint[1].y, obj->abs_edge[5]);
    get_win_coord(xpoint[2].x, xpoint[2].y, obj->abs_edge[1]);
    get_win_coord(xpoint[3].x, xpoint[3].y, obj->abs_edge[0]);
    get_win_coord(xpoint[4].x, xpoint[4].y, obj->abs_edge[2]);
    get_win_coord(xpoint[5].x, xpoint[5].y, obj->abs_edge[3]);
    get_win_coord(xpoint[6].x, xpoint[6].y, obj->abs_edge[7]);
    get_win_coord(xpoint[7].x, xpoint[7].y, obj->abs_edge[6]);

    xpoint[8] = xpoint[0];
    xpoint[9] = xpoint[3];

    XDrawLines(display, window, gc, xpoint, 10, CoordModeOrigin);

    xpoint[0] = xpoint[6];
    xpoint[3] = xpoint[5];
    xpoint[5] = xpoint[7];

    XDrawSegments(display, window, gc, (XSegment *)xpoint, 3);
}


static void draw2d_rectangle(mw_object obj)
{
    get_win_coord(xpoint[0].x, xpoint[0].y, obj->abs_edge[0]);
    get_win_coord(xpoint[1].x, xpoint[1].y, obj->abs_edge[1]);
    get_win_coord(xpoint[2].x, xpoint[2].y, obj->abs_edge[2]);
    get_win_coord(xpoint[3].x, xpoint[3].y, obj->abs_edge[3]);
    xpoint[4] = xpoint[0];

    XDrawLines(display, window, gc, xpoint, 5, CoordModeOrigin);
}

static void draw2d_triangle(mw_object obj)
{
    get_win_coord(xpoint[0].x, xpoint[0].y, obj->abs_edge[0]);
    get_win_coord(xpoint[1].x, xpoint[1].y, obj->abs_edge[1]);
    get_win_coord(xpoint[2].x, xpoint[2].y, obj->abs_edge[2]);
    xpoint[3] = xpoint[0];

    XDrawLines(display, window, gc, xpoint, 4, CoordModeOrigin);
}




void draw2d_app(app_2d *app, hom_matrix t)
{
    vector3 v;
    int *codes, old_code;
    int i;

    
    if (app->nr_points > MAX_NR_XEDGES) {
	warning("xdraw: draw2d_app() - max number of edges per app_2d exceeded");
	return ;
    }
    if (app->nr_points <= 0) return;

    for (i=(app->nr_points)-1; i>=0; i--) {
	hm_vmul(&v, t, app->point[i]);
	get_win_coord(xpoint[i].x, xpoint[i].y, v);
    }

    codes = app->code;
    while (*codes && ((*codes)!=END_2D)) {
        switch(*codes) {
	case POINTS_2D:
	    codes++;
	    while ((*codes)>=0) {
		XDrawPoint(display, window, gc,
			   (int)xpoint[*codes].x, (int)xpoint[*codes].y);
		codes++;
	    }
	    break;

        case POLYLINE_2D:
	    ++codes;
            if (*codes < 0)
		error("xdraw: polydraw2d() - invalid index for 1.point of a polyline");

            old_code = *codes;

            ++codes;
	    while ((*codes)>=0) {
		XDrawLine(display, window, gc,
			   (int)xpoint[old_code].x, (int)xpoint[old_code].y,
			   (int)xpoint[*codes].x, (int)xpoint[*codes].y);
		old_code = *codes;
		codes++;
	    }

	case END_2D:
	    break;

	default:
            error("xdraw: polydraw2d() - unknown draw code");
	    break;
	}
    }
}

static void draw2d_object(mw_object obj)
{
    change_gc(obj->marked, obj->highlighted);

    if (obj->appearance_2d && (obj->appearance_2d != &def_2d_empty)) {
	    draw2d_app(obj->appearance_2d, obj->htm);
    }

    if (obj->type == MWO_GROUP)	{
	mw_object o;
	
	o=list_first(obj->obj.group.members);
	while (o) {
	    draw2d_object(o);
	    o=list_next(obj->obj.group.members);
	}
    }
    
    if (obj->appearance_2d == 0) switch (obj->type) {
    case MWO_GROUP:
	break;

    case MWO_CUBOID:
	draw2d_cuboid(obj);
	break;

    case MWO_RECTANGLE:
	draw2d_rectangle(obj);
	break;

    case MWO_TRIANGLE:
	draw2d_triangle(obj);
	break;

    default:
	error("xdraw: draw2d_obstacle() - unknown obstacle type");
	break;
    }

    if (obj->group_head == 0) {
	int x,y;
	vector3 v;

	v3_set(v, obj->htm[3][0], obj->htm[3][1], obj->htm[3][2]);
	get_win_coord(x,y, v);
	XDrawPoint(display, window, gc, x, y);
    }

    if (draw_names && obj->name) {
	int x,y;
	int len = strlen(obj->name);
	vector3 v;

	if (len>10) len = 10;

	v3_set(v, obj->htm[3][0], obj->htm[3][1], obj->htm[3][2]);
	get_win_coord(x,y, v);
	XDrawString(display, window, gc, x + 2, y + 4, obj->name, len);
	
    }
}

boolean draw2d_us_sensor_result(void *vp)
{
    us_sensor_check_point *sr = *((us_sensor_check_point **)vp);


    if (sr->detailed) {
	if (sr->path) {
	    vector3 v;
	    int i;

	    if (queue_first(&v, sr->path)) {
		get_win_coord(xpoint[0].x, xpoint[0].y, v);
		queue_add(sr->path, &v);
		
		i=queue_size(sr->path);
		for (; i>1; i--) {
		    queue_first(&v, sr->path);
		    get_win_coord(xpoint[1].x, xpoint[1].y, v);
		    queue_add(sr->path, &v);
		    XDrawLine(display, window, gc,
			      (int)(xpoint[0].x), (int)(xpoint[0].y),
			      (int)(xpoint[1].x), (int)(xpoint[1].y));
		    xpoint[0] = xpoint[1];
		}
		    
	    }
	    get_win_coord(xpoint[1].x, xpoint[1].y, sr->end_pos);
	}
	else {
	    get_win_coord(xpoint[0].x, xpoint[0].y, sr->position);
	    get_win_coord(xpoint[1].x, xpoint[1].y, sr->end_pos);
	    
	    XDrawLine(display, window, gc,
		      (int)(xpoint[0].x), (int)(xpoint[0].y),
		      (int)(xpoint[1].x), (int)(xpoint[1].y));
	}
    } else {
	get_win_coord(xpoint[1].x, xpoint[1].y, sr->end_pos);
    }
    XDrawArc(display, window, gc, (int)(xpoint[1].x)-3, (int)(xpoint[1].y)-3,
		 6, 6, 0, 360*64);
	
    return TRUE;
}

boolean draw2d_mob_check_point(void *vp)
{
    mob_check_point *cp = (mob_check_point *)vp;
    vector3 v;

    v3_set(v, cp->state.x_real, cp->state.y_real, 0.0);
    get_win_coord(xpoint[0].x, xpoint[0].y, v);
    XDrawPoint(display, window, gc, (int)xpoint[0].x, (int)xpoint[0].y);
        
    return TRUE;
}

void draw2d(mob_world mw)
{
    mw_object obj;

    XClearWindow(display, window);

    change_gc(FALSE, FALSE);

    /* draw mob-track
     */
    queue_first_to_last(mw->mob_check_points, draw2d_mob_check_point);

    /* draw us-sensor-results
     */
    queue_first_to_last(mw->us_sensor_results, draw2d_us_sensor_result);
    
    /* draw objects
     */
    for (obj=list_first(mw->objects); obj; obj=list_next(mw->objects)) {
	draw2d_object(obj);
    }

    XFlush(display);
}


void draw2d_space(mob_world mw)
{
    XClearWindow(display, window);

    change_gc(FALSE, FALSE);

    triangles_do(mw->space, draw2d_striangle);

    /* draw mob-track
     */
    queue_first_to_last(mw->mob_check_points, draw2d_mob_check_point);

    /* draw us-sensor-results
     */
    queue_first_to_last(mw->us_sensor_results, draw2d_us_sensor_result);
    
    XFlush(display);
}


int draw2d_opt(int type, int value)
{
    int old_value = 0;

    switch(type) {
    case D2D_DRAW_NAMES:
	old_value = draw_names;
	draw_names = value;
	break;

    default:
	break;
    }
    return old_value;
}
